#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <gdkmm.h>
#include <gtkmm.h>

#ifdef HAVE_X11_XLIB_H
#include <X11/Xlib.h>
#undef None
#endif

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#ifdef HAVE_GRP_H
#include <grp.h>
#endif

#ifdef HAVE_SYSTEMD_SD_DAEMON_H
#include <systemd/sd-daemon.h>
#endif

#include <errno.h>
#include <string.h>
#include <libxml/HTMLparser.h>

#include "cfmuvalidateserver.hh"
#include "fplan.h"
#include "icaofpl.h"

static Glib::RefPtr<Glib::MainLoop> mainloop;

static void signalterminate(int signr)
{
	if (mainloop)
		mainloop->quit();
}

void CFMUValidateServer::Results::print(const TrafficFlowRestrictions::Message& m)
{
	std::ostringstream oss;
	oss << m.get_type_char() << ':';
	if (!m.get_rule().empty())
		oss << " R:" << m.get_rule();
	if (!m.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_vertexset().begin()), e(m.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!m.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_edgeset().begin()), e(m.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << m.get_text();
	push_back(oss.str());
}

void CFMUValidateServer::Results::print(const TrafficFlowRestrictions::Result& res)
{
	// print Messages
	for (unsigned int i = 0; i < res.messages_size(); ++i)
		print(res.get_message(i));
	push_back(res ? "NO ERRORS" : "FAIL");
	for (unsigned int i = 0; i < res.size(); ++i)
		print(res[i]);
}

void CFMUValidateServer::Results::print(const TrafficFlowRestrictions::RuleResult& rule)
{
	std::ostringstream oss;
	oss << "Rule: " << rule.get_rule() << ' ';
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		oss << "Mandatory";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		oss << "Forbidden";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		oss << "ClosedForCruising";
		break;

	default:
		oss << '?';
		break;
	}
	if (rule.is_dct())
		oss << " (DCT)";
	if (!rule.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_reflocset().empty()) {
		oss << " L:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_reflocset().begin()), e(rule.get_reflocset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	push_back(oss.str());
	push_back("Desc: " + rule.get_desc());
	push_back("OprGoal: " + rule.get_oprgoal());
	push_back("Cond: " + rule.get_cond());
	for (unsigned int i = 0; i < rule.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
		{
			std::ostringstream oss;
			oss << "  Alternative " << i;
			if (!alt.get_vertexset().empty()) {
				oss << " V:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!alt.get_edgeset().empty()) {
				oss << " E:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
		        push_back(oss.str());
		}
		for (unsigned int j = 0; j < alt.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
			std::ostringstream oss;
			oss << "    " << el.get_type_string();
			switch (el.get_type()) {
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
				oss << ' ' << el.get_ident();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
				oss << ' ' << el.get_ident() << '/'
				    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
			default:
				break;
			}
			oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100) << "...F"
			    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
		        push_back(oss.str());
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				push_back(oss.str());
			}
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				push_back(oss.str());
			}
		}
	}
}

void CFMUValidateServer::Results::print(const rules_t& rules, const Glib::ustring& pfx)
{
	std::string r;
	bool subseq(false);
	for (rules_t::const_iterator ti(rules.begin()), te(rules.end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	push_back(pfx + r);
}

void CFMUValidateServer::Results::annotate(TrafficFlowRestrictions& m_tfr)
{
	Results x;
	for (Results::const_iterator ri(begin()), re(end()); ri != re; ++ri) {
		x.push_back(*ri);
		Glib::ustring::size_type n(ri->rfind('['));
		if (n == Glib::ustring::npos)
			continue;
		Glib::ustring::size_type ne(ri->find(']', n + 1));
		if (ne == Glib::ustring::npos || ne - n < 2 || ne - n > 11)
			continue;
		std::vector<TrafficFlowRestrictions::RuleResult> rr(m_tfr.find_rules(ri->substr(n + 1, ne - n - 1), true));
		if (rr.empty())
			continue;
		x.print(rr[0]);
	}
	swap(x);
}

bool CFMUValidateServer::OptionalArgFilename::on_option_arg_filename(const Glib::ustring& option_name, const std::string& value, bool has_value)
{
	m_value = has_value;
	if (has_value)
		m_arg = value;
	else
		m_arg.clear();
	m_set = true;
	return true;
}
	
bool CFMUValidateServer::OptionalArgString::on_option_arg_string(const Glib::ustring& option_name, const Glib::ustring& value, bool has_value)
{
	m_value = has_value;
	if (has_value)
		m_arg = value;
	else
		m_arg.clear();
	m_set = true;
	return true;
}

CFMUValidateServer::Client::Client(CFMUValidateServer *server)
	: m_server(server), m_refcount(1),
#ifdef HAVE_CURL
	  m_curl(0), m_curl_headers(0),
#endif
	  m_validator(validator_cfmu), m_annotate(false)
{
        try {
		Glib::KeyFile kf;
                kf.load_from_file(PACKAGE_SYSCONF_DIR "/cfmu.ini", Glib::KEY_FILE_KEEP_COMMENTS | Glib::KEY_FILE_KEEP_TRANSLATIONS);
		if (kf.has_group("b2bop")) {
			if (kf.has_key("b2bop", "url"))
				m_cfmub2bopurl = kf.get_string("b2bop", "url");
			if (kf.has_key("b2bop", "url"))
				m_cfmub2bopcert = kf.get_string("b2bop", "cert");
		}
		if (kf.has_group("b2bpreop")) {
			if (kf.has_key("b2bpreop", "url"))
				m_cfmub2bpreopurl = kf.get_string("b2bpreop", "url");
			if (kf.has_key("b2bpreop", "url"))
				m_cfmub2bpreopcert = kf.get_string("b2bpreop", "cert");
		}
        } catch (const Glib::FileError& e) {
                std::cerr << "Unable to load CFMU settings from " PACKAGE_SYSCONF_DIR "/cfmu.ini: " << e.what() << std::endl;
        } catch (const Gio::Error& e) {
                std::cerr << "Unable to load CFMU settings from " PACKAGE_SYSCONF_DIR "/cfmu.ini: " << e.what() << std::endl;
        }
	if (!m_cfmub2bopcert.empty())
		if (!Glib::file_test(m_cfmub2bopcert, Glib::FILE_TEST_EXISTS) ||
		    !Glib::file_test(m_cfmub2bopcert, Glib::FILE_TEST_IS_REGULAR)) {
			std::cerr << "B2B Op Cert " << m_cfmub2bopcert << " unreadable" << std::endl;
			m_cfmub2bopcert.clear();
		}
	if (!m_cfmub2bpreopcert.empty())
		if (!Glib::file_test(m_cfmub2bpreopcert, Glib::FILE_TEST_EXISTS) ||
		    !Glib::file_test(m_cfmub2bpreopcert, Glib::FILE_TEST_IS_REGULAR)) {
			std::cerr << "B2B Preop Cert " << m_cfmub2bpreopcert << " unreadable" << std::endl;
			m_cfmub2bpreopcert.clear();
		}
	if (!m_cfmub2bopurl.empty() && !m_cfmub2bopcert.empty())
		m_validator = validator_cfmub2bop;
	else if (!m_cfmub2bpreopurl.empty() && !m_cfmub2bpreopcert.empty())
		m_validator = validator_cfmub2bpreop;
}

CFMUValidateServer::Client::~Client()
{
#ifdef HAVE_CURL
	close_curl();
#endif
}

void CFMUValidateServer::Client::reference(void) const
{
        g_atomic_int_inc(&m_refcount);
	if (false)
		std::cerr << "client " << (unsigned long long)this << " reference " << m_refcount << std::endl;
}

void CFMUValidateServer::Client::unreference(void) const
{
	bool nz(!g_atomic_int_dec_and_test(&m_refcount));
	if (false)
		std::cerr << "client " << (unsigned long long)this << " unreference " << m_refcount << std::endl;
        if (nz)
                return;
        delete this;
}

const std::string& CFMUValidateServer::Client::to_str(validator_t v)
{
	switch (v) {
	case validator_local:
	{
		static const std::string r("local");
		return r;
	}

	case validator_cfmu:
	{
		static const std::string r("cfmu");
		return r;
	}

	case validator_eurofpl:
	{
		static const std::string r("eurofpl");
		return r;
	}

	default:
	{
		static const std::string r("?");
		return r;
	}
	}
}

void CFMUValidateServer::Client::handle_input(void)
{
	if (false)
		std::cerr << "handle_input: " << m_fplan << std::endl;
	if (m_fplan.empty())
		return;
	// check for pseudo instructions
	if (!m_fplan.compare(0, 6, "trace?")) {
		Results res;
		res.print(m_tracerules, "trace:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 6, "trace:") || !m_fplan.compare(0, 6, "trace+")) {
		Glib::ustring txt1(m_fplan);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tracerules.insert(txt1.substr(6));
		Results res;
		res.print(m_tracerules, "trace:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 6, "trace-")) {
		Glib::ustring txt1(m_fplan);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_tracerules.erase(txt1.substr(6));
		Results res;
		res.print(m_tracerules, "trace:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 6, "trace*")) {
		m_tracerules.clear();
		Results res;
		res.print(m_tracerules, "trace:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "disabled?")) {
	        Results res;
		res.print(m_disabledrules, "disabled:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "disabled:") || !m_fplan.compare(0, 9, "disabled+")) {
		Glib::ustring txt1(m_fplan);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_disabledrules.insert(txt1.substr(9));
		Results res;
		res.print(m_disabledrules, "disabled:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "disabled-")) {
		Glib::ustring txt1(m_fplan);
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		m_disabledrules.erase(txt1.substr(9));
		Results res;
		res.print(m_disabledrules, "disabled:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "disabled*")) {
		m_disabledrules.clear();
		Results res;
		res.print(m_disabledrules, "disabled:");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "validate?")) {
		Results res;
		res.push_back("validate:" + to_str(m_validator));
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 9, "validate:")) {
		Glib::ustring txt1(m_fplan.substr(9));
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		for (validator_t v = validator_cfmu; v <= validator_local; v = (validator_t)(v + 1)) {
			if (txt1 != to_str(v))
				continue;
			m_validator = v;
			break;
		}
		Results res;
		res.push_back("validate:" + to_str(m_validator));
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 10, "validate*:")) {
		Glib::ustring txt1(m_fplan.substr(10));
		Glib::ustring::size_type n(txt1.find('\n'));
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		n = txt1.find(' ');
		if (n != Glib::ustring::npos)
			txt1.erase(n);
		for (validator_t v = validator_cfmu; v <= validator_local; v = (validator_t)(v + 1)) {
			if (txt1 != to_str(v))
				continue;
			m_validator = v;
			break;
		}
		m_fplan.clear();
		return;
	}
	if (!m_fplan.compare(0, 9, "annotate?")) {
		Results res;
		res.push_back((std::string)"annotate:" + (m_annotate ? "on" : "off"));
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 11, "annotate:on")) {
		m_annotate = true;
		Results res;
		if (m_server)
			m_server->annotate(res);
		res.push_back("annotate:on");
		set_result(res);
		return;
	}
	if (!m_fplan.compare(0, 12, "annotate:off")) {
		m_annotate = false;
		Results res;
		res.push_back("annotate:off");
		set_result(res);
		return;
	}
	if (!m_server) {
		Results res;
		res.push_back("FAIL:Internal Error");
		set_result(res);
		return;
	}
	switch (m_validator) {
	case validator_cfmu:
		m_server->validate_cfmu(get_ptr());
		break;

	case validator_local:
		m_server->validate_locally(get_ptr(), m_tracerules, m_disabledrules);
		break;
	
#ifdef HAVE_CURL
	case validator_eurofpl:
	{
		close_curl();
		m_curlrxbuffer.clear();
		m_curltxbuffer.clear();
		m_curl = curl_easy_init();
		if (!m_curl) {
			Results res;
			res.push_back("FAIL:Curl Context");
			set_result(res);
			return;
		}
		//m_curl_headers = curl_slist_append(m_curl_headers, "Content-Type: application/xml");
		curl_easy_setopt(m_curl, CURLOPT_HTTPHEADER, m_curl_headers);
		curl_easy_setopt(m_curl, CURLOPT_URL, "http://validation.eurofpl.eu");
		curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, my_write_func_1);
		curl_easy_setopt(m_curl, CURLOPT_READDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_READFUNCTION, my_read_func_1);
		curl_easy_setopt(m_curl, CURLOPT_NOPROGRESS, 0L);
		curl_easy_setopt(m_curl, CURLOPT_PROGRESSFUNCTION, my_progress_func_1);
		curl_easy_setopt(m_curl, CURLOPT_PROGRESSDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_REFERER, "http://validation.eurofpl.eu");
		curl_easy_setopt(m_curl, CURLOPT_POST, 1L);
		std::string query("freeEntry=");
		{
			unsigned int skipdash(!m_fplan.empty() && m_fplan[0] == '-');
			char *postfld(curl_easy_escape(m_curl, m_fplan.c_str() + skipdash, m_fplan.size() - skipdash));
			query += postfld;
			curl_free(postfld);
		}
		curl_easy_setopt(m_curl, CURLOPT_COPYPOSTFIELDS, query.c_str());
		m_server->curl_add(m_curl, get_ptr());
		break;
	}

	case validator_cfmub2bop:
	case validator_cfmub2bpreop:
	{
		const std::string *url(&m_cfmub2bopurl);
		const std::string *cert(&m_cfmub2bopcert);
		if (m_validator == validator_cfmub2bpreop) {
			url = &m_cfmub2bpreopurl;
			cert = &m_cfmub2bpreopcert;
		}
		if (url->empty() || cert->empty()) {
			Results res;
			res.push_back("FAIL:Validator not configured");
			set_result(res);
			return;
		}
		close_curl();
		m_curlrxbuffer.clear();
		m_curltxbuffer.clear();
		m_curl = curl_easy_init();
		if (!m_curl) {
			Results res;
			res.push_back("FAIL:Curl Context");
			set_result(res);
			return;
		}
		{
			std::auto_ptr<xmlpp::Document> doc(new xmlpp::Document());
			xmlpp::Element *rootel(doc->create_root_node("Envelope", "http://schemas.xmlsoap.org/soap/envelope/", "S"));
			rootel->set_namespace_declaration("eurocontrol/cfmu/b2b/Flight_icao2012Services", "ns0");
			xmlpp::Element *bodyel(rootel->add_child("Body", "S"));
			xmlpp::Element *valreqel(bodyel->add_child("FlightPlanValidationRequest", "ns0"));
			xmlpp::Element *sndtel(valreqel->add_child("sendTime"));
			xmlpp::Element *fplel(valreqel->add_child("fplMessage"));
			fplel->set_child_text(m_fplan.c_str());
			Glib::DateTime dt(Glib::DateTime::create_now_utc());
			sndtel->set_child_text(dt.format("%Y-%m-%d %H:%M:%S").c_str());
			if (false)
				std::cerr << "transmitting: " << doc->write_to_string_formatted() << std::endl;
			Glib::ustring q(doc->write_to_string());
			//m_curltxbuffer.insert(m_curltxbuffer.end(), q.begin(), q.end());
			m_curltxbuffer.resize(q.size());
			if (m_curltxbuffer.size())
				memcpy(&m_curltxbuffer[0], q.c_str(), m_curltxbuffer.size());
		}
		m_curl_headers = curl_slist_append(m_curl_headers, "Content-Type: text/xml; charset=utf-8");
		{
			std::ostringstream oss;
			oss << "Content-Length: " << m_curltxbuffer.size();
			m_curl_headers = curl_slist_append(m_curl_headers, oss.str().c_str());
		}
		m_curl_headers = curl_slist_append(m_curl_headers, "SOAPAction: \"\"");
		curl_easy_setopt(m_curl, CURLOPT_HTTPHEADER, m_curl_headers);
		curl_easy_setopt(m_curl, CURLOPT_URL, url->c_str());
		curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, my_write_func_1);
		curl_easy_setopt(m_curl, CURLOPT_READDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_READFUNCTION, my_read_func_1);
		curl_easy_setopt(m_curl, CURLOPT_NOPROGRESS, 0L);
		curl_easy_setopt(m_curl, CURLOPT_PROGRESSFUNCTION, my_progress_func_1);
		curl_easy_setopt(m_curl, CURLOPT_PROGRESSDATA, this);
		curl_easy_setopt(m_curl, CURLOPT_POST, 1L);
		if (false && (curl_easy_setopt(m_curl, CURLOPT_SSLENGINE, 0) != CURLE_OK ||
			      curl_easy_setopt(m_curl, CURLOPT_SSLENGINE_DEFAULT, 1L) != CURLE_OK)) {
			close_curl();
			Results res;
			res.push_back("FAIL:Curl SSL Setup");
			set_result(res);
			return;
		}
		curl_easy_setopt(m_curl, CURLOPT_SSLCERTTYPE, "PEM");
		curl_easy_setopt(m_curl, CURLOPT_SSLCERT, cert->c_str());
		//curl_easy_setopt(m_curl, CURLOPT_KEYPASSWD, "");
		curl_easy_setopt(m_curl, CURLOPT_SSLKEYTYPE, "PEM");
		curl_easy_setopt(m_curl, CURLOPT_SSLKEY, cert->c_str());
		curl_easy_setopt(m_curl, CURLOPT_CAINFO, "/etc/pki/tls/certs/ca-bundle.crt");
		//curl_easy_setopt(m_curl, CURLOPT_SSL_VERIFYPEER, 1L);
		curl_easy_setopt(m_curl, CURLOPT_SSL_VERIFYPEER, 1L);
		//curl_easy_setopt(m_curl, CURLOPT_VERBOSE, 1L);
		m_server->curl_add(m_curl, get_ptr());
		break;
	}
#endif

	default:
	{
		Results res;
		res.push_back("FAIL:Invalid Validator");
		set_result(res);
		return;
	}
	}
}

#ifdef HAVE_CURL
void CFMUValidateServer::Client::close_curl(void)
{
	if (m_curl_headers)
		curl_slist_free_all(m_curl_headers);
	m_curl_headers = 0;
	if (!m_curl)
		return;
	if (m_server)
		m_server->curl_remove(m_curl);
	curl_easy_cleanup(m_curl);
	m_curl = 0;
}

size_t CFMUValidateServer::Client::my_write_func_1(const void *ptr, size_t size, size_t nmemb, Client *clnt)
{
        return clnt->my_write_func(ptr, size, nmemb);
}

size_t CFMUValidateServer::Client::my_read_func_1(void *ptr, size_t size, size_t nmemb, Client *clnt)
{
        return clnt->my_read_func(ptr, size, nmemb);
}

int CFMUValidateServer::Client::my_progress_func_1(Client *clnt, double t, double d, double ultotal, double ulnow)
{
        return clnt->my_progress_func(t, d, ultotal, ulnow);
}

size_t CFMUValidateServer::Client::my_write_func(const void *ptr, size_t size, size_t nmemb)
{
        unsigned int sz(size * nmemb);
        unsigned int p(m_curlrxbuffer.size());
        m_curlrxbuffer.resize(p + sz);
        memcpy(&m_curlrxbuffer[p], ptr, sz);
        return sz;
}

size_t CFMUValidateServer::Client::my_read_func(void *ptr, size_t size, size_t nmemb)
{
        unsigned int sz(size * nmemb);
	if (sz > m_curltxbuffer.size())
		sz = m_curltxbuffer.size();
	if (sz) {
		memcpy(ptr, &m_curltxbuffer[0], sz);
		m_curltxbuffer.erase(m_curltxbuffer.begin(), m_curltxbuffer.begin() + sz);
	}
        return sz;
}

int CFMUValidateServer::Client::my_progress_func(double t, double d, double ultotal, double ulnow)
{
        return 0;
}

void CFMUValidateServer::Client::find_eurofpl_results(std::string& result, xmlNode *a_node, unsigned int level)
{
	for (xmlNode *cur_node = a_node; cur_node; cur_node = cur_node->next) {
		if (cur_node->type == XML_ELEMENT_NODE) {
			xmlChar *cls(xmlGetProp(cur_node, BAD_CAST "class"));
			if (cls) {
				if (!xmlStrcmp(cls, BAD_CAST "ifpuv_result")) {
					for (xmlNode *cld_node = cur_node->children; cld_node; cld_node = cld_node->next) {
						if (cld_node->type == XML_ELEMENT_NODE && !xmlStrcmp(cld_node->name, BAD_CAST "br")) {
							result += "\r\n";
							continue;
						}
						if (cld_node->type == XML_TEXT_NODE) {
							result += (const char *)cld_node->content;
						}
					}
				}
			}
			if (false) {
				std::cerr << std::string(level, ' ') << "node type: Element, name: " << cur_node->name;
				if (cls)
					std::cerr << " class: " << (char *)cls;
				std::cerr << std::endl;
			}
			if (cls)
				xmlFree(cls);
			if (!result.empty())
				break;
		}
		find_eurofpl_results(result, cur_node->children, level + 1);
		if (!result.empty())
			break;
	}
}

void CFMUValidateServer::Client::parse_eurofpl(void)
{
	xmlDocPtr doc(htmlReadMemory(const_cast<char *>(&m_curlrxbuffer[0]), m_curlrxbuffer.size(), "http://validation.eurofpl.eu", 0, 0));
	xmlCleanupParser();
	std::string resultstr;
	find_eurofpl_results(resultstr, xmlDocGetRootElement(doc));
	xmlFreeDoc(doc);
	xmlCleanupParser();
	if (resultstr.empty()) {
		Results res;
		res.push_back("FAIL:ifpuv_result not found");
		set_result(res);
		return;
	}
	Results res;
	res.push_back(resultstr);
	resultstr.clear();
	for (;;) {
		Glib::ustring::size_type n(res.back().find_first_of("\r\n"));
		if (n == Glib::ustring::npos)
			break;
		Glib::ustring::size_type n1(n);
		++n1;
		while (n1 < res.back().size() && (res.back()[n1] == '\n' || res.back()[n1] == '\r'))
			++n1;
		Glib::ustring x(res.back().substr(n1));
		res.back().erase(n);
		res.push_back(x);
	}
	for (Results::size_type i(0); i < res.size(); ) {
		if (res[i].empty()) {
			res.erase(res.begin() + i);
			continue;
		}
		++i;
	}
	set_result(res);
}

void CFMUValidateServer::Client::parse_cfmub2b(void)
{
	xmlpp::DomParser p;
	p.set_validate(false);
	p.set_substitute_entities(false);
	//p.set_throw_messages(false);
 	p.parse_memory_raw((const unsigned char *)&m_curlrxbuffer[0], m_curlrxbuffer.size());
        if (!p) {
		Results res;
		res.push_back("FAIL:Unparseable XML");
		set_result(res);
                return;
	}
	if (false)
		std::cerr << "received: " << p.get_document()->write_to_string_formatted() << std::endl;
	const xmlpp::Element *rootel(p.get_document()->get_root_node());
	if (rootel->get_name() != "Envelope") {
		Results res;
		res.push_back("FAIL:XML:root element not Envelope");
		set_result(res);
                return;
	}
	xmlpp::Node::NodeList nlbody(rootel->get_children("Body"));
	for (xmlpp::Node::NodeList::const_iterator ni(nlbody.begin()), ne(nlbody.end()); ni != ne; ++ni) {
                const xmlpp::Element *bodyel(static_cast<const xmlpp::Element *>(*ni));
		if (!bodyel)
			continue;
		xmlpp::Node::NodeList nlval(bodyel->get_children("FlightPlanValidationReply"));
		for (xmlpp::Node::NodeList::const_iterator ni(nlval.begin()), ne(nlval.end()); ni != ne; ++ni) {
			const xmlpp::Element *valel(static_cast<const xmlpp::Element *>(*ni));
			if (!valel)
				continue;
			{
				xmlpp::Node::NodeList nlstat(valel->get_children("status"));
				xmlpp::Node::NodeList::const_iterator ni(nlstat.begin()), ne(nlstat.end());
				for (; ni != ne; ++ni) {
					std::string t(get_text(*ni));
					if (t.empty())
						continue;
					if (t == "OK")
						break;
					Results res;
					res.push_back("FAIL:CFMU:Status:" + t);
					set_result(res);
					return;
				}
				if (ni == ne)
					continue;
			}
			Results res;
			xmlpp::Node::NodeList nldata(valel->get_children("data"));
			for (xmlpp::Node::NodeList::const_iterator ni(nldata.begin()), ne(nldata.end()); ni != ne; ++ni) {
				const xmlpp::Element *datael(static_cast<const xmlpp::Element *>(*ni));
				if (!datael)
					continue;
				xmlpp::Node::NodeList nlerr(datael->get_children("ifpsErrors"));
				for (xmlpp::Node::NodeList::const_iterator ni(nlerr.begin()), ne(nlerr.end()); ni != ne; ++ni) {
					const xmlpp::Element *errel(static_cast<const xmlpp::Element *>(*ni));
					if (!errel)
						continue;
					std::string code;
					std::string descr;
					xmlpp::Node::NodeList nl(errel->get_children("code"));
					for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
						code += get_text(*ni);
					nl = errel->get_children("description");
					for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
						descr += get_text(*ni);
					res.push_back(code + ": " + descr);
				}
			}
			if (res.empty())
				res.push_back("NO ERRORS");
			set_result(res);
			return;
		}
	}
}

std::string CFMUValidateServer::Client::get_text(const xmlpp::Node *n)
{
        const xmlpp::Element *el(static_cast<const xmlpp::Element *>(n));
        if (!el)
                return "";
        const xmlpp::TextNode *t(el->get_child_text());
        if (!t)
                return "";
        return t->get_content();
}

void CFMUValidateServer::Client::curl_done(CURLcode result)
{
	if (result != CURLE_OK) {
		std::ostringstream oss;
		oss << "FAIL:CURL " << curl_easy_strerror(result);
		Results res;
		res.push_back(oss.str());
		set_result(res);
		m_curlrxbuffer.clear();
		return;
	}
	if (false)
		std::cerr << "received: " << std::string((char *)&m_curlrxbuffer[0], ((char *)&m_curlrxbuffer[0]) + m_curlrxbuffer.size()) << std::endl;
	switch (m_validator) {
	case validator_eurofpl:
		parse_eurofpl();
		break;

	case validator_cfmub2bop:
	case validator_cfmub2bpreop:
		parse_cfmub2b();
		break;

	default:
	{
		Results res;
		res.push_back("FAIL:Invalid Validator");
		set_result(res);
		break;
	}
	}
	m_curlrxbuffer.clear();
}

#endif

CFMUValidateServer::StdioClient::StdioClient(CFMUValidateServer *server)
	: Client(server), m_inbufptr(0)
{
#ifdef G_OS_WIN32
        m_inputchan = Glib::IOChannel::create_from_win32_fd(0);
        m_outputchan = Glib::IOChannel::create_from_win32_fd(1);	
#else
        m_inputchan = Glib::IOChannel::create_from_fd(0);
        m_outputchan = Glib::IOChannel::create_from_fd(1);
#endif
        m_inputchan->set_encoding();
        m_outputchan->set_encoding();
#ifndef G_OS_WIN32
	m_inputchan->set_flags(m_inputchan->get_flags() | Glib::IO_FLAG_NONBLOCK);
#endif
	m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(this, &StdioClient::input_line_handler),
						      m_inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
}

CFMUValidateServer::StdioClient::~StdioClient()
{
	m_inputreaderconn.disconnect();
	m_inputchan.reset();
	m_outputchan.reset();
}

void CFMUValidateServer::StdioClient::set_result(const Results& res)
{
	if (!m_outputchan) {
		m_inputreaderconn.disconnect();
		Client::ptr_t p(get_ptr());
		if (m_server)
			m_server->close(p);
		return;
	}
	Results res1(res);
	if (m_annotate && m_server && m_validator != validator_local)
		m_server->annotate(res1);
	for (Results::const_iterator i(res1.begin()), e(res1.end()); i != e; ++i)
		m_outputchan->write(*i + "\n");
	m_outputchan->write("\n");
	m_outputchan->flush();
	m_fplan.clear();
	handle_input_loop();
}

bool CFMUValidateServer::StdioClient::input_line_handler(Glib::IOCondition iocond)
{
	if (!m_inputchan) {
		handle_input_loop();
		return false;
        }
	gsize nrread;
#ifdef G_OS_WIN32
	Glib::IOStatus iostat(m_inputchan->read(m_inbuffer + m_inbufptr, 1, nrread));
#else
	Glib::IOStatus iostat(m_inputchan->read(m_inbuffer + m_inbufptr, sizeof(m_inbuffer) - m_inbufptr, nrread));
#endif
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
		m_inputreaderconn.disconnect();
		m_inputchan.reset();
		m_outputchan.reset();
		handle_input_loop();
		return false;
        }
	m_inbufptr += nrread;
	handle_input_loop();
	return true;
}

void CFMUValidateServer::StdioClient::handle_input_loop(void)
{
	for (;;) {
		if (!m_inputchan || !m_outputchan) {
			m_inputreaderconn.disconnect();
			Client::ptr_t p(get_ptr());
			if (m_server)
				m_server->close(p);
			return;
		}
		if (!m_fplan.empty()) {
			m_inputreaderconn.disconnect();
			return;
		}
		if (false)
			std::cerr << "handle_input_loop: buffer count " << m_inbufptr << std::endl;
		char *cp(static_cast<char *>(memchr(m_inbuffer, '\n', m_inbufptr)));
		{
			char *cp2(static_cast<char *>(memchr(m_inbuffer, '\r', m_inbufptr)));
			if (!cp || (cp2 && cp2 < cp))
				cp = cp2;
		}
		if (!cp) {
			if (m_inbufptr >= sizeof(m_inbuffer))
				m_inbufptr = 0;
			m_inputreaderconn.disconnect();
			if (m_inputchan)
				m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(this, &StdioClient::input_line_handler),
									      m_inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
			return;
		}
		{
			std::string fpl(m_inbuffer, cp);
			m_fplan.swap(fpl);
		}
		while (cp < m_inbuffer + m_inbufptr && (*cp == '\n' || *cp == '\r'))
			++cp;
		m_inbufptr -= cp - m_inbuffer;
		if (m_inbufptr)
			memmove(m_inbuffer, cp, m_inbufptr);
		handle_input();
	}
}

CFMUValidateServer::SocketClient::SocketClient(CFMUValidateServer *server, const Glib::RefPtr<Gio::SocketConnection>& connection)
	: Client(server), m_connection(connection), m_inbufptr(0)
{
	if (m_connection)
		m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(*this, &SocketClient::input_line_handler),
							      m_connection->get_socket()->get_fd(),
							      Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
}

CFMUValidateServer::SocketClient::~SocketClient()
{
	m_inputreaderconn.disconnect();
	m_connection.reset();
}

void CFMUValidateServer::SocketClient::set_result(const Results& res)
{
	if (!m_connection) {
		m_inputreaderconn.disconnect();
		Client::ptr_t p(get_ptr());
		if (m_server)
			m_server->close(p);
		return;
	}
	Results res1(res);
	if (m_annotate && m_server && m_validator != validator_local)
		m_server->annotate(res1);
	std::string tx;
	for (Results::const_iterator i(res1.begin()), e(res1.end()); i != e; ++i)
		tx += *i + "\n";
	tx += "\n";
	try {
		m_connection->get_socket()->property_blocking() = true;
		m_connection->get_socket()->send(tx.c_str(), tx.size());
	} catch (const Glib::Error& e) {
		m_inputreaderconn.disconnect();
		m_connection.reset();
		std::cerr << "error sending reply: " << e.what() << std::endl;
	}
	m_fplan.clear();
	handle_input_loop();
}

bool CFMUValidateServer::SocketClient::input_line_handler(Glib::IOCondition iocond)
{
	if (!m_connection) {
		handle_input_loop();
		return false;
        }
	m_connection->get_socket()->property_blocking() = false;
	gssize nrread(m_connection->get_socket()->receive(m_inbuffer + m_inbufptr, sizeof(m_inbuffer) - m_inbufptr));
	if (nrread <= 0) {
		m_inputreaderconn.disconnect();
		m_connection.reset();
		handle_input_loop();
		return false;
        }
	m_inbufptr += nrread;
	handle_input_loop();
	return true;
}

void CFMUValidateServer::SocketClient::handle_input_loop(void)
{
	for (;;) {
		if (!m_connection) {
			m_inputreaderconn.disconnect();
			Client::ptr_t p(get_ptr());
			if (m_server)
				m_server->close(p);
			return;
		}
		if (!m_fplan.empty()) {
			m_inputreaderconn.disconnect();
			return;
		}
		if (false)
			std::cerr << "handle_input_loop: buffer count " << m_inbufptr << std::endl;
		char *cp(static_cast<char *>(memchr(m_inbuffer, '\n', m_inbufptr)));
		{
			char *cp2(static_cast<char *>(memchr(m_inbuffer, '\r', m_inbufptr)));
			if (!cp || (cp2 && cp2 < cp))
				cp = cp2;
		}
		if (!cp) {
			if (m_inbufptr >= sizeof(m_inbuffer))
				m_inbufptr = 0;
			m_inputreaderconn.disconnect();
			if (m_connection)
				m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(*this, &SocketClient::input_line_handler),
									      m_connection->get_socket()->get_fd(),
									      Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
			return;
		}
		{
			std::string fpl(m_inbuffer, cp);
			m_fplan.swap(fpl);
		}
		while (cp < m_inbuffer + m_inbufptr && (*cp == '\n' || *cp == '\r'))
			++cp;
		m_inbufptr -= cp - m_inbuffer;
		if (m_inbufptr)
			memmove(m_inbuffer, cp, m_inbufptr);
		handle_input();
	}
}

CFMUValidateServer::SingleFplClient::SingleFplClient(CFMUValidateServer *server)
	: Client(server)
{
        m_outputchan = Glib::IOChannel::create_from_fd(1);
        m_outputchan->set_encoding();
}

CFMUValidateServer::SingleFplClient::~SingleFplClient()
{
	m_outputchan.reset();
}

void CFMUValidateServer::SingleFplClient::handle_input_loop(void)
{
	for (;;) {
		if (!m_fplan.empty())
			return;
		if (m_fplans.empty()) {
			if (!is_closeatend())
				return;
			m_outputchan.reset();
			Client::ptr_t p(get_ptr());
			if (m_server)
				m_server->close(p);
			return;
		}
		m_fplan = *m_fplans.begin();
		m_fplans.erase(m_fplans.begin());
		handle_input();
	}
}

void CFMUValidateServer::SingleFplClient::set_result(const Results& res)
{
	if (!m_outputchan) {
		Client::ptr_t p(get_ptr());
		if (m_server)
			m_server->close(p);
		return;
	}
	Results res1(res);
	if (m_annotate && m_server && m_validator != validator_local)
		m_server->annotate(res1);
	for (Results::const_iterator i(res1.begin()), e(res1.end()); i != e; ++i)
		m_outputchan->write(*i + "\n");
	m_outputchan->write("\n");
	m_outputchan->flush();
	m_fplan.clear();
	handle_input_loop();
}

void CFMUValidateServer::SingleFplClient::set_fplan(const std::string& fpl)
{
	m_fplans.push_back(fpl);
	handle_input_loop();
}

CFMUValidateServer::CFMUValidateServer(int& argc, char**& argv)
	: m_engine(0), m_statecnt(0), m_portal_webview(0), m_validate_webview(0),
	  m_portal_webview_close_sighandler(0), m_portal_webview_create_sighandler(0),
	  m_portal_webview_ready_sighandler(0), m_portal_webview_loadstatus_sighandler(0),
	  m_portal_webview_navpolicy_sighandler(0), m_portal_webview_newwndpolicy_sighandler(0),
	  m_portal_webview_rsrcload_sighandler(0), m_portal_webview_consolemsg_sighandler(0),
	  m_validate_webview_close_sighandler(0), m_validate_webview_create_sighandler(0),
	  m_validate_webview_ready_sighandler(0), m_validate_webview_loadstatus_sighandler(0),
	  m_validate_webview_navpolicy_sighandler(0), m_validate_webview_newwndpolicy_sighandler(0),
	  m_validate_webview_rsrcload_sighandler(0), m_validate_webview_consolemsg_sighandler(0),
	  m_verbose(0),
#ifdef HAVE_X11_XLIB_H
	  m_xdisplay(-1), m_x11srvrun(false),
#endif
#ifdef HAVE_CURL
	  m_curl(0),
#endif
	  m_wvstate(wvstate_idle), m_tfrloaded(false), m_quit(true)
{
#ifdef HAVE_CURL
	m_curl = curl_multi_init();
#endif
        bool version(false), local(false), eurofpl(false), b2bop(false), b2bpreop(false);
	Glib::ustring fpl;
	int verbose(0);
        OptionalArgFilename sockaddr;
#ifdef G_OS_UNIX
	uid_t sockservuid(0);
	gid_t sockservgid(0);
	uid_t socketuid(getuid());
	gid_t socketgid(getgid());
	mode_t socketmode(S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP);
#endif
	bool noterminate(false);
	// option group
	{
		bool dis_aux(false);
#ifdef G_OS_UNIX
		Glib::ustring strsockservuid;
		Glib::ustring strsockservgid;
		Glib::ustring strsocketuid;
		Glib::ustring strsocketgid;
		Glib::ustring strsocketmode;
#endif
		Glib::OptionGroup options("cfmuvalidateserver", "CFMU Flight Plan Validation Server");
		// version
		Glib::OptionEntry entry_version;
		entry_version.set_long_name(_("version"));
		entry_version.set_description(_("Show version and exit."));
		options.add_entry(entry_version, version);
		// fpl
                Glib::OptionEntry entry_fpl;
                entry_fpl.set_long_name(_("fpl"));
                entry_fpl.set_description(_("Query flight plan given as argument and return."));
                options.add_entry(entry_fpl, fpl);
 		// verbose      
		Glib::OptionEntry entry_verbose;
		entry_verbose.set_long_name(_("verbose"));
		entry_verbose.set_description(_("Set Verbosity."));
		options.add_entry(entry_verbose, verbose);
		// maindir
                Glib::OptionEntry entry_maindir;
                entry_maindir.set_short_name('m');
                entry_maindir.set_long_name("maindir");
                entry_maindir.set_description("Directory containing the main database");
                entry_maindir.set_arg_description("DIR");
                options.add_entry_filename(entry_maindir, m_dir_main);
		// auxdir
                Glib::OptionEntry entry_auxdir;
                entry_auxdir.set_short_name('a');
                entry_auxdir.set_long_name("auxdir");
                entry_auxdir.set_description("Directory containing the auxilliary database");
                entry_auxdir.set_arg_description("DIR");
                options.add_entry_filename(entry_auxdir, m_dir_aux);
		// disableaux
                Glib::OptionEntry entry_disableaux;
                entry_disableaux.set_short_name('A');
                entry_disableaux.set_long_name("disableaux");
                entry_disableaux.set_description("Disable the auxilliary database");
                entry_disableaux.set_arg_description("BOOL");
                options.add_entry(entry_disableaux, dis_aux);
		// local
                Glib::OptionEntry entry_local;
                entry_local.set_short_name('l');
                entry_local.set_long_name("local");
                entry_local.set_description("Local Checker (default CFMU Checks)");
                entry_local.set_arg_description("BOOL");
                options.add_entry(entry_local, local);
		// EuroFPL
                Glib::OptionEntry entry_eurofpl;
                entry_eurofpl.set_short_name('e');
                entry_eurofpl.set_long_name("eurofpl");
                entry_eurofpl.set_description("EuroFPL Checker (default CFMU Checks)");
                entry_eurofpl.set_arg_description("BOOL");
                options.add_entry(entry_eurofpl, eurofpl);
		// B2B Ops
                Glib::OptionEntry entry_b2bop;
		entry_b2bop.set_long_name("b2bop");
                entry_b2bop.set_description("CFMU B2B Ops Checker (default CFMU Checks)");
                entry_b2bop.set_arg_description("BOOL");
                options.add_entry(entry_b2bop, b2bop);
		// B2B Preop
                Glib::OptionEntry entry_b2bpreop;
		entry_b2bpreop.set_long_name("b2bpreop");
                entry_b2bpreop.set_description("CFMU B2B Preop Checker (default CFMU Checks)");
                entry_b2bpreop.set_arg_description("BOOL");
                options.add_entry(entry_b2bpreop, b2bpreop);
#ifdef HAVE_X11_XLIB_H
 		// xdisplay      
		Glib::OptionEntry entry_xdisplay;
		entry_xdisplay.set_long_name(_("xdisplay"));
		entry_xdisplay.set_description(_("Set X Display Server to be spawned."));
                entry_xdisplay.set_arg_description("INT");
		options.add_entry(entry_xdisplay, m_xdisplay);
#endif
		Glib::OptionEntry entry_socketserver;
		entry_socketserver.set_long_name("socketserver");
                entry_socketserver.set_description("Operate as socket server");
		entry_socketserver.set_arg_description("SOCKADDR");
		entry_socketserver.set_flags(entry_socketserver.get_flags() | Glib::OptionEntry::FLAG_OPTIONAL_ARG);
		options.add_entry(entry_socketserver, sigc::mem_fun(&sockaddr, &OptionalArgFilename::on_option_arg_filename));

#ifdef G_OS_UNIX
		Glib::OptionEntry entry_uid;
		entry_uid.set_long_name(_("uid"));
		entry_uid.set_description(_("Set Daemon User ID."));
                entry_uid.set_arg_description("UID");
		options.add_entry(entry_uid, strsockservuid);

		Glib::OptionEntry entry_gid;
		entry_gid.set_long_name(_("gid"));
		entry_gid.set_description(_("Set Daemon Group ID."));
                entry_gid.set_arg_description("GID");
		options.add_entry(entry_gid, strsockservgid);

		Glib::OptionEntry entry_sockuid;
		entry_sockuid.set_long_name(_("sockuid"));
		entry_sockuid.set_description(_("Set Daemon User ID."));
                entry_sockuid.set_arg_description("UID");
		options.add_entry(entry_sockuid, strsocketuid);

		Glib::OptionEntry entry_sockgid;
		entry_sockgid.set_long_name(_("sockgid"));
		entry_sockgid.set_description(_("Set Daemon Group ID."));
                entry_sockgid.set_arg_description("GID");
		options.add_entry(entry_sockgid, strsocketgid);

		Glib::OptionEntry entry_sockmode;
		entry_sockmode.set_long_name(_("sockmode"));
		entry_sockmode.set_description(_("Set Daemon Group ID."));
                entry_sockmode.set_arg_description("MODE");
		options.add_entry(entry_sockmode, strsocketmode);
#endif

		Glib::OptionEntry entry_noterminate;
		entry_noterminate.set_long_name(_("noterminate"));
		entry_noterminate.set_description(_("Do not terminate when socket activated."));
                entry_noterminate.set_arg_description("BOOL");
		options.add_entry(entry_noterminate, noterminate);

		// option context
		Glib::OptionContext context(_("-- CFMU Flight Plan Validation Server"));
		//context.add_group(options);
		context.set_main_group(options);
		context.set_help_enabled(true);
		context.set_ignore_unknown_options(true);
		Glib::set_prgname(Glib::path_get_basename(*argv));
		if (!context.parse(argc, argv)) {
			std::cerr << "Command Line Option Error: Usage: " << Glib::get_prgname() << std::endl
				  << context.get_help(false);
			exit(1);
		}
		m_auxdbmode = Engine::auxdb_prefs;
		if (dis_aux)
                        m_auxdbmode = Engine::auxdb_none;
                else if (!m_dir_aux.empty())
                        m_auxdbmode = Engine::auxdb_override;
#ifdef G_OS_UNIX
		if (!strsockservuid.empty()) {
			char *cp;
			uid_t uid1(strtoul(strsockservuid.c_str(), &cp, 0));
			if (*cp) {
				struct passwd *pw = getpwnam(strsockservuid.c_str());
				if (pw) {
					sockservuid = pw->pw_uid;
					sockservgid = pw->pw_gid;
				} else {
					std::cerr << "User " << strsockservuid << " not found" << std::endl;
				}
			} else {
				sockservuid = uid1;
			}
		}
		if (!strsockservgid.empty()) {
			char *cp;
			uid_t gid1(strtoul(strsockservgid.c_str(), &cp, 0));
			if (*cp) {
				struct group *gr = getgrnam(strsockservgid.c_str());
				if (gr) {
					sockservgid = gr->gr_gid;
				} else {
					std::cerr << "Group " << strsockservgid << " not found" << std::endl;
				}
			} else {
				sockservgid = gid1;
			}
		}
		if (!strsocketuid.empty()) {
			char *cp;
			uid_t uid1(strtoul(strsocketuid.c_str(), &cp, 0));
			if (*cp) {
				struct passwd *pw = getpwnam(strsocketuid.c_str());
				if (pw) {
					socketuid = pw->pw_uid;
					socketgid = pw->pw_gid;
				} else {
					std::cerr << "User " << strsocketuid << " not found" << std::endl;
				}
			} else {
				socketuid = uid1;
			}
		}
		if (!strsocketgid.empty()) {
			char *cp;
			uid_t gid1(strtoul(strsocketgid.c_str(), &cp, 0));
			if (*cp) {
				struct group *gr = getgrnam(strsocketgid.c_str());
				if (gr) {
					socketgid = gr->gr_gid;
				} else {
					std::cerr << "Group " << strsocketgid << " not found" << std::endl;
				}
			} else {
				socketgid = gid1;
			}
		}
		if (!strsocketmode.empty()) {
			socketmode = strtoul(strsocketmode.c_str(), 0, 0);
		}
#endif
 	}
        if (version) {
                std::cout << Glib::get_prgname() << (" Version " VERSION) << std::endl;
		exit(0);
	}
#ifdef HAVE_X11_XLIB_H
	if (m_xdisplay >= 0) {
		std::ostringstream disp;
		disp << ':' << m_xdisplay;
		Glib::setenv("DISPLAY", disp.str());
	}
#endif
	m_verbose = std::max(verbose, 0);
        m_mainloop = Glib::MainLoop::create(Glib::MainContext::get_default());
	if (sockaddr.is_set()) {
		m_quit = m_quit && !noterminate;
		std::string path(sockaddr.get_arg());
#ifdef G_OS_UNIX
		if (path.empty())
			path = PACKAGE_RUN_DIR "/validator/socket";
	        listen(path, socketuid, socketgid, socketmode);
		if (sockservgid && setgid(sockservgid))
			std::cerr << "Cannot set gid to " << sockservgid << ": "
				  << strerror(errno) << " (" << errno << ')' << std::endl;
		if (sockservuid && setuid(sockservuid))
			std::cerr << "Cannot set uid to " << sockservuid << ": "
				  << strerror(errno) << " (" << errno << ')' << std::endl;
		/* Cancel certain signals */
		signal(SIGTSTP, SIG_IGN); /* Various TTY signals */
		signal(SIGTTOU, SIG_IGN);
		signal(SIGTTIN, SIG_IGN);
		signal(SIGHUP,  SIG_IGN); /* Ignore hangup signal */
		signal(SIGTERM, signalterminate); /* terminate mainloop on SIGTERM */
		/* Create a new SID for the child process */
		if (getpgrp() != getpid() && setsid() == (pid_t)-1)
			std::cerr << "Cannot create new session: "
				  << strerror(errno) << " (" << errno << ')' << std::endl;
		/* Change the file mode mask */
		umask(0);
		/* Change the current working directory.  This prevents the current
		   directory from being locked; hence not being able to remove it. */
		if (chdir("/tmp") < 0)
			std::cerr << "Cannot change directory to /tmp: "
				  << strerror(errno) << " (" << errno << ')' << std::endl;
#else
		if (path.empty())
			path = "53181";
		listen(path);
#endif
		std::cout << "Starting Validation service on " << path << std::endl;
		return;
	}
 	if (fpl.empty()) {
		StdioClient::ptr_t p(new StdioClient(this));
		if (local)
			p->set_validator(Client::validator_local);
		else if (eurofpl)
			p->set_validator(Client::validator_eurofpl);
		else if (b2bop)
			p->set_validator(Client::validator_cfmub2bop);
		else if (b2bpreop)
			p->set_validator(Client::validator_cfmub2bpreop);
		else
			p->set_validator(Client::validator_cfmu);
		m_clients.insert(p);
	} else {
		SingleFplClient::ptr_t p(new SingleFplClient(this));
		if (local)
			p->set_validator(Client::validator_local);
		else if (eurofpl)
			p->set_validator(Client::validator_eurofpl);
		else if (b2bop)
			p->set_validator(Client::validator_cfmub2bop);
		else if (b2bpreop)
			p->set_validator(Client::validator_cfmub2bpreop);
		else
			p->set_validator(Client::validator_cfmu);
		m_clients.insert(p);
		p->set_closeatend(true);
		p->set_fplan(fpl);
	}
	if (!local && !eurofpl && !b2bop && !b2bpreop)
		start_cfmu();
}

CFMUValidateServer::~CFMUValidateServer()
{
	if (m_listenservice)
		m_listenservice->stop();
	m_clients.clear();
	m_cfmuclients.clear();
	destroy_webview();
	m_listenservice.reset();
	m_listensock.reset();
	remove_listensockaddr();
#ifdef HAVE_X11_XLIB_H
	stop_x11();
#endif
#ifdef HAVE_CURL
	curl_removeio();
	for (curlhandlemap_t::iterator i(m_curlhandlemap.begin()), e(m_curlhandlemap.end()); i != e; ) {
		curlhandlemap_t::iterator i2(i);
		++i;
		if (false)
			std::cerr << "curl handle " << (long long)i2->first << std::endl;
		if (i2->second)
			i2->second->curl_done(CURLE_OPERATION_TIMEDOUT);
	}
	if (m_curl) {
		while (!m_curlhandlemap.empty()) {
			CURL *h(m_curlhandlemap.begin()->first);
			m_curlhandlemap.erase(m_curlhandlemap.begin());
			curl_multi_remove_handle(m_curl, h);
		}
		curl_multi_cleanup(m_curl);
	}
#endif
}

void CFMUValidateServer::run(void)
{
        m_mainloop->run();
}

void CFMUValidateServer::add_portal_sighandlers(void)
{
	remove_portal_sighandlers();
	m_portal_webview_close_sighandler = g_signal_connect(m_portal_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_portal_webview_create_sighandler = g_signal_connect(m_portal_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_portal_webview_ready_sighandler = g_signal_connect(m_portal_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_portal_webview_loadstatus_sighandler = g_signal_connect(m_portal_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_portal_webview_navpolicy_sighandler = g_signal_connect(m_portal_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_newwndpolicy_sighandler = g_signal_connect(m_portal_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_rsrcload_sighandler = g_signal_connect(m_portal_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_portal_webview_consolemsg_sighandler = g_signal_connect(m_portal_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_portal_sighandlers(void)
{
        if (m_portal_webview_close_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_close_sighandler);
		m_portal_webview_close_sighandler = 0;
        }
        if (m_portal_webview_create_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_create_sighandler);
		m_portal_webview_create_sighandler = 0;
        }
	if (m_portal_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_portal_webview, m_portal_webview_ready_sighandler);
		m_portal_webview_ready_sighandler = 0;
	}
        if (m_portal_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_loadstatus_sighandler);
		m_portal_webview_loadstatus_sighandler = 0;
        }
        if (m_portal_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_navpolicy_sighandler);
		m_portal_webview_navpolicy_sighandler = 0;
        }
        if (m_portal_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_newwndpolicy_sighandler);
		m_portal_webview_newwndpolicy_sighandler = 0;
        }
        if (m_portal_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_rsrcload_sighandler);
		m_portal_webview_rsrcload_sighandler = 0;
        }
        if (m_portal_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_consolemsg_sighandler);
		m_portal_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::add_validate_sighandlers(void)
{
	remove_validate_sighandlers();
	m_validate_webview_close_sighandler = g_signal_connect(m_validate_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_validate_webview_create_sighandler = g_signal_connect(m_validate_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_validate_webview_ready_sighandler = g_signal_connect(m_validate_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_validate_webview_loadstatus_sighandler = g_signal_connect(m_validate_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_validate_webview_navpolicy_sighandler = g_signal_connect(m_validate_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_newwndpolicy_sighandler = g_signal_connect(m_validate_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_rsrcload_sighandler = g_signal_connect(m_validate_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_validate_webview_consolemsg_sighandler = g_signal_connect(m_validate_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_validate_sighandlers(void)
{
        if (m_validate_webview_close_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_close_sighandler);
		m_validate_webview_close_sighandler = 0;
        }
        if (m_validate_webview_create_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_create_sighandler);
		m_validate_webview_create_sighandler = 0;
        }
	if (m_validate_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_validate_webview, m_validate_webview_ready_sighandler);
		m_validate_webview_ready_sighandler = 0;
	}
        if (m_validate_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_loadstatus_sighandler);
		m_validate_webview_loadstatus_sighandler = 0;
        }
        if (m_validate_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_navpolicy_sighandler);
		m_validate_webview_navpolicy_sighandler = 0;
        }
        if (m_validate_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_newwndpolicy_sighandler);
		m_validate_webview_newwndpolicy_sighandler = 0;
        }
        if (m_validate_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_rsrcload_sighandler);
		m_validate_webview_rsrcload_sighandler = 0;
        }
        if (m_validate_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_consolemsg_sighandler);
		m_validate_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::prepare_webview(void)
{
	remove_portal_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	// initialize Gtk
	{
		std::string prgname(Glib::get_prgname());
		char *argv[3] = { 0, 0, 0 };
		int argc(1);
		argv[0] = const_cast<char *>(prgname.c_str());
#ifdef HAVE_X11_XLIB_H
		std::ostringstream disp;
		disp << "--display=:" << m_xdisplay;
		if (m_xdisplay >= 0)
			argv[argc++] = const_cast<char *>(disp.str().c_str());
#endif
		char **argvp(argv);
		if (false) {
			std::cerr << "gtk args: " << argv[0];
			if (argv[1])
				std::cerr << ' ' << argv[1];
			std::cerr << std::endl;
		}
		gtk_init(&argc, &argvp);
	}
	// add cookie jar
	std::string cookiefile(Glib::build_filename(FPlan::get_userdbpath(), "cookies.txt"));
	SoupSession *session(webkit_get_default_session());
	SoupCookieJar *cookiejar(soup_cookie_jar_text_new(cookiefile.c_str(), false));
	soup_session_add_feature(session, SOUP_SESSION_FEATURE(cookiejar));
	SoupCookie *bwarn(soup_cookie_new("eurocontrol_nm_portal_hideBroswerWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1));
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	bwarn = soup_cookie_new("eurocontrol_nm_portal_hideCookieWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1);
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	g_object_unref(cookiejar);
	// open webview
	m_portal_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_portal_webview);
	add_portal_sighandlers();
	WebKitWebView *webview = WEBKIT_WEB_VIEW(m_portal_webview);
	webkit_web_view_load_uri(webview, "https://www.public.cfmu.eurocontrol.int/PUBPORTAL/gateway/spec/index.html");
	set_wvstate(wvstate_loadmainpage);
}

void CFMUValidateServer::destroy_webview(void)
{
	remove_portal_sighandlers();
	remove_validate_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	m_portal_webview = 0;
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
	m_validate_webview = 0;
}

void CFMUValidateServer::submit_webview(void)
{
	if (m_wvstate != wvstate_ready && m_wvstate != wvstate_failed)
		return;
	while (!m_cfmuclients.empty()) {
		const Client::ptr_t& p(*m_cfmuclients.begin());
		if (!p) {
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		const std::string& fplan(p->get_fplan());
		if (fplan.empty()) {
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		if (m_wvstate == wvstate_failed) {
			Results res;
			res.push_back("FAIL: CFMU Webpage");
			p->set_result(res);
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		if (!m_validate_webview) {
			Results res;
			res.push_back("FAIL: Internal Error (webview)");
			p->set_result(res);
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		if (!document) {
			Results res;
			res.push_back("FAIL: Internal Error (document)");
			p->set_result(res);
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"));
		if (!element || !WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element)) {
			Results res;
			res.push_back("FAIL: Internal Error (cannot find FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD)");
			p->set_result(res);
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		webkit_dom_html_text_area_element_set_value(WEBKIT_DOM_HTML_TEXT_AREA_ELEMENT(element), fplan.c_str());
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL");
		if (!element) {
			Results res;
			res.push_back("FAIL: Internal Error (cannot find FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL)");
			p->set_result(res);
			m_cfmuclients.erase(m_cfmuclients.begin());
			continue;
		}
		zap_errors();
		webkit_web_view_execute_script(webview, "document.getElementById(\"FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL\").click();");
		set_wvstate(wvstate_query);
		break;
	}
}

void CFMUValidateServer::start_cfmu(void)
{
	if (m_wvstate != wvstate_idle)
		return;
#ifdef HAVE_X11_XLIB_H
	if (m_xdisplay >= 0) {
		m_x11srvrun = false;
		start_x11();
		if (!m_x11srvrun) {
			set_wvstate(wvstate_failed);
			return;
		}
		set_wvstate(wvstate_startxserver);
		m_statecnt = 0;
		m_stateconn.disconnect();
		m_stateconn = Glib::signal_timeout().connect_seconds(sigc::mem_fun(*this, &CFMUValidateServer::poll_xserver), 5);
		return;
	}
#endif
	prepare_webview();
}

void CFMUValidateServer::set_wvstate(wvstate_t newstate)
{
	bool chg(m_wvstate != newstate);
	if (chg && m_verbose >= 1U)
		std::cerr << "state change " << (unsigned int)m_wvstate << " -> " << (unsigned int)newstate << std::endl;
	m_wvstate = newstate;
	switch (m_wvstate) {
	case wvstate_idle:
		if (chg) {
			destroy_webview();
#ifdef HAVE_X11_XLIB_H
			//stop_x11();
#endif
			if (m_quit && m_listenservice && m_clients.empty())
				m_mainloop->quit();
		}
		if (!m_cfmuclients.empty())
			start_cfmu();
		break;

	case wvstate_loadmainpage:
		if (chg) {
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_failed), false), 120);
		}
		break;
	
	case wvstate_loadquerypage:
		if (chg) {
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_failed), false), 120);
		}
		break;

	case wvstate_ready:
		if (chg) {
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_idle), false), 9 * 60);
		}
	        submit_webview();
		break;

	case wvstate_query:
		if (chg) {
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_failed), false), 30);
		}
		break;

	case wvstate_waitresult:
		if (chg) {
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_failed), false), 30);
		}
		break;

	case wvstate_failed:
		if (chg) {
			destroy_webview();
#ifdef HAVE_X11_XLIB_H
			//stop_x11();
#endif
			m_stateconn.disconnect();
			m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_idle), false), 30);
			if (m_quit && m_listenservice && m_clients.empty())
				m_mainloop->quit();
		}
	        submit_webview();
		break;

	default:
		break;
	}
}

void CFMUValidateServer::parse_results(void)
{
	if (!m_validate_webview)
		return;
	typedef std::pair<guint,bool> lvlvisible_t;
	typedef std::pair<WebKitDOMNode *,lvlvisible_t> nodelvlvisible_t;
	std::stack<nodelvlvisible_t> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(std::make_pair(WEBKIT_DOM_NODE(element), std::make_pair(0U, true)));
	}
	Results results;
	while (!subtree.empty()) {
		WebKitDOMNode *node;
		guint level;
		bool visible;
		{
			nodelvlvisible_t nl(subtree.top());
			node = nl.first;
			level = nl.second.first;
			visible = nl.second.second;
		}
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					visible = !!strcmp(disp, "none");
					g_free(disp);
				}
			}
		}
		if (visible) {
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(std::make_pair(ptr, std::make_pair(level + 1U, visible)));
			}
		}
		if (!WEBKIT_DOM_IS_HTML_ELEMENT(node))
			continue;
		char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
		if (m_verbose > 2U) {
			std::cerr << std::string(2U * level, ' ') << "element:";
			gchar *lname(webkit_dom_node_get_local_name(WEBKIT_DOM_NODE(node)));
			if (lname) {
				if (*lname)
					std::cerr << " name \"" << lname << "\"";
				g_free(lname);
			}
			lname = webkit_dom_html_element_get_id(WEBKIT_DOM_HTML_ELEMENT(node));
			if (lname) {
				if (*lname)
					std::cerr << " id \"" << lname << "\"";
				g_free(lname);
			}
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(node)));
			if (typ && *typ)
				std::cerr << " type \"" << typ << "\"";
			if (class_name && *class_name)
				std::cerr << " class \"" << class_name << "\"";
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					if (*disp)
						std::cerr << " display \"" << disp << "\"" << std::endl;
					g_free(disp);
				}
			}
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt)
					std::cerr << " text \"" << txt << "\"";
				g_free(txt);
			}
			std::cerr << std::endl;
		}
		if (!class_name)
			continue;
		if (!visible) {
			g_free(class_name);
			continue;
		}
		if (!strcmp(class_name, "portal_dataValue") && WEBKIT_DOM_IS_HTML_DIV_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (false && !strcmp(class_name, "portal_decotableCell ") && WEBKIT_DOM_IS_HTML_TABLE_CELL_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (true && (!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
		    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		g_free(class_name);
	}
	if (results.empty())
		return;
	if (!m_cfmuclients.empty()) {
		const Client::ptr_t& p(*m_cfmuclients.begin());
		if (p) {
			p->set_result(results);
		}
		m_cfmuclients.erase(m_cfmuclients.begin());
	}
	set_wvstate(wvstate_ready);
}

void CFMUValidateServer::zap_errors(void)
{
	if (!m_validate_webview)
		return;
	std::stack<WebKitDOMNode *> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(WEBKIT_DOM_NODE(element));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node(subtree.top());
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (class_name) {
		     		if ((!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
				    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
					g_free(class_name);
					while (webkit_dom_node_has_child_nodes(node)) {
						GError *error(0);
						webkit_dom_node_remove_child(node, webkit_dom_node_get_first_child(node), &error);
						if (error) {
							std::cerr << "Cannot remove child: " << error->message << std::endl;
							g_error_free(error);
						}
					}
					continue;
				}
				g_free(class_name);
			}
		}
		{
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(ptr);
			}
		}
	}
}

gboolean CFMUValidateServer::close_webview_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->close_webview(webview);
}

gboolean CFMUValidateServer::close_webview(WebKitWebView* webview)
{
        set_wvstate(wvstate_failed);
	return true;
}

WebKitWebView *CFMUValidateServer::create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->create_webview(webview, frame);
}

WebKitWebView *CFMUValidateServer::create_webview(WebKitWebView *webview, WebKitWebFrame *frame)
{
	if (m_verbose > 1U)
		std::cerr << "new webview!" << std::endl;
	if (webview == WEBKIT_WEB_VIEW(m_validate_webview))
		return 0;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview));
	if (m_wvstate != wvstate_loadmainpage)
		return 0;
	remove_validate_sighandlers();
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
	m_validate_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_validate_webview);
	webkit_web_view_set_settings(WEBKIT_WEB_VIEW(m_validate_webview), webkit_web_view_get_settings(WEBKIT_WEB_VIEW(m_portal_webview)));
	add_validate_sighandlers();
	set_wvstate(wvstate_loadquerypage);
	WebKitWebView *wv(WEBKIT_WEB_VIEW(m_validate_webview));
	return wv;
}

gboolean CFMUValidateServer::webview_ready_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->webview_ready(webview);
}

gboolean CFMUValidateServer::webview_ready(WebKitWebView* webview)
{
	if (m_verbose > 1U)
		std::cerr << "WebView: ready" << std::endl;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	gtk_widget_show_all(GTK_WIDGET(webview));
	return false;
}

void CFMUValidateServer::loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->loadstatus_webview(webview, pspec);
}

void CFMUValidateServer::loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec)
{
        WebKitLoadStatus status(webkit_web_view_get_load_status(webview));
        GObject *object(G_OBJECT(webview));
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
        g_object_freeze_notify(object);
	switch (status) {
        case WEBKIT_LOAD_FINISHED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load finished" << std::endl;
		gtk_widget_show_all(GTK_WIDGET(webview));
		break;

        case WEBKIT_LOAD_PROVISIONAL:
		if (m_verbose > 1U)
			std::cerr << "WebView: load provisional" << std::endl;
                break;

        case WEBKIT_LOAD_COMMITTED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load committed" << std::endl;
                break;

        case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
		if (m_verbose > 1U)
			std::cerr << "WebView: load first visually nonempty layout" << std::endl;
                break;

        case WEBKIT_LOAD_FAILED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load failed" << std::endl;
		set_wvstate(wvstate_failed);
                break;

        default:
                break;
        }
        g_object_thaw_notify(object);
}

gboolean CFMUValidateServer::policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							      WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->policy_decision_required(webview, webframe, request, action, decision);
}

gboolean CFMUValidateServer::policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							    WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	WebKitWebNavigationReason reason(webkit_web_navigation_action_get_reason(action));
        gint button(webkit_web_navigation_action_get_button(action));
        gint state(webkit_web_navigation_action_get_modifier_state(action));
        const char *uri(webkit_network_request_get_uri(request));
	if (m_verbose > 1U)
		std::cerr << "policy_decision_required: reason " << reason << " button " << button << " state " << state
			  << " uri " << uri << std::endl;
	return false;
}

void CFMUValidateServer::resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->resourceload_finished(webview, webframe, webresource);
}

void CFMUValidateServer::resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	if (m_verbose > 1U)
		std::cerr << "resource load finished! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_loadmainpage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_portal_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL"));
		if (!element)
			break;
		if (m_verbose > 2U)
			std::cerr << "Button found! Injecting script" << std::endl;
		webkit_web_view_execute_script(webview, "document.getElementById(\"IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL\").click();");
		set_wvstate(wvstate_loadquerypage);
		break;
	}

	case wvstate_loadquerypage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL"));
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Validate element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "FPL entry element type " << typ << std::endl;
		}
		if (!WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Result Errors element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA");
		if (!element)
			break;
		webkit_dom_event_target_add_event_listener(WEBKIT_DOM_EVENT_TARGET(element), "DOMSubtreeModified", G_CALLBACK(&result_change_1), true, this);
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Results element type " << typ << std::endl;
		}
		m_stateconn.disconnect();
		m_stateconn = Glib::signal_timeout().connect_seconds(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_ready), false), 3);
		break;
	}

	case wvstate_query:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		set_wvstate(wvstate_waitresult);
		parse_results();
		break;
	}

	default:
		break;
	}
}

gboolean CFMUValidateServer::result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->result_change(target, e);
}

gboolean CFMUValidateServer::result_change(WebKitDOMElement *target, WebKitDOMEvent *e)
{
	std::cerr << "result change! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_waitresult:
		parse_results();
		break;

	default:
		break;
	}
	return true;
}

gboolean CFMUValidateServer::console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->console_message(webview, message, line, source_id);
}

gboolean CFMUValidateServer::console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id)
{
	if (m_verbose > 0U) {
		std::cerr << message << " (" << (source_id ? "*null*" : source_id) << ':' << line << ')' << std::endl;
	}
	return true;
}

#ifdef HAVE_X11_XLIB_H
void CFMUValidateServer::start_x11(void)
{
	if (m_x11srvrun)
		return;
	if (m_xdisplay < 0)
		return;
	try {
		std::vector<std::string> argv, env;
#ifdef __MINGW32__
		std::string workdir;
		{
			gchar *instdir(g_win32_get_package_installation_directory_of_module(0));
			if (instdir) {
				workdir = instdir;
				argv.push_back(Glib::build_filename(instdir, "bin", "Xvfb.exe"));
			} else {
				argv.push_back(Glib::build_filename(PACKAGE_PREFIX_DIR "/bin", "Xvfb.exe"));
				workdir = PACKAGE_PREFIX_DIR;
			}
		}
#else
		std::string workdir("/tmp");
		argv.push_back("/usr/bin/Xvfb");
		env.push_back("PATH=/bin:/usr/bin");
		{
			std::ostringstream disp;
			disp << ':' << m_xdisplay;
			argv.push_back(disp.str());
		}
		argv.push_back("-screen");
		argv.push_back("0");
		argv.push_back("100x100x16");
#endif
		if (false) {
			std::cerr << "Spawn: workdir " << workdir << " argv:" << std::endl;
			for (std::vector<std::string>::const_iterator i(argv.begin()), e(argv.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
			std::cerr << "env:" << std::endl;
			for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
				std::cerr << "  " << *i << std::endl;
		}
#ifdef __MINGW32__
		// inherit environment; otherwise, this may fail on WindowsXP with STATUS_SXS_ASSEMBLY_NOT_FOUND (C0150004)
		Glib::spawn_async_with_pipes(workdir, argv,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_x11srvpid, 0, 0, 0);
#else
		Glib::spawn_async_with_pipes(workdir, argv, env,
					     Glib::SPAWN_DO_NOT_REAP_CHILD | Glib::SPAWN_STDERR_TO_DEV_NULL,
					     sigc::slot<void>(), &m_x11srvpid, 0, 0, 0);
#endif
		m_x11srvrun = true;
	} catch (Glib::SpawnError& e) {
		m_x11srvrun = false;
		std::cerr << "Cannot run X Server: " + e.what() << std::endl;
		return;
	}
	m_x11srvconnwatch.disconnect();
	m_x11srvconnwatch = Glib::signal_child_watch().connect(sigc::mem_fun(this, &CFMUValidateServer::x11srv_child_watch), m_x11srvpid);
}

void CFMUValidateServer::x11srv_child_watch(GPid pid, int child_status)
{
	if (!m_x11srvrun)
		return;
	if (m_x11srvpid != pid) {
		Glib::spawn_close_pid(pid);
		return;
	}
	Glib::spawn_close_pid(pid);
	m_x11srvrun = false;
	m_x11srvconnwatch.disconnect();
}

void CFMUValidateServer::stop_x11(void)
{
	if (!m_x11srvrun)
		return;
	kill(m_x11srvpid, SIGHUP);
}

bool CFMUValidateServer::x11_fail;

int CFMUValidateServer::x11_error(Display *dpy, XErrorEvent *err)
{
	if (true) {
		char msg[80];
		XGetErrorText(dpy, err->error_code, msg, sizeof(msg));
		std::cerr << "X11 Error " << err->error_code << " (" <<  msg << "): request "
			  << err->request_code << '.' << err->minor_code << std::endl;
	}
	x11_fail = true;
	return 0;
}

int CFMUValidateServer::x11_ioerror(Display *dpy)
{
	if (true) {
		std::cerr << "X11 I/O Error" << std::endl;
	}
	x11_fail = true;
	return 0;
}

bool CFMUValidateServer::poll_xserver(void)
{
	{
		x11_fail = false;
		std::ostringstream dpyn;
		dpyn << ':' << m_xdisplay;
		Display *dpy(XOpenDisplay(dpyn.str().c_str()));
		if (!dpy) {
			x11_fail = true;
		} else {
			int (*old_error_handler)(Display *, XErrorEvent *);
			int (*old_ioerror_handler)(Display *);
			old_error_handler = XSetErrorHandler(&CFMUValidateServer::x11_error);
			old_ioerror_handler = XSetIOErrorHandler(&CFMUValidateServer::x11_ioerror);
			int screen(DefaultScreen(dpy));
			Window win(XCreateSimpleWindow(dpy, RootWindow(dpy, screen), 0, 0, 10, 10, 5, BlackPixel(dpy, screen), WhitePixel(dpy, screen)));
			if (!win) {
				x11_fail = true;
			} else {
				XMapWindow(dpy, win);
				XFlush(dpy);
				XDestroyWindow(dpy, win);
			}
			XSetErrorHandler(old_error_handler);
			XSetIOErrorHandler(old_ioerror_handler);
			XCloseDisplay(dpy);
		}
		if (x11_fail) {
			++m_statecnt;
			if (m_statecnt < 60)
				return true;
			set_wvstate(wvstate_failed);
			return false;
		}
	}
	m_stateconn.disconnect();
	prepare_webview();
	return false;
}
#endif

#ifdef HAVE_CURL
void CFMUValidateServer::curl_add(CURL *handle, const Client::ptr_t& p)
{
	if (!handle || !m_curl)
		return;
	if (m_curlhandlemap.find(handle) != m_curlhandlemap.end())
		return;
	m_curlhandlemap[handle] = p;
	curl_multi_add_handle(m_curl, handle);
	if (false)
		std::cerr << "curl_add handle " << (long long)handle << " map size " << m_curlhandlemap.size() << std::endl;
	curl_io();
}

void CFMUValidateServer::curl_remove(CURL *handle)
{
	if (!handle || !m_curl)
		return;
	curlhandlemap_t::iterator i(m_curlhandlemap.find(handle));
	if (i == m_curlhandlemap.end())
		return;
	m_curlhandlemap.erase(i);
	curl_multi_remove_handle(m_curl, handle);
	if (false)
		std::cerr << "curl_remove handle " << (long long)handle << " map size " << m_curlhandlemap.size() << std::endl;
	curl_io();
}

void CFMUValidateServer::curl_removeio(void)
{
	while (!m_curlio.empty()) {
		m_curlio.begin()->disconnect();
		m_curlio.erase(m_curlio.begin());
	}
}

void CFMUValidateServer::curl_io(void)
{
	curl_removeio();
	if (!m_curl)
		return;
	int running_handles(0);
	while (curl_multi_perform(m_curl, &running_handles) == CURLM_CALL_MULTI_PERFORM);
	int msgs_in_queue;
	while (CURLMsg *msg = curl_multi_info_read(m_curl, &msgs_in_queue)) {
		if (msg->msg != CURLMSG_DONE)
			continue;
		curlhandlemap_t::iterator i(m_curlhandlemap.find(msg->easy_handle));
		if (i == m_curlhandlemap.end())
			continue;
		if (i->second)
			i->second->curl_done(msg->data.result);
	}
	fd_set readset, writeset, exceptset;
	int maxfd(0);
	FD_ZERO(&readset);
	FD_ZERO(&writeset);
	FD_ZERO(&exceptset);
	if (CURLM_OK == curl_multi_fdset(m_curl, &readset, &writeset, &exceptset, &maxfd)) {
		for (int fd = 0; fd <= maxfd; ++fd) {
			Glib::IOCondition cond((Glib::IOCondition)0);
			if (FD_ISSET(fd, &readset))
				cond |= Glib::IO_IN;
			if (FD_ISSET(fd, &writeset))
				cond |= Glib::IO_OUT;
			if (FD_ISSET(fd, &exceptset))
				cond |= Glib::IO_ERR;
			if (cond == (Glib::IOCondition)0)
				continue;
			m_curlio.push_back(Glib::signal_io().connect(sigc::bind_return(sigc::hide(sigc::mem_fun(*this, &CFMUValidateServer::curl_io)), false), fd, cond));
		}
	}
	long timeout;
	if (CURLM_OK == curl_multi_timeout(m_curl, &timeout)) {
		if (timeout > 0) {
			m_curlio.push_back(Glib::signal_timeout().connect(sigc::bind_return(sigc::mem_fun(*this, &CFMUValidateServer::curl_io), false), timeout));
		} else if (!timeout) {
			curl_io();
			return;
		}
	}
}
#endif

void CFMUValidateServer::annotate(Results& res)
{
	if (m_tfrloaded) {
		res.annotate(m_tfr);
		return;
	}
	Results res1;
	res.swap(res1);
	// first open the databases
	delete m_engine;
	m_engine = new Engine(m_dir_main, m_auxdbmode, m_dir_aux, false, false);
	std::string dir_aux1 = m_engine->get_aux_dir(m_auxdbmode, m_dir_aux);
	m_airportdb.open(m_dir_main);
	if (!dir_aux1.empty() && m_airportdb.is_dbfile_exists(dir_aux1))
		m_airportdb.attach(dir_aux1);
	if (m_airportdb.is_aux_attached())
		std::cerr << "Auxillary airports database attached" << std::endl;
	m_navaiddb.open(m_dir_main);
	if (!dir_aux1.empty() && m_navaiddb.is_dbfile_exists(dir_aux1))
		m_navaiddb.attach(dir_aux1);
	if (m_navaiddb.is_aux_attached())
		std::cerr << "Auxillary navaids database attached" << std::endl;
	m_waypointdb.open(m_dir_main);
	if (!dir_aux1.empty() && m_waypointdb.is_dbfile_exists(dir_aux1))
		m_waypointdb.attach(dir_aux1);
	if (m_waypointdb.is_aux_attached())
		std::cerr << "Auxillary waypoints database attached" << std::endl;
	m_airwaydb.open(m_dir_main);
	if (!dir_aux1.empty() && m_airwaydb.is_dbfile_exists(dir_aux1))
		m_airwaydb.attach(dir_aux1);
	if (m_airwaydb.is_aux_attached())
		std::cerr << "Auxillary airways database attached" << std::endl;
	m_airspacedb.open(m_dir_main);
	if (!dir_aux1.empty() && m_airspacedb.is_dbfile_exists(dir_aux1))
		m_airspacedb.attach(dir_aux1);
	if (m_airspacedb.is_aux_attached())
		std::cerr << "Auxillary airspaces database attached" << std::endl;
	Glib::TimeVal tv;
	tv.assign_current_time();
	m_tfr.start_add_rules();
	std::vector<std::string> tfrfnames;
	bool ok(false);
	{
		std::string fname(Glib::build_filename(dir_aux1, "eRAD.bin"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			ok = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (ok)
				tfrfnames.push_back(fname);
		}
	}
	if (!ok) {
		std::string fname(Glib::build_filename(dir_aux1, "eRAD.xml"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			ok = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (ok)
				tfrfnames.push_back(fname);
		}
	}
	bool okloc(false);
	{
		std::string fname(Glib::build_filename(dir_aux1, "eRAD_local.bin"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			okloc = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (okloc)
				tfrfnames.push_back(fname);
		}
	}
	if (!okloc) {
		std::string fname(Glib::build_filename(dir_aux1, "eRAD_local.xml"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			okloc = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (okloc)
				tfrfnames.push_back(fname);
		}
	}
	bool okuser(false);
	{
		std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD.bin"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			okuser = m_tfr.add_binary_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (okuser)
				tfrfnames.push_back(fname);
		}
	}
	if (!okuser) {
		std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD.xml"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			res.push_back("Loading traffic flow restrictions " + fname);
			std::vector<TrafficFlowRestrictions::Message> msg;
			okuser = m_tfr.add_rules(msg, fname, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb);
			for (unsigned int i = 0; i < msg.size(); ++i)
				res.print(msg[i]);
			if (okuser)
				tfrfnames.push_back(fname);
		}
	}
	{
		Glib::TimeVal tv1;
		tv1.assign_current_time();
		tv = tv1 - tv;
	}
	if (!ok) {
		res.push_back("Cannot load traffic flow restrictions " + Glib::build_filename(dir_aux1, "eRAD.bin") + ", " +
			       Glib::build_filename(dir_aux1, "eRAD.xml") + "\n");
		return;
	}
	m_tfr.end_add_rules();
	if (ok || okloc || okuser) {
		res.push_back("Origin:    " + m_tfr.get_origin());
		res.push_back("Created:   " + m_tfr.get_created());
		res.push_back("Effective: " + m_tfr.get_effective());
		m_tfr.reset_rules();
		{
			std::ostringstream oss;
			oss << m_tfr.count_rules() << " rules loaded traffic flow restrictions ";
			for (std::vector<std::string>::const_iterator i(tfrfnames.begin()), e(tfrfnames.end()); i != e; ) {
				oss << *i;
				++i;
				if (i == e)
					break;
				oss << ", ";
			}
			oss << ", created " << m_tfr.get_created() << ", effective " << m_tfr.get_effective() << ", "
			    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
			res.push_back(oss.str());
		}
	}
	{
		std::string fname(Glib::build_filename(dir_aux1, "eRAD_rules.xml"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			if (m_tfr.load_disabled_trace_rules_file(fname))
				res.push_back("Loaded global disabled/trace rules from " + fname);
		}
	}
	{
		std::string fname(Glib::build_filename(FPlan::get_userdbpath(), "eRAD_rules.xml"));
		if (Glib::file_test(fname, Glib::FILE_TEST_EXISTS) && !Glib::file_test(fname, Glib::FILE_TEST_IS_DIR)) {
			if (m_tfr.load_disabled_trace_rules_file(fname))
				res.push_back("Loaded user disabled/trace rules from " + fname);
		}
	}
	m_tfr.reset_rules();
	if (true) {
		Glib::TimeVal tv;
		tv.assign_current_time();
		m_tfr.simplify_rules_typeofflight('G');
		m_tfr.simplify_rules_mil(false);
		m_tfr.simplify_rules_gat(true);
		{
			Glib::TimeVal tv1;
			tv1.assign_current_time();
			tv = tv1 - tv;
		}
		std::ostringstream oss;
		oss << m_tfr.count_rules() << " rules after fixing type of flight G CIV GAT, "
		    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
		res.push_back(oss.str());
	}
	m_tfrloaded = true;
	res1.annotate(m_tfr);
	res.insert(res.end(), res1.begin(), res1.end());
}

void CFMUValidateServer::validate_locally(const Client::ptr_t& p, const rules_t& tracerules, const rules_t& disabledrules)
{
	if (!p)
		return;
	Results res1;
	annotate(res1);
	// parse flight plan
	FPlanRoute fpl(*(FPlan *)0);
	char type_of_flight('G');
	std::string acft_type;
	std::string equipment;
	Aircraft::pbn_t pbn(Aircraft::pbn_none);
	{
		IcaoFlightPlan icaofpl(*m_engine);
	        IcaoFlightPlan::errors_t err(icaofpl.parse(p->get_fplan(), false));
		if (!err.empty()) {
			std::string er("EFPL: ");
			bool subseq(false);
			for (IcaoFlightPlan::errors_t::const_iterator i(err.begin()), e(err.end()); i != e; ++i) {
				if (subseq)
					er += "; ";
				subseq = true;
				er += *i;
			}
			res1.push_back(er);
			p->set_result(res1);
			return;
		}
		icaofpl.set_route(fpl);
		type_of_flight = icaofpl.get_flighttype();
		acft_type = icaofpl.get_aircrafttype();
		equipment = icaofpl.get_equipment();
		pbn = icaofpl.get_pbn();
		if (true) {
			std::ostringstream oss;
			oss << "I: FPL Parsed, " << fpl.get_nrwpt() << " waypoints";
			res1.push_back(oss.str());
		}
	}
	// print Parsed Flight Plan
	if (false) {
		for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
			const FPlanWaypoint& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "PFPL[" << i << "]: " << wpt.get_icao() << '/' << wpt.get_name() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
			case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

			case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			res1.push_back(oss.str());
		}
	}
	m_tfr.tracerules_clear();
	for (rules_t::const_iterator i(tracerules.begin()), e(tracerules.end()); i != e; ++i)
		m_tfr.tracerules_add(*i);
	m_tfr.disabledrules_clear();
	for (rules_t::const_iterator i(disabledrules.begin()), e(disabledrules.end()); i != e; ++i)
		m_tfr.disabledrules_add(*i);
	TrafficFlowRestrictions::Result res(m_tfr.check_fplan(fpl, type_of_flight, acft_type, equipment, pbn,
							      m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb));
	// print Flight Plan
	{
		{
			std::ostringstream oss;
			oss << "FR: " << res.get_aircrafttype() << ' ' << res.get_aircrafttypeclass() << ' '
			    << res.get_equipment() << " PBN/" << res.get_pbn_string() << ' ' << res.get_typeofflight();
			res1.push_back(oss.str());
		}
		const TrafficFlowRestrictions::Fpl& fpl(res.get_fplan());
		for (unsigned int i = 0; i < fpl.size(); ++i) {
			const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
			std::ostringstream oss;
			oss << "F[" << i << '/' << wpt.get_wptnr() << "]: " << wpt.get_ident() << ' ' << wpt.get_type_string() << ' ';
			switch (wpt.get_pathcode()) {
			case FPlanWaypoint::pathcode_sid:
			case FPlanWaypoint::pathcode_star:
			case FPlanWaypoint::pathcode_airway:
				oss << wpt.get_pathname();
				break;

			case FPlanWaypoint::pathcode_directto:
				oss << "DCT";
				break;

			default:
				oss << '-';
				break;
			}
			oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
			    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
			    << ' ' << wpt.get_icao() << '/' << wpt.get_name() << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			{
				time_t t(wpt.get_time_unix());
				struct tm tm;
				if (gmtime_r(&t, &tm)) {
					oss << ' ' << std::setw(2) << std::setfill('0') << tm.tm_hour
					    << ':' << std::setw(2) << std::setfill('0') << tm.tm_min;
					if (t >= 48 * 60 * 60)
						oss << ' ' << std::setw(4) << std::setfill('0') << (tm.tm_year + 1900)
						    << '-' << std::setw(2) << std::setfill('0') << (tm.tm_mon + 1)
						    << '-' << std::setw(2) << std::setfill('0') << tm.tm_mday;
				}
			}
			res1.push_back(oss.str());
		}
	}
	res1.print(res);
	p->set_result(res1);
}

void CFMUValidateServer::validate_cfmu(const Client::ptr_t& p)
{
	if (!p)
		return;
	if (p->get_fplan().empty())
		return;
	if (m_verbose >= 1U)
		std::cerr << "Queue fpl " << p->get_fplan() << std::endl;
	m_cfmuclients.push_back(p);
	start_cfmu();
	submit_webview();
}

void CFMUValidateServer::close(const Client::ptr_t& p)
{
	m_clients.erase(p);
	if (!m_listenservice && m_clients.empty())
		m_mainloop->quit();
}

void CFMUValidateServer::logmessage(unsigned int loglevel, const Glib::ustring& msg)
{
	if (m_verbose < loglevel)
		return;
	std::cerr << msg << std::endl;
}

void CFMUValidateServer::remove_listensockaddr(void)
{
#ifdef G_OS_UNIX
	if (!m_listensockaddr)
		return;
	if (m_listensockaddr->property_address_type() == Gio::UNIX_SOCKET_ADDRESS_PATH) {
		Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(m_listensockaddr->property_path()));
		try {
			f->remove();
		} catch (...) {
		}
	}
	m_listensockaddr.reset();
#endif
}

#ifdef G_OS_UNIX
void CFMUValidateServer::listen(const std::string& path, uid_t socketuid, gid_t socketgid, mode_t socketmode)
#else
void CFMUValidateServer::listen(const std::string& path)
#endif
{
	m_listenservice.reset();
	m_listensock.reset();
	remove_listensockaddr();
#ifdef HAVE_SYSTEMDDAEMON
	{
		int n = sd_listen_fds(0);
		if (n > 1)
			throw std::runtime_error("SystemD: too many file descriptors received");
		if (n == 1) {
			int fd = SD_LISTEN_FDS_START + 0;
			m_listensock = Gio::Socket::create_from_fd(fd);
		}
	}
#endif
#ifdef G_OS_UNIX
	if (!m_listensock) {
		m_listensock = Gio::Socket::create(Gio::SOCKET_FAMILY_UNIX, Gio::SOCKET_TYPE_STREAM, Gio::SOCKET_PROTOCOL_DEFAULT);
	        m_listensockaddr = Gio::UnixSocketAddress::create(path, Gio::UNIX_SOCKET_ADDRESS_PATH);
		if (true) {
			Glib::RefPtr<Gio::File> f(Gio::File::create_for_path(path));
			try {
				f->remove();
			} catch (...) {
			}
		}
		m_listensock->bind(m_listensockaddr, true);
		m_listensock->listen();
		//if (fchown(m_listensock->get_fd(), socketuid, socketgid))
		if (chown(path.c_str(), socketuid, socketgid))
			std::cerr << "Cannot set socket " << path << " uid/gid to " << socketuid << '/' << socketgid
				  << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
		//if (fchmod(m_listensock->get_fd(), socketmode))
		if (chmod(path.c_str(), socketmode))
			std::cerr << "Cannot set socket " << path << " mode to " << std::oct << '0'
				  << (unsigned int)socketmode << std::dec << ": " << strerror(errno) << " (" << errno << ')' << std::endl;
		m_quit = false;
	}
#else
	if (!m_listensock) {
		Glib::RefPtr<Gio::InetAddress> addr(Gio::InetAddress::create_loopback(Gio::SOCKET_FAMILY_IPV4));
		Glib::RefPtr<Gio::InetSocketAddress> sockaddr(Gio::InetSocketAddress::create(addr, strtoul(path.c_str(), 0, 0)));
		m_listensock = Gio::Socket::create(Gio::SOCKET_FAMILY_IPV4, Gio::SOCKET_TYPE_STREAM, Gio::SOCKET_PROTOCOL_DEFAULT);
		m_listensock->bind(sockaddr, true);
		m_listensock->listen();
		m_quit = false;
	}
#endif
	m_listenservice = Gio::SocketService::create();
	m_listenservice->signal_incoming().connect(sigc::mem_fun(*this, &CFMUValidateServer::on_connect));
	if (!m_listenservice->add_socket(m_listensock))
		throw std::runtime_error("Cannot listen to socket");
	m_listenservice->start();
}

bool CFMUValidateServer::on_connect(const Glib::RefPtr<Gio::SocketConnection>& connection, const Glib::RefPtr<Glib::Object>& source_object)
{
	Client::ptr_t p(new SocketClient(this, connection));
	m_clients.insert(p);
	return true;
}

int main(int argc, char *argv[])
{
        try {
#if !GLIB_CHECK_VERSION(2,35,0)
		g_type_init();
#endif
                Glib::init();
		Gio::init();
                Glib::set_application_name("CFMU Validate Server");
                Glib::thread_init();
                CFMUValidateServer app(argc, argv);
		mainloop = app.get_mainloop();
                app.run();
                return 0;
        } catch(const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return 1;
        } catch(const Gtk::BuilderError& ex) {
                std::cerr << "BuilderError: " << ex.what() << std::endl;
                return 1;
        }
        return 1;
}
