#ifndef CFMUVALIDATESERVER_H
#define CFMUVALIDATESERVER_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdepsgui.h"

#include <gtkmm.h>
#include <giomm/socketservice.h>
#include <sigc++/sigc++.h>
#ifdef HAVE_WEBKITGTK
#include <webkit/webkit.h>
#endif
#ifdef HAVE_CURL
#include <curl/curl.h>
#endif
#include <libxml/tree.h>

#include "tfr.hh"

class CFMUValidateServer : public sigc::trackable {
public:
	CFMUValidateServer(int& argc, char**& argv);
	~CFMUValidateServer();
	void run(void);
        const Glib::RefPtr<Glib::MainLoop>& get_mainloop(void) const { return m_mainloop; }

protected:
	typedef std::set<Glib::ustring> rules_t;

	class Results : public std::vector<Glib::ustring> {
	public:
		void print(const TrafficFlowRestrictions::Message& m);
		void print(const TrafficFlowRestrictions::Result& res);
		void print(const TrafficFlowRestrictions::RuleResult& rule);
		void print(const rules_t& r, const Glib::ustring& pfx);
		void annotate(TrafficFlowRestrictions& m_tfr);
	};

	class OptionalArgFilename : public sigc::trackable {
	public:
		OptionalArgFilename(void) : m_value(false), m_set(false) {}
		bool on_option_arg_filename(const Glib::ustring& option_name, const std::string& value, bool has_value);
		const std::string& get_arg(void) const { return m_arg; }
		bool is_value(void) const { return m_value; }
		bool is_set(void) const { return m_set; }

	protected:
		std::string m_arg;
		bool m_value;
		bool m_set;
	};

	class OptionalArgString : public sigc::trackable {
	public:
		OptionalArgString(void) : m_value(false), m_set(false) {}
		bool on_option_arg_string(const Glib::ustring& option_name, const Glib::ustring& value, bool has_value);
		const Glib::ustring& get_arg(void) const { return m_arg; }
		bool is_value(void) const { return m_value; }
		bool is_set(void) const { return m_set; }

	protected:
		Glib::ustring m_arg;
		bool m_value;
		bool m_set;
	};

	class Client : public sigc::trackable {
	public:
		typedef Glib::RefPtr<Client> ptr_t;
		typedef Glib::RefPtr<const Client> const_ptr_t;
		Client(CFMUValidateServer *server);
		virtual ~Client();
		void reference(void) const;
		void unreference(void) const;
		ptr_t get_ptr(void) { reference(); return ptr_t(this); }
		const_ptr_t get_ptr(void) const { reference(); return const_ptr_t(this); }

		typedef enum {
			validator_cfmu,
			validator_cfmub2bop,
			validator_cfmub2bpreop,
			validator_eurofpl,
			validator_local
		} validator_t;
		validator_t get_validator(void) const { return m_validator; }
		void set_validator(validator_t v) { m_validator = v; }
		static const std::string& to_str(validator_t v);

		const std::string& get_fplan(void) const { return m_fplan; }
		virtual void set_result(const Results& res) = 0;
#ifdef HAVE_CURL
		void curl_done(CURLcode result);
#endif

	protected:
		CFMUValidateServer *m_server;
		rules_t m_tracerules;
		rules_t m_disabledrules;
		mutable gint m_refcount;
		std::string m_fplan;
#ifdef HAVE_CURL
		CURL *m_curl;
		struct curl_slist *m_curl_headers;
		std::vector<char> m_curlrxbuffer;
		std::vector<char> m_curltxbuffer;
		std::string m_cfmub2bopurl;
		std::string m_cfmub2bopcert;
		std::string m_cfmub2bpreopurl;
		std::string m_cfmub2bpreopcert;
#endif
		validator_t m_validator;
		bool m_annotate;

		void handle_input(void);
#ifdef HAVE_CURL
		void close_curl(void);
		void find_eurofpl_results(std::string& result, xmlNode *a_node, unsigned int level = 0);
		void parse_eurofpl(void);
 		void parse_cfmub2b(void);
		static std::string get_text(const xmlpp::Node *n);
		static size_t my_write_func_1(const void *ptr, size_t size, size_t nmemb, Client *clnt);
		static size_t my_read_func_1(void *ptr, size_t size, size_t nmemb, Client *clnt);
		static int my_progress_func_1(Client *clnt, double t, double d, double ultotal, double ulnow);
		size_t my_write_func(const void *ptr, size_t size, size_t nmemb);
		size_t my_read_func(void *ptr, size_t size, size_t nmemb);
		int my_progress_func(double t, double d, double ultotal, double ulnow);
#endif
	};

	class StdioClient : public Client {
	public:
		typedef Glib::RefPtr<StdioClient> ptr_t;
		typedef Glib::RefPtr<const StdioClient> const_ptr_t;
		StdioClient(CFMUValidateServer *server);
		virtual ~StdioClient();

		virtual void set_result(const Results& res);

	protected:
		char m_inbuffer[16384];
		Glib::RefPtr<Glib::IOChannel> m_inputchan;
		Glib::RefPtr<Glib::IOChannel> m_outputchan;
		sigc::connection m_inputreaderconn;
		unsigned int m_inbufptr;

		bool input_line_handler(Glib::IOCondition iocond);
		void handle_input_loop(void);
	};

	class SocketClient : public Client {
	public:
		typedef Glib::RefPtr<SocketClient> ptr_t;
		typedef Glib::RefPtr<const SocketClient> const_ptr_t;
		SocketClient(CFMUValidateServer *server, const Glib::RefPtr<Gio::SocketConnection>& connection);
		virtual ~SocketClient();

		virtual void set_result(const Results& res);

	protected:
		char m_inbuffer[16384];
		Glib::RefPtr<Gio::SocketConnection> m_connection;
		sigc::connection m_inputreaderconn;
		unsigned int m_inbufptr;

		bool input_line_handler(Glib::IOCondition iocond);
		void handle_input_loop(void);
	};

	class SingleFplClient : public Client {
	public:
		typedef Glib::RefPtr<SingleFplClient> ptr_t;
		typedef Glib::RefPtr<const SingleFplClient> const_ptr_t;
		SingleFplClient(CFMUValidateServer *server);
		virtual ~SingleFplClient();

		virtual void set_result(const Results& res);

		void set_fplan(const std::string& fpl);

		bool is_closeatend(void) const { return m_closeatend; }
		void set_closeatend(bool c = true) { m_closeatend = c; }

	protected:
		Glib::RefPtr<Glib::IOChannel> m_outputchan;
		typedef std::list<std::string> fplans_t;
		fplans_t m_fplans;
		bool m_closeatend;

		void handle_input_loop(void);
	};



	AirportsDb m_airportdb;
	NavaidsDb m_navaiddb;
	WaypointsDb m_waypointdb;
	AirwaysDb m_airwaydb;
	AirspacesDb m_airspacedb;
	Engine *m_engine;
	TrafficFlowRestrictions m_tfr;
	std::string m_dir_main;
	std::string m_dir_aux;
        Glib::RefPtr<Glib::MainLoop> m_mainloop;
	typedef std::set<Client::ptr_t> clients_t;
	clients_t m_clients;
	typedef std::list<Client::ptr_t> cfmuclients_t;
	cfmuclients_t m_cfmuclients;
        sigc::connection m_stateconn;
	unsigned int m_statecnt;
	GtkWidget *m_portal_webview;
	GtkWidget *m_validate_webview;
	gulong m_portal_webview_close_sighandler;
	gulong m_portal_webview_create_sighandler;
	gulong m_portal_webview_ready_sighandler;
	gulong m_portal_webview_loadstatus_sighandler;
	gulong m_portal_webview_navpolicy_sighandler;
	gulong m_portal_webview_newwndpolicy_sighandler;
	gulong m_portal_webview_rsrcload_sighandler;
	gulong m_portal_webview_consolemsg_sighandler;
	gulong m_validate_webview_close_sighandler;
	gulong m_validate_webview_create_sighandler;
	gulong m_validate_webview_ready_sighandler;
	gulong m_validate_webview_loadstatus_sighandler;
	gulong m_validate_webview_navpolicy_sighandler;
	gulong m_validate_webview_newwndpolicy_sighandler;
	gulong m_validate_webview_rsrcload_sighandler;
	gulong m_validate_webview_consolemsg_sighandler;
	guint m_verbose;
#ifdef HAVE_X11_XLIB_H
	sigc::connection m_x11srvconnwatch;
	int m_xdisplay;
        Glib::Pid m_x11srvpid;
	bool m_x11srvrun;
#endif
#ifdef HAVE_CURL
	typedef std::map<CURL *,Client::ptr_t> curlhandlemap_t;
	curlhandlemap_t m_curlhandlemap;
	typedef std::list<sigc::connection> curlio_t;
	curlio_t m_curlio;
	CURLM *m_curl;
#endif
	Glib::RefPtr<Gio::Socket> m_listensock;
#ifdef G_OS_UNIX
	Glib::RefPtr<Gio::UnixSocketAddress> m_listensockaddr;
#endif
	Glib::RefPtr<Gio::SocketService> m_listenservice;
	typedef enum {
		wvstate_idle,
		wvstate_startxserver,
		wvstate_loadmainpage,
		wvstate_loadquerypage,
		wvstate_ready,
		wvstate_query,
		wvstate_waitresult,
		wvstate_failed
	} wvstate_t;
	wvstate_t m_wvstate;
	Engine::auxdb_mode_t m_auxdbmode;
	bool m_tfrloaded;
	bool m_quit;

	void add_portal_sighandlers(void);
	void remove_portal_sighandlers(void);
	void add_validate_sighandlers(void);
	void remove_validate_sighandlers(void);
	void start_cfmu(void);
	void prepare_webview(void);
	void destroy_webview(void);
	void submit_webview(void);
	void set_wvstate(wvstate_t newstate);
	void set_wvstate_nochange(void) { set_wvstate(m_wvstate); }
	void parse_results(void);
	void zap_errors(void);
	static gboolean close_webview_1(WebKitWebView* webview, gpointer userdata);
	gboolean close_webview(WebKitWebView* webview);
	static WebKitWebView *create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata);
	WebKitWebView *create_webview(WebKitWebView *webview, WebKitWebFrame *frame);
	static gboolean webview_ready_1(WebKitWebView* webview, gpointer userdata);
	gboolean webview_ready(WebKitWebView* webview);
	static void loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata);
	void loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec);
	static gboolean policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
						   WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata);
	gboolean policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
					  WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision);
	static void resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata);
	void resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource);
	static gboolean result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata);
	gboolean result_change(WebKitDOMElement *target, WebKitDOMEvent *e);
	static gboolean console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata);
	gboolean console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id);

#ifdef HAVE_X11_XLIB_H
	void start_x11(void);
	void stop_x11(void);
	void x11srv_child_watch(GPid pid, int child_status);
	bool poll_xserver(void);
	static int x11_error(Display *dpy, XErrorEvent *err);
	static int x11_ioerror(Display *dpy);
	static bool x11_fail;
#endif
#ifdef HAVE_CURL
	void curl_add(CURL *handle, const Client::ptr_t& p);
	void curl_remove(CURL *handle);
	void curl_removeio(void);
	void curl_io(void);
#endif

	void annotate(Results& res);
	void validate_locally(const Client::ptr_t& p, const rules_t& tracerules = rules_t(), const rules_t& disabledrules = rules_t());
	void validate_cfmu(const Client::ptr_t& p);
	void close(const Client::ptr_t& p);
	void logmessage(unsigned int loglevel, const Glib::ustring& msg);

	void remove_listensockaddr(void);
#ifdef G_OS_UNIX
	void listen(const std::string& path, uid_t socketuid, gid_t socketgid, mode_t socketmode);
#else
	void listen(const std::string& path);
#endif
	bool on_connect(const Glib::RefPtr<Gio::SocketConnection>& connection, const Glib::RefPtr<Glib::Object>& source_object);
};

#endif /* CFMUVALIDATESERVER_H */
