#ifndef INTERVAL_H
#define INTERVAL_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <limits>
#include <string>
#include <set>
//#include <cstdint>
#include <stdint.h>

template <typename T>
class IntervalSet {
public:
	typedef T type_t;

	class Interval {
	public:
		Interval(void);
		Interval(type_t val);
		Interval(type_t val0, type_t val1);

		type_t get_lower(void) const { return m_lower; }
		type_t get_upper(void) const { return m_upper; }
		bool is_empty(void) const;
		bool is_full(void) const;
		bool is_inside(type_t x) const;
		bool is_overlap(const Interval& x) const;
		bool is_adjacent(const Interval& x) const;
		bool is_overlap_or_adjacent(const Interval& x) const;
		void set_empty(void);
		void set_full(void);
		std::string to_str(void) const;
		int compare(const Interval& x) const;
		bool operator==(const Interval& x) const { return compare(x) == 0; }
		bool operator!=(const Interval& x) const { return compare(x) != 0; }
		bool operator<(const Interval& x) const { return compare(x) < 0; }
		bool operator<=(const Interval& x) const { return compare(x) <= 0; }
		bool operator>(const Interval& x) const { return compare(x) > 0; }
		bool operator>=(const Interval& x) const { return compare(x) >= 0; }

		template<class Archive> void hibernate(Archive& ar) {
			ar.io(m_lower);
			ar.io(m_upper);
		}

	protected:
		type_t m_lower;
		type_t m_upper;
	};

	IntervalSet(void);
	IntervalSet(const Interval& x);
	IntervalSet(const IntervalSet& x);
	bool is_empty(void) const;
	bool is_full(void) const;
	bool is_inside(type_t x) const;
	void set_empty(void);
	void set_full(void);
	void swap(IntervalSet& x);
	IntervalSet& operator&=(type_t x);
	IntervalSet& operator|=(type_t x);
	IntervalSet& operator^=(type_t x);
	IntervalSet& operator-=(type_t x);
	IntervalSet& operator&=(const IntervalSet::Interval& x);
	IntervalSet& operator|=(const IntervalSet::Interval& x);
	IntervalSet& operator^=(const IntervalSet::Interval& x);
	IntervalSet& operator-=(const IntervalSet::Interval& x);
	IntervalSet& operator&=(const IntervalSet& x);
	IntervalSet& operator|=(const IntervalSet& x);
	IntervalSet& operator^=(const IntervalSet& x);
	IntervalSet& operator-=(const IntervalSet& x);
	IntervalSet operator~() const;
	std::string to_str(void) const;
	int compare(const IntervalSet& x) const;
	bool operator==(const IntervalSet& x) const { return compare(x) == 0; }
	bool operator!=(const IntervalSet& x) const { return compare(x) != 0; }
	bool operator<(const IntervalSet& x) const { return compare(x) < 0; }
	bool operator<=(const IntervalSet& x) const { return compare(x) <= 0; }
	bool operator>(const IntervalSet& x) const { return compare(x) > 0; }
	bool operator>=(const IntervalSet& x) const { return compare(x) >= 0; }

	template<class Archive> void hibernate(Archive& ar) {
		uint32_t sz(m_set.size());
		ar.ioleb(sz);
		if (ar.is_load()) {
			m_set.clear();
			for (; sz; --sz) {
				IntervalSet::Interval x;
				x.hibernate(ar);
				m_set.insert(x);
			}
		} else {
			for (typename set_t::iterator i(m_set.begin()), e(m_set.end()); i != e && sz > 0; ++i, --sz)
				const_cast<Interval&>(*i).hibernate(ar);
		}
	}

protected:
	typedef std::set<Interval> set_t;

public:
	typedef typename set_t::const_iterator const_iterator;
	const_iterator begin(void) const { return m_set.begin(); }
	const_iterator end(void) const { return m_set.end(); }

protected:
	set_t m_set;

	void normalize(void);
};

template <typename T> IntervalSet<T> operator&(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r &= y;
	return r;
}

template <typename T> IntervalSet<T> operator|(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r |= y;
	return r;
}

template <typename T> IntervalSet<T> operator^(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r ^= y;
	return r;
}

template <typename T> IntervalSet<T> operator-(const IntervalSet<T>& x, const IntervalSet<T>& y)
{
	IntervalSet<T> r(x);
	r -= y;
	return r;
}

#endif /* INTERVAL_H */
