/*****************************************************************************/

/*
 *      aircraft.cc  --  Aircraft Model.
 *
 *      Copyright (C) 2012, 2013  Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include "aircraft.h"
#include "airdata.h"
#include "baro.h"
#include "geom.h"

#include <cmath>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <iomanip>

#ifdef HAVE_EIGEN3
#include <Eigen/Dense>
#endif

template <typename T> T Aircraft::Poly1D<T>::eval(T x) const
{
	if (this->empty())
		return std::numeric_limits<T>::quiet_NaN();
	T ret(0);
	for (typename Poly1D<T>::const_reverse_iterator pi(this->rbegin()), pe(this->rend()); pi != pe; ++pi)
		ret = ret * x + *pi;
	return ret;
}

template <typename T> T Aircraft::Poly1D<T>::evalderiv(T x) const
{
	if (this->empty())
		return std::numeric_limits<T>::quiet_NaN();
	T ret(0), x1(1);
	for (typename Poly1D<T>::size_type i(1); i < this->size(); ++i, x1 *= x)
		ret += i * x1 * (*this)[i];
	return ret;
}

template <typename T> T Aircraft::Poly1D<T>::inveval(T x, T tol, unsigned int maxiter, T t) const
{
	for (unsigned int i = 0; i < maxiter; ++i) {
		T x0(this->eval(t));
		if (std::fabs(x - x0) <= tol)
			break;
		T xd(this->evalderiv(t));
		if (xd == 0)
			break;
		if (false)
			std::cerr << "inveval: y " << x << " yest " << x0 << " yderiv " << xd
				  << " x " << t << " tol " << tol << std::endl;
		t += (x - x0) / xd;
	}
	return t;
}

template <typename T> T Aircraft::Poly1D<T>::inveval(T x, T tol, unsigned int maxiter) const
{
	if (this->size() < 2 || (*this)[1] == 0)
		return std::numeric_limits<T>::quiet_NaN();
	return this->inveval(x, tol, maxiter, (x - (*this)[0]) / (*this)[1]);
}

template <typename T> T Aircraft::Poly1D<T>::inveval(T x) const
{
	static const T abstol(1e-8);
	static const T reltol(1e-4);
	return this->inveval(x, std::max(abstol, std::fabs(x) * reltol));
}

template <typename T> void Aircraft::Poly1D<T>::reduce(void)
{
	typename Poly1D<T>::size_type n(this->size());
	while (n > 1) {
		typename Poly1D<T>::size_type n1(n - 1);
		if ((*this)[n1] != (T)0)
			break;
		n = n1;
	}
	this->resize(n);
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator+=(T x)
{
	this->resize(std::max(this->size(), (typename Poly1D<T>::size_type)1U), (T)0);
	(*this)[0] += x;
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator-=(T x)
{
	this->resize(std::max(this->size(), (typename Poly1D<T>::size_type)1U), (T)0);
	(*this)[0] -= x;
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator*=(T x)
{
	for (typename Poly1D<T>::size_type i(0); i < this->size(); ++i)
		(*this)[i] *= x;
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator/=(T x)
{
	for (typename Poly1D<T>::size_type i(0); i < this->size(); ++i)
		(*this)[i] /= x;
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator+=(const Poly1D<T>& x)
{
	this->resize(std::max(this->size(), x.size()), (T)0);
	for (typename Poly1D<T>::size_type i(0); i < x.size(); ++i)
		(*this)[i] += x[i];
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T>& Aircraft::Poly1D<T>::operator-=(const Poly1D<T>& x)
{
	this->resize(std::max(this->size(), x.size()), (T)0);
	for (typename Poly1D<T>::size_type i(0); i < x.size(); ++i)
		(*this)[i] -= x[i];
	reduce();
	return *this;
}

template <typename T> Aircraft::Poly1D<T> Aircraft::Poly1D<T>::operator*(const Poly1D<T>& x) const
{
	Poly1D<T> r;
	{
		typename Poly1D<T>::size_type n(this->size() + x.size());
		if (!n)
			return r;
		--n;
		r.resize(n, (T)0);
	}
	for (typename Poly1D<T>::size_type i(0); i < this->size(); ++i)
		for (typename Poly1D<T>::size_type j(0); j < x.size(); ++j)
			r[i + j] += (*this)[i] * x[j];
	r.reduce();
	return r;
}

template <typename T> Aircraft::Poly1D<T> Aircraft::Poly1D<T>::differentiate(void) const
{
	if (this->size() < 2U)
		return Poly1D<T>();
	Poly1D<T> r(this->size() - 1);
	for (typename Poly1D<T>::size_type i(1); i < this->size(); ++i)
		r[i - 1] = i * (*this)[i];
	r.reduce();
	return r;
}

template <typename T> Aircraft::Poly1D<T> Aircraft::Poly1D<T>::integrate(void) const
{
	if (!this->size())
		return Poly1D<T>();
	Poly1D<T> r(this->size() + 1);
	for (typename Poly1D<T>::size_type i(1); i <= this->size(); ++i)
		r[i] = (*this)[i - 1] / (T)i;
	r.reduce();
	return r;
}

template <typename T> void Aircraft::Poly1D<T>::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	const xmlpp::TextNode* txt(el->get_child_text());
	if (!txt)
		return;
	this->clear();
	std::istringstream iss(txt->get_content());
	while (!iss.eof()) {
		double v;
		iss >> v;
		if (iss.fail()) {
			std::cerr << "Aircraft::Poly1D::load_xml: poly parse error: "
				  << txt->get_content() << " order " << this->size() << std::endl;
			break;
		}
		this->push_back(v);
		if (iss.eof())
			break;
		char ch;
		iss >> ch;
		if (iss.fail()) {
			std::cerr << "Aircraft::Poly1D::load_xml: poly parse error: "
				  << txt->get_content() << " order " << this->size() << std::endl;
			break;
		}
		if (ch != ',') {
			std::cerr << "Aircraft::Poly1D::load_xml: poly delimiter error: "
				  << txt->get_content() << " order " << this->size() << std::endl;
			break;
		}
	}
}

template <typename T> void Aircraft::Poly1D<T>::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	std::ostringstream oss;
	oss << std::setprecision(10);
	bool subseq(false);
	for (typename Poly1D<T>::const_iterator pi(this->begin()), pe(this->end()); pi != pe; ++pi) {
		if (subseq)
			oss << ',';
		subseq = true;
		oss << *pi;
	}
	el->set_child_text(oss.str());
}

template <typename T> std::ostream& Aircraft::Poly1D<T>::save_octave(std::ostream& os, const Glib::ustring& name) const
{
	os << "# name: " << name << std::endl
	   << "# type: matrix" << std::endl
	   << "# rows: " << this->size() << std::endl
	   << "# columns: 1" << std::endl;
	for (typename Poly1D<T>::const_iterator i(this->begin()), e(this->end()); i != e; ++i) {
		std::ostringstream oss;
		oss << std::setprecision(10) << *i;
		os << oss.str() << std::endl;
	}
	return os;
}

template class Aircraft::Poly1D<double>;

Aircraft::WeightBalance::Element::Unit::Unit(const Glib::ustring& name, double factor, double offset)
	: m_name(name), m_factor(factor), m_offset(offset)
{
}

void Aircraft::WeightBalance::Element::Unit::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_name.clear();
	m_factor = 0;
	m_offset = 0;
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("name")))
		m_name = attr->get_value();
	if ((attr = el->get_attribute("factor")))
		m_factor = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("offset")))
		m_offset = Glib::Ascii::strtod(attr->get_value());
}

void Aircraft::WeightBalance::Element::Unit::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_name.empty())
		el->remove_attribute("name");
	else
		el->set_attribute("name", m_name);
	if (m_factor == 0) {
		el->remove_attribute("factor");
	} else {
		std::ostringstream oss;
		oss << m_factor;
		el->set_attribute("factor", oss.str());
	}
	if (m_offset == 0) {
		el->remove_attribute("offset");
	} else {
		std::ostringstream oss;
		oss << m_offset;
		el->set_attribute("offset", oss.str());
	}
}

const Aircraft::WeightBalance::Element::FlagName Aircraft::WeightBalance::Element::flagnames[] =
{
	{ "fixed", flag_fixed },
	{ "binary", flag_binary },
	{ "avgas", flag_avgas },
	{ "oil", flag_oil },
	{ "gear", flag_gear },
	{ 0, flag_none }
};

Aircraft::WeightBalance::Element::Element(const Glib::ustring& name, double massmin, double massmax, double arm, flags_t flags)
	: m_name(name), m_massmin(massmin), m_massmax(massmax), m_arm(arm), m_flags(flags)
{
}

void Aircraft::WeightBalance::Element::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_units.clear();
	m_name.clear();
	m_massmin = 0;
	m_massmax = 0;
	m_arm = 0;
	m_flags = flag_none;
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("name")))
		m_name = attr->get_value();
	if ((attr = el->get_attribute("massmin")))
		m_massmin = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("massmax")))
		m_massmax = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("arm")))
		m_arm = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("flags"))) {
		Glib::ustring val(attr->get_value());
		Glib::ustring::size_type vn(0);
		while (vn < val.size()) {
			Glib::ustring::size_type vn2(val.find('|', vn));
			Glib::ustring v1(val.substr(vn, vn2 - vn));
			vn = vn2;
			if (vn < val.size())
				++vn;
			if (v1.empty())
				continue;
			if (*std::min_element(v1.begin(), v1.end()) >= (Glib::ustring::value_type)'0' &&
			    *std::max_element(v1.begin(), v1.end()) <= (Glib::ustring::value_type)'9') {
				m_flags |= (flags_t)strtol(v1.c_str(), 0, 10);
				continue;
			}
			const FlagName *fn = flagnames;
			for (; fn->m_name; ++fn)
				if (fn->m_name == v1)
					break;
			if (!fn->m_name) {
				std::cerr << "WeightBalance::Element::load_xml: invalid flag " << v1 << std::endl;
				continue;
			}
			m_flags |= fn->m_flag;
		}
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("unit"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_units.push_back(Unit());
			m_units.back().load_xml(static_cast<xmlpp::Element *>(*ni));
		}
	}
}

void Aircraft::WeightBalance::Element::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	{
		xmlpp::Node::NodeList nl(el->get_children());
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (m_name.empty())
		el->remove_attribute("name");
	else
		el->set_attribute("name", m_name);
	if (m_massmin == 0) {
		el->remove_attribute("massmin");
	} else {
		std::ostringstream oss;
		oss << m_massmin;
		el->set_attribute("massmin", oss.str());
	}
	if (m_massmax == 0) {
		el->remove_attribute("massmax");
	} else {
		std::ostringstream oss;
		oss << m_massmax;
		el->set_attribute("massmax", oss.str());
	}
	if (m_arm == 0) {
		el->remove_attribute("arm");
	} else {
		std::ostringstream oss;
		oss << m_arm;
		el->set_attribute("arm", oss.str());
	}
	{
		Glib::ustring flg;
		flags_t flags(m_flags);
		for (const FlagName *fn = flagnames; fn->m_name; ++fn) {
			if (fn->m_flag & ~flags)
				continue;
			flags &= ~fn->m_flag;
			flg += "|";
			flg += fn->m_name;
		}
		if (flags) {
			std::ostringstream oss;
			oss << '|' << (unsigned int)flags;
			flg += oss.str();
		}
		if (flg.empty())
			el->remove_attribute("flags");
		else
			el->set_attribute("flags", flg.substr(1));
	}
	for (units_t::const_iterator ui(m_units.begin()), ue(m_units.end()); ui != ue; ++ui)
		ui->save_xml(el->add_child("unit"));
}

void Aircraft::WeightBalance::Element::clear_units(void)
{
	m_units.clear();
}

void Aircraft::WeightBalance::Element::add_unit(const Glib::ustring& name, double factor, double offset)
{
	m_units.push_back(Unit(name, factor, offset));
	if (m_units.back().is_fixed())
		m_flags |= flag_fixed;
}

void Aircraft::WeightBalance::Element::add_auto_units(double kg_to_unit)
{
	if (get_flags() & flag_oil) {
		add_unit("Quart", 0.94635295 * 0.860 * kg_to_unit, 0);
		add_unit("l", 0.860 * kg_to_unit, 0);
	}
	if (get_flags() & flag_avgas) {
		add_unit("l", 0.718 * kg_to_unit, 0);
		add_unit("USgal", 3.7854118 * 0.718 * kg_to_unit, 0);
	}
	add_unit("kg", kg_to_unit, 0);
	add_unit("lb", 0.45359237 * kg_to_unit, 0);
}

Aircraft::WeightBalance::Envelope::Point::Point(double arm, double mass)
	: m_arm(arm), m_mass(mass)
{
}

/* twice the signed area of the triangle p0, p1, p2 */
double Aircraft::WeightBalance::Envelope::Point::area2(const Point& p0, const Point& p1, const Point& p2)
{
        double p2x = p2.get_arm();
        double p2y = p2.get_mass();
        double p0x = p0.get_arm() - p2x;
        double p0y = p0.get_mass() - p2y;
        double p1x = p1.get_arm() - p2x;
        double p1y = p1.get_mass() - p2y;
        return p0x * (p1y - p0y) - p0y * (p1x - p0x);
}

void Aircraft::WeightBalance::Envelope::Point::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_arm = 0;
	m_mass = 0;
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("arm")))
		m_arm = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("mass")))
		m_mass = Glib::Ascii::strtod(attr->get_value());
}

void Aircraft::WeightBalance::Envelope::Point::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_arm == 0) {
		el->remove_attribute("arm");
	} else {
		std::ostringstream oss;
		oss << m_arm;
		el->set_attribute("arm", oss.str());
	}
	if (m_mass == 0) {
		el->remove_attribute("mass");
	} else {
		std::ostringstream oss;
		oss << m_mass;
		el->set_attribute("mass", oss.str());
	}
}

Aircraft::WeightBalance::Envelope::Envelope(const Glib::ustring& name)
	: m_name(name)
{
}

int Aircraft::WeightBalance::Envelope::windingnumber(const Point& pt) const
{
        int wn = 0; /* the winding number counter */

        /* loop through all edges of the polygon */
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++) {
		const Point& pj(operator[](j));
 		const Point& pi(operator[](i));
                double lon = pt.area2(pj, pi);
		/* edge from V[j] to V[i] */
                if (pj.get_mass() <= pt.get_mass()) {
                        if (pi.get_mass() > pt.get_mass())
                                /* an upward crossing */
                                if (lon > 0)
                                        /* P left of edge: have a valid up intersect */
                                        wn++;
                } else {
                        if (pi.get_mass() <= pt.get_mass())
                                /* a downward crossing */
                                if (lon < 0)
                                        /* P right of edge: have a valid down intersect */
                                        wn--;
                }
        }
        return wn;
}

void Aircraft::WeightBalance::Envelope::get_bounds(double& minarm, double& maxarm, double& minmass, double& maxmass) const
{
	double mina(std::numeric_limits<double>::max());
	double maxa(std::numeric_limits<double>::min());
	double minm(mina);
	double maxm(maxa);
	for (env_t::const_iterator ei(m_env.begin()), ee(m_env.end()); ei != ee; ++ei) {
		const Point& el(*ei);
		mina = std::min(mina, el.get_arm());
		maxa = std::max(maxa, el.get_arm());
		minm = std::min(minm, el.get_mass());
		maxm = std::max(maxm, el.get_mass());
	}
	minarm = mina;
	maxarm = maxa;
	minmass = minm;
	maxmass = maxm;
}

void Aircraft::WeightBalance::Envelope::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_env.clear();
	m_name.clear();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("name")))
		m_name = attr->get_value();
	{
		xmlpp::Node::NodeList nl(el->get_children("point"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_env.push_back(Point());
			m_env.back().load_xml(static_cast<xmlpp::Element *>(*ni));
		}
	}
}

void Aircraft::WeightBalance::Envelope::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_name.empty())
		el->remove_attribute("name");
	else
		el->set_attribute("name", m_name);
	for (env_t::const_iterator ei(m_env.begin()), ee(m_env.end()); ei != ee; ++ei)
		ei->save_xml(el->add_child("point"));
}

void Aircraft::WeightBalance::Envelope::add_point(const Point& pt)
{
	m_env.push_back(pt);
}

void Aircraft::WeightBalance::Envelope::add_point(double arm, double mass)
{
	add_point(Point(arm, mass));
}

Aircraft::WeightBalance::WeightBalance(void)
{
	// Default: P28R-200
	m_elements.push_back(Element("Pilot and Front Pax", 0, 600, 85.5, Element::flag_none));
	m_elements.back().add_auto_units(2.2046226);
	m_elements.push_back(Element("Rear Pax", 0, 600, 118.1, Element::flag_none));
	m_elements.back().add_auto_units(2.2046226);
	m_elements.push_back(Element("Cargo", 0, 200, 142.8, Element::flag_none));
	m_elements.back().add_auto_units(2.2046226);
	m_elements.push_back(Element("Fuel", 0, 300, 95, Element::flag_avgas));
	m_elements.back().add_auto_units(2.2046226);
	m_elements.push_back(Element("Oil", 0, 15, 29.5, Element::flag_oil));
	m_elements.back().add_auto_units(2.2046226);
	m_elements.push_back(Element("Gear", 0, 0.1, 8190, Element::flag_gear));
	m_elements.back().add_unit("Extended", 0, 0);
	m_elements.back().add_unit("Retracted", 0, 0.1);
	m_elements.push_back(Element("Basic Empty Mass", 1653.0, 1653.0, 84.69, Element::flag_fixed));
	m_envelopes.push_back(Envelope("Normal"));
	m_envelopes.back().add_point(81.0, 1600);
	m_envelopes.back().add_point(95.9, 1600);
	m_envelopes.back().add_point(95.9, 2600);
	m_envelopes.back().add_point(90.0, 2600);
	m_envelopes.back().add_point(81.0, 1925);
	m_armunitname = "in";
	m_massunitname = "lb";
}

void Aircraft::WeightBalance::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_elements.clear();
	m_envelopes.clear();
	m_armunitname.clear();
	m_massunitname.clear();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("armunitname")))
		m_armunitname = attr->get_value();
 	if ((attr = el->get_attribute("massunitname")))
		m_massunitname = attr->get_value();
	{
		xmlpp::Node::NodeList nl(el->get_children("element"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_elements.push_back(Element());
			m_elements.back().load_xml(static_cast<xmlpp::Element *>(*ni));
		}
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("envelope"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_envelopes.push_back(Envelope());
			m_envelopes.back().load_xml(static_cast<xmlpp::Element *>(*ni));
		}
	}
}

void Aircraft::WeightBalance::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_armunitname.empty())
		el->remove_attribute("armunitname");
	else
		el->set_attribute("armunitname", m_armunitname);
	if (m_massunitname.empty())
		el->remove_attribute("massunitname");
	else
		el->set_attribute("massunitname", m_massunitname);
	for (elements_t::const_iterator ei(m_elements.begin()), ee(m_elements.end()); ei != ee; ++ei)
		ei->save_xml(el->add_child("element"));
	for (envelopes_t::const_iterator ei(m_envelopes.begin()), ee(m_envelopes.end()); ei != ee; ++ei)
		ei->save_xml(el->add_child("envelope"));
}

void Aircraft::WeightBalance::add_element(const Element& el)
{
	m_elements.push_back(el);
}

void Aircraft::WeightBalance::add_envelope(const Envelope& el)
{
	m_envelopes.push_back(el);
}

double Aircraft::WeightBalance::get_useable_fuelmass(void) const
{
	double m(0);
	for (elements_t::const_iterator i(m_elements.begin()), e(m_elements.end()); i != e; ++i) {
		const Aircraft::WeightBalance::Element& el(*i);
		if (!(el.get_flags() & Aircraft::WeightBalance::Element::flag_avgas))
			continue;
		m += el.get_massmax() - el.get_massmin();
	}
	return m;
}

double Aircraft::WeightBalance::get_total_fuelmass(void) const
{
	double m(0);
	for (elements_t::const_iterator i(m_elements.begin()), e(m_elements.end()); i != e; ++i) {
		const Aircraft::WeightBalance::Element& el(*i);
		if (!(el.get_flags() & Aircraft::WeightBalance::Element::flag_avgas))
			continue;
		m += el.get_massmax();
	}
	return m;
}

double Aircraft::WeightBalance::get_useable_oilmass(void) const
{
	double m(0);
	for (elements_t::const_iterator i(m_elements.begin()), e(m_elements.end()); i != e; ++i) {
		const Aircraft::WeightBalance::Element& el(*i);
		if (!(el.get_flags() & Aircraft::WeightBalance::Element::flag_oil))
			continue;
		m += el.get_massmax() - el.get_massmin();
	}
	return m;
}

double Aircraft::WeightBalance::get_total_oilmass(void) const
{
	double m(0);
	for (elements_t::const_iterator i(m_elements.begin()), e(m_elements.end()); i != e; ++i) {
		const Aircraft::WeightBalance::Element& el(*i);
		if (!(el.get_flags() & Aircraft::WeightBalance::Element::flag_oil))
			continue;
		m += el.get_massmax();
	}
	return m;
}

Aircraft::Distances::Distance::Point::Point(double da, double pa, double temp, double gnddist, double obstdist)
	: m_da(da), m_pa(pa), m_temp(temp), m_gnddist(gnddist), m_obstdist(obstdist)
{
}

void Aircraft::Distances::Distance::Point::load_xml(const xmlpp::Element *el, double altfactor, double tempfactor, double tempoffset, double distfactor)
{
	if (!el)
		return;
	m_da = m_pa = m_temp = m_gnddist = m_obstdist = std::numeric_limits<double>::quiet_NaN();
	xmlpp::Attribute *attr;
	if ((attr = el->get_attribute("da")))
		m_da = Glib::Ascii::strtod(attr->get_value()) * altfactor;
	if ((attr = el->get_attribute("pa")))
		m_pa = Glib::Ascii::strtod(attr->get_value()) * altfactor;
	if ((attr = el->get_attribute("temp")))
		m_temp = Glib::Ascii::strtod(attr->get_value()) * tempfactor + tempoffset;
	if ((attr = el->get_attribute("gnddist")))
		m_gnddist = Glib::Ascii::strtod(attr->get_value()) * distfactor;
	if ((attr = el->get_attribute("obstdist")))
		m_obstdist = Glib::Ascii::strtod(attr->get_value()) * distfactor;

}

void Aircraft::Distances::Distance::Point::save_xml(xmlpp::Element *el, double altfactor, double tempfactor, double tempoffset, double distfactor) const
{
	if (!el)
		return;
	if (std::isnan(m_da) || altfactor == 0) {
		el->remove_attribute("da");
	} else {
		std::ostringstream oss;
		oss << (m_da / altfactor);
		el->set_attribute("da", oss.str());
	}
	if (std::isnan(m_pa) || altfactor == 0) {
		el->remove_attribute("pa");
	} else {
		std::ostringstream oss;
		oss << (m_pa / altfactor);
		el->set_attribute("pa", oss.str());
	}
	if (std::isnan(m_temp) || tempfactor == 0) {
		el->remove_attribute("temp");
	} else {
		std::ostringstream oss;
		oss << ((m_temp - tempoffset) / tempfactor);
		el->set_attribute("temp", oss.str());
	}
	if (std::isnan(m_gnddist) || distfactor == 0) {
		el->remove_attribute("gnddist");
	} else {
		std::ostringstream oss;
		oss << (m_gnddist / distfactor);
		el->set_attribute("gnddist", oss.str());
	}
	if (std::isnan(m_obstdist) || distfactor == 0) {
		el->remove_attribute("obstdist");
	} else {
		std::ostringstream oss;
		oss << (m_obstdist / distfactor);
		el->set_attribute("obstdist", oss.str());
	}
}

double Aircraft::Distances::Distance::Point::get_densityalt(void) const
{
	if (!std::isnan(m_da))
		return m_da;
	if (std::isnan(m_pa))
		return std::numeric_limits<double>::quiet_NaN();
	float temp(m_temp);
	if (std::isnan(m_temp)) {
		float press;
		IcaoAtmosphere::std_altitude_to_pressure(&press, &temp, m_pa);
	}
	float press;
	IcaoAtmosphere::std_altitude_to_pressure(&press, 0, m_pa);
	float density(IcaoAtmosphere::pressure_to_density(press, temp));
	float densityalt;
	for (unsigned int i = 0; i < 4; ++i) {
		float press1(IcaoAtmosphere::density_to_pressure(density, temp));
		IcaoAtmosphere::std_pressure_to_altitude(&densityalt, &temp, press1);
	}
	return densityalt;
}

Aircraft::Distances::Distance::Distance(const Glib::ustring& name, double vrotate, double mass)
	: m_name(name), m_vrotate(vrotate), m_mass(mass)
{
}

#ifdef __WIN32__
__attribute__((force_align_arg_pointer))
#endif
bool Aircraft::Distances::Distance::recalculatepoly(bool force)
{
#ifdef HAVE_EIGEN3
	if (!force && !m_gndpoly.empty() && !m_obstpoly.empty())
		return false;
	if (m_points.empty()) {
		m_gndpoly.clear();
		m_obstpoly.clear();
		return false;
	}
	unsigned int polyorder(4);
	unsigned int ptgnd(0);
	unsigned int ptobst(0);
	Eigen::MatrixXd mg(m_points.size(), polyorder);
	Eigen::MatrixXd mo(m_points.size(), polyorder);
	Eigen::VectorXd vg(m_points.size());
	Eigen::VectorXd vo(m_points.size());
	for (points_t::const_iterator pi(m_points.begin()), pe(m_points.end()); pi != pe; ++pi) {
		double da(pi->get_densityalt());
		if (std::isnan(da))
			continue;
		double d(pi->get_gnddist());
		if (!std::isnan(d)) {
			double da1(1);
			for (unsigned int i = 0; i < polyorder; ++i, da1 *= da)
				mg(ptgnd, i) = da1;
			vg(ptgnd) = d;
			++ptgnd;
		}
		d = pi->get_obstdist();
		if (!std::isnan(d)) {
			double da1(1);
			for (unsigned int i = 0; i < polyorder; ++i, da1 *= da)
				mo(ptobst, i) = da1;
			vo(ptobst) = d;
			++ptobst;
		}
	}
	bool ret(false);
	if (force || m_gndpoly.empty()) {
		m_gndpoly.clear();
		if (ptgnd) {
			ret = true;
			mg.conservativeResize(ptgnd, std::min(ptgnd, polyorder));
			vg.conservativeResize(ptgnd);
			Eigen::VectorXd x((mg.transpose() * mg).inverse() * mg.transpose() * vg);
			poly_t p(x.data(), x.data() + x.size());
			m_gndpoly.swap(p);
		}
	}
	if (force || m_obstpoly.empty()) {
		m_obstpoly.clear();
		if (ptobst) {
			ret = true;
			mo.conservativeResize(ptobst, std::min(ptobst, polyorder));
			vo.conservativeResize(ptobst);
			Eigen::VectorXd x((mo.transpose() * mo).inverse() * mo.transpose() * vo);
			poly_t p(x.data(), x.data() + x.size());
			m_obstpoly.swap(p);
		}
	}
	return ret;
#else
	return false;
#endif
}

void Aircraft::Distances::Distance::add_point_da(double gnddist, double obstdist, double da)
{
	m_points.push_back(Point(da, std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), gnddist, obstdist));
}

void Aircraft::Distances::Distance::add_point_pa(double gnddist, double obstdist, double pa, double temp)
{
	m_points.push_back(Point(std::numeric_limits<double>::quiet_NaN(), pa, temp, gnddist, obstdist));
}

void Aircraft::Distances::Distance::load_xml(const xmlpp::Element *el, double altfactor, double tempfactor, double tempoffset, double distfactor)
{
	if (!el)
		return;
	m_name.clear();
	m_vrotate = 0;
	m_gndpoly.clear();
	m_obstpoly.clear();
	m_points.clear();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("name")))
		m_name = attr->get_value();
 	if ((attr = el->get_attribute("vrotate")))
		m_vrotate = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("mass")))
		m_mass = Glib::Ascii::strtod(attr->get_value());
	{
		xmlpp::Node::NodeList nl(el->get_children("point"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_points.push_back(Point());
			m_points.back().load_xml(static_cast<xmlpp::Element *>(*ni), altfactor, tempfactor, tempoffset, distfactor);
		}
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("gndpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_gndpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
 	}
	{
		xmlpp::Node::NodeList nl(el->get_children("obstpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_obstpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
}

void Aircraft::Distances::Distance::save_xml(xmlpp::Element *el, double altfactor, double tempfactor, double tempoffset, double distfactor) const
{
	if (!el)
		return;
	if (m_name.empty())
		el->remove_attribute("name");
	else
		el->set_attribute("name", m_name);
	if (m_vrotate == 0) {
		el->remove_attribute("vrotate");
	} else {
		std::ostringstream oss;
		oss << m_vrotate;
		el->set_attribute("vrotate", oss.str());
	}
	if (m_mass == 0) {
		el->remove_attribute("mass");
	} else {
		std::ostringstream oss;
		oss << m_mass;
		el->set_attribute("mass", oss.str());
	}
	for (points_t::const_iterator di(m_points.begin()), de(m_points.end()); di != de; ++di)
		di->save_xml(el->add_child("point"), altfactor, tempfactor, tempoffset, distfactor);
	{
		xmlpp::Node::NodeList nl(el->get_children("gndpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_gndpoly.empty())
		m_gndpoly.save_xml(el->add_child("gndpoly"));
 	{
		xmlpp::Node::NodeList nl(el->get_children("obstpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_obstpoly.empty())
		m_obstpoly.save_xml(el->add_child("obstpoly"));
}

double Aircraft::Distances::Distance::calculate_gnddist(double da) const
{
	return m_gndpoly.eval(da);
}

double Aircraft::Distances::Distance::calculate_obstdist(double da) const
{
	return m_obstpoly.eval(da);
}

void Aircraft::Distances::Distance::calculate(double& gnddist, double& obstdist, double da, double headwind) const
{
	calculate(gnddist, obstdist, da, headwind, get_mass());
}

void Aircraft::Distances::Distance::calculate(double& gnddist, double& obstdist, double da, double headwind, double mass) const
{
	double mmul(1);
	if (get_mass() > 0)
		mmul = std::max(0.1, std::min(10.0, mass / get_mass()));
	double mul(1);
	if (get_vrotate() > 0)
		mul = std::max(0.5, (get_vrotate() - headwind) / get_vrotate());
	double mul2(mul * mul);
	double gd(calculate_gnddist(da));
	double od(calculate_obstdist(da));
	if (std::isnan(gd)) {
		gnddist = gd;
		obstdist = od;
		return;
	}
	if (false)
		std::cerr << "Aircraft: Distance: " << m_name << ": da " << da << " MTOM gnd: " << gd << " obst " << od << std::endl;
	od -= gd;
	od *= mul;
	gd *= mul2 * mmul;
	od += gd;
	gnddist = gd;
	obstdist = od;
	if (false)
		std::cerr << "Aircraft: Distance: " << m_name << ": da " << da << " mass: " << mass << " headwind: " << headwind
			  << " mmul: " << mmul << " mul: " << mul << " gnd: " << gd << " obst " << od << std::endl;
}

Aircraft::Distances::Distances(double altfactor, double tempfactor, double tempoffset, double distfactor)
	: m_altfactor(altfactor), m_tempfactor(tempfactor), m_tempoffset(tempoffset), m_distfactor(distfactor)
{
	// Default: P28R-200
	static const int16_t p28r200_tkoffdist[][4] = {
		{  756, 1022, 1578, 1800 },
		{  867, 1156, 1800, 2022 },
		{  956, 1289, 2000, 2267 },
		{ 1067, 1422, 2244, 2533 },
		{ 1200, 1600, 2533, 2867 },
		{ 1311, 1755, 2889, 3244 },
		{ 1489, 1978, 3311, 3689 },
		{ 1667, 2200, 3867, 4244 },
		{ 1867, 2467, 4556, 5000 }
	};
	static const int16_t p28r200_ldgdist[][2] = {
		{ 773, 1351 },
		{ 800, 1369 },
		{ 827, 1396 },
		{ 853, 1431 },
		{ 880, 1449 },
		{ 898, 1476 },
		{ 924, 1502 },
		{ 951, 1520 }
	};
	m_takeoffdist.push_back(Distance("Flaps 25", 56, 1179.3402));
	for (unsigned int i = 0; i < sizeof(p28r200_tkoffdist)/sizeof(p28r200_tkoffdist[0]); ++i)
		m_takeoffdist.back().add_point_da(p28r200_tkoffdist[i][0] * 0.3048, p28r200_tkoffdist[i][2] * 0.3048, i * 1000);
	m_takeoffdist.push_back(Distance("Flaps 0", 56, 1179.3402));
	for (unsigned int i = 0; i < sizeof(p28r200_tkoffdist)/sizeof(p28r200_tkoffdist[0]); ++i)
		m_takeoffdist.back().add_point_da(p28r200_tkoffdist[i][1] * 0.3048, p28r200_tkoffdist[i][3] * 0.3048, i * 1000);
	m_ldgdist.push_back(Distance("Flaps 40", 56, 1179.3402));
	for (unsigned int i = 0; i < sizeof(p28r200_ldgdist)/sizeof(p28r200_ldgdist[0]); ++i)
		m_ldgdist.back().add_point_da(p28r200_ldgdist[i][0] * 0.3048, p28r200_ldgdist[i][1] * 0.3048, i * 1000);
	recalculatepoly(true);
	m_distfactor = 0.3048;	
}

bool Aircraft::Distances::recalculatepoly(bool force)
{
	bool ret(false);
	for (distances_t::iterator di(m_takeoffdist.begin()), de(m_takeoffdist.end()); di != de; ++di)
		ret = di->recalculatepoly(force) || ret;
	for (distances_t::iterator di(m_ldgdist.begin()), de(m_ldgdist.end()); di != de; ++di)
		ret = di->recalculatepoly(force) || ret;
	return ret;
}

void Aircraft::Distances::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_takeoffdist.clear();
	m_ldgdist.clear();
	m_altfactor = 1;
	m_tempfactor = 1;
	m_tempoffset = 273.15;
	m_distfactor = 1;
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("altfactor")))
		m_altfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("tempfactor")))
		m_tempfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("tempoffset")))
		m_tempoffset = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("distfactor")))
		m_distfactor = Glib::Ascii::strtod(attr->get_value());
	{
		xmlpp::Node::NodeList nl(el->get_children("tkoffdist"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_takeoffdist.push_back(Distance());
			m_takeoffdist.back().load_xml(static_cast<xmlpp::Element *>(*ni), m_altfactor, m_tempfactor, m_tempoffset, m_distfactor);
		}
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("ldgdist"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_ldgdist.push_back(Distance());
			m_ldgdist.back().load_xml(static_cast<xmlpp::Element *>(*ni), m_altfactor, m_tempfactor, m_tempoffset, m_distfactor);
		}
	}
}

void Aircraft::Distances::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_altfactor == 1) {
		el->remove_attribute("altfactor");
	} else {
		std::ostringstream oss;
		oss << m_altfactor;
		el->set_attribute("altfactor", oss.str());
	}
	if (m_tempfactor == 1) {
		el->remove_attribute("tempfactor");
	} else {
		std::ostringstream oss;
		oss << m_tempfactor;
		el->set_attribute("tempfactor", oss.str());
	}
	if (m_tempoffset == 273.15) {
		el->remove_attribute("tempoffset");
	} else {
		std::ostringstream oss;
		oss << m_tempoffset;
		el->set_attribute("tempoffset", oss.str());
	}
	if (m_distfactor == 1) {
		el->remove_attribute("distfactor");
	} else {
		std::ostringstream oss;
		oss << m_distfactor;
		el->set_attribute("distfactor", oss.str());
	}
	for (distances_t::const_iterator di(m_takeoffdist.begin()), de(m_takeoffdist.end()); di != de; ++di)
		di->save_xml(el->add_child("tkoffdist"), m_altfactor, m_tempfactor, m_tempoffset, m_distfactor);
	for (distances_t::const_iterator di(m_ldgdist.begin()), de(m_ldgdist.end()); di != de; ++di)
		di->save_xml(el->add_child("ldgdist"), m_altfactor, m_tempfactor, m_tempoffset, m_distfactor);
}

Aircraft::Climb::Point::Point(double da, double rate, double fuelflow)
	: m_da(da), m_rate(rate), m_fuelflow(fuelflow)
{
}
			
void Aircraft::Climb::Point::load_xml(const xmlpp::Element *el, double altfactor, double ratefactor, double fuelfactor)
{
	if (!el)
		return;
	m_da = m_rate = m_fuelflow = std::numeric_limits<double>::quiet_NaN();
	xmlpp::Attribute *attr;
	if ((attr = el->get_attribute("da")))
		m_da = Glib::Ascii::strtod(attr->get_value()) * altfactor;
	if ((attr = el->get_attribute("rate")))
		m_rate = Glib::Ascii::strtod(attr->get_value()) * ratefactor;
	if ((attr = el->get_attribute("fuelflow")))
		m_fuelflow = Glib::Ascii::strtod(attr->get_value()) * fuelfactor;

}

void Aircraft::Climb::Point::save_xml(xmlpp::Element *el, double altfactor, double ratefactor, double fuelfactor) const
{
	if (!el)
		return;
	if (std::isnan(m_da) || altfactor == 0) {
		el->remove_attribute("da");
	} else {
		std::ostringstream oss;
		oss << (m_da / altfactor);
		el->set_attribute("da", oss.str());
	}
	if (std::isnan(m_rate) || ratefactor == 0) {
		el->remove_attribute("rate");
	} else {
		std::ostringstream oss;
		oss << (m_rate / ratefactor);
		el->set_attribute("rate", oss.str());
	}
	if (std::isnan(m_fuelflow) || fuelfactor == 0) {
		el->remove_attribute("fuelflow");
	} else {
		std::ostringstream oss;
		oss << (m_fuelflow / fuelfactor);
		el->set_attribute("fuelflow", oss.str());
	}
}

Aircraft::Climb::Climb(double altfactor, double ratefactor, double fuelfactor)
	: m_altfactor(altfactor), m_ratefactor(ratefactor), m_fuelfactor(fuelfactor), m_ceiling(10000), m_climbtime(3600)
{
	// PA28R-200
	m_points.push_back(Point(0, 910, 14));
	m_points.push_back(Point(18000, 0, 8));
	recalculatepoly(95 * 0.86897624, false);
}

void Aircraft::Climb::calculate(double& rate, double& fuelflow, double da)
{
	rate = m_ratepoly.eval(da);
	fuelflow = m_fuelflowpoly.eval(da);
}

double Aircraft::Climb::time_to_altitude(double t) const
{
	return m_climbaltpoly.eval(t);
}

double Aircraft::Climb::time_to_distance(double t) const
{
	return m_climbdistpoly.eval(t);
}

double Aircraft::Climb::time_to_fuel(double t) const
{
	return m_climbfuelpoly.eval(t);
}

double Aircraft::Climb::time_to_climbrate(double t) const
{
	return m_climbaltpoly.evalderiv(t) * 60.0;
}

double Aircraft::Climb::time_to_tas(double t) const
{
	return m_climbdistpoly.evalderiv(t) * 3600.0;
}

double Aircraft::Climb::time_to_fuelflow(double t) const
{
	return m_climbfuelpoly.evalderiv(t) * 3600.0;
}

double Aircraft::Climb::altitude_to_time(double a) const
{
	return m_climbaltpoly.inveval(a);
}

double Aircraft::Climb::distance_to_time(double d) const
{
	return m_climbdistpoly.inveval(d);
}

#ifdef __WIN32__
__attribute__((force_align_arg_pointer))
#endif
bool Aircraft::Climb::recalculatepoly(double vy, bool force)
{
#ifdef HAVE_EIGEN3
	if (!force && !m_ratepoly.empty() && !m_fuelflowpoly.empty())
		return false;
	if (m_points.empty()) {
		m_ratepoly.clear();
		m_fuelflowpoly.clear();
		return false;
	}
	unsigned int polyorder(4);
	unsigned int ptrate(0);
	unsigned int ptfuel(0);
	Eigen::MatrixXd mr(m_points.size(), polyorder);
	Eigen::MatrixXd mf(m_points.size(), polyorder);
	Eigen::VectorXd vr(m_points.size());
	Eigen::VectorXd vf(m_points.size());
	for (points_t::const_iterator pi(m_points.begin()), pe(m_points.end()); pi != pe; ++pi) {
		double da(pi->get_densityalt());
		if (std::isnan(da))
			continue;
		double d(pi->get_rate());
		if (!std::isnan(d)) {
			double da1(1);
			for (unsigned int i = 0; i < polyorder; ++i, da1 *= da)
				mr(ptrate, i) = da1;
			vr(ptrate) = d;
			++ptrate;
		}
		d = pi->get_fuelflow();
		if (!std::isnan(d)) {
			double da1(1);
			for (unsigned int i = 0; i < polyorder; ++i, da1 *= da)
				mf(ptfuel, i) = da1;
			vf(ptfuel) = d;
			++ptfuel;
		}
	}
	bool ret(false);
	if (force || m_ratepoly.empty()) {
		m_ratepoly.clear();
		if (ptrate) {
			ret = true;
			//mr.conservativeResize(ptrate, std::min(ptrate, polyorder));
			//vr.conservativeResize(ptrate);
			mr = mr.topLeftCorner(ptrate, std::min(ptrate, polyorder)).eval();
			vr = vr.head(ptrate).eval();
			Eigen::VectorXd x((mr.transpose() * mr).inverse() * mr.transpose() * vr);
			if (false)
				std::cerr << "mr [ " << mr << " ]; vr [ " << vr << " ]" << std::endl;
			poly_t p(x.data(), x.data() + x.size());
			m_ratepoly.swap(p);
		}
	}
	if (force || m_fuelflowpoly.empty()) {
		m_fuelflowpoly.clear();
		if (ptfuel) {
			ret = true;
			//mf.conservativeResize(ptfuel, std::min(ptfuel, polyorder));
			//vf.conservativeResize(ptfuel);
			mf = mf.topLeftCorner(ptfuel, std::min(ptfuel, polyorder)).eval();
			vf = vf.head(ptfuel).eval();
			Eigen::VectorXd x((mf.transpose() * mf).inverse() * mf.transpose() * vf);
			if (false)
				std::cerr << "mf [ " << mr << " ]; vf [ " << vr << " ]" << std::endl;
			poly_t p(x.data(), x.data() + x.size());
			m_fuelflowpoly.swap(p);
		}
	}
	// recompute climb profile polynomials
	if (ret) {
		Poly1D<double> altv, distv, fuelv;
		double dist(0), fuel(0);
		AirData ad;
		ad.set_qnh_temp(); // standard atmosphere
		ad.set_pressure_altitude(0);
		ad.set_cas(vy);
		for (unsigned int i = 0; ; ++i) {
			altv.push_back(ad.get_pressure_altitude());
			distv.push_back(dist);
			fuelv.push_back(fuel);
			// limit simulated climb to one hour
			if (i >= 3600)
				break;
			fuel += m_fuelflowpoly.eval(ad.get_pressure_altitude()) * (1.0 / 3600.0);
			dist += ad.get_tas() * (1.0 / 3600.0);
			{
				double altr(m_ratepoly.eval(ad.get_pressure_altitude()));
				if (altr <= 0 || std::isnan(altr))
					break;
				ad.set_pressure_altitude(ad.get_pressure_altitude() + altr * (1.0 / 60.0));
			}
		}
		m_ceiling = altv.back();
		m_climbtime = altv.size();
		if (altv.size() > 10) {
			const bool force_origin(true);
			{
				const unsigned int polyorder(7);
				Eigen::MatrixXd m(altv.size(), polyorder);
				Eigen::VectorXd v(altv.size());
				for (std::vector<double>::size_type i(0); i < altv.size(); ++i) {
					v(i) = i;
					double x(1);
					if (force_origin)
						x *= altv[i];
					for (unsigned int j = 0; j < polyorder; ++j, x *= altv[i])
						m(i, j) = x;
				}
				Eigen::VectorXd x((m.transpose() * m).inverse() * m.transpose() * v);
				poly_t p(x.data(), x.data() + x.size());
				if (force_origin)
					p.insert(p.begin(), 0);
				m_climbtimepoly.swap(p);
			}
			{
				const unsigned int polyorder(5);
				Eigen::MatrixXd m(altv.size(), polyorder);
				Eigen::VectorXd v(altv.size());
				for (std::vector<double>::size_type i(0); i < altv.size(); ++i) {
					v(i) = altv[i];
					double t(1);
					if (force_origin)
						t *= i;
					for (unsigned int j = 0; j < polyorder; ++j, t *= i)
						m(i, j) = t;
				}
				Eigen::VectorXd x((m.transpose() * m).inverse() * m.transpose() * v);
				poly_t p(x.data(), x.data() + x.size());
				if (force_origin)
					p.insert(p.begin(), 0);
				m_climbaltpoly.swap(p);
			}
			{
				const unsigned int polyorder(5);
				Eigen::MatrixXd m(distv.size(), polyorder);
				Eigen::VectorXd v(distv.size());
				for (std::vector<double>::size_type i(0); i < distv.size(); ++i) {
					v(i) = distv[i];
					double t(1);
					if (force_origin)
						t *= i;
					for (unsigned int j = 0; j < polyorder; ++j, t *= i)
						m(i, j) = t;
				}
				Eigen::VectorXd x((m.transpose() * m).inverse() * m.transpose() * v);
				poly_t p(x.data(), x.data() + x.size());
				if (force_origin)
					p.insert(p.begin(), 0);
				m_climbdistpoly.swap(p);
			}
			{
				const unsigned int polyorder(5);
				Eigen::MatrixXd m(fuelv.size(), polyorder);
				Eigen::VectorXd v(fuelv.size());
				for (std::vector<double>::size_type i(0); i < fuelv.size(); ++i) {
					v(i) = fuelv[i];
					double t(1);
					if (force_origin)
						t *= i;
					for (unsigned int j = 0; j < polyorder; ++j, t *= i)
						m(i, j) = t;
				}
				Eigen::VectorXd x((m.transpose() * m).inverse() * m.transpose() * v);
				poly_t p(x.data(), x.data() + x.size());
				if (force_origin)
					p.insert(p.begin(), 0);
				m_climbfuelpoly.swap(p);
			}
			if (false) {
				std::ofstream os("/tmp/arrowclimb.m");
				altv.save_octave(os, "altv");
				distv.save_octave(os, "distv");
				fuelv.save_octave(os, "fuelv");
				m_ratepoly.save_octave(os, "ratepoly");
				m_fuelflowpoly.save_octave(os, "fuelflowpoly");
				m_climbtimepoly.save_octave(os, "climbtimepoly");
				m_climbaltpoly.save_octave(os, "climbaltpoly");
				m_climbdistpoly.save_octave(os, "climbdistpoly");
				m_climbfuelpoly.save_octave(os, "climbfuelpoly");
			}
		}
	}
	return ret;
#else
	return false;
#endif
}

void Aircraft::Climb::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_altfactor = 1;
	m_ratefactor = 1;
	m_fuelfactor = 1;
	m_ceiling = 10000;
	m_climbtime = 3600;
	m_ratepoly.clear();
	m_fuelflowpoly.clear();
	m_climbtimepoly.clear();
	m_climbaltpoly.clear();
	m_climbdistpoly.clear();
	m_climbfuelpoly.clear();
	m_points.clear();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("altfactor")))
		m_altfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("ratefactor")))
		m_ratefactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("fuelfactor")))
		m_fuelfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("ceiling")))
		m_ceiling = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("climbtime")))
		m_climbtime = Glib::Ascii::strtod(attr->get_value());
	{
		xmlpp::Node::NodeList nl(el->get_children("point"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_points.push_back(Point());
			m_points.back().load_xml(static_cast<xmlpp::Element *>(*ni), m_altfactor, m_ratefactor, m_fuelfactor);
		}
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("ratepoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_ratepoly.load_xml(static_cast<xmlpp::Element *>(*ni));
 	}
	{
		xmlpp::Node::NodeList nl(el->get_children("fuelflowpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_fuelflowpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("climbtimepoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_climbtimepoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("climbaltpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_climbaltpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("climbdistpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_climbdistpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
	{
		xmlpp::Node::NodeList nl(el->get_children("climbfuelpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			m_climbfuelpoly.load_xml(static_cast<xmlpp::Element *>(*ni));
	}
}

void Aircraft::Climb::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_altfactor == 1) {
		el->remove_attribute("altfactor");
	} else {
		std::ostringstream oss;
		oss << m_altfactor;
		el->set_attribute("altfactor", oss.str());
	}
	if (m_ratefactor == 1) {
		el->remove_attribute("ratefactor");
	} else {
		std::ostringstream oss;
		oss << m_ratefactor;
		el->set_attribute("ratefactor", oss.str());
	}
	if (m_fuelfactor == 1) {
		el->remove_attribute("fuelfactor");
	} else {
		std::ostringstream oss;
		oss << m_fuelfactor;
		el->set_attribute("fuelfactor", oss.str());
	}
	if (m_ceiling == 1) {
		el->remove_attribute("ceiling");
	} else {
		std::ostringstream oss;
		oss << m_ceiling;
		el->set_attribute("ceiling", oss.str());
	}
	if (m_climbtime == 1) {
		el->remove_attribute("climbtime");
	} else {
		std::ostringstream oss;
		oss << m_climbtime;
		el->set_attribute("climbtime", oss.str());
	}
	for (points_t::const_iterator di(m_points.begin()), de(m_points.end()); di != de; ++di)
		di->save_xml(el->add_child("point"), m_altfactor, m_ratefactor, m_fuelfactor);
	{
		xmlpp::Node::NodeList nl(el->get_children("ratepoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_ratepoly.empty())
		m_ratepoly.save_xml(el->add_child("ratepoly"));
 	{
		xmlpp::Node::NodeList nl(el->get_children("fuelflowpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_fuelflowpoly.empty())
		m_fuelflowpoly.save_xml(el->add_child("fuelflowpoly"));
 	{
		xmlpp::Node::NodeList nl(el->get_children("climbtimepoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_climbtimepoly.empty())
		m_climbtimepoly.save_xml(el->add_child("climbtimepoly"));
 	{
		xmlpp::Node::NodeList nl(el->get_children("climbaltpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_climbaltpoly.empty())
		m_climbaltpoly.save_xml(el->add_child("climbaltpoly"));
  	{
		xmlpp::Node::NodeList nl(el->get_children("climbdistpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_climbdistpoly.empty())
		m_climbdistpoly.save_xml(el->add_child("climbdistpoly"));
	{
		xmlpp::Node::NodeList nl(el->get_children("climbfuelpoly"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
			el->remove_child(*ni);
	}
	if (!m_climbfuelpoly.empty())
		m_climbfuelpoly.save_xml(el->add_child("climbfuelpoly"));
}

void Aircraft::Climb::add_point(const Point& pt)
{
	m_points.push_back(pt);
}

void Aircraft::Climb::clear_points(void)
{
	m_points.clear();
}

Aircraft::Cruise::Point::Point(double da, double rpm, double mp, double bhp, double tas, double fuelflow)
	: m_da(da), m_rpm(rpm), m_mp(mp), m_bhp(bhp), m_tas(tas), m_fuelflow(fuelflow)
{
}
			
void Aircraft::Cruise::Point::load_xml(const xmlpp::Element *el, double altfactor, double tasfactor, double fuelfactor)
{
	if (!el)
		return;
	m_da = m_rpm = m_mp = m_bhp = m_tas = m_fuelflow = std::numeric_limits<double>::quiet_NaN();
	xmlpp::Attribute *attr;
	if ((attr = el->get_attribute("da")))
		m_da = Glib::Ascii::strtod(attr->get_value()) * altfactor;
	if ((attr = el->get_attribute("rpm")))
		m_rpm = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("mp")))
		m_mp = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("bhp")))
		m_bhp = Glib::Ascii::strtod(attr->get_value());
	if ((attr = el->get_attribute("tas")))
		m_tas = Glib::Ascii::strtod(attr->get_value()) * tasfactor;
	if ((attr = el->get_attribute("fuelflow")))
		m_fuelflow = Glib::Ascii::strtod(attr->get_value()) * fuelfactor;
}

void Aircraft::Cruise::Point::save_xml(xmlpp::Element *el, double altfactor, double tasfactor, double fuelfactor) const
{
	if (!el)
		return;
	if (std::isnan(m_da) || altfactor == 0) {
		el->remove_attribute("da");
	} else {
		std::ostringstream oss;
		oss << (m_da / altfactor);
		el->set_attribute("da", oss.str());
	}
	if (std::isnan(m_rpm)) {
		el->remove_attribute("rpm");
	} else {
		std::ostringstream oss;
		oss << m_rpm;
		el->set_attribute("rpm", oss.str());
	}
	if (std::isnan(m_mp)) {
		el->remove_attribute("mp");
	} else {
		std::ostringstream oss;
		oss << m_mp;
		el->set_attribute("mp", oss.str());
	}
	if (std::isnan(m_bhp)) {
		el->remove_attribute("bhp");
	} else {
		std::ostringstream oss;
		oss << m_bhp;
		el->set_attribute("bhp", oss.str());
	}
	if (std::isnan(m_tas) || tasfactor == 0) {
		el->remove_attribute("tas");
	} else {
		std::ostringstream oss;
		oss << (m_tas / tasfactor);
		el->set_attribute("tas", oss.str());
	}
	if (std::isnan(m_fuelflow) || fuelfactor == 0) {
		el->remove_attribute("fuelflow");
	} else {
		std::ostringstream oss;
		oss << (m_fuelflow / fuelfactor);
		el->set_attribute("fuelflow", oss.str());
	}
}

const std::string& to_str(Aircraft::Cruise::propulsion_t prop)
{
	switch (prop) {
	case Aircraft::Cruise::propulsion_fixedpitch:
	{
		static const std::string r("fixedpitch");
		return r;
	}

	case Aircraft::Cruise::propulsion_constantspeed:
	{
		static const std::string r("constantspeed");
		return r;
	}

	case Aircraft::Cruise::propulsion_turboprop:
	{
		static const std::string r("turboprop");
		return r;
	}

	case Aircraft::Cruise::propulsion_turbojet:
	{
		static const std::string r("turbojet");
		return r;
	}

	default:
	{
		static const std::string r("?");
		return r;
	}
	}
}

Aircraft::Cruise::Cruise(propulsion_t prop, double altfactor, double tasfactor, double fuelfactor)
	: m_altfactor(altfactor), m_tasfactor(tasfactor), m_fuelfactor(fuelfactor), m_propulsion(prop)
{
	// PA28R-200
	m_points.push_back(Point(    0, 2100, 22.9, 110, 131.5 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 1000, 2100, 22.7, 110, 133.0 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 2000, 2100, 22.4, 110, 134.5 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 3000, 2100, 22.2, 110, 136.0 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 4000, 2100, 21.9, 110, 137.5 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 5000, 2100, 21.7, 110, 139.0 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 6000, 2100, 21.4, 110, 140.5 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 7000, 2100, 21.2, 110, 142.0 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 8000, 2100, 21.0, 110, 143.5 * 0.86897624, 54.0/6));
	m_points.push_back(Point( 9000, 2100, 20.8, 110, 145.0 * 0.86897624, 54.0/6));
	m_points.push_back(Point(    0, 2400, 20.4, 110, 131.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 1000, 2400, 20.2, 110, 133.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 2000, 2400, 20.0, 110, 134.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 3000, 2400, 19.8, 110, 136.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 4000, 2400, 19.5, 110, 137.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 5000, 2400, 19.3, 110, 139.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 6000, 2400, 19.1, 110, 140.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 7000, 2400, 18.9, 110, 142.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 8000, 2400, 18.7, 110, 143.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point( 9000, 2400, 18.5, 110, 145.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point(10000, 2400, 18.3, 110, 146.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point(11000, 2400, 18.1, 110, 148.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point(12000, 2400, 17.8, 110, 149.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point(13000, 2400, 17.6, 110, 151.0 * 0.86897624, 57.0/6));
	m_points.push_back(Point(14000, 2400, 17.4, 110, 152.5 * 0.86897624, 57.0/6));
	m_points.push_back(Point(    0, 2100, 25.9, 130, 143.0 * 0.86897624, 61.0/6));
	m_points.push_back(Point( 1000, 2100, 25.6, 130, 144.8 * 0.86897624, 61.0/6));
	m_points.push_back(Point( 2000, 2100, 25.4, 130, 146.6 * 0.86897624, 61.0/6));
	m_points.push_back(Point( 3000, 2100, 25.1, 130, 148.3 * 0.86897624, 61.0/6));
	m_points.push_back(Point( 4000, 2100, 24.8, 130, 150.1 * 0.86897624, 61.0/6));
	m_points.push_back(Point( 5000, 2100, 24.5, 130, 151.9 * 0.86897624, 61.0/6));
	m_points.push_back(Point(    0, 2400, 22.9, 130, 143.0 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 1000, 2400, 22.7, 130, 144.8 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 2000, 2400, 22.5, 130, 146.6 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 3000, 2400, 22.2, 130, 148.3 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 4000, 2400, 22.0, 130, 150.1 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 5000, 2400, 21.7, 130, 151.9 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 6000, 2400, 21.5, 130, 153.7 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 7000, 2400, 21.3, 130, 155.4 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 8000, 2400, 21.0, 130, 157.2 * 0.86897624, 64.0/6));
	m_points.push_back(Point( 9000, 2400, 20.8, 130, 159.0 * 0.86897624, 64.0/6));
	m_points.push_back(Point(    0, 2400, 25.5, 150, 153.0 * 0.86897624, 72.0/6));
	m_points.push_back(Point( 1000, 2400, 25.2, 150, 155.0 * 0.86897624, 72.0/6));
	m_points.push_back(Point( 2000, 2400, 25.0, 150, 157.0 * 0.86897624, 72.0/6));
	m_points.push_back(Point( 3000, 2400, 24.7, 150, 159.0 * 0.86897624, 72.0/6));
	m_points.push_back(Point( 4000, 2400, 24.4, 150, 161.0 * 0.86897624, 72.0/6));
	m_points.push_back(Point( 5000, 2400, 24.2, 150, 163.0 * 0.86897624, 72.0/6));
	// Extrapolated
	m_points.push_back(Point(    0, 2400, 17.9,  90, 120.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 1000, 2400, 17.7,  90, 122.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 2000, 2400, 17.5,  90, 123.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 3000, 2400, 17.3,  90, 125.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 4000, 2400, 17.0,  90, 126.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 5000, 2400, 16.8,  90, 128.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 6000, 2400, 16.6,  90, 129.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 7000, 2400, 16.4,  90, 131.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 8000, 2400, 16.2,  90, 132.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point( 9000, 2400, 16.0,  90, 134.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point(10000, 2400, 15.8,  90, 135.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point(11000, 2400, 15.6,  90, 137.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point(12000, 2400, 15.3,  90, 138.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point(13000, 2400, 15.1,  90, 140.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point(14000, 2400, 14.9,  90, 141.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point(15000, 2400, 14.7,  90, 143.0 * 0.86897624, 50.0/6));
	m_points.push_back(Point(16000, 2400, 14.5,  90, 144.5 * 0.86897624, 50.0/6));
	m_points.push_back(Point(17000, 2400, 14.3,  90, 145.0 * 0.86897624, 50.0/6));
	build_altmap();
}

void Aircraft::Cruise::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_propulsion = propulsion_fixedpitch;
	m_altfactor = 1;
	m_tasfactor = 1;
	m_fuelfactor = 1;
	bool hasmp(true);
	bool hasprop(false);
	m_points.clear();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("propulsion"))) {
		static const propulsion_t propval[] = {
			propulsion_fixedpitch,
			propulsion_constantspeed,
			propulsion_turboprop,
			propulsion_turbojet
		};
		unsigned int i;
		for (i = 0; i < sizeof(propval)/sizeof(propval[0]); ++i) {
			if (attr->get_value() == to_str(propval[i]))
				break;
		}
		if (i < sizeof(propval)/sizeof(propval[0])) {
			m_propulsion = propval[i];
			hasprop = true;
		}
	}
 	if ((attr = el->get_attribute("altfactor")))
		m_altfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("tasfactor")))
		m_tasfactor = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("fuelfactor")))
		m_fuelfactor = Glib::Ascii::strtod(attr->get_value());
	{
		xmlpp::Node::NodeList nl(el->get_children("point"));
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			m_points.push_back(Point());
			m_points.back().load_xml(static_cast<xmlpp::Element *>(*ni), m_altfactor, m_tasfactor, m_fuelfactor);
			hasmp = hasmp && !std::isnan(m_points.back().get_mp());
		}
	}
	if (!hasprop && hasmp)
		m_propulsion = propulsion_constantspeed;
	if (hasprop && m_propulsion == propulsion_constantspeed && !hasmp)
		m_propulsion = propulsion_fixedpitch;
	build_altmap();
}

void Aircraft::Cruise::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	el->set_attribute("propulsion", to_str(m_propulsion));
	if (m_altfactor == 1) {
		el->remove_attribute("altfactor");
	} else {
		std::ostringstream oss;
		oss << m_altfactor;
		el->set_attribute("altfactor", oss.str());
	}
	if (m_tasfactor == 1) {
		el->remove_attribute("tasfactor");
	} else {
		std::ostringstream oss;
		oss << m_tasfactor;
		el->set_attribute("tasfactor", oss.str());
	}
	if (m_fuelfactor == 1) {
		el->remove_attribute("fuelfactor");
	} else {
		std::ostringstream oss;
		oss << m_fuelfactor;
		el->set_attribute("fuelfactor", oss.str());
	}
	for (points_t::const_iterator pi(m_points.begin()), pe(m_points.end()); pi != pe; ++pi)
		pi->save_xml(el->add_child("point"), m_altfactor, m_tasfactor, m_fuelfactor);
}

void Aircraft::Cruise::add_point(const Point& pt)
{
	m_points.push_back(pt);	
}

void Aircraft::Cruise::clear_points(void)
{
	m_points.clear();
}

void Aircraft::Cruise::build_altmap(void)
{
	m_altmap.clear();
	m_bhpaltmap.clear();
	for (points_t::const_iterator pi(m_points.begin()), pe(m_points.end()); pi != pe; ++pi) {
		if (std::isnan(pi->get_densityalt()) || std::isnan(pi->get_rpm()))
			continue;
		double mp(pi->get_mp());
		if (!is_constantspeed())
			mp = 0;
		if (!std::isnan(mp))
			m_altmap[pi->get_densityalt()][pi->get_rpm()][mp] = Point1(pi->get_bhp(), pi->get_tas(), pi->get_fuelflow());
		double rpm(pi->get_rpm());
		if (!std::isnan(rpm))
			m_bhpaltmap[pi->get_densityalt()][pi->get_bhp()][rpm] = Point1BHP(pi->get_mp(), pi->get_tas(), pi->get_fuelflow());
	}
	if (false) {
		for (altmap_t::const_iterator i(m_altmap.begin()), e(m_altmap.end()); i != e; ++i) {
			std::cout << "Cruise DA = " << i->first << std::endl;
			for (rpmmap_t::const_iterator i2(i->second.begin()), e2(i->second.end()); i2 != e2; ++i2) {
				std::cout << "  RPM = " << i2->first << std::endl;
				for (mpmap_t::const_iterator i3(i2->second.begin()), e3(i2->second.end()); i3 != e3; ++i3) {
					std::cout << "    MP = " << i3->first << " BHP = " << i3->second.get_bhp()
						  << " TAS = " << i3->second.get_tas() << " FF = " << i3->second.get_fuelflow() << std::endl;
				}
			}
		}
		for (bhpaltmap_t::const_iterator i(m_bhpaltmap.begin()), e(m_bhpaltmap.end()); i != e; ++i) {
			std::cout << "Cruise DA = " << i->first << std::endl;
			for (bhpbhpmap_t::const_iterator i2(i->second.begin()), e2(i->second.end()); i2 != e2; ++i2) {
				std::cout << "  BHP = " << i2->first << std::endl;
				for (bhprpmmap_t::const_iterator i3(i2->second.begin()), e3(i2->second.end()); i3 != e3; ++i3) {
					std::cout << "    RPM = " << i3->first << " MP = " << i3->second.get_mp()
						  << " TAS = " << i3->second.get_tas() << " FF = " << i3->second.get_fuelflow() << std::endl;
				}
			}
		}
	}
}

void Aircraft::Cruise::calculate(double& tas, double& fuelflow, double& bhp, double& mp, rpmmap_t::const_iterator it) const
{
	tas = fuelflow = bhp = std::numeric_limits<double>::quiet_NaN();
	if (!is_constantspeed()) {
		mp = std::numeric_limits<double>::quiet_NaN();
		if (it->second.empty())
			return;
		mpmap_t::const_iterator it1(it->second.begin());
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		bhp = it1->second.get_bhp();
		if (false)
			std::cerr << "Cruise::calculate mp: fixed pitch: rpm " << it->first << " tas* " << tas
				  << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
		return;
	}
	mpmap_t::const_iterator it1(it->second.lower_bound(mp));
	if (it1 == it->second.end()) {
		if (it1 == it->second.begin()) {
			mp = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate mp: no record" << std::endl;
			return;
		}
		--it1;
		mp = it1->first;
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		bhp = it1->second.get_bhp();
		if (false)
			std::cerr << "Cruise::calculate mp: rpm " << it->first << " mp* " << mp
				  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
		return;
	}
	if (it1 == it->second.begin()) {
		mp = it1->first;
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		bhp = it1->second.get_bhp();
		if (false)
			std::cerr << "Cruise::calculate mp: rpm " << it->first << " mp* " << mp
				  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
		return;
	}
	mpmap_t::const_iterator it2(it1);
	--it1;
	double m2((mp - it1->first) / (it2->first - it1->first));
	double m1(1 - m2);
	tas = it1->second.get_tas() * m1 + it2->second.get_tas() * m2;
	fuelflow = it1->second.get_fuelflow() * m1 + it2->second.get_fuelflow() * m2;
	bhp = it1->second.get_bhp() * m1 + it2->second.get_bhp() * m2;
	if (false)
		std::cerr << "Cruise::calculate mp: rpm " << it->first << " mp " << mp
			  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
}

void Aircraft::Cruise::calculate(double& tas, double& fuelflow, double& bhp, double& mp, double& rpm, altmap_t::const_iterator it) const
{
	rpmmap_t::const_iterator it1(it->second.lower_bound(rpm));
	if (it1 == it->second.end()) {
		if (it1 == it->second.begin()) {
			tas = fuelflow = bhp = rpm = mp = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate rpm: no record" << std::endl;
			return;
		}
		--it1;
		rpm = it1->first;
		calculate(tas, fuelflow, bhp, mp, it1);
		if (false)
			std::cerr << "Cruise::calculate rpm: da " << it->first << " rpm* " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	if (it1 == it->second.begin()) {
		rpm = it1->first;
		calculate(tas, fuelflow, bhp, mp, it1);
		if (false)
			std::cerr << "Cruise::calculate rpm: da: da " << it->first << " rpm* " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	rpmmap_t::const_iterator it2(it1);
	--it1;
	double m2((rpm - it1->first) / (it2->first - it1->first));
	double m1(1 - m2), mp1(mp), mp2(mp);
	double tas1, fuelflow1, bhp1, tas2, fuelflow2, bhp2;
	calculate(tas1, fuelflow1, bhp1, mp1, it1);
	calculate(tas2, fuelflow2, bhp2, mp2, it2);
	tas = tas1 * m1 + tas2 * m2;
	fuelflow = fuelflow1 * m1 + fuelflow2 * m2;
	bhp = bhp1 * m1 + bhp2 * m2;
	mp = mp1 * m1 + mp2 * m2;
	if (false)
		std::cerr << "Cruise::calculate rpm: da " << it->first << " rpm " << rpm << " mp* " << mp
			  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
}

void Aircraft::Cruise::calculate(double& tas, double& fuelflow, double& bhp, double& rpm, double& mp, double& da) const
{
	if (std::isnan(is_constantspeed() ? mp : rpm)) {
		calculate_bhp(tas, fuelflow, bhp, rpm, mp, da);
		return;
	}
	altmap_t::const_iterator it1(m_altmap.lower_bound(da));
	if (it1 == m_altmap.end()) {
		if (it1 == m_altmap.begin()) {
			tas = fuelflow = bhp = rpm = mp = da = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate da: no record" << std::endl;
			return;
		}
		--it1;
		da = it1->first;
		calculate(tas, fuelflow, bhp, mp, rpm, it1);
		if (false)
			std::cerr << "Cruise::calculate da: da* " << da << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	if (it1 == m_altmap.begin()) {
		da = it1->first;
		calculate(tas, fuelflow, bhp, mp, rpm, it1);
		if (false)
			std::cerr << "Cruise::calculate da: da* " << da << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	altmap_t::const_iterator it2(it1);
	--it1;
	double m2((da - it1->first) / (it2->first - it1->first));
	double m1(1 - m2), mp1(mp), mp2(mp), rpm1(rpm), rpm2(rpm);
	double tas1, fuelflow1, bhp1, tas2, fuelflow2, bhp2;
	calculate(tas1, fuelflow1, bhp1, mp1, rpm1, it1);
	calculate(tas2, fuelflow2, bhp2, mp2, rpm2, it2);
	tas = tas1 * m1 + tas2 * m2;
	fuelflow = fuelflow1 * m1 + fuelflow2 * m2;
	bhp = bhp1 * m1 + bhp2 * m2;
	mp = mp1 * m1 + mp2 * m2;
	rpm = rpm1 * m1 + rpm2 * m2;
	if (false)
		std::cerr << "Cruise::calculate da: da " << da << " rpm* " << rpm << " mp* " << mp
			  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
}

// alt -> alt
// rpm -> bhp
// mp -> rpm
// bhp -> mp

void Aircraft::Cruise::calculate_bhp(double& tas, double& fuelflow, double& rpm, double& mp, bhpbhpmap_t::const_iterator it) const
{
	tas = fuelflow = mp = std::numeric_limits<double>::quiet_NaN();
	if (!is_constantspeed()) {
		rpm = std::numeric_limits<double>::quiet_NaN();
		if (it->second.empty())
			return;
		bhprpmmap_t::const_iterator it1(it->second.begin());
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		rpm = it1->first;
		mp = it1->second.get_mp();
		if (false)
			std::cerr << "Cruise::calculate_bhp rpm: fixed pitch: bhp " << it->first << " rpm* " << rpm
				  << " tas* " << tas << " ff* " << fuelflow << " mp* " << mp << std::endl;
		return;
	}
	bhprpmmap_t::const_iterator it1(it->second.lower_bound(rpm));
	if (it1 == it->second.end()) {
		if (it1 == it->second.begin()) {
			rpm = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate_bhp rpm: no record" << std::endl;
			return;
		}
		--it1;
		rpm = it1->first;
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		mp = it1->second.get_mp();
		if (false)
			std::cerr << "Cruise::calculate_bhp rpm: bhp " << it->first << " rpm* " << rpm
				  << " tas* " << tas << " ff* " << fuelflow << " mp* " << mp << std::endl;
		return;
	}
	if (it1 == it->second.begin()) {
		rpm = it1->first;
		tas = it1->second.get_tas();
		fuelflow = it1->second.get_fuelflow();
		mp = it1->second.get_mp();
		if (false)
			std::cerr << "Cruise::calculate_bhp rpm: bhp " << it->first << " rpm* " << rpm
				  << " tas* " << tas << " ff* " << fuelflow << " mp* " << mp << std::endl;
		return;
	}
	bhprpmmap_t::const_iterator it2(it1);
	--it1;
	double m2((rpm - it1->first) / (it2->first - it1->first));
	double m1(1 - m2);
	tas = it1->second.get_tas() * m1 + it2->second.get_tas() * m2;
	fuelflow = it1->second.get_fuelflow() * m1 + it2->second.get_fuelflow() * m2;
	mp = it1->second.get_mp() * m1 + it2->second.get_mp() * m2;
	if (false)
		std::cerr << "Cruise::calculate_bhp rpm: bhp " << it->first << " rpm* " << rpm
			  << " tas " << tas << " ff* " << fuelflow << " mp* " << mp << std::endl;
}

void Aircraft::Cruise::calculate_bhp(double& tas, double& fuelflow, double& bhp, double& mp, double& rpm, bhpaltmap_t::const_iterator it) const
{
	bhpbhpmap_t::const_iterator it1(it->second.lower_bound(bhp));
	if (it1 == it->second.end()) {
		if (it1 == it->second.begin()) {
			tas = fuelflow = bhp = rpm = mp = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate_bhp bhp: no record" << std::endl;
			return;
		}
		--it1;
		bhp = it1->first;
		calculate_bhp(tas, fuelflow, rpm, mp, it1);
		if (false)
			std::cerr << "Cruise::calculate_bhp bhp: da " << it->first << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp* " << bhp << std::endl;
		return;
	}
	if (it1 == it->second.begin()) {
		bhp = it1->first;
		calculate_bhp(tas, fuelflow, rpm, mp, it1);
		if (false)
			std::cerr << "Cruise::calculate_bhp bhp: da: da " << it->first << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp* " << bhp << std::endl;
		return;
	}
	bhpbhpmap_t::const_iterator it2(it1);
	--it1;
	double m2((bhp - it1->first) / (it2->first - it1->first));
	double m1(1 - m2), rpm1(rpm), rpm2(rpm);
	double tas1, fuelflow1, mp1, tas2, fuelflow2, mp2;
	calculate_bhp(tas1, fuelflow1, rpm1, mp1, it1);
	calculate_bhp(tas2, fuelflow2, rpm2, mp2, it2);
	tas = tas1 * m1 + tas2 * m2;
	fuelflow = fuelflow1 * m1 + fuelflow2 * m2;
	rpm = rpm1 * m1 + rpm2 * m2;
	mp = mp1 * m1 + mp2 * m2;
	if (false)
		std::cerr << "Cruise::calculate_bhp bhp: da " << it->first << " rpm* " << rpm << " mp* " << mp
			  << " tas* " << tas << " ff* " << fuelflow << " bhp " << bhp << std::endl;
}

void Aircraft::Cruise::calculate_bhp(double& tas, double& fuelflow, double& bhp, double& rpm, double& mp, double& da) const
{
	bhpaltmap_t::const_iterator it1(m_bhpaltmap.lower_bound(da));
	if (it1 == m_bhpaltmap.end()) {
		if (it1 == m_bhpaltmap.begin()) {
			tas = fuelflow = bhp = rpm = mp = da = std::numeric_limits<double>::quiet_NaN();
			if (false)
				std::cerr << "Cruise::calculate_bhp da: no record" << std::endl;
			return;
		}
		--it1;
		da = it1->first;
		calculate_bhp(tas, fuelflow, bhp, mp, rpm, it1);
		if (false)
			std::cerr << "Cruise::calculate_bhp da: da* " << da << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	if (it1 == m_bhpaltmap.begin()) {
		da = it1->first;
		calculate_bhp(tas, fuelflow, bhp, mp, rpm, it1);
		if (false)
			std::cerr << "Cruise::calculate_bhp da: da* " << da << " rpm " << rpm << " mp " << mp
				  << " tas " << tas << " ff " << fuelflow << " bhp " << bhp << std::endl;
		return;
	}
	bhpaltmap_t::const_iterator it2(it1);
	--it1;
	double m2((da - it1->first) / (it2->first - it1->first));
	double m1(1 - m2),bhp1(bhp), bhp2(bhp), rpm1(rpm), rpm2(rpm);
	double tas1, fuelflow1, mp1, tas2, fuelflow2, mp2;
	calculate_bhp(tas1, fuelflow1, bhp1, mp1, rpm1, it1);
	calculate_bhp(tas2, fuelflow2, bhp2, mp2, rpm2, it2);
	tas = tas1 * m1 + tas2 * m2;
	fuelflow = fuelflow1 * m1 + fuelflow2 * m2;
	bhp = bhp1 * m1 + bhp2 * m2;
	mp = mp1 * m1 + mp2 * m2;
	rpm = rpm1 * m1 + rpm2 * m2;
	if (false)
		std::cerr << "Cruise::calculate_bhp da: da " << da << " rpm* " << rpm << " mp* " << mp
			  << " tas* " << tas << " ff* " << fuelflow << " bhp* " << bhp << std::endl;
}

Aircraft::Aircraft(void)
	: m_callsign("ZZZZZ"), m_manufacturer("Piper"), m_model("PA28R-200"), m_year("1971"), m_icaotype("P28R"),
	  m_equipment("SDFGLOR"), m_transponder("S"), m_colormarking("WHITE BLUE STRIPES"), m_emergencyradio("E"),
	  m_survival(""), m_lifejackets(""), m_dinghies(""), m_picname(""), m_crewcontact(""),
	  m_mtom(2600 * 0.45359237), m_vr0(62 * 0.86897624), m_vr1(68 * 0.86897624), m_vx0(80 * 0.86897624), m_vx1(80 * 0.86897624),
	  m_vy(95 * 0.86897624), m_vs0(64 * 0.86897624), m_vs1(70 * 0.86897624), m_vno(170 * 0.86897624), m_vne(214 * 0.86897624),
	  m_vref0(64 * 1.4 * 0.86897624), m_vref1(70 * 1.4 * 0.86897624), m_va(134 * 0.86897624), m_vfe(125 * 0.86897624),
	  m_vgearext(150 * 0.86897624), m_vgearret(125 * 0.86897624), m_vbestglide(105 * 0.86897624), m_glideslope(8.1086),
	  m_fuelmass(6), m_maxbhp(200), m_pbn(pbn_b2)
{
	//check_aircraft_type_class();
}

void Aircraft::load_xml(const xmlpp::Element *el)
{
	if (!el)
		return;
	m_otherinfo.clear();
	m_callsign.clear();
	m_manufacturer.clear();
 	m_model.clear();
 	m_year.clear();
	m_icaotype.clear();
	m_equipment.clear();
	m_transponder.clear();
	m_pbn = pbn_none;
	m_colormarking.clear();
	m_emergencyradio.clear();
	m_survival.clear();
	m_lifejackets.clear();
	m_dinghies.clear();
	m_picname.clear();
	m_crewcontact.clear();
	m_mtom = m_vr0 = m_vr1 = m_vx0 = m_vx1 = m_vy = m_vs0 = m_vs1 = m_vno = m_vne =
		m_vref0 = m_vref1 = m_va = m_vfe = m_vgearext = m_vgearret = m_vbestglide = m_glideslope =
		m_fuelmass = m_maxbhp = std::numeric_limits<double>::quiet_NaN();
	xmlpp::Attribute *attr;
 	if ((attr = el->get_attribute("callsign")))
		m_callsign = attr->get_value();
 	if ((attr = el->get_attribute("manufacturer")))
		m_manufacturer = attr->get_value();
 	if ((attr = el->get_attribute("model")))
		m_model = attr->get_value();
 	else if ((attr = el->get_attribute("aircrafttype")))
		m_model = attr->get_value();
 	if ((attr = el->get_attribute("year")))
		m_year = attr->get_value();
 	if ((attr = el->get_attribute("icaotype")))
		m_icaotype = attr->get_value();
 	if ((attr = el->get_attribute("equipment")))
		m_equipment = attr->get_value();
 	if ((attr = el->get_attribute("transponder")))
		m_transponder = attr->get_value();
 	if ((attr = el->get_attribute("pbn")))
		set_pbn(attr->get_value());
 	if ((attr = el->get_attribute("colormarking")))
		m_colormarking = attr->get_value();
 	if ((attr = el->get_attribute("emergencyradio")))
		m_emergencyradio = attr->get_value();
 	if ((attr = el->get_attribute("survival")))
		m_survival = attr->get_value();
 	if ((attr = el->get_attribute("picname")))
		m_picname = attr->get_value();
 	if ((attr = el->get_attribute("crewcontact")))
		m_crewcontact = attr->get_value();
 	if ((attr = el->get_attribute("lifejackets")))
		m_lifejackets = attr->get_value();
 	if ((attr = el->get_attribute("dinghies")))
		m_dinghies = attr->get_value();
 	if ((attr = el->get_attribute("mtom")))
		m_mtom = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vr0")))
		m_vr0 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vr1")))
		m_vr1 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vx0")))
		m_vx0 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vx1")))
		m_vx1 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vy")))
		m_vy = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vs0")))
		m_vs0 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vs1")))
		m_vs1 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vno")))
		m_vno = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vne")))
		m_vne = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vref0")))
		m_vref0 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vref1")))
		m_vref1 = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("va")))
		m_va = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vfe")))
		m_vfe = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vgearext")))
		m_vgearext = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vgearret")))
		m_vgearret = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("vbestglide")))
		m_vbestglide = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("glideslope")))
		m_glideslope = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("fuelmass")))
		m_fuelmass = Glib::Ascii::strtod(attr->get_value());
 	if ((attr = el->get_attribute("maxbhp")))
		m_maxbhp = Glib::Ascii::strtod(attr->get_value());
	xmlpp::Node::NodeList nl(el->get_children("wb"));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
		m_wb.load_xml(static_cast<const xmlpp::Element *>(*ni));
	nl = el->get_children("distances");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
		m_dist.load_xml(static_cast<const xmlpp::Element *>(*ni));
	nl = el->get_children("climb");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
		m_climb.load_xml(static_cast<const xmlpp::Element *>(*ni));
	nl = el->get_children("cruise");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni)
		m_cruise.load_xml(static_cast<const xmlpp::Element *>(*ni));
	nl = el->get_children("otherinfo");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		const xmlpp::Element *el2(static_cast<const xmlpp::Element *>(*ni));
		Glib::ustring category, text;
		if ((attr = el2->get_attribute("category")))
			category = attr->get_value();
		if (category.empty())
			continue;
		if ((attr = el2->get_attribute("text")))
			text = attr->get_value();
		otherinfo_add(category, text);
	}
	// Fix up equipment
	m_equipment = m_equipment.uppercase();
	m_transponder = m_transponder.uppercase();
	pbn_fix_equipment();
	// try to load from FS xml file
	nl = el->get_children("aircraft_setup_info");
	if (nl.begin() != nl.end())
		load_xml_fs(el);
}

void Aircraft::save_xml(xmlpp::Element *el) const
{
	if (!el)
		return;
	if (m_callsign.empty())
		el->remove_attribute("callsign");
	else
		el->set_attribute("callsign", m_callsign);
	if (m_manufacturer.empty())
		el->remove_attribute("manufacturer");
	else
		el->set_attribute("manufacturer", m_manufacturer);
	if (m_model.empty())
		el->remove_attribute("model");
	else
		el->set_attribute("model", m_model);
	if (m_year.empty())
		el->remove_attribute("year");
	else
		el->set_attribute("year", m_year);
	if (m_icaotype.empty())
		el->remove_attribute("icaotype");
	else
		el->set_attribute("icaotype", m_icaotype);
	if (m_equipment.empty())
		el->remove_attribute("equipment");
	else
		el->set_attribute("equipment", m_equipment);
	if (m_transponder.empty())
		el->remove_attribute("transponder");
	else
		el->set_attribute("transponder", m_transponder);
	if (m_pbn == pbn_none)
		el->remove_attribute("pbn");
	else
		el->set_attribute("pbn", get_pbn_string());
	if (m_colormarking.empty())
		el->remove_attribute("colormarking");
	else
		el->set_attribute("colormarking", m_colormarking);
	if (m_emergencyradio.empty())
		el->remove_attribute("emergencyradio");
	else
		el->set_attribute("emergencyradio", m_emergencyradio);
	if (m_survival.empty())
		el->remove_attribute("survival");
	else
		el->set_attribute("survival", m_survival);
	if (m_lifejackets.empty())
		el->remove_attribute("lifejackets");
	else
		el->set_attribute("lifejackets", m_lifejackets);
	if (m_dinghies.empty())
		el->remove_attribute("dinghies");
	else
		el->set_attribute("dinghies", m_dinghies);
	if (m_picname.empty())
		el->remove_attribute("picname");
	else
		el->set_attribute("picname", m_picname);
	if (m_crewcontact.empty())
		el->remove_attribute("crewcontact");
	else
		el->set_attribute("crewcontact", m_crewcontact);
	if (std::isnan(m_mtom)) {
		el->remove_attribute("mtom");
	} else {
		std::ostringstream oss;
		oss << m_mtom;
		el->set_attribute("mtom", oss.str());
	}
	if (std::isnan(m_vr0)) {
		el->remove_attribute("vr0");
	} else {
		std::ostringstream oss;
		oss << m_vr0;
		el->set_attribute("vr0", oss.str());
	}
	if (std::isnan(m_vr1)) {
		el->remove_attribute("vr1");
	} else {
		std::ostringstream oss;
		oss << m_vr1;
		el->set_attribute("vr1", oss.str());
	}
	if (std::isnan(m_vx0)) {
		el->remove_attribute("vx0");
	} else {
		std::ostringstream oss;
		oss << m_vx0;
		el->set_attribute("vx0", oss.str());
	}
	if (std::isnan(m_vx1)) {
		el->remove_attribute("vx1");
	} else {
		std::ostringstream oss;
		oss << m_vx1;
		el->set_attribute("vx1", oss.str());
	}
	if (std::isnan(m_vy)) {
		el->remove_attribute("vy");
	} else {
		std::ostringstream oss;
		oss << m_vy;
		el->set_attribute("vy", oss.str());
	}
	if (std::isnan(m_vs0)) {
		el->remove_attribute("vs0");
	} else {
		std::ostringstream oss;
		oss << m_vs0;
		el->set_attribute("vs0", oss.str());
	}
	if (std::isnan(m_vs1)) {
		el->remove_attribute("vs1");
	} else {
		std::ostringstream oss;
		oss << m_vs1;
		el->set_attribute("vs1", oss.str());
	}
	if (std::isnan(m_vno)) {
		el->remove_attribute("vno");
	} else {
		std::ostringstream oss;
		oss << m_vno;
		el->set_attribute("vno", oss.str());
	}
	if (std::isnan(m_vne)) {
		el->remove_attribute("vne");
	} else {
		std::ostringstream oss;
		oss << m_vne;
		el->set_attribute("vne", oss.str());
	}
	if (std::isnan(m_vref0)) {
		el->remove_attribute("vref0");
	} else {
		std::ostringstream oss;
		oss << m_vref0;
		el->set_attribute("vref0", oss.str());
	}
	if (std::isnan(m_vref1)) {
		el->remove_attribute("vref1");
	} else {
		std::ostringstream oss;
		oss << m_vref1;
		el->set_attribute("vref1", oss.str());
	}
	if (std::isnan(m_va)) {
		el->remove_attribute("va");
	} else {
		std::ostringstream oss;
		oss << m_va;
		el->set_attribute("va", oss.str());
	}
	if (std::isnan(m_vfe)) {
		el->remove_attribute("vfe");
	} else {
		std::ostringstream oss;
		oss << m_vfe;
		el->set_attribute("vfe", oss.str());
	}
	if (std::isnan(m_vgearext)) {
		el->remove_attribute("vgearext");
	} else {
		std::ostringstream oss;
		oss << m_vgearext;
		el->set_attribute("vgearext", oss.str());
	}
	if (std::isnan(m_vgearret)) {
		el->remove_attribute("vgearret");
	} else {
		std::ostringstream oss;
		oss << m_vgearret;
		el->set_attribute("vgearret", oss.str());
	}
	if (std::isnan(m_vbestglide)) {
		el->remove_attribute("vbestglide");
	} else {
		std::ostringstream oss;
		oss << m_vbestglide;
		el->set_attribute("vbestglide", oss.str());
	}
	if (std::isnan(m_glideslope)) {
		el->remove_attribute("glideslope");
	} else {
		std::ostringstream oss;
		oss << m_glideslope;
		el->set_attribute("glideslope", oss.str());
	}
	if (std::isnan(m_fuelmass)) {
		el->remove_attribute("fuelmass");
	} else {
		std::ostringstream oss;
		oss << m_fuelmass;
		el->set_attribute("fuelmass", oss.str());
	}
	if (std::isnan(m_maxbhp)) {
		el->remove_attribute("maxbhp");
	} else {
		std::ostringstream oss;
		oss << m_maxbhp;
		el->set_attribute("maxbhp", oss.str());
	}
	m_wb.save_xml(el->add_child("wb"));
	m_dist.save_xml(el->add_child("distances"));
	m_climb.save_xml(el->add_child("climb"));
	m_cruise.save_xml(el->add_child("cruise"));
	for (otherinfo_const_iterator_t oii(otherinfo_begin()), oie(otherinfo_end()); oii != oie; ++oii) {
		if (!oii->is_valid())
			continue;
		xmlpp::Element *el2(el->add_child("otherinfo"));
		el2->set_attribute("category", oii->get_category());
		if (!oii->get_text().empty())
			el2->set_attribute("text", oii->get_text());
	}
}

std::string Aircraft::get_text(const xmlpp::Node *n)
{
	const xmlpp::Element *el(static_cast<const xmlpp::Element *>(n));
	if (!el)
		return "";
	const xmlpp::TextNode *t(el->get_child_text());
	if (!t)
		return "";
	return t->get_content();
}

void Aircraft::load_xml_fs(const xmlpp::Element *el)
{
	Cruise::propulsion_t prop(Cruise::propulsion_fixedpitch);
	double speedfactor(1); // kts
	double fuelfactor(1); // USgal
	double massfactor(0.45359237); // lb->kg
	double armfactor(1); // In
	xmlpp::Node::NodeList nl(el->get_children("aircraft_setup_info"));
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		const xmlpp::Element *el2(static_cast<const xmlpp::Element *>(*ni));
		xmlpp::Node::NodeList nl2(el2->get_children("aircraft_type"));
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			} else if (txt == "Piston fixed pitch") {
				prop = Cruise::propulsion_fixedpitch;
			} else if (txt == "Piston variable pitch") {
				prop = Cruise::propulsion_constantspeed;
			} else {
				std::cerr << "load_xml_fs: unknown aircraft_type " << txt << std::endl;
			}
		}
		nl2 = el2->get_children("speed_distance_units");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			} else if (txt == "Knots") {
				speedfactor = 1;
			} else if (txt == "MPH") {
				speedfactor = 0.86897624;
			} else {
				std::cerr << "load_xml_fs: unknown speed_distance_units " << txt << std::endl;
			}
		}
		nl2 = el2->get_children("fuel_units");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			} else if (txt == "Gallons") {
				fuelfactor = 1;
			} else if (txt == "Liters") {
				fuelfactor = 0.26417152;
			} else {
				std::cerr << "load_xml_fs: unknown fuel_units " << txt << std::endl;
			}
		}
		nl2 = el2->get_children("weight_units");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			} else if (txt == "Pounds") {
				massfactor = 0.45359237;
			} else if (txt == "kg") {
				massfactor = 1;
			} else {
				std::cerr << "load_xml_fs: unknown aircraft_type " << txt << std::endl;
			}
		}
		nl2 = el2->get_children("cg_units");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			} else if (txt == "Inches") {
				armfactor = 1;
			} else {
				std::cerr << "load_xml_fs: unknown cg_units " << txt << std::endl;
			}
		}
	}
	double maxfuel(0);
	{
		double servceiling(0), servceilingroc(0), servceilingff(0), slroc(0), slff(0);
		nl = el->get_children("basic_info");
		for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
			const xmlpp::Element *el2(static_cast<const xmlpp::Element *>(*ni));
			xmlpp::Node::NodeList nl2(el2->get_children("make"));
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_model = txt;
			}
			nl2 = el2->get_children("model");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_icaotype = txt;
			}
			nl2 = el2->get_children("registration");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_callsign = txt;
			}
			nl2 = el2->get_children("color");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_colormarking = txt;
			}
			nl2 = el2->get_children("equipment_icao");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_equipment = txt;
			}
			nl2 = el2->get_children("fuel_capacity");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				maxfuel = Glib::Ascii::strtod(txt) * fuelfactor;
			}
			nl2 = el2->get_children("service_ceiling");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				servceiling = Glib::Ascii::strtod(txt);
			}
			nl2 = el2->get_children("service_ceiling_roc");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				servceilingroc = Glib::Ascii::strtod(txt);
			}
			nl2 = el2->get_children("sea_level_roc");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				slroc = Glib::Ascii::strtod(txt);
			}
			nl2 = el2->get_children("indicated_climb_speed");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				m_vy = Glib::Ascii::strtod(txt) * speedfactor;
			}
			nl2 = el2->get_children("sea_level_climb_fuel_rate");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				slff = Glib::Ascii::strtod(txt) * fuelfactor;
			}
			nl2 = el2->get_children("service_ceiling_climb_fuel_rate");
			for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
				std::string txt(get_text(*ni2));
				if (txt.empty()) {
					continue;
				}
				servceilingff = Glib::Ascii::strtod(txt) * fuelfactor;
			}
		}
		if (servceiling > 0 && slroc > 0) {
			m_climb.add_point(Climb::Point(0, slroc, slff));
			m_climb.add_point(Climb::Point(servceiling, servceilingroc, servceilingff));
		}
	}
	nl = el->get_children("weight_balance_info");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		const xmlpp::Element *el2(static_cast<const xmlpp::Element *>(*ni));
		xmlpp::Node::NodeList nl2(el2->get_children("fuel_density"));
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			std::string txt(get_text(*ni2));
			if (txt.empty()) {
				continue;
			}
			m_fuelmass = Glib::Ascii::strtod(txt) / fuelfactor;
		}
		nl2 = el2->get_children("main_envelope");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			const xmlpp::Element *el3(static_cast<const xmlpp::Element *>(*ni2));
			{
				WeightBalance::Envelope ev;
				xmlpp::Node::NodeList nl3(el3->get_children("point"));
				for (xmlpp::Node::NodeList::const_iterator ni3(nl3.begin()), ne3(nl3.end()); ni3 != ne3; ++ni3) {
					const xmlpp::Element *el4(static_cast<const xmlpp::Element *>(*ni3));
					double w, a;
					xmlpp::Attribute *attr;
					if ((attr = el4->get_attribute("weight")))
						w = Glib::Ascii::strtod(attr->get_value());
					else
						continue;
					if ((attr = el4->get_attribute("arm")))
						a = Glib::Ascii::strtod(attr->get_value());
					else
						continue;
					w *= massfactor;
					a *= armfactor;
					ev.add_point(a, w);
				}
				if (ev.size()) {
					double minarm, maxarm, minmass, maxmass;
					ev.get_bounds(minarm, maxarm, minmass, maxmass);
					m_mtom = maxmass;
					m_wb.add_envelope(ev);
				}
			}
		}
		nl2 = el2->get_children("moment_item_list");
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			const xmlpp::Element *el3(static_cast<const xmlpp::Element *>(*ni2));
			xmlpp::Node::NodeList nl3(el3->get_children("moment_item"));
			for (xmlpp::Node::NodeList::const_iterator ni3(nl3.begin()), ne3(nl3.end()); ni3 != ne3; ++ni3) {
				const xmlpp::Element *el4(static_cast<const xmlpp::Element *>(*ni3));
				Glib::ustring name;
				double arm(std::numeric_limits<double>::quiet_NaN());
				double massmin(arm), massmax(arm), massdflt(arm);
				WeightBalance::Element::flags_t flg(WeightBalance::Element::flag_none);
				xmlpp::Node::NodeList nl4(el4->get_children("type"));
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
					if (txt == "Fixed")
						flg |= WeightBalance::Element::flag_fixed;
				}
				nl4 = el4->get_children("desc");
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
					name = txt;
				}
				nl4 = el4->get_children("default_weight");
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
				        massdflt = Glib::Ascii::strtod(txt) * massfactor;
				}
				nl4 = el4->get_children("min_weight");
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
				        massmin = Glib::Ascii::strtod(txt) * massfactor;
				}
				nl4 = el4->get_children("max_weight");
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
				        massmax = Glib::Ascii::strtod(txt) * massfactor;
				}
				nl4 = el4->get_children("arm");
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					std::string txt(get_text(*ni2));
					if (txt.empty()) {
						continue;
					}
				        arm = Glib::Ascii::strtod(txt) * armfactor;
				}
				if (flg & WeightBalance::Element::flag_fixed)
					massmin = massmax = massdflt;
				if (!std::isnan(massmin) && !std::isnan(massmax) && !std::isnan(arm)) {
					WeightBalance::Element el(name, massmin, massmax, arm, flg);
					el.add_auto_units(1);
					m_wb.add_element(el);
				}
			}
		}
	}
	nl = el->get_children("performance_info");
	for (xmlpp::Node::NodeList::const_iterator ni(nl.begin()), ne(nl.end()); ni != ne; ++ni) {
		m_cruise = Cruise(prop);
		const xmlpp::Element *el2(static_cast<const xmlpp::Element *>(*ni));
		xmlpp::Node::NodeList nl2(el2->get_children("constant_power_model"));
		for (xmlpp::Node::NodeList::const_iterator ni2(nl2.begin()), ne2(nl2.end()); ni2 != ne2; ++ni2) {
			const xmlpp::Element *el3(static_cast<const xmlpp::Element *>(*ni2));
			xmlpp::Node::NodeList nl3(el3->get_children("power_table"));
			for (xmlpp::Node::NodeList::const_iterator ni3(nl3.begin()), ne3(nl3.end()); ni3 != ne3; ++ni3) {
				const xmlpp::Element *el4(static_cast<const xmlpp::Element *>(*ni3));
				double pwr;
				xmlpp::Attribute *attr;
				if ((attr = el4->get_attribute("power")))
					pwr = Glib::Ascii::strtod(attr->get_value());
				else
					continue;
				xmlpp::Node::NodeList nl4(el4->get_children("power_entry"));
				for (xmlpp::Node::NodeList::const_iterator ni4(nl4.begin()), ne4(nl4.end()); ni4 != ne4; ++ni4) {
					const xmlpp::Element *el5(static_cast<const xmlpp::Element *>(*ni4));
					double alt, rpm;
					if ((attr = el5->get_attribute("altitude")))
						alt = Glib::Ascii::strtod(attr->get_value());
					else
						continue;
					if ((attr = el5->get_attribute("rpm")))
						rpm = Glib::Ascii::strtod(attr->get_value());
					else
						continue;
					xmlpp::Node::NodeList nl5(el5->get_children("isa_standard"));
					for (xmlpp::Node::NodeList::const_iterator ni5(nl5.begin()), ne5(nl5.end()); ni5 != ne5; ++ni5) {
						const xmlpp::Element *el6(static_cast<const xmlpp::Element *>(*ni5));
						double mp(std::numeric_limits<double>::quiet_NaN()), fuel, tas;
						if ((attr = el6->get_attribute("mp")))
							mp = Glib::Ascii::strtod(attr->get_value());
						if ((attr = el6->get_attribute("fuel")))
							fuel = Glib::Ascii::strtod(attr->get_value()) * fuelfactor;
						else
							continue;
						if ((attr = el6->get_attribute("tas")))
							tas = Glib::Ascii::strtod(attr->get_value()) * speedfactor;
						else
							continue;
						m_maxbhp = 200;
						m_cruise.add_point(Cruise::Point(alt, rpm, mp, m_maxbhp * pwr * 0.01, tas, fuel));
					}
				}
			}
		}
	}
	pbn_fix_equipment();
}


bool Aircraft::load_file(const Glib::ustring& filename)
{
 	xmlpp::DomParser p(filename);
	if (!p)
		return false;
	xmlpp::Document *doc(p.get_document());
	xmlpp::Element *el(doc->get_root_node());
	if (el->get_name() != "aircraft")
		return false;
	load_xml(el);
	return true;
}

void Aircraft::save_file(const Glib::ustring& filename) const
{
	std::auto_ptr<xmlpp::Document> doc(new xmlpp::Document());
	save_xml(doc->create_root_node("aircraft"));
	doc->write_to_file_formatted(filename);
}

bool Aircraft::load_string(const Glib::ustring& data)
{
 	xmlpp::DomParser p;
	p.parse_memory(data);
	if (!p)
		return false;
	xmlpp::Document *doc(p.get_document());
	xmlpp::Element *el(doc->get_root_node());
	if (el->get_name() != "aircraft")
		return false;
	load_xml(el);
	return true;
}

Glib::ustring Aircraft::save_string(void) const
{
	std::auto_ptr<xmlpp::Document> doc(new xmlpp::Document());
	save_xml(doc->create_root_node("aircraft"));
	return doc->write_to_string_formatted();
}

Glib::ustring Aircraft::get_description(void) const
{
	Glib::ustring r(get_manufacturer());
	if (!r.empty() || !get_model().empty())
		r += " " + get_model();
	if (!get_year().empty())
		r += " (" + get_year() + ")";
	return r;
}

char Aircraft::get_wakecategory(void) const
{
	if (get_mtom() <= 7000)
		return 'L';
	if (get_mtom() <= 136000)
		return 'M';
	return 'H';
}

void Aircraft::otherinfo_add(const OtherInfo& oi)
{
	if (!oi.is_valid())
		return;
	std::pair<otherinfo_t::iterator,bool> ins(m_otherinfo.insert(oi));
	if (ins.second)
		return;
	Glib::ustring t(ins.first->get_text());
	if (!t.empty() || !oi.get_text().empty())
		t += " ";
	t += oi.get_text();
	OtherInfo oi2(ins.first->get_category(), t);
	m_otherinfo.erase(ins.first);
	m_otherinfo.insert(oi2);
}

const Aircraft::OtherInfo& Aircraft::otherinfo_find(const Glib::ustring& category) const
{
	static const OtherInfo notfound("", "");
	otherinfo_t::const_iterator i(m_otherinfo.find(OtherInfo(category)));
	if (i == m_otherinfo.end())
		return notfound;
	return *i;
}

// PBN Classification

Glib::ustring Aircraft::get_pbn_string(pbn_t pbn)
{
	static const char pbnstr[][4] = {
		"A1",
		"B1",
		"B2",
		"B3",
		"B4",
		"B5",
		"B6",
		"C1",
		"C2",
		"C3",
		"C4",
		"D1",
		"D2",
		"D3",
		"D4",
		"L1",
		"O1",
		"O2",
		"O3",
		"O4",
		"S1",
		"S2",
		"T1",
		"T2"
	};
	Glib::ustring ret;
	for (unsigned int i = 0; i < sizeof(pbnstr)/sizeof(pbnstr[0]); ++i)
		if (pbn & (pbn_t)(1U << i))
			ret += (const char *)pbnstr[i];
	return ret;
}

Aircraft::pbn_t Aircraft::parse_pbn(const Glib::ustring& x)
{
	pbn_t pbn(pbn_none);
	for (Glib::ustring::const_iterator i(x.begin()), e(x.end()); i != e;) {
		char c1(*i++);
		if (i == e)
			break;
		char c2(*i);
		switch (c1) {
		case 'a':
		case 'A':
			if (c2 != '1')
				break;
			++i;
			pbn |= pbn_a1;
			break;

		case 'l':
		case 'L':
			if (c2 != '1')
				break;
			++i;
			pbn |= pbn_l1;
			break;

		case 'b':
		case 'B':
			if (c2 < '1' || c2 > '6')
				break;
			++i;
			pbn |= (pbn_t)(pbn_b1 << (c2 - '1'));
			break;

		case 'c':
		case 'C':
			if (c2 < '1' || c2 > '4')
				break;
			++i;
			pbn |= (pbn_t)(pbn_c1 << (c2 - '1'));
			break;

		case 'd':
		case 'D':
			if (c2 < '1' || c2 > '4')
				break;
			++i;
			pbn |= (pbn_t)(pbn_d1 << (c2 - '1'));
			break;

		case 'o':
		case 'O':
			if (c2 < '1' || c2 > '4')
				break;
			++i;
			pbn |= (pbn_t)(pbn_o1 << (c2 - '1'));
			break;

		case 's':
		case 'S':
			if (c2 < '1' || c2 > '2')
				break;
			++i;
			pbn |= (pbn_t)(pbn_s1 << (c2 - '1'));
			break;

		case 't':
		case 'T':
			if (c2 < '1' || c2 > '2')
				break;
			++i;
			pbn |= (pbn_t)(pbn_t1 << (c2 - '1'));
			break;

		default:
			break;
		}
	}
	return pbn;
}

void Aircraft::pbn_fix_equipment(Glib::ustring& equipment, pbn_t pbn)
{
	equipment = equipment.uppercase();
	{
		Glib::ustring::size_type n(equipment.find('R'));
		if (n == Glib::ustring::npos) {
			if (pbn != pbn_none)
				equipment += 'R';
		} else {
			if (pbn == pbn_none)
				equipment.erase(n, 1);
		}
	}
	if (pbn & pbn_gnss) {
		Glib::ustring::size_type n(equipment.find('G'));
		if (n == Glib::ustring::npos)
			equipment += 'G';
	}
	if (pbn & pbn_dme) {
		Glib::ustring::size_type n(equipment.find('D'));
		if (n == Glib::ustring::npos)
			equipment += 'D';
	}
	if (pbn & pbn_dmedmeiru) {
		Glib::ustring::size_type n(equipment.find('I'));
		if (n == Glib::ustring::npos)
			equipment += 'I';
	}
	if (pbn & pbn_vordme) {
		Glib::ustring::size_type n(equipment.find_first_of("SO"));
		if (n == Glib::ustring::npos)
			equipment += 'O';
	}
}

void Aircraft::pbn_fix_equipment(std::string& equipment, pbn_t pbn)
{
	Glib::ustring e(equipment);
	pbn_fix_equipment(e, pbn);
	equipment = e;
}

double Aircraft::get_useable_fuel(void) const
{
	double d(get_fuelmass());
	if (std::isnan(d) || d <= 0)
		return std::numeric_limits<double>::quiet_NaN();
	return get_useable_fuelmass() / d;
}

double Aircraft::get_total_fuel(void) const
{
	double d(get_fuelmass());
	if (std::isnan(d) || d <= 0)
		return std::numeric_limits<double>::quiet_NaN();
	return get_total_fuelmass() / d;
}

Aircraft::Endurance::Endurance(void)
	: m_dist(std::numeric_limits<double>::quiet_NaN()),
	  m_fuel(std::numeric_limits<double>::quiet_NaN()),
	  m_tas(std::numeric_limits<double>::quiet_NaN()),
	  m_fuelflow(std::numeric_limits<double>::quiet_NaN()),
	  m_bhp(std::numeric_limits<double>::quiet_NaN()),
	  m_rpm(std::numeric_limits<double>::quiet_NaN()),
	  m_mp(std::numeric_limits<double>::quiet_NaN()),
	  m_da(std::numeric_limits<double>::quiet_NaN()),
	  m_tkoffelev(std::numeric_limits<double>::quiet_NaN()),
	  m_endurance(0)
{
}

class Aircraft::EnduranceW : public Aircraft::Endurance {
public:
	void set_dist(double x) { m_dist = x; }
	void set_fuel(double x) { m_fuel = x; }
	void set_tas(double x) { m_tas = x; }
	void set_fuelflow(double x) { m_fuelflow = x; }
	void set_bhp(double x) { m_bhp = x; }
	void set_rpm(double x) { m_rpm = x; }
	void set_mp(double x) { m_mp = x; }
	void set_da(double x) { m_da = x; }
	void set_tkoffelev(double x) { m_tkoffelev = x; }
	void set_endurance(time_t x) { m_endurance = x; }
	void add_dist(double x) { m_dist += x; }
	void add_endurance(time_t x) { m_endurance += x; }
};

Aircraft::Endurance Aircraft::compute_endurance(double da, double bhp, double rpm, double mp, double tkoffelev)
{
	EnduranceW e;
	double fuel(get_useable_fuel());
	e.set_fuel(fuel);
	if (std::isnan(fuel) || fuel <= 0)
		return e;
	if (std::isnan(da))
		return e;
	if (da < 0)
		da = 0;
	if (da > get_climb().get_ceiling())
		da = get_climb().get_ceiling();
	double tas(0), fuelflow(0);
	get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
	e.set_tas(tas);
	e.set_fuelflow(fuelflow);
	e.set_bhp(bhp);
	e.set_rpm(rpm);
	e.set_mp(mp);
	e.set_da(da);
	if (std::isnan(da) || std::isnan(fuelflow) || fuelflow <= 0)
		return e;
	e.set_dist(0);
	e.set_endurance(0);
	if (!std::isnan(tkoffelev)) {
		double t0(get_climb().altitude_to_time(tkoffelev));
		double t1(get_climb().altitude_to_time(da));
		if (!std::isnan(t0) && !std::isnan(t1) && t1 >= t0) {
			e.set_tkoffelev(get_climb().time_to_altitude(t0));
			e.add_dist(get_climb().time_to_distance(t1) - get_climb().time_to_distance(t0));
			e.add_endurance(Point::round<time_t,double>(t1 - t0));
			fuel -= get_climb().time_to_fuel(t1) - get_climb().time_to_fuel(t0);
		}
	}
	{
		double t(fuel / fuelflow);
		e.add_endurance(Point::round<time_t,double>(t * 3600.0));
		e.add_dist(t * tas);
	}
	return e;
}


// Aircraft Type Classification "Database", ICAO Doc 8643
// First character:
// L = Landplane
// S = Seaplane
// A = Amphibian
// H = Helicopter
// G = Gyrocopter
// T = Tilt-wing aircraft
// Second character:
// 1,2,3,4,6, 8 = the number of engines
// C = 2 coupled engines driving 1 propeller
// Third character:
// P = Piston Engine
// T = Turbo Engine
// J = Jet Engine
// Fourth Character:
// Approach Category
// Fifth Character:
// Wake Turbulence Category

const Aircraft::aircraft_type_class_t Aircraft::aircraft_type_class_db[] = {
	"A318L2JCM",
	"A319L2JCM",
	"A320L2JCM",
	"A321L2JCM",
	"A332L2JCH",
	"A333L2JCH",
	"A342L4JCH",
	"A343L4JCH",
	"A345L4JCH",
	"A346L4JCH",
	"A388L4JCH",
	"AN2 L1PAL",
	"B461L4JCM",
	"B462L4JCM",
	"B463L4JCM",
	"B731L2JCM",
	"B732L2JCM",
	"B733L2JCM",
	"B734L2JCM",
	"B735L2JCM",
	"B736L2JCM",
	"B737L2JCM",
	"B738L2JCM",
	"B739L2JCM",
	"B741L4JDH",
	"B742L4JDH",
	"B743L4JDH",
	"B744L4JDH",
	"B748L4JCH",
	"B752L2JCM",
	"B753L2JCM",
	"B762L2JDH",
	"B763L2JDH",
	"B764L2JDH",
	"BE40L2JBM",
	"C150L1PAL",
	"C152L1PAL",
	"C17 L4JBH",
	"C170L1PAL",
	"C172L1PAL",
	"C177L1PAL",
	"C180L1PAL",
	"C182L1PAL",
	"C25AL2JBL",
	"C25BL2JBL",
	"C5  L4JCH",
	"C501L2JBL",
	"C510L2JBL",
	"C525L2JBL",
	"C550L2JBL",
	"C551L2JBL",
	"C560L2JBM",
	"C56XL2JBM",
	"C650L2JBM",
	"C680L2JBM",
	"C77RL1PAL",
	"CL30L2JCM", // ?
	"CL60L2JCM",
	"CRJ1L2JCM",
	"CRJ2L2JCM",
	"CRJ7L2JCM",
	"CRJ9L2JCM",
	"D328L2TBM",
	"DH8DL2TCM",
	"E135L2JCM",
	"E145L2JCM",
	"E170L2JCM",
	"E190L2JCM",
	"E50PL2JBM", // Embraer Phenom 100 ?
	"F100L2JCM",
	"F2THL2JCM",
	"F70 L2JBM",
	"F900L3JBM",
	"FA20L2JBM",
	"FA50L3JCM",
	"GLEXL2JCM",
	"GLF4L2JCM",
	"GLF5L2JCM",
	"H25BL2JCM",
	"J328L2JCM",
	"LJ31L2JBM",
	"LJ35L2JCM",
	"LJ40L2JCM",
	"LJ45L2JCM",
	"LJ55L2JCM",
	"LJ60L2JCM",
	"M20LL1PAL",
	"M20PL1PAL",
	"M20TL1PAL",
	"P28AL1PAL",
	"P28BL1PAL",
	"P28RL1PAL",
	"P28TL1PAL",
	"P32RL1PAL",
	"P32TL1PAL",
	"PA18L1PAL",
	"PA27L1PAL",
	"PA31L1PAL",
	"PA32L1PAL",
	"PA34L1PAL",
	"PA38L1PAL",
	"PA44L1PAL",
	"PA46L1PAL",
	"PA47L1PAL",
	"PC12L1TAL",
	"PC6TL1TAL",
	"PC7 L1TAL",
	"PC9 L1TAL",
      	"PRM1L2JBM", // Beech Premier 1 ?
	"RJ1HL4JCM",
	"RJ70L4JCM",
	"RJ85L4JCM",
	"SB20L2TBM"
};

#if 0

const char *Aircraft::get_aircrafttypeclass(const Glib::ustring& acfttype)
{
	static const char dfltclass[6] = "L1PAL";
	std::string at(acfttype);
	at.resize(4, ' ');
	AircraftTypeClassDbCompare cmp;
        const aircraft_type_class_t *it(std::lower_bound(&aircraft_type_class_db[0],
							 &aircraft_type_class_db[sizeof(aircraft_type_class_db)/sizeof(aircraft_type_class_db[0])],
							 at.c_str(), cmp));
	if (it == &aircraft_type_class_db[sizeof(aircraft_type_class_db)/sizeof(aircraft_type_class_db[0])])
		return dfltclass;
	if (cmp(*it, acfttype.c_str()) || cmp(acfttype.c_str(), *it))
		return dfltclass;
	return (*it) + 4;
}

#else

const char *Aircraft::get_aircrafttypeclass(const Glib::ustring& acfttype)
{
	static const char dfltclass[6] = "L1PAL";
	std::string at(acfttype);
	at.resize(4, ' ');
	AircraftTypeClassDbCompare cmp;
	std::pair<const aircraft_type_class_t *, const aircraft_type_class_t *> p(std::equal_range(&aircraft_type_class_db[0],
												   &aircraft_type_class_db[sizeof(aircraft_type_class_db)/sizeof(aircraft_type_class_db[0])],
												   at.c_str(), cmp));
	if (p.first == p.second)
		return dfltclass;
	return (*p.first) + 4;
}

#endif

void Aircraft::check_aircraft_type_class(void)
{
	for (unsigned int i = 1; i < sizeof(aircraft_type_class_db)/sizeof(aircraft_type_class_db[0]); ++i) {
		AircraftTypeClassDbCompare cmp;
		if (cmp(aircraft_type_class_db[i - 1], aircraft_type_class_db[i]))
			continue;
		std::ostringstream oss;
		oss << "Aircraft Type/Class Database not ascending: " << aircraft_type_class_db[i - 1] << ", " << aircraft_type_class_db[i];
		std::cerr << oss.str() << std::endl;
		throw std::runtime_error(oss.str());
	}
}

// http://www.aircraftguru.com/aircraft/aircraft-information.php?craftid=127
// VR (no flaps, rotation): 45-55 kias (83-102 km/h)
// VR (25° flaps, rotation): 40-52 kias (74-96 km/h)
// VX (no flaps, best angle of climb): 63 kias (117 km/h)
// VX (25° flaps, best angle of climb): 44-57 kias (82-106 km/h)
// VY (best rate of climb): 79 kias (146 km/h)
// VA (maneuvering): 88-111 kias (163-206 km/h)
// VNO (max cruise): 126 kias (233 km/h)
// VNE (never exceed): 160 kias (296 km/h)
// VFE (flaps extended): 103 kias (191 km/h)
// VREF (no flaps, approach): 70 kias (130 km/h)
// VREF (40° flaps, approach): 63 kias (117 km/h)
// VS (stall, clean): 50 kias (93 km/h)
// VS0 (stall, dirty): 44 kias (82 km/h)
