//
// C++ Interface: aircraft
//
// Description: Aircraft Model
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2012, 2013, 2014
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef AIRCRAFT_H
#define AIRCRAFT_H

#include "sysdeps.h"

#include <limits>
#include <set>
#include <glibmm.h>
#include <libxml++/libxml++.h>

class FPlanRoute;
class Engine;

class Aircraft {
  protected:
	template <typename T> class Poly1D : public std::vector<T> {
	  public:
		Poly1D(void) : std::vector<T>() {}
		Poly1D(typename Poly1D<T>::size_type n) : std::vector<T>(n) {}
		Poly1D(typename Poly1D<T>::size_type n, const T& t) : std::vector<T>(n, t) {}
		Poly1D(const std::vector<T>& t) : std::vector<T>(t) {}
		template <class InputIterator> Poly1D(InputIterator a, InputIterator b) : std::vector<T>(a, b) {}

		T eval(T x) const;
		T evalderiv(T x) const;
		T inveval(T x, T tol, unsigned int maxiter, T t) const;
		T inveval(T x, T tol, unsigned int maxiter = 16) const;
		T inveval(T x) const;

		void reduce(void);

		Poly1D<T>& operator+=(T x);
		Poly1D<T>& operator-=(T x);
		Poly1D<T>& operator*=(T x);
		Poly1D<T>& operator/=(T x);
		Poly1D<T>& operator+=(const Poly1D<T>& x);
		Poly1D<T>& operator-=(const Poly1D<T>& x);
		Poly1D<T> operator*(const Poly1D<T>& x) const;

		inline Poly1D<T> operator+(T x) const { Poly1D<T> y(*this); y += x; return y; }
		inline Poly1D<T> operator-(T x) const { Poly1D<T> y(*this); y -= x; return y; }
		inline Poly1D<T> operator*(T x) const { Poly1D<T> y(*this); y *= x; return y; }
		inline Poly1D<T> operator/(T x) const { Poly1D<T> y(*this); y /= x; return y; }
		inline Poly1D<T> operator+(const Poly1D<T>& x) const { Poly1D<T> y(*this); y += x; return y; }
		inline Poly1D<T> operator-(const Poly1D<T>& x) const { Poly1D<T> y(*this); y -= x; return y; }
		inline Poly1D<T>& operator*=(const Poly1D<T>& x) { Poly1D<T> y((*this) * x); this->swap(y); return *this; }

		Poly1D<T> differentiate(void) const;
		Poly1D<T> integrate(void) const;

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

		std::ostream& save_octave(std::ostream& os, const Glib::ustring& name) const;

		template<class Archive> void hibernate(Archive& ar) {
			uint32_t n(this->size());
			ar.ioleb(n);
			if (ar.is_load())
				this->resize(n);
			for (typename Poly1D<T>::iterator i(this->begin()), e(this->end()); i != e; ++i)
				ar.io(*i);
		}
	};

  public:
	class WeightBalance {
	  public:
		class Element {
		  public:
			class Unit {
			  public:
				Unit(const Glib::ustring& name = "", double factor = 1, double offset = 0);

				const Glib::ustring& get_name(void) const { return m_name; }
				double get_factor(void) const { return m_factor; }
				double get_offset(void) const { return m_offset; }
				bool is_fixed(void) const { return get_factor() == 0; }

				void load_xml(const xmlpp::Element *el);
				void save_xml(xmlpp::Element *el) const;

				template<class Archive> void hibernate(Archive& ar) {
					ar.io(m_name);
					ar.io(m_factor);
					ar.io(m_offset);
				}

			  protected:
				Glib::ustring m_name;
				double m_factor;
				double m_offset;
			};

			typedef enum {
				flag_none = 0,
				flag_fixed = (1 << 0),
				flag_binary = (1 << 1),
				flag_avgas = (1 << 2),
				flag_oil = (1 << 3),
				flag_gear = (1 << 4),
				flag_all = flag_fixed | flag_binary | flag_avgas | flag_oil | flag_gear
			} flags_t;

			Element(const Glib::ustring& name = "", double massmin = 0, double massmax = 0, double arm = 0, flags_t flags = flag_none);
			const Glib::ustring& get_name(void) const { return m_name; }
			double get_massmin(void) const { return m_massmin; }
			double get_massmax(void) const { return m_massmax; }
			double get_arm(void) const { return m_arm; }
			flags_t get_flags(void) const { return m_flags; }

			unsigned int get_nrunits(void) const { return m_units.size(); }
			const Unit& get_unit(unsigned int idx) const { return m_units[idx]; }

			void clear_units(void);
			void add_unit(const Glib::ustring& name = "", double factor = 1, double offset = 0);
			void add_auto_units(double kg_to_unit);

			void load_xml(const xmlpp::Element *el);
			void save_xml(xmlpp::Element *el) const;

			template<class Archive> void hibernate(Archive& ar) {
				uint32_t n(m_units.size());
				ar.ioleb(n);
				if (ar.is_load())
					m_units.resize(n);
				for (units_t::iterator i(m_units.begin()), e(m_units.end()); i != e; ++i)
					i->hibernate(ar);
				ar.io(m_name);
				ar.io(m_massmin);
				ar.io(m_massmax);
				ar.io(m_arm);
				ar.iouint8(m_flags);
			}

		  protected:
			class FlagName {
			  public:
				const char *m_name;
				flags_t m_flag;
			};
			static const FlagName flagnames[];

			typedef std::vector<Unit> units_t;
			units_t m_units;
			Glib::ustring m_name;
			double m_massmin;
			double m_massmax;
			double m_arm;
			flags_t m_flags;
		};

		class ElementValue {
		  public:
			ElementValue(double value = 0, unsigned int unit = 0) : m_value(value), m_unit(unit) {}
			double get_value(void) const { return m_value; }
			unsigned int get_unit(void) const { return m_unit; }
			void set_value(double v) { m_value = v; }
			void set_unit(unsigned int u) { m_unit = u; }

			template<class Archive> void hibernate(Archive& ar) {
				ar.io(m_value);
				ar.ioleb(m_unit);
			}

		  protected:
			double m_value;
			unsigned int m_unit;
		};

		typedef std::vector<ElementValue> elementvalues_t;

		class Envelope {
		  public:
			class Point {
			  public:
				Point(double arm = 0, double mass = 0);
				double get_arm(void) const { return m_arm; }
				double get_mass(void) const { return m_mass; }

				static double area2(const Point& p0, const Point& p1, const Point& p2);
				double area2(const Point& p0, const Point& p1) const { return area2(p0, p1, *this); }

				void load_xml(const xmlpp::Element *el);
				void save_xml(xmlpp::Element *el) const;

				template<class Archive> void hibernate(Archive& ar) {
					ar.io(m_arm);
					ar.io(m_mass);
				}

			  protected:
				double m_arm;
				double m_mass;
			};

			Envelope(const Glib::ustring& name = "");

			const Glib::ustring& get_name(void) const { return m_name; }

			unsigned int size(void) const { return m_env.size(); }
			const Point& operator[](unsigned int idx) const { return m_env[idx]; }
			int windingnumber(const Point& pt) const;
			bool is_inside(const Point& pt) const { return !!windingnumber(pt); }
			void get_bounds(double& minarm, double& maxarm, double& minmass, double& maxmass) const;

			void load_xml(const xmlpp::Element *el);
			void save_xml(xmlpp::Element *el) const;

			void add_point(const Point& pt);
			void add_point(double arm = 0, double mass = 0);

		  public:
			template<class Archive> void hibernate(Archive& ar) {
				uint32_t n(m_env.size());
				ar.ioleb(n);
				if (ar.is_load())
					m_env.resize(n);
				for (env_t::iterator i(m_env.begin()), e(m_env.end()); i != e; ++i)
					i->hibernate(ar);
				ar.io(m_name);
			}

		  protected:
			typedef std::vector<Point> env_t;
			env_t m_env;
			Glib::ustring m_name;
		};

		WeightBalance(void);

		const Glib::ustring& get_armunitname(void) const { return m_armunitname; }
		const Glib::ustring& get_massunitname(void) const { return m_massunitname; }

		unsigned int get_nrelements(void) const { return m_elements.size(); }
		const Element& get_element(unsigned int idx) const { return m_elements[idx]; }
		void add_element(const Element& el);

		unsigned int get_nrenvelopes(void) const { return m_envelopes.size(); }
		const Envelope& get_envelope(unsigned int idx) const { return m_envelopes[idx]; }
		void add_envelope(const Envelope& el);

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

		double get_useable_fuelmass(void) const;
		double get_total_fuelmass(void) const;

		double get_useable_oilmass(void) const;
		double get_total_oilmass(void) const;

		template<class Archive> void hibernate(Archive& ar) {
			uint32_t n(m_elements.size());
			ar.ioleb(n);
			if (ar.is_load())
				m_elements.resize(n);
			for (elements_t::iterator i(m_elements.begin()), e(m_elements.end()); i != e; ++i)
				i->hibernate(ar);
			n = m_envelopes.size();
			ar.ioleb(n);
			if (ar.is_load())
				m_envelopes.resize(n);
			for (envelopes_t::iterator i(m_envelopes.begin()), e(m_envelopes.end()); i != e; ++i)
				i->hibernate(ar);
			ar.io(m_armunitname);
			ar.io(m_massunitname);
		}

	  protected:
		typedef std::vector<Element> elements_t;
		elements_t m_elements;
		typedef std::vector<Envelope> envelopes_t;
		envelopes_t m_envelopes;
		Glib::ustring m_armunitname;
		Glib::ustring m_massunitname;
	};

	class Distances {
	  public:
		class Distance {
		  public:
			class Point {
			  public:
				Point(double da = 0, double pa = 0, double temp = 0, double gnddist = 0, double obstdist = 0);
				
				void load_xml(const xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
					      double tempoffset = 273.15, double distfactor = 1);
				void save_xml(xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
					      double tempoffset = 273.15, double distfactor = 1) const;

				double get_densityalt(void) const;
				double get_gnddist(void) const { return m_gnddist; }
				double get_obstdist(void) const { return m_obstdist; }

				template<class Archive> void hibernate(Archive& ar) {
					ar.io(m_da);
					ar.io(m_pa);
					ar.io(m_temp);
					ar.io(m_gnddist);
					ar.io(m_obstdist);
				}

			  protected:
				double m_da;
				double m_pa;
				double m_temp;
				double m_gnddist;
				double m_obstdist;
			};

			Distance(const Glib::ustring& name = "", double vrotate = 0, double mass = 0);

			bool recalculatepoly(bool force = false);
			void add_point_da(double gnddist, double obstdist, double da);
			void add_point_pa(double gnddist, double obstdist, double pa, double temp = std::numeric_limits<double>::quiet_NaN());

			void load_xml(const xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
				      double tempoffset = 273.15, double distfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
				      double tempoffset = 273.15, double distfactor = 1) const;

			const Glib::ustring& get_name(void) const { return m_name; }
			double get_vrotate(void) const { return m_vrotate; }
			double get_mass(void) const { return m_mass; }
			const Poly1D<double>& get_gndpoly(void) const { return m_gndpoly; }
			const Poly1D<double>& get_obstpoly(void) const { return m_obstpoly; }
			double calculate_gnddist(double da) const;
			double calculate_obstdist(double da) const;
			void calculate(double& gnddist, double& obstdist, double da, double headwind = 0) const;
			void calculate(double& gnddist, double& obstdist, double da, double headwind, double mass) const;

			template<class Archive> void hibernate(Archive& ar) {
				ar.io(m_name);
				ar.io(m_vrotate);
				ar.io(m_mass);
			        m_gndpoly.hibernate(ar);
				m_obstpoly.hibernate(ar);
				uint32_t n(m_points.size());
				ar.ioleb(n);
				if (ar.is_load())
					m_points.resize(n);
				for (points_t::iterator i(m_points.begin()), e(m_points.end()); i != e; ++i)
					i->hibernate(ar);
			}

		  protected:
			Glib::ustring m_name;
			double m_vrotate;
			double m_mass;
			typedef Poly1D<double> poly_t;
			poly_t m_gndpoly;
			poly_t m_obstpoly;
			typedef std::vector<Point> points_t;
			points_t m_points;
		};

		Distances(double altfactor = 1, double tempfactor = 1, double tempoffset = 273.15, double distfactor = 1);

		unsigned int get_nrtakeoffdist(void) const { return m_takeoffdist.size(); }
		const Distance& get_takeoffdist(unsigned int i) const { return m_takeoffdist[i]; }
		Distance& get_takeoffdist(unsigned int i) { return m_takeoffdist[i]; }

		unsigned int get_nrldgdist(void) const { return m_ldgdist.size(); }
		const Distance& get_ldgdist(unsigned int i) const { return m_ldgdist[i]; }
		Distance& get_ldgdist(unsigned int i) { return m_ldgdist[i]; }

		bool recalculatepoly(bool force = false);

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

		template<class Archive> void hibernate(Archive& ar) {
			uint32_t n(m_takeoffdist.size());
			ar.ioleb(n);
			if (ar.is_load())
				m_takeoffdist.resize(n);
			for (distances_t::iterator i(m_takeoffdist.begin()), e(m_takeoffdist.end()); i != e; ++i)
				i->hibernate(ar);
			n = m_ldgdist.size();
			ar.ioleb(n);
			if (ar.is_load())
				m_ldgdist.resize(n);
			for (distances_t::iterator i(m_ldgdist.begin()), e(m_ldgdist.end()); i != e; ++i)
				i->hibernate(ar);
			ar.io(m_altfactor);
			ar.io(m_tempfactor);
			ar.io(m_tempoffset);
			ar.io(m_distfactor);
		}

	  protected:
		typedef std::vector<Distance> distances_t;
		distances_t m_takeoffdist;
		distances_t m_ldgdist;
		double m_altfactor;
		double m_tempfactor;
		double m_tempoffset;
		double m_distfactor;
	};

	class Climb {
	  public:
		class Point {
		  public:
			Point(double da = 0, double rate = 0, double fuelflow = 0);
			
			void load_xml(const xmlpp::Element *el, double altfactor = 1, double ratefactor = 1, double fuelfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double ratefactor = 1, double fuelfactor = 1) const;

			double get_densityalt(void) const { return m_da; }
			double get_rate(void) const { return m_rate; }
			double get_fuelflow(void) const { return m_fuelflow; }

			template<class Archive> void hibernate(Archive& ar) {
				ar.io(m_da);
				ar.io(m_rate);
				ar.io(m_fuelflow);
			}

		  protected:
			double m_da;
			double m_rate;
			double m_fuelflow;
		};

		Climb(double altfactor = 1, double ratefactor = 1, double fuelfactor = 1);

		void calculate(double& rate, double& fuelflow, double da);
		double get_ceiling(void) const { return m_ceiling; }
		double get_climbtime(void) const { return m_climbtime; }

		double time_to_altitude(double t) const;
		double time_to_distance(double t) const;
		double time_to_fuel(double t) const;
		double time_to_climbrate(double t) const;
		double time_to_tas(double t) const;
		double time_to_fuelflow(double t) const;
		double altitude_to_time(double a) const;
		double distance_to_time(double d) const;

		bool recalculatepoly(double vy, bool force = false);

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;
		void add_point(const Point& pt);
		void clear_points(void);

		template<class Archive> void hibernate(Archive& ar) {
			m_ratepoly.hibernate(ar);
			m_fuelflowpoly.hibernate(ar);
			m_climbtimepoly.hibernate(ar);
			m_climbaltpoly.hibernate(ar);
			m_climbdistpoly.hibernate(ar);
			m_climbfuelpoly.hibernate(ar);
			uint32_t n(m_points.size());
			ar.ioleb(n);
			if (ar.is_load())
				m_points.resize(n);
			for (points_t::iterator i(m_points.begin()), e(m_points.end()); i != e; ++i)
				i->hibernate(ar);
			ar.io(m_altfactor);
			ar.io(m_ratefactor);
			ar.io(m_fuelfactor);
			ar.io(m_ceiling);
			ar.io(m_climbtime);
		}

	  protected:
		typedef Poly1D<double> poly_t;
		poly_t m_ratepoly;
		poly_t m_fuelflowpoly;
		poly_t m_climbtimepoly;
		poly_t m_climbaltpoly;
		poly_t m_climbdistpoly;
		poly_t m_climbfuelpoly;
		typedef std::vector<Point> points_t;
		points_t m_points;
		double m_altfactor;
		double m_ratefactor;
		double m_fuelfactor;
		double m_ceiling;
		double m_climbtime;
	};

	class Cruise {
	  public:
		typedef enum {
			propulsion_fixedpitch,
			propulsion_constantspeed,
			propulsion_turboprop, // future extension
			propulsion_turbojet // future extension
		} propulsion_t;

		class Point {
		  public:
			Point(double da = 0, double rpm = 0, double mp = 0, double bhp = 0, double tas = 0, double fuelflow = 0);
			
			void load_xml(const xmlpp::Element *el, double altfactor = 1, double tasfactor = 1, double fuelfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double tasfactor = 1, double fuelfactor = 1) const;

			double get_densityalt(void) const { return m_da; }
			double get_rpm(void) const { return m_rpm; }
			double get_mp(void) const { return m_mp; }
			double get_bhp(void) const { return m_bhp; }
			double get_tas(void) const { return m_tas; }
			double get_fuelflow(void) const { return m_fuelflow; }

			template<class Archive> void hibernate(Archive& ar) {
				ar.io(m_da);
				ar.io(m_rpm);
				ar.io(m_mp);
				ar.io(m_bhp);
				ar.io(m_tas);
				ar.io(m_fuelflow);
			}

		  protected:
			double m_da;
			double m_rpm;
			double m_mp;
			double m_bhp;
			double m_tas;
			double m_fuelflow;
		};

		Cruise(propulsion_t prop = propulsion_constantspeed, double altfactor = 1, double tasfactor = 1, double fuelfactor = 1);

		propulsion_t get_propulsion(void) const { return m_propulsion; }
		bool is_constantspeed(void) const { return get_propulsion() == propulsion_constantspeed; }
		void calculate(double& tas, double& fuelflow, double& bhp, double& rpm, double& mp, double& da) const;

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;
		void add_point(const Point& pt);
		void clear_points(void);

		template<class Archive> void hibernate(Archive& ar) {
			uint32_t n(m_points.size());
			ar.ioleb(n);
			if (ar.is_load())
				m_points.resize(n);
			for (points_t::iterator i(m_points.begin()), e(m_points.end()); i != e; ++i)
				i->hibernate(ar);
			ar.io(m_altfactor);
			ar.io(m_tasfactor);
			ar.io(m_fuelfactor);
			ar.iouint8(m_propulsion);
			if (ar.is_load())
				build_altmap();
		}

	  protected:
		class Point1 {
		  public:
			Point1(double bhp = 0, double tas = 0, double fuelflow = 0) : m_bhp(bhp), m_tas(tas), m_fuelflow(fuelflow) {}
			double get_bhp(void) const { return m_bhp; }
			double get_tas(void) const { return m_tas; }
			double get_fuelflow(void) const { return m_fuelflow; }

		  protected:
			double m_bhp;
			double m_tas;
			double m_fuelflow;
		};

		class Point1BHP {
		  public:
			Point1BHP(double mp = 0, double tas = 0, double fuelflow = 0) : m_mp(mp), m_tas(tas), m_fuelflow(fuelflow) {}
			double get_mp(void) const { return m_mp; }
			double get_tas(void) const { return m_tas; }
			double get_fuelflow(void) const { return m_fuelflow; }

		  protected:
			double m_mp;
			double m_tas;
			double m_fuelflow;
		};

		typedef std::map<double,Point1> mpmap_t;
		typedef std::map<double,mpmap_t> rpmmap_t;
		typedef std::map<double,rpmmap_t> altmap_t;
		altmap_t m_altmap;
		typedef std::map<double,Point1BHP> bhprpmmap_t;
		typedef std::map<double,bhprpmmap_t> bhpbhpmap_t;
		typedef std::map<double,bhpbhpmap_t> bhpaltmap_t;
		bhpaltmap_t m_bhpaltmap;
		typedef std::vector<Point> points_t;
		points_t m_points;
		double m_altfactor;
		double m_tasfactor;
		double m_fuelfactor;
		propulsion_t m_propulsion;

		void build_altmap(void);
		void calculate(double& tas, double& fuelflow, double& bhp, double& mp, double& rpm, altmap_t::const_iterator it) const;
		void calculate(double& tas, double& fuelflow, double& bhp, double& mp, rpmmap_t::const_iterator it) const;
		void calculate_bhp(double& tas, double& fuelflow, double& bhp, double& rpm, double& mp, double& da) const;
		void calculate_bhp(double& tas, double& fuelflow, double& bhp, double& mp, double& rpm, bhpaltmap_t::const_iterator it) const;
		void calculate_bhp(double& tas, double& fuelflow, double& bhp, double& mp, bhpbhpmap_t::const_iterator it) const;
	};

	class OtherInfo {
	  public:
		OtherInfo(const Glib::ustring& category = "", const Glib::ustring& text = "") : m_category(category), m_text(text) {}
		const Glib::ustring& get_category(void) const { return m_category; }
		const Glib::ustring& get_text(void) const { return m_text; }
		bool operator<(const OtherInfo& x) const { return m_category < x.m_category; }
		bool is_valid(void) const { return !m_category.empty(); }

		template<class Archive> void hibernate(Archive& ar) {
			ar.io(m_category);
			ar.io(m_text);
		}

	  protected:
		Glib::ustring m_category;
		Glib::ustring m_text;
	};

  protected:
	typedef std::set<OtherInfo> otherinfo_t;

  public:
	typedef enum {
		pbn_a1 = 1 << 0,
		pbn_b1 = 1 << 1,
		pbn_b2 = 1 << 2,
		pbn_b3 = 1 << 3,
		pbn_b4 = 1 << 4,
		pbn_b5 = 1 << 5,
		pbn_b6 = 1 << 6,
		pbn_c1 = 1 << 7,
		pbn_c2 = 1 << 8,
		pbn_c3 = 1 << 9,
		pbn_c4 = 1 << 10,
		pbn_d1 = 1 << 11,
		pbn_d2 = 1 << 12,
		pbn_d3 = 1 << 13,
		pbn_d4 = 1 << 14,
		pbn_l1 = 1 << 15,
		pbn_o1 = 1 << 16,
		pbn_o2 = 1 << 17,
		pbn_o3 = 1 << 18,
		pbn_o4 = 1 << 19,
		pbn_s1 = 1 << 20,
		pbn_s2 = 1 << 21,
		pbn_t1 = 1 << 22,
		pbn_t2 = 1 << 23,
		pbn_rnav = pbn_b1 | pbn_b2 | pbn_b3 | pbn_b4 | pbn_b5, // B1 B2 B3 B4 B5
		pbn_gnss = pbn_b1 | pbn_b2 | pbn_c1 | pbn_c2 | pbn_d1 | pbn_d2 | pbn_o1 | pbn_o2, // B1 B2 C1 C2 D1 D2 O1 O2
		pbn_dmedme = pbn_b1 | pbn_b3 | pbn_c1 | pbn_c3 | pbn_d1 | pbn_d3 | pbn_o1 | pbn_o3, // B1 B3 C1 C3 D1 D3 O1 O3
		pbn_vordme = pbn_b1 | pbn_b4, // B1 B4
		pbn_dmedmeiru = pbn_b1 | pbn_b5 | pbn_c1 | pbn_c4 | pbn_d1 | pbn_d4 | pbn_o1 | pbn_o4, //  B1 B5 C1 C4 D1 D4 O1 O4
		pbn_loran = pbn_b5, // B5
		pbn_dme = pbn_dmedme | pbn_vordme | pbn_dmedmeiru,
		pbn_none = 0,
		pbn_all = pbn_a1 | pbn_b1 | pbn_b2 | pbn_b3 | pbn_b4 | pbn_b5 | pbn_b6
		| pbn_c1 | pbn_c2 | pbn_c3 | pbn_c4
		| pbn_d1 | pbn_d2 | pbn_d3 | pbn_d4
		| pbn_l1
		| pbn_o1 | pbn_o2 | pbn_o3 | pbn_o4
		| pbn_s1 | pbn_s2 | pbn_t1 | pbn_t2
	} pbn_t;

	Aircraft(void);

	WeightBalance& get_wb(void) { return m_wb; }
	const WeightBalance& get_wb(void) const { return m_wb; }

	Distances& get_dist(void) { return m_dist; }
	const Distances& get_dist(void) const { return m_dist; }

	Climb& get_climb(void) { return m_climb; }
	const Climb& get_climb(void) const { return m_climb; }

	Cruise& get_cruise(void) { return m_cruise; }
	const Cruise& get_cruise(void) const { return m_cruise; }

	typedef otherinfo_t::const_iterator otherinfo_const_iterator_t;
	otherinfo_const_iterator_t otherinfo_begin(void) const { return m_otherinfo.begin(); }
	otherinfo_const_iterator_t otherinfo_end(void) const { return m_otherinfo.end(); }
	void otherinfo_add(const OtherInfo& oi);
	void otherinfo_add(const Glib::ustring& category, const Glib::ustring& text = "") { otherinfo_add(OtherInfo(category, text)); }
	const OtherInfo& otherinfo_find(const Glib::ustring& category) const;

	const Glib::ustring& get_callsign(void) const { return m_callsign; }
 	const Glib::ustring& get_manufacturer(void) const { return m_manufacturer; }
 	const Glib::ustring& get_model(void) const { return m_model; }
 	const Glib::ustring& get_year(void) const { return m_year; }
	Glib::ustring get_description(void) const;
	static const char *get_aircrafttypeclass(const Glib::ustring& acfttype);
	const char *get_aircrafttypeclass(void) const { return get_aircrafttypeclass(get_icaotype()); }
	const Glib::ustring& get_icaotype(void) const { return m_icaotype; }
	const Glib::ustring& get_equipment(void) const { return m_equipment; }
	const Glib::ustring& get_transponder(void) const { return m_transponder; }
	pbn_t get_pbn(void) const { return m_pbn; }
	static Glib::ustring get_pbn_string(pbn_t pbn);
	Glib::ustring get_pbn_string(void) const { return get_pbn_string(m_pbn); }
	const Glib::ustring& get_colormarking(void) const { return m_colormarking; }
	const Glib::ustring& get_emergencyradio(void) const { return m_emergencyradio; }
	const Glib::ustring& get_survival(void) const { return m_survival; }
	const Glib::ustring& get_lifejackets(void) const { return m_lifejackets; }
	const Glib::ustring& get_dinghies(void) const { return m_dinghies; }
	const Glib::ustring& get_picname(void) const { return m_picname; }
	const Glib::ustring& get_crewcontact(void) const { return m_crewcontact; }
	char get_wakecategory(void) const;
	double get_mtom(void) const { return m_mtom; }
	double get_vr0(void) const { return m_vr0; }
	double get_vr1(void) const { return m_vr1; }
	double get_vx0(void) const { return m_vx0; }
	double get_vx1(void) const { return m_vx1; }
	double get_vy(void) const { return m_vy; }
	double get_vs0(void) const { return m_vs0; }
	double get_vs1(void) const { return m_vs1; }
	double get_vno(void) const { return m_vno; }
	double get_vne(void) const { return m_vne; }
	double get_vref0(void) const { return m_vref0; }
	double get_vref1(void) const { return m_vref1; }
	double get_va(void) const { return m_va; }
	double get_vfe(void) const { return m_vfe; }
	double get_vgearext(void) const { return m_vgearext; }
	double get_vgearret(void) const { return m_vgearret; }
	double get_vbestglide(void) const { return m_vbestglide; }
	double get_glideslope(void) const { return m_glideslope; }
	double get_fuelmass(void) const { return m_fuelmass; }
	double get_maxbhp(void) const { return m_maxbhp; }

	void set_callsign(const Glib::ustring& x) { m_callsign = x; }
 	void set_manufacturer(const Glib::ustring& x) { m_manufacturer = x; }
 	void set_model(const Glib::ustring& x) { m_model = x; }
 	void set_year(const Glib::ustring& x) { m_year = x; }
	void set_icaotype(const Glib::ustring& x) { m_icaotype = x; }
	void set_equipment(const Glib::ustring& x) { m_equipment = x; }
	void set_transponder(const Glib::ustring& x) { m_transponder = x; }
	void set_pbn(pbn_t x) { m_pbn = x; }
	void set_pbn(const Glib::ustring& x) { m_pbn = parse_pbn(x); }
	void set_colormarking(const Glib::ustring& x) { m_colormarking = x; }
	void set_emergencyradio(const Glib::ustring& x) { m_emergencyradio = x; }
	void set_survival(const Glib::ustring& x) { m_survival = x; }
	void set_lifejackets(const Glib::ustring& x) { m_lifejackets = x; }
	void set_dinghies(const Glib::ustring& x) { m_dinghies = x; }
	void set_picname(const Glib::ustring& x) { m_picname = x; }
	void set_crewcontact(const Glib::ustring& x) { m_crewcontact = x; }
	void set_mtom(double x) { m_mtom = x; }
	void set_vr0(double x) { m_vr0 = x; }
	void set_vr1(double x) { m_vr1 = x; }
	void set_vx0(double x) { m_vx0 = x; }
	void set_vx1(double x) { m_vx1 = x; }
	void set_vy(double x) { m_vy = x; }
	void set_vs0(double x) { m_vs0 = x; }
	void set_vs1(double x) { m_vs1 = x; }
	void set_vno(double x) { m_vno = x; }
	void set_vne(double x) { m_vne = x; }
	void set_vref0(double x) { m_vref0 = x; }
	void set_vref1(double x) { m_vref1 = x; }
	void set_va(double x) { m_va = x; }
	void set_vfe(double x) { m_vfe = x; }
	void set_vgearext(double x) { m_vgearext = x; }
	void set_vgearret(double x) { m_vgearret = x; }
	void set_vbestglide(double x) { m_vbestglide = x; }
	void set_glideslope(double x) { m_glideslope = x; }
	void set_fuelmass(double x) { m_fuelmass = x; }
	void set_maxbhp(double x) { m_maxbhp = x; }

	void load_xml(const xmlpp::Element *el);
	void save_xml(xmlpp::Element *el) const;

	bool load_file(const Glib::ustring& filename);
	void save_file(const Glib::ustring& filename) const;

	bool load_string(const Glib::ustring& data);
	Glib::ustring save_string(void) const;

	static pbn_t parse_pbn(const Glib::ustring& x);
	static void pbn_fix_equipment(Glib::ustring& equipment, pbn_t pbn);
	static void pbn_fix_equipment(std::string& equipment, pbn_t pbn);
	void pbn_fix_equipment(void) { pbn_fix_equipment(m_equipment, m_pbn); }

	double get_useable_fuelmass(void) const { return m_wb.get_useable_fuelmass(); }
	double get_total_fuelmass(void) const { return m_wb.get_total_fuelmass(); }

	double get_useable_fuel(void) const;
	double get_total_fuel(void) const;

	double get_useable_oilmass(void) const { return m_wb.get_useable_oilmass(); }
	double get_total_oilmass(void) const { return m_wb.get_total_oilmass(); }

	class Endurance {
	  public:
		Endurance(void);
		double get_dist(void) const { return m_dist; }
		double get_fuel(void) const { return m_fuel; }
		double get_tas(void) const { return m_tas; }
		double get_fuelflow(void) const { return m_fuelflow; }
		double get_bhp(void) const { return m_bhp; }
		double get_rpm(void) const { return m_rpm; }
		double get_mp(void) const { return m_mp; }
		double get_da(void) const { return m_da; }
		double get_tkoffelev(void) const { return m_tkoffelev; }
		time_t get_endurance(void) const { return m_endurance; }

	  protected:
		double m_dist;
		double m_fuel;
		double m_tas;
		double m_fuelflow;
		double m_bhp;
		double m_rpm;
		double m_mp;
		double m_da;
		double m_tkoffelev;
		time_t m_endurance;
	};

	Endurance compute_endurance(double da, double bhp,
				    double rpm = std::numeric_limits<double>::quiet_NaN(),
				    double mp = std::numeric_limits<double>::quiet_NaN(),
				    double tkoffelev = std::numeric_limits<double>::quiet_NaN());

	void navfplan(const Glib::ustring& templatefile, Engine& engine, const FPlanRoute& fplan,
		      double cruise_rpm = std::numeric_limits<double>::quiet_NaN(),
		      double cruise_mp = std::numeric_limits<double>::quiet_NaN(),
		      double cruise_bhp = std::numeric_limits<double>::quiet_NaN(),
		      const WeightBalance::elementvalues_t& wbev = WeightBalance::elementvalues_t()) const;

	void navfplan_gnumeric(const std::string& outfile, Engine& engine, const FPlanRoute& fplan,
			       double cruise_rpm = std::numeric_limits<double>::quiet_NaN(),
			       double cruise_mp = std::numeric_limits<double>::quiet_NaN(),
			       double cruise_bhp = std::numeric_limits<double>::quiet_NaN(),
			       const WeightBalance::elementvalues_t& wbev = WeightBalance::elementvalues_t(),
			       const std::string& templatefile = PACKAGE_DATA_DIR "/navlog.xml",
			       gint64 gfsminreftime = -1, gint64 gfsmaxreftime = -1,
			       gint64 gfsminefftime = -1, gint64 gfsmaxefftime = -1, bool hidewbpage = false) const;

	template<class Archive> void hibernate(Archive& ar) {
		m_wb.hibernate(ar);
		m_dist.hibernate(ar);
		m_climb.hibernate(ar);
		m_cruise.hibernate(ar);
		uint32_t n(m_otherinfo.size());
		ar.ioleb(n);
		if (ar.is_load()) {
			m_otherinfo.clear();
			for (; n; --n) {
				OtherInfo o;
				o.hibernate(ar);
				m_otherinfo.insert(o);
			}
		} else {
			for (otherinfo_t::iterator i(m_otherinfo.begin()), e(m_otherinfo.end()); i != e; ++i)
				const_cast<OtherInfo&>(*i).hibernate(ar);
		}
		ar.io(m_callsign);
		ar.io(m_manufacturer);
		ar.io(m_model);
		ar.io(m_year);
		ar.io(m_icaotype);
		ar.io(m_equipment);
		ar.io(m_transponder);
		ar.io(m_colormarking);
		ar.io(m_emergencyradio);
		ar.io(m_survival);
		ar.io(m_lifejackets);
		ar.io(m_dinghies);
		ar.io(m_picname);
		ar.io(m_crewcontact);
		ar.io(m_mtom);
		ar.io(m_vr0);
		ar.io(m_vr1);
		ar.io(m_vx0);
		ar.io(m_vx1);
		ar.io(m_vy);
		ar.io(m_vs0);
		ar.io(m_vs1);
		ar.io(m_vno);
		ar.io(m_vne);
		ar.io(m_vref0);
		ar.io(m_vref1);
		ar.io(m_va);
		ar.io(m_vfe);
		ar.io(m_vgearext);
		ar.io(m_vgearret);
		ar.io(m_vbestglide);
		ar.io(m_glideslope);
		ar.io(m_fuelmass);
		ar.io(m_maxbhp);
		ar.iouint8(m_pbn);
	}

  protected:
	typedef char aircraft_type_class_t[10];
	static const aircraft_type_class_t aircraft_type_class_db[];

	class AircraftTypeClassDbCompare {
	  public:
		bool operator()(const char *a, const char *b) const { return strncmp(a, b, 4) < 0; }
	};

	WeightBalance m_wb;
	Distances m_dist;
	Climb m_climb;
	Cruise m_cruise;
	otherinfo_t m_otherinfo;
	// some basic data
	Glib::ustring m_callsign;
	Glib::ustring m_manufacturer;
 	Glib::ustring m_model;
 	Glib::ustring m_year;
	Glib::ustring m_icaotype;
	Glib::ustring m_equipment;
	Glib::ustring m_transponder;
	Glib::ustring m_colormarking;
	Glib::ustring m_emergencyradio;
	Glib::ustring m_survival;
	Glib::ustring m_lifejackets;
	Glib::ustring m_dinghies;
	Glib::ustring m_picname;
	Glib::ustring m_crewcontact;
	double m_mtom;
	double m_vr0;
	double m_vr1;
	double m_vx0;
	double m_vx1;
	double m_vy;
	double m_vs0;
	double m_vs1;
	double m_vno;
	double m_vne;
	double m_vref0;
	double m_vref1;
	double m_va;
	double m_vfe;
	double m_vgearext;
	double m_vgearret;
	double m_vbestglide;
	double m_glideslope;
	double m_fuelmass;
	double m_maxbhp;
	pbn_t m_pbn;

	class NavFPlan;
	class EnduranceW;
	class GnmCellSpec;
	class GnmCellSpecs;

	static std::string get_text(const xmlpp::Node *n);
	void load_xml_fs(const xmlpp::Element *el);
	void check_aircraft_type_class(void);
};

inline Aircraft::WeightBalance::Element::flags_t operator|(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x | (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator&(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x & (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator^(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x ^ (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator~(Aircraft::WeightBalance::Element::flags_t x){ return (Aircraft::WeightBalance::Element::flags_t)~(unsigned int)x; }
inline Aircraft::WeightBalance::Element::flags_t& operator|=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x | y; return x; }
inline Aircraft::WeightBalance::Element::flags_t& operator&=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x & y; return x; }
inline Aircraft::WeightBalance::Element::flags_t& operator^=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x ^ y; return x; }

inline Aircraft::pbn_t operator|(Aircraft::pbn_t x, Aircraft::pbn_t y) { return (Aircraft::pbn_t)((unsigned int)x | (unsigned int)y); }
inline Aircraft::pbn_t operator&(Aircraft::pbn_t x, Aircraft::pbn_t y) { return (Aircraft::pbn_t)((unsigned int)x & (unsigned int)y); }
inline Aircraft::pbn_t operator^(Aircraft::pbn_t x, Aircraft::pbn_t y) { return (Aircraft::pbn_t)((unsigned int)x ^ (unsigned int)y); }
inline Aircraft::pbn_t operator~(Aircraft::pbn_t x){ return (Aircraft::pbn_t)~(unsigned int)x; }
inline Aircraft::pbn_t& operator|=(Aircraft::pbn_t& x, Aircraft::pbn_t y) { x = x | y; return x; }
inline Aircraft::pbn_t& operator&=(Aircraft::pbn_t& x, Aircraft::pbn_t y) { x = x & y; return x; }
inline Aircraft::pbn_t& operator^=(Aircraft::pbn_t& x, Aircraft::pbn_t y) { x = x ^ y; return x; }

const std::string& to_str(Aircraft::Cruise::propulsion_t prop);

#endif /* AIRCRAFT_H */
