/*****************************************************************************/

/*
 *      airdata.c  --  Air Data related Computations.
 *
 *      Copyright (C) 2012, 2014  Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include "airdata.h"
#include "fplan.h"
#include <cmath>
#include <stdexcept>
#include <limits>

const float AirData::gamma;

AirData::AirData(void)
	: m_pressalt(std::numeric_limits<float>::quiet_NaN()),
	  m_truealt(std::numeric_limits<float>::quiet_NaN()),
	  m_densityalt(std::numeric_limits<float>::quiet_NaN()),
	  m_p(std::numeric_limits<float>::quiet_NaN()),
	  m_q(std::numeric_limits<float>::quiet_NaN()),
	  m_cas(std::numeric_limits<float>::quiet_NaN()),
	  m_eas(std::numeric_limits<float>::quiet_NaN()),
	  m_tas(std::numeric_limits<float>::quiet_NaN()),
	  m_mach(std::numeric_limits<float>::quiet_NaN()),
	  m_ct(1.0),
	  m_temp(std::numeric_limits<float>::quiet_NaN()),
	  m_tempstd(std::numeric_limits<float>::quiet_NaN())
{
}

void AirData::set_qnh(float qnh)
{
	m_atmosphere.set(qnh, get_temp());
	m_densityalt = std::numeric_limits<float>::quiet_NaN();
}

void AirData::set_qnh_temp(float qnh, float temp)
{
	m_atmosphere.set(qnh, temp);
	set_pressure_altitude(m_pressalt);
}

float AirData::true_to_pressure_altitude(float truealt) const
{
	if (std::isnan(truealt))
		return truealt;
	float press, alt;
	m_atmosphere.altitude_to_pressure(&press, 0, truealt * FPlanWaypoint::ft_to_m);
	m_atmosphere.std_pressure_to_altitude(&alt, 0, press);
	return alt * FPlanWaypoint::m_to_ft;
}

float AirData::pressure_to_true_altitude(float baroalt) const
{
	if (std::isnan(baroalt))
		return baroalt;
	float press, alt;
	m_atmosphere.std_altitude_to_pressure(&press, 0, baroalt * FPlanWaypoint::ft_to_m);
	m_atmosphere.pressure_to_altitude(&alt, 0, press);
	return alt * FPlanWaypoint::m_to_ft;
}

void AirData::set_pressure_altitude(float pressalt)
{
	m_densityalt = std::numeric_limits<float>::quiet_NaN();
	if (std::isnan(pressalt)) {
		m_pressalt = m_truealt = m_p = std::numeric_limits<float>::quiet_NaN();
		return;
	}
	m_pressalt = pressalt;
	m_atmosphere.std_altitude_to_pressure(&m_p, &m_tempstd, m_pressalt * FPlanWaypoint::ft_to_m);
	float alt;
	m_atmosphere.pressure_to_altitude(&alt, &m_temp, m_p);
	m_truealt = alt * FPlanWaypoint::m_to_ft;
	compute_mach_eas_tas();
}

void AirData::set_true_altitude(float truealt)
{
	m_densityalt = std::numeric_limits<float>::quiet_NaN();
	if (std::isnan(truealt)) {
		m_truealt = m_pressalt = m_p = std::numeric_limits<float>::quiet_NaN();
		return;
	}
	m_truealt = truealt;
	m_atmosphere.altitude_to_pressure(&m_p, &m_temp, m_truealt * FPlanWaypoint::ft_to_m);
	float alt;
	m_atmosphere.std_pressure_to_altitude(&alt, &m_tempstd, m_p);
	m_pressalt = alt * FPlanWaypoint::m_to_ft;
	compute_mach_eas_tas();
}

float AirData::get_density_altitude(void) const
{
	if (!std::isnan(m_densityalt))
		return m_densityalt;
	if (std::isnan(m_p) || std::isnan(m_temp))
		return std::numeric_limits<float>::quiet_NaN();
	static const float c0(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::const_l0 * FPlanWaypoint::m_to_ft);
	static const float c1(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::std_sealevel_pressure);
	static const float c2((IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r) / (IcaoAtmosphere::const_g * IcaoAtmosphere::const_m / 1000.0 -
										      IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r));
	m_densityalt = c0 * (1 - powf(m_p / m_temp * c1, c2));
	return m_densityalt;
}

void AirData::set_density_altitude(float da)
{
	if (std::isnan(da)) {
		m_truealt = m_pressalt = m_p = m_densityalt = std::numeric_limits<float>::quiet_NaN();
		return;
	}
	static const float c0(IcaoAtmosphere::const_r / (IcaoAtmosphere::const_g * IcaoAtmosphere::const_m / 1000.0 -
							 IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r) *
			      FPlanWaypoint::m_to_ft);
	set_pressure_altitude(da + c0 * get_tempoffs());
	m_densityalt = da;
}

void AirData::set_tatprobe_ct(float ct)
{
	if (!std::isnan(ct))
		m_ct = std::max(std::min(ct, 1.0f), 0.5f);
}

void AirData::set_oat(float oat)
{
	if (std::isnan(oat) || std::isnan(m_tempstd)) {
		m_truealt = m_p = std::numeric_limits<float>::quiet_NaN();
		compute_mach_eas_tas();
		return;
	}
	set_qnh_tempoffs(get_qnh(), oat - m_tempstd);
}

float AirData::get_tat_sat_ratio(void) const
{
	if (std::isnan(m_mach))
		return std::numeric_limits<float>::quiet_NaN();
	return 1 + ((gamma - 1) / 2.0) * m_ct * m_mach * m_mach;
}

float AirData::get_rat(void) const
{
	if (std::isnan(m_temp) || m_temp <= 0)
		return std::numeric_limits<float>::quiet_NaN();
	float x(get_tat_sat_ratio());
	if (std::isnan(x))
		return std::numeric_limits<float>::quiet_NaN();
	return m_temp * x - IcaoAtmosphere::degc_to_kelvin;	
}

void AirData::set_rat(float rat)
{
	if (!std::isnan(rat)) {
		float x(get_tat_sat_ratio());
		if (false)
			std::cerr << "RAT: " << rat << " TAT/SAT ratio " << x << std::endl;
		if (!std::isnan(x)) {
			rat += IcaoAtmosphere::degc_to_kelvin;
			rat = std::max(rat, 0.f);
			set_oat(rat / x);
		}
	}
}

void AirData::set_cas(float cas)
{
	if (std::isnan(cas) || cas < 0) {
		m_q = m_cas = std::numeric_limits<float>::quiet_NaN();
	} else {
		m_cas = cas;
		cas *= (1.0 / speed_of_sound_std);
		m_q = IcaoAtmosphere::std_sealevel_pressure * (powf(((gamma - 1) / 2.0) * cas * cas + 1, gamma / (gamma - 1)) - 1);
	}
	compute_mach_eas_tas();
	if (false)
		std::cerr << "set_cas: tas " << m_tas << " eas " << m_eas << " mach " << m_mach << " q " << m_q << " cas " << m_cas
			  << " p " << m_p << " T " << m_temp << std::endl;
	if (false)
		std::cerr << "CAT/SSS: " << cas << " q " << get_q() << " p " << get_p() << " mach " << get_mach() << std::endl;
}

void AirData::set_tas(float tas)
{
	if (std::isnan(tas) || tas < 0) {
		m_q = m_cas = m_eas = m_tas = std::numeric_limits<float>::quiet_NaN();
	} else {
		m_tas = tas;
		{
			static const float c0(sqrtf(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::std_sealevel_pressure));
			m_eas = m_tas * c0 * sqrtf(m_p / m_temp);
		}
		{
			static const float c0(sqrtf(IcaoAtmosphere::degc_to_kelvin) / speed_of_sound_0degc);
			m_mach = m_tas * c0 / sqrtf(m_temp);
		}
		{
			static const float c0((gamma - 1) / 2.0);
			static const float c1(gamma / (gamma - 1));
			m_q = m_p * (powf(m_mach * m_mach * c0 + 1.f, c1) - 1.f);
		}
		{
			static const float c0(speed_of_sound_std * sqrtf(2.0 / (gamma - 1)));
			static const float c1(1 / IcaoAtmosphere::std_sealevel_pressure);
			static const float c2((gamma - 1) / gamma);
			m_cas = c0 * sqrtf(powf(m_q * c1 + 1, c2) - 1);
		}
		if (false)
			std::cerr << "set_tas: tas " << tas << " eas " << m_eas << " mach " << m_mach << " q " << m_q << " cas " << m_cas
				  << " p " << m_p << " T " << m_temp << std::endl;
	}
	if (false)
		std::cerr << "CAT/SSS: " << m_cas << " q " << get_q() << " p " << get_p() << " mach " << get_mach() << std::endl;
}

float AirData::get_lss(float temp)
{
	return speed_of_sound_0degc * sqrtf(1 + temp * (1.0 / IcaoAtmosphere::degc_to_kelvin));
}

void AirData::compute_mach_eas_tas(void)
{
	if (std::isnan(m_p) || std::isnan(m_q) || m_p <= 0) {
		m_mach = m_eas = m_tas = std::numeric_limits<float>::quiet_NaN();
		return;
	}
	m_mach = sqrtf((2 / (gamma - 1)) * (powf(m_q / m_p + 1, (gamma - 1) / gamma) - 1));
	m_eas = m_mach * sqrtf(m_p) * (speed_of_sound_std / sqrtf(IcaoAtmosphere::std_sealevel_pressure));
	if (std::isnan(m_temp) || m_temp <= 0) {
		m_tas = std::numeric_limits<float>::quiet_NaN();
	} else {
		m_tas = m_mach * sqrtf(m_temp) * (speed_of_sound_0degc / sqrtf(IcaoAtmosphere::degc_to_kelvin));
	}
}

void AirData::compute_tas(void)
{
	if (std::isnan(m_mach) || std::isnan(m_temp) || m_temp <= 0) {
		m_tas = std::numeric_limits<float>::quiet_NaN();
		return;
	}
	m_tas = m_mach * sqrtf(m_temp) * (speed_of_sound_0degc / sqrtf(IcaoAtmosphere::degc_to_kelvin));
}

void AirData::compute_cas(void)
{
	m_cas = speed_of_sound_std * sqrt(2 / (gamma - 1) * (pow(m_q * (1.0f / IcaoAtmosphere::std_sealevel_pressure) + 1, (gamma - 1) / gamma) - 1));
}
