//
// C++ Implementation: dbobjwpt
//
// Description: Database Objects: Waypoints (Intersections)
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include "dbobj.h"

const char DbBaseElements::Waypoint::usage_invalid;
const char DbBaseElements::Waypoint::usage_highlevel;
const char DbBaseElements::Waypoint::usage_lowlevel;
const char DbBaseElements::Waypoint::usage_bothlevel;
const char DbBaseElements::Waypoint::usage_rnav;
const char DbBaseElements::Waypoint::usage_terminal;
const char DbBaseElements::Waypoint::usage_vfr;
const char DbBaseElements::Waypoint::usage_user;

const char DbBaseElements::Waypoint::type_invalid;
const char DbBaseElements::Waypoint::type_unknown;
const char DbBaseElements::Waypoint::type_adhp;
const char DbBaseElements::Waypoint::type_icao;
const char DbBaseElements::Waypoint::type_coord;
const char DbBaseElements::Waypoint::type_other;

const unsigned int DbBaseElements::Waypoint::subtables_all;
const unsigned int DbBaseElements::Waypoint::subtables_none;

const char *DbBaseElements::Waypoint::db_query_string = "waypoints.ID AS 'ID',0 AS 'TABLE',waypoints.SRCID AS 'SRCID',waypoints.ICAO,waypoints.NAME,waypoints.AUTHORITY,waypoints.USAGE,waypoints.TYPE,waypoints.LAT,waypoints.LON,waypoints.LABELPLACEMENT,waypoints.MODIFIED";
const char *DbBaseElements::Waypoint::db_aux_query_string = "aux.waypoints.ID AS 'ID',1 AS 'TABLE',aux.waypoints.SRCID AS 'SRCID',aux.waypoints.ICAO,aux.waypoints.NAME,aux.waypoints.AUTHORITY,aux.waypoints.USAGE,aux.waypoints.TYPE,aux.waypoints.LAT,aux.waypoints.LON,aux.waypoints.LABELPLACEMENT,aux.waypoints.MODIFIED";
const char *DbBaseElements::Waypoint::db_text_fields[] = { "SRCID", "ICAO", "NAME", 0 };
const char *DbBaseElements::Waypoint::db_time_fields[] = { "MODIFIED", 0 };

DbBaseElements::Waypoint::Waypoint(void)
        : m_usage(usage_invalid), m_type(type_invalid)
{
}

void DbBaseElements::Waypoint::load(sqlite3x::sqlite3_cursor & cursor, sqlite3x::sqlite3_connection & db, unsigned int loadsubtables)
{
        m_id = 0;
        m_table = table_main;
        m_usage = usage_invalid;
        m_type = type_invalid;
        if (!cursor.step())
                return;
        m_id = cursor.getint(0);
        m_table = (table_t)cursor.getint(1);
        m_sourceid = cursor.getstring(2);
        m_icao = cursor.getstring(3);
        m_name = cursor.getstring(4);
        m_authority = cursor.getstring(5);
        m_usage = cursor.getint(6);
        m_type = cursor.getint(7);
        m_coord = Point(cursor.getint(9), cursor.getint(8));
        m_label_placement = (label_placement_t)cursor.getint(10);
        m_modtime = cursor.getint(11);
}

void DbBaseElements::Waypoint::update_index(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (m_table != table_main || !m_id)
                return;
        sqlite3x::sqlite3_transaction tran(db);
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints SET TILE=?2 WHERE ID=?1;");
                cmd.bind(1, (long long int)m_id);
                cmd.bind(2, (int)TileNumber::to_tilenumber(get_coord()));
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Waypoint::save(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO waypoints_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
                m_id = 0;
        }
        if (!m_id) {
                m_id = 1;
                {
                        sqlite3x::sqlite3_command cmd(db, "SELECT MAX(ID) FROM waypoints;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO waypoints (ID,SRCID) VALUES(?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_sourceid);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "INSERT INTO waypoints_rtree (ID,SWLAT,NELAT,SWLON,NELON) VALUES(?,?,?,?,?);");
                        cmd.bind(1, (long long int)m_id);
                        cmd.bind(2, m_coord.get_lat());
                        cmd.bind(3, m_coord.get_lat());
                        cmd.bind(4, m_coord.get_lon());
                        cmd.bind(5, m_coord.get_lon());
                        cmd.executenonquery();
                }
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints SET SRCID=?,ICAO=?,NAME=?,AUTHORITY=?,USAGE=?,TYPE=?,LAT=?,LON=?,LABELPLACEMENT=?,MODIFIED=?,TILE=? WHERE ID=?;");
                cmd.bind(1, m_sourceid);
                cmd.bind(2, m_icao);
                cmd.bind(3, m_name);
		cmd.bind(4, m_authority);
                cmd.bind(5, (int)m_usage);
                cmd.bind(6, (int)m_type);
                cmd.bind(7, m_coord.get_lat());
                cmd.bind(8, m_coord.get_lon());
                cmd.bind(9, m_label_placement);
                cmd.bind(10, (long long int)m_modtime);
                cmd.bind(11, (int)TileNumber::to_tilenumber(m_coord));
                cmd.bind(12, (long long int)m_id);
                cmd.executenonquery();
        }
        if (rtree) {
                sqlite3x::sqlite3_command cmd(db, "UPDATE waypoints_rtree SET SWLAT=?,NELAT=?,SWLON=?,NELON=? WHERE ID=?;");
                cmd.bind(1, m_coord.get_lat());
                cmd.bind(2, m_coord.get_lat());
                cmd.bind(3, m_coord.get_lon());
                cmd.bind(4, m_coord.get_lon());
                cmd.bind(5, (long long int)m_id);
                cmd.executenonquery();
        }
        tran.commit();
}

void DbBaseElements::Waypoint::erase(sqlite3x::sqlite3_connection & db, bool rtree, bool aux_rtree)
{
        if (!is_valid())
                return;
        sqlite3x::sqlite3_transaction tran(db);
        if (m_table == table_aux) {
                sqlite3x::sqlite3_command cmd(db, "INSERT OR REPLACE INTO waypoints_deleted (SRCID) VALUES (?);");
                cmd.bind(1, m_sourceid);
                cmd.executenonquery();
                m_table = table_main;
        } else {
                {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM waypoints WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
                if (rtree) {
                        sqlite3x::sqlite3_command cmd(db, "DELETE FROM waypoints_rtree WHERE ID=?;");
                        cmd.bind(1, (long long int)m_id);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        m_id = 0;
}

const Glib::ustring& DbBaseElements::Waypoint::get_usage_name( char usage )
{
        switch (usage) {
	default:
	case usage_invalid:
	{
		static const Glib::ustring r("Invalid");
		return r;
	}

	case usage_highlevel:
	{
		static const Glib::ustring r("High Level");
		return r;
	}

	case usage_lowlevel:
	{
		static const Glib::ustring r("Low Level");
		return r;
	}

	case usage_bothlevel:
	{
		static const Glib::ustring r("Low and High Level");
		return r;
	}

	case usage_rnav:
	{
		static const Glib::ustring r("RNAV");
		return r;
	}

	case usage_terminal:
	{
		static const Glib::ustring r("Terminal");
		return r;
	}

	case usage_vfr:
	{
		static const Glib::ustring r("VFR");
		return r;
	}

	case usage_user:
	{
		static const Glib::ustring r("User");
		return r;
	}
        }
}

const Glib::ustring& DbBaseElements::Waypoint::get_type_name(char t)
{
        switch (t) {
	default:
	case type_invalid:
	{
		static const Glib::ustring r("Invalid");
		return r;
	}

	case type_unknown:
	{
		static const Glib::ustring r("Unknown");
		return r;
	}

	case type_adhp:
	{
		static const Glib::ustring r("ADHP");
		return r;
	}

	case type_icao:
	{
		static const Glib::ustring r("ICAO");
		return r;
	}

	case type_coord:
	{
		static const Glib::ustring r("COORD");
		return r;
	}

	case type_other:
	{
		static const Glib::ustring r("Other");
		return r;
	}
	}
}

template<> void DbBase<DbBaseElements::Waypoint>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS waypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS waypoints_deleted;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Waypoint>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS waypoints (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "SRCID TEXT UNIQUE NOT NULL,"
                                              "ICAO TEXT,"
                                              "NAME TEXT,"
                                              "AUTHORITY TEXT,"
                                              "USAGE CHAR,"
                                              "TYPE CHAR,"
                                              "LAT INTEGER,"
                                              "LON INTEGER,"
                                              "LABELPLACEMENT INTEGER,"
                                              "MODIFIED INTEGER,"
                                              "TILE INTEGER);");
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS waypoints_deleted(SRCID TEXT PRIMARY KEY NOT NULL);");
                cmd.executenonquery();
        }
}

template<> void DbBase<DbBaseElements::Waypoint>::drop_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_srcid;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_icao;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_name;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_latlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_modified;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_tile;"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_lat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS waypoints_lon;"); cmd.executenonquery(); }
}

template<> void DbBase<DbBaseElements::Waypoint>::create_indices(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_srcid ON waypoints(SRCID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_icao ON waypoints(ICAO COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_name ON waypoints(NAME COLLATE NOCASE);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_latlon ON waypoints(LAT,LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_modified ON waypoints(MODIFIED);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_tile ON waypoints(TILE);"); cmd.executenonquery(); }
        // old indices
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_lat ON waypoints(LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS waypoints_lon ON waypoints(LON);"); cmd.executenonquery(); }
}

template<> const char *DbBase<DbBaseElements::Waypoint>::main_table_name = "waypoints";
template<> const char *DbBase<DbBaseElements::Waypoint>::database_file_name = "waypoints.db";
template<> const char *DbBase<DbBaseElements::Waypoint>::order_field = "SRCID";
template<> const char *DbBase<DbBaseElements::Waypoint>::delete_field = "SRCID";
template<> const bool DbBase<DbBaseElements::Waypoint>::area_data = false;
