//
// C++ Implementation: fplan
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009, 2012, 2013, 2014
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <libxml++/libxml++.h>
#include <glibmm.h>
#ifdef HAVE_GIOMM
#include <giomm.h>
#endif
#include <unistd.h>
#include <getopt.h>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <sqlite3x.hpp>
#include <sqlite3.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <stdexcept>
#include <stdarg.h>
#include <limits>
#include <cstring>
#include <glib.h>
//#include <glib/gstdio.h>

#include "fplan.h"
#include "baro.h"
#include "aircraft.h"
#include "airdata.h"
#include "wmm.h"
#include "dbobj.h"
#include "wind.h"
#include "nwxweather.h"
#include "grib2.h"
#include "dbobj.h"
#include "engine.h"

FPlanLeg::FPlanLeg(void)
        : m_fromicao(), m_toicao(), m_fromname(), m_toname(), m_fromcoord(), m_tocoord(),
          m_fromalt(0), m_toalt(0), m_fromtruealt(0), m_totruealt(0), m_fromflags(0), m_toflags(0),
          m_fromfreq(0), m_tofreq(0), m_fromtime(0), m_legtime(0), m_winddir(0), m_windspeed(0),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype_invalid), m_future(false),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
}

FPlanLeg::FPlanLeg(const Glib::ustring& fromicao, const Glib::ustring & fromname, const Point & fromcoord,
                   int16_t fromalt, uint16_t fromflags, uint32_t fromtime, uint32_t fromfreq,
                   const Glib::ustring& toicao, const Glib::ustring & toname, const Point & tocoord,
                   int16_t toalt, uint16_t toflags, uint32_t totime, uint32_t tofreq,
		   float winddir, float windspeed, float qnh, float isaoffset,
                   legtype_t legtype, bool future)
        : m_fromicao(fromicao), m_toicao(toicao), m_fromname(fromname), m_toname(toname), m_fromcoord(fromcoord), m_tocoord(tocoord),
          m_fromalt(fromalt), m_toalt(toalt), m_fromtruealt(fromalt), m_totruealt(toalt), m_fromflags(fromflags), m_toflags(toflags),
          m_fromfreq(fromfreq), m_tofreq(tofreq), m_fromtime(fromtime), m_legtime(totime - fromtime), m_winddir(winddir), m_windspeed(windspeed),
          m_dist(0), m_vdist(0), m_truetrack(0), m_var(0), m_wca(0),
          m_legtype(legtype), m_future(future),
          m_crosstrack(0), m_disttodest(0), m_vgnd(0), m_legfraction(0),
          m_altdiff(0), m_referencealt(0), m_mapalt(0),
          m_mapcoord(), m_coursetodest(0), m_gpsvalid(false), m_gpsaltvalid(false)
{
        m_dist = m_fromcoord.spheric_distance_nmi(m_tocoord);
        m_truetrack = (uint16_t)(m_fromcoord.spheric_true_course(m_tocoord) * from_deg);
        WMM wmm;
        if (wmm.compute(get_fromaltitude() * (FPlanWaypoint::ft_to_m * 1e-3), get_fromcoord(), get_fromtime_unix()))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
	baro_correction(qnh, isaoffset);
}

Glib::ustring FPlanLeg::get_fromicaoname(void) const
{
        Glib::ustring r(m_fromicao);
        if (!m_fromicao.empty() && !m_fromname.empty())
                r += " ";
        r += m_fromname;
        return r;
}

Glib::ustring FPlanLeg::get_toicaoname(void) const
{
        Glib::ustring r(m_toicao);
        if (!m_toicao.empty() && !m_toname.empty())
                r += " ";
        r += m_toname;
        return r;
}

void FPlanLeg::baro_correction(float qnh, float isaoffset)
{
        m_fromtruealt = m_fromalt;
        m_totruealt = m_toalt;
        if (!((m_fromflags | m_toflags) & FPlanWaypoint::altflag_standard))
                return;
	IcaoAtmosphere atmo(qnh, isaoffset + IcaoAtmosphere::std_sealevel_temperature);
        float press, alt;
        if (m_fromflags & FPlanWaypoint::altflag_standard) {
                atmo.std_altitude_to_pressure(&press, 0, get_fromaltitude() * FPlanWaypoint::ft_to_m);
                atmo.pressure_to_altitude(&alt, 0, press);
                m_fromtruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
        if (m_toflags & FPlanWaypoint::altflag_standard) {
                atmo.std_altitude_to_pressure(&press, 0, get_toaltitude() * FPlanWaypoint::ft_to_m);
                atmo.pressure_to_altitude(&alt, 0, press);
                m_totruealt = (int16_t)(alt * FPlanWaypoint::m_to_ft);
        }
}

int16_t FPlanLeg::baro_correction(float qnh, float isaoffset, int16_t alt, uint16_t altflags)
{
        if (!(altflags & FPlanWaypoint::altflag_standard))
                return alt;
	IcaoAtmosphere atmo(qnh, isaoffset + IcaoAtmosphere::std_sealevel_temperature);
        float press, alt1;
        atmo.std_altitude_to_pressure(&press, 0, alt * FPlanWaypoint::ft_to_m);
        atmo.pressure_to_altitude(&alt1, 0, press);
        return Point::round<int16_t,float>(alt1 * FPlanWaypoint::m_to_ft);
}

void FPlanLeg::wind_correction(float vcruise_kts)
{
	wind_correction(vcruise_kts, m_winddir, m_windspeed);
}

/*
 * Winddreieck:
 * a = vdist
 * b = vwind
 * c = vcruise
 * alpha = ?
 * beta = WCA
 * gamma = 180 + TT - winddir
 *
 * Sinussatz: vwind/sin(WCA) = vcruise/sin(winddir-TT) = vdist/sin(winddir-TT-WCA)
 * Cosinussatz: vdist^2 = vwind^2 + vcruise^2 - 2*vwind*vcruise*cos(winddir-TT-WCA)
 */

void FPlanLeg::wind_correction(float vcruise_kts, float winddir_deg, float vwind_kts)
{
        uint16_t wdirminustt = (uint16_t)(winddir_deg * from_deg);
        wdirminustt -= m_truetrack;
        m_wca = (uint16_t)(from_rad * asin(sin(wdirminustt * to_rad) * vwind_kts / vcruise_kts));
        m_vdist = sqrt(vwind_kts * vwind_kts + vcruise_kts * vcruise_kts -
                       2.0 * vwind_kts * vcruise_kts * cos(to_rad * (wdirminustt - m_wca)));
        m_vgnd = m_vdist;
        switch (m_legtype) {
                case legtype_offblock_takeoff:
                case legtype_landing_onblock:
                        m_legtime = 5 * 60;
                        break;

                case legtype_normal:
                        m_legtime = (uint32_t)(3600 * m_dist / std::max(m_vdist, 0.1f));
                        break;

                default:
                        m_legtime = 0;
                        break;
        }
}

void FPlanLeg::update( time_t curtime, const Point & coord, bool coordvalid, int16_t alt, bool altvalid, float vgnd )
{
        m_crosstrack = m_disttodest = 0.0;
        m_legfraction = 0;
        m_altdiff = m_mapalt = 0;
        m_coursetodest = m_truetrack;
        m_gpsvalid = m_gpsaltvalid = false;
        CartesianVector3Df cfrom = PolarVector3Df(m_fromcoord, 0);
        CartesianVector3Df cto = PolarVector3Df(m_tocoord, 0);
        m_gpsvalid = coordvalid;
        m_gpsaltvalid = coordvalid && altvalid;
        float t = 0;
        if (m_gpsvalid) {
                m_mapcoord = coord;
                m_vgnd = vgnd;
                if (get_legtype() == legtype_normal) {
                        CartesianVector3Df cm = PolarVector3Df(m_mapcoord, 0);
                        t = cm.nearest_fraction(cfrom, cto);
                        CartesianVector3Df cproj = cfrom + (cto - cfrom) * t;
                        m_crosstrack = (cm - cproj).length();
                        m_coursetodest = (uint16_t)(m_mapcoord.spheric_true_course(m_tocoord) * from_deg);
                        if ((int16_t)(m_coursetodest - m_truetrack) < 0)
                                m_crosstrack = -m_crosstrack;
                }
        } else {
                if (!m_legtime) {
                        t = 0.5;
                } else {
                        t = curtime - get_fromtime_unix();
                        t = std::min(std::max(t / m_legtime, 0.0f), 1.0f);
                }
                m_mapcoord = ((PolarVector3Df)(cfrom + (cto - cfrom) * t)).getcoord();
                m_vgnd = m_vdist;
        }
        m_referencealt = m_fromtruealt + (int16_t)(t * (m_totruealt - m_fromtruealt));
        m_legfraction = (uint16_t)((1U << 15) * t);
        m_mapalt = m_gpsaltvalid ? alt : m_referencealt;
        m_altdiff = m_mapalt - m_referencealt;
        m_disttodest = m_mapcoord.spheric_distance_nmi(m_tocoord);
        WMM wmm;
        if (wmm.compute(get_mapalt() * (FPlanWaypoint::ft_to_m * 1e-3), get_mapcoord(), curtime))
                m_var = (uint16_t)(wmm.get_dec() * from_deg);
}

float FPlanLeg::get_verticalspeedtodest(void) const
{
        return (get_totruealtitude() - get_mapalt()) / std::max(1.0f / 60.0f, 60.0f * get_disttodest() / std::max(get_vgnd(), 1.0f));
}

float FPlanLeg::get_verticalspeed(void) const
{
        return (get_totruealtitude() - get_fromtruealtitude()) / std::max(1.0f / 60.0f, 60.0f * get_dist() / std::max(get_vgnd(), 1.0f));
}

const time_t FPlanWaypoint::unix_minus_palmos_epoch;
const uint16_t FPlanWaypoint::altflag_standard;
const uint16_t FPlanWaypoint::altflag_climb;
const uint16_t FPlanWaypoint::altflag_descent;
const uint16_t FPlanWaypoint::altflag_turnpoint;
const uint16_t FPlanWaypoint::altflag_ifr;
const int32_t FPlanWaypoint::invalid_altitude = std::numeric_limits<int32_t>::min();
const float FPlanWaypoint::ft_to_m;
const float FPlanWaypoint::m_to_ft;
const float FPlanWaypoint::to_rad;
const float FPlanWaypoint::to_deg;
const float FPlanWaypoint::from_rad;
const float FPlanWaypoint::from_deg;
const std::string FPlanWaypoint::reader_fields = "ICAO,NAME,LON,LAT,ALT,FREQ,TIME,SAVETIME,FLAGS,PATHNAME,PATHCODE,NOTE,WINDDIR,WINDSPEED,QFF,ISAOFFS,TRUEALT,TRUETRACK,TRUEHEADING,DECLINATION,DIST,FUEL,TAS,RPM,MP,TYPE,NAVAID,FLIGHTTIME";

FPlanWaypoint::FPlanWaypoint(void)
        : m_icao(), m_name(), m_pathname(), m_note(), m_coord(), m_time(0), m_savetime(0), m_flighttime(0), m_frequency(0), m_alt(0), m_truealt(0),
	  m_dist(0), m_fuel(0), m_flags(0), m_winddir(0), m_windspeed(0), m_qff(1013.25 * 32), m_isaoffset(0), m_tt(0), m_th(0),
	  m_decl(0), m_tas(0), m_rpm(0), m_mp(0), m_pathcode(pathcode_none), m_type(type_undefined), m_navaid(navaid_invalid), m_dirty(false)
{
}

FPlanWaypoint::FPlanWaypoint(sqlite3x::sqlite3_cursor& cursor)
        : m_icao(cursor.getstring(0)), m_name(cursor.getstring(1)), m_pathname(cursor.getstring(9)), m_note(cursor.getstring(11)),
          m_coord(Point(cursor.getint(2), cursor.getint(3))), m_time(cursor.getint64(6)), m_savetime(cursor.getint64(7)),
	  m_flighttime(cursor.getint(27)), m_frequency(cursor.getint(5)), m_alt(cursor.getint(4)), m_truealt(cursor.getint(16)),
	  m_dist(cursor.getint(20)), m_fuel(cursor.getint(21)), m_flags(cursor.getint(8)),
	  m_winddir(cursor.getint(12)), m_windspeed(cursor.getint(13)), m_qff(cursor.getint(14)), m_isaoffset(cursor.getint(15)),
	  m_tt(cursor.getint(17)), m_th(cursor.getint(18)), m_decl(cursor.getint(19)), m_tas(cursor.getint(22)),
	  m_rpm(cursor.getint(23)), m_mp(cursor.getint(24)), m_pathcode(static_cast<pathcode_t>(cursor.getint(10))),
	  m_type(static_cast<type_t>(cursor.getint(25))), m_navaid(static_cast<navaid_type_t>(cursor.getint(26))), m_dirty(false)
{
	if (cursor.isnull(25))
		m_type = type_undefined;
	if (cursor.isnull(26))
		m_navaid = navaid_invalid;
}

Glib::ustring FPlanWaypoint::get_icao_name(void) const
{
        Glib::ustring r(m_icao);
        if (!m_icao.empty() && !m_name.empty())
                r += " ";
        r += m_name;
        return r;
}

const Glib::ustring& FPlanWaypoint::get_ident(void) const
{
	if (!m_icao.empty())
		return m_icao;
	return m_name;
}

const std::string& FPlanWaypoint::get_pathcode_name(pathcode_t pc)
{
	switch (pc) {
	case pathcode_none:
	{
		static const std::string r("None");
		return r;
	}

	case pathcode_sid:
	{
		static const std::string r("SID");
		return r;
	}

	case pathcode_star:
	{
		static const std::string r("STAR");
		return r;
	}

	case pathcode_approach:
	{
		static const std::string r("Approach");
		return r;
	}

	case pathcode_transition:
	{
		static const std::string r("Transition");
		return r;
	}

	case pathcode_vfrdeparture:
	{
		static const std::string r("VFR Departure");
		return r;
	}

	case pathcode_vfrarrival:
	{
		static const std::string r("VFR Arrival");
		return r;
	}

	case pathcode_vfrtransition:
	{
		static const std::string r("VFR Transition");
		return r;
	}

	case pathcode_airway:
	{
		static const std::string r("Airway");
		return r;
	}

	case pathcode_directto:
	{
		static const std::string r("Direct To");
		return r;
	}

	default:
	{
		static const std::string r("Invalid");
		return r;
	}

	}
}

bool FPlanWaypoint::set_pathcode_name(const std::string& p)
{
	for (pathcode_t pc(pathcode_none); pc <= pathcode_directto; pc = static_cast<FPlanWaypoint::pathcode_t>(pc + 1)) {
		if (get_pathcode_name(pc) != p)
			continue;
		set_pathcode(pc);
		return true;
	}
	char *cp;
	unsigned int n(strtoul(p.c_str(), &cp, 10));
	if (cp != p.c_str() && !*cp && n >= pathcode_none && n <= pathcode_directto) {
		set_pathcode(static_cast<pathcode_t>(n));
		return true;
	}
	return false;
}

bool FPlanWaypoint::is_stay(const Glib::ustring& pn, unsigned int& nr, unsigned int& tm)
{
	if (pn.size() != 10)
		return false;
	if (!((pn[0] == 'S' || pn[0] == 's') && (pn[1] == 'T' || pn[1] == 't') &&
	      (pn[2] == 'A' || pn[2] == 'a') && (pn[3] == 'Y' || pn[3] == 'y') &&
	      (pn[4] >= '1' && pn[4] <= '9') && pn[5] == '/' && std::isdigit(pn[6]) &&
	      std::isdigit(pn[7]) && std::isdigit(pn[8]) && std::isdigit(pn[9])))
		return false;
	nr = pn[4] - '1';
	tm = (((pn[6] - '0') * 10 + pn[7] - '0') * 60 + ((pn[8] - '0') * 10 + pn[9] - '0')) * 60;
	return true;
}

bool FPlanWaypoint::is_stay(const Glib::ustring& pn)
{
	unsigned int nr, tm;
	return is_stay(pn, nr, tm);
}

bool FPlanWaypoint::is_stay(unsigned int& nr, unsigned int& tm) const
{
	if (get_pathcode() != pathcode_airway)
		return false;
	return is_stay(get_pathname(), nr, tm);
}

bool FPlanWaypoint::is_stay(void) const
{
	unsigned int nr, tm;
	return is_stay(nr, tm);
}

std::string FPlanWaypoint::get_fpl_altstr(void) const
{
	if (!is_altitude_valid() || get_altitude() < 0)
		return "VFR";
	std::ostringstream oss;
	oss << (is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((get_altitude() + 50) / 100);
	return oss.str();
}

bool FPlanWaypoint::enforce_pathcode_vfrifr(void)
{
	Glib::ustring pn(get_pathname());
	{
		Glib::ustring::size_type n(pn.size()), i(0);
		while (i < n && std::isspace(pn[i]))
			++i;
		while (n > i) {
			--n;
			if (std::isspace(pn[n]))
				continue;
			++n;
			break;
		}
		pn = pn.substr(i, n - i);
	}
	FPlanWaypoint::pathcode_t pc(get_pathcode());
	if (is_ifr()) {
		pn = pn.uppercase();
		switch (pc) {
		case FPlanWaypoint::pathcode_none:
			pc = FPlanWaypoint::pathcode_airway;
			// fall through

		case FPlanWaypoint::pathcode_airway:
			if (pn.empty() || pn == "DCT")
				pc = FPlanWaypoint::pathcode_directto;
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_directto;
			break;
		}
	} else {
		switch (pc) {
		case FPlanWaypoint::pathcode_vfrdeparture:
		case FPlanWaypoint::pathcode_vfrarrival:
		case FPlanWaypoint::pathcode_vfrtransition:
			break;

		default:
			pn.clear();
			pc = FPlanWaypoint::pathcode_none;
			break;
		}
	}
	bool work(false);
	if (get_pathcode() != pc) {
		set_pathcode(pc);
		work = true;
	}
	if (get_pathname() != pn) {
		set_pathname(pn);
		work = true;
	}
	return work;
}

char FPlanWaypoint::get_ruleschar(uint16_t flags)
{
	return (flags & altflag_ifr) ? 'I' : 'V';
}

char FPlanWaypoint::get_altchar(uint16_t flags)
{
	return (flags & altflag_standard) ? 'F' : 'A';
}

float FPlanWaypoint::get_sltemp_kelvin(void) const
{
	return IcaoAtmosphere::std_sealevel_temperature + get_isaoffset_kelvin();
}

void FPlanWaypoint::set_sltemp_kelvin(float temp)
{
	set_isaoffset_kelvin(temp - IcaoAtmosphere::std_sealevel_temperature);
}

float FPlanWaypoint::get_oat_kelvin(void) const
{
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa(), get_sltemp_kelvin());
	float oat(0);
	atmo.altitude_to_pressure(0, &oat, get_altitude() * FPlanWaypoint::ft_to_m);
	return oat;
}

void FPlanWaypoint::set_oat_kelvin(float temp)
{
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa());
	float oat(0);
	atmo.altitude_to_pressure(0, &oat, get_altitude() * FPlanWaypoint::ft_to_m);
	set_isaoffset_kelvin(temp - oat);
}

float FPlanWaypoint::get_pressure_altitude(void) const
{
	if (get_flags() & altflag_standard)
		return get_altitude();
	IcaoAtmosphere atmo(get_qff_hpa(), get_sltemp_kelvin());
        float p(0), temp(0), tempstd(0), alt(0);
        atmo.altitude_to_pressure(&p, &temp, get_altitude() * FPlanWaypoint::ft_to_m);
        atmo.std_pressure_to_altitude(&alt, &tempstd, p);
        return alt * FPlanWaypoint::m_to_ft;
}

float FPlanWaypoint::get_true_altitude(void) const
{
	if (!(get_flags() & altflag_standard))
		return get_altitude();
	IcaoAtmosphere atmo(get_qff_hpa(), get_sltemp_kelvin());
	float p(0), tempstd(0), temp(0), alt(0);
	atmo.std_altitude_to_pressure(&p, &tempstd, get_altitude() * FPlanWaypoint::ft_to_m);
        atmo.pressure_to_altitude(&alt, &temp, p);
        return alt * FPlanWaypoint::m_to_ft;
}

float FPlanWaypoint::get_density_altitude(void) const
{
        static const float c0(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::const_l0 * FPlanWaypoint::m_to_ft);
        static const float c1(IcaoAtmosphere::std_sealevel_temperature / IcaoAtmosphere::std_sealevel_pressure);
        static const float c2((IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r) / (IcaoAtmosphere::const_g * IcaoAtmosphere::const_m / 1000.0 -
                                                                                      IcaoAtmosphere::const_l0 * IcaoAtmosphere::const_r));
	IcaoAtmosphere atmo((get_flags() & altflag_standard) ? IcaoAtmosphere::std_sealevel_pressure : get_qff_hpa());
	float p(0), temp(0);
	atmo.altitude_to_pressure(&p, &temp, get_altitude() * FPlanWaypoint::ft_to_m);
        return c0 * (1 - powf(p / temp * c1, c2));
}

const std::string& FPlanWaypoint::get_type_string(type_t typ, navaid_type_t navaid)
{
	switch (typ) {
	case type_airport:
	{
		static const std::string r("ARPT");
		return r;
	}

	case type_navaid:
		switch (navaid) {
		default:
		case navaid_invalid:
		{
			static const std::string r("NAVAID");
			return r;
		}

		case navaid_vor:
		{
			static const std::string r("VOR");
			return r;
		}

		case navaid_vordme:
		{
			static const std::string r("VORDME");
			return r;
		}

		case navaid_dme:
		{
			static const std::string r("DME");
			return r;
		}

		case navaid_ndb:
		{
			static const std::string r("NDB");
			return r;
		}

		case navaid_ndbdme:
		{
			static const std::string r("NDBDME");
			return r;
		}

		case navaid_vortac:
		{
			static const std::string r("VORTAC");
			return r;
		}

		case navaid_tacan:
		{
			static const std::string r("TACAN");
			return r;
		}
		}


	case type_intersection:
	{
		static const std::string r("INT");
		return r;
	}

	case type_mapelement:
	{
		static const std::string r("MAP");
		return r;
	}

	default:
	case type_undefined:
	{
		static const std::string r("UNDEF");
		return r;
	}

	case type_fplwaypoint:
	{
		static const std::string r("FPLWPT");
		return r;
	}

	case type_vfrreportingpt:
	{
		static const std::string r("VFRREPPT");
		return r;
	}

	case type_user:
	{
		static const std::string r("USER");
		return r;
	}
	}
}

bool FPlanWaypoint::set_type_string(const std::string& x)
{
	std::string xu(Glib::ustring(x).uppercase());
	for (type_t typ = type_airport; typ <= type_user; typ = (type_t)(typ + 1)) {
		if (typ != type_navaid) {
			if (get_type_string(typ) != xu)
				continue;
			set_type(typ);
			return true;
		}
		for (navaid_type_t nav = navaid_invalid; nav <= navaid_tacan; nav = (navaid_type_t)(nav + 1)) {
			if (get_type_string(typ, nav) != xu)
				continue;
			set_type(typ, nav);
			return true;
		}
	}
	return false;
}

void FPlanWaypoint::set_airport(const AirportsDb::Airport& el)
{
	set_type(type_airport);
	set_icao(el.get_icao());
	set_name(el.get_name());
	set_pathcode(FPlanWaypoint::pathcode_none);
	set_pathname("");
	set_coord(el.get_coord());
	set_altitude(el.get_elevation());
	set_flags(get_flags() & ~FPlanWaypoint::altflag_standard);
	set_frequency(0);
	Glib::ustring note("Airport Type: " + el.get_type_string());
	if (el.get_nrrwy() > 0)
		note += "\n\nRunways:";
	for (unsigned int nr(0); nr < el.get_nrrwy(); ++nr) {
		const AirportsDb::Airport::Runway& rwy(el.get_rwy(nr));
		if (rwy.get_flags() & AirportsDb::Airport::Runway::flag_le_usable) {
			std::ostringstream oss;
			oss << "\n  " << std::setw(4) << rwy.get_ident_le() << ' '
			    << std::setw(4) << Conversions::track_str(rwy.get_le_hdg() * FPlanLeg::to_deg) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_length() * Point::ft_to_m_dbl) << 'x'
			    << std::fixed << std::setprecision(0) << (rwy.get_width() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_le_tda() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_le_lda() * Point::ft_to_m_dbl) << ' '
			    << std::setw(4) << Conversions::alt_str(rwy.get_le_elev(), 0) << ' ' << rwy.get_surface();
			note += oss.str();
			Glib::ustring lights;
			for (unsigned int i = 0; i < AirportsDb::Airport::Runway::nr_lights; ++i) {
				AirportsDb::Airport::Runway::light_t l(rwy.get_le_light(i));
				if (l == AirportsDb::Airport::Runway::light_off)
					continue;
				if (!lights.empty())
					lights += "; ";
				lights += AirportsDb::Airport::Runway::get_light_string(l);
			}
			note += " " + lights;
		}
		if (rwy.get_flags() & AirportsDb::Airport::Runway::flag_he_usable) {
			std::ostringstream oss;
			oss << "\n  " << std::setw(4) << rwy.get_ident_he() << ' '
			    << std::setw(4) << Conversions::track_str(rwy.get_he_hdg() * FPlanLeg::to_deg) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_length() * Point::ft_to_m_dbl) << 'x'
			    << std::fixed << std::setprecision(0) << (rwy.get_width() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_he_tda() * Point::ft_to_m_dbl) << ' '
			    << std::fixed << std::setw(4) << std::setprecision(0) << (rwy.get_he_lda() * Point::ft_to_m_dbl) << ' '
			    << std::setw(4) << Conversions::alt_str(rwy.get_he_elev(), 0) << ' ' << rwy.get_surface();
			note += oss.str();
			Glib::ustring lights;
			for (unsigned int i = 0; i < AirportsDb::Airport::Runway::nr_lights; ++i) {
				AirportsDb::Airport::Runway::light_t l(rwy.get_he_light(i));
				if (l == AirportsDb::Airport::Runway::light_off)
					continue;
				if (!lights.empty())
					lights += "; ";
				lights += AirportsDb::Airport::Runway::get_light_string(l);
			}
			note += " " + lights;
		}
	}
	if (el.get_nrhelipads() > 0)
		note += "\n\nHelipads:";
	for (unsigned int nr(0); nr < el.get_nrhelipads(); ++nr) {
		const AirportsDb::Airport::Helipad& helipad(el.get_helipad(nr));
		if (!(helipad.get_flags() & AirportsDb::Airport::Helipad::flag_usable))
			continue;
		 std::ostringstream oss;
		 oss << "\n  " << std::setw(4) << helipad.get_ident() << ' '
		     << std::setw(4) << Conversions::track_str(helipad.get_hdg() * FPlanLeg::to_deg) << ' '
		     << std::fixed << std::setw(4) << std::setprecision(0) << (helipad.get_length() * Point::ft_to_m_dbl) << 'x'
		     << std::fixed << std::setprecision(0) << (helipad.get_width() * Point::ft_to_m_dbl) << ' '
		     << std::setw(4) << Conversions::alt_str(helipad.get_elev(), 0);
		 note += oss.str();
	}
	if (el.get_nrcomm() > 0) {
		note += "\n\nComm:";
		unsigned int prio(0);
		Glib::ustring twr("twr");
		twr = twr.casefold();
		for (unsigned int nr(0); nr < el.get_nrcomm(); ++nr) {
			const AirportsDb::Airport::Comm& comm(el.get_comm(nr));
			note += "\n  " + comm.get_name() + " " + Conversions::freq_str(comm[0]) + " " + Conversions::freq_str(comm[1]) + " " +
				Conversions::freq_str(comm[2]) + " " + Conversions::freq_str(comm[3]) + " " + Conversions::freq_str(comm[4]) + " " +
				comm.get_sector() + " " + comm.get_oprhours();
			if (!comm[0])
				continue;
			unsigned int p(1);
			Glib::ustring name(comm.get_name().casefold());
			if (name == twr)
				p = 3;
			else if (name.find(twr) != Glib::ustring::npos)
				p = 2;
			if (p <= prio)
				continue;
			set_frequency(comm[0]);
		}
	}
	if (!el.get_vfrrmk().empty())
		note += "\n\n" + el.get_vfrrmk();
	set_note(note);
}

void FPlanWaypoint::set_navaid(const NavaidsDb::Navaid& el)
{
	set_type(type_navaid, static_cast<navaid_type_t>(el.get_navaid_type()));
	set_flags(get_flags() & ~FPlanWaypoint::altflag_standard);
	set_icao(el.get_icao());
	set_name(el.get_name());
	set_pathcode(FPlanWaypoint::pathcode_none);
	set_pathname("");
	set_coord(el.get_coord());
	set_altitude(el.get_elev());
	set_frequency(el.get_frequency());
	std::ostringstream range;
	range << el.get_range();
	Glib::ustring note(el.get_navaid_typename() + " Range " + range.str() + " ");
	if (el.is_inservice())
		note += "in";
	else
		note += "out of";
	note += " service";
	set_note(note);
}

void FPlanWaypoint::set_waypoint(const WaypointsDb::Waypoint& el)
{
	set_type(type_intersection);
	set_icao(el.get_icao());
	set_name(el.get_name());
	set_pathcode(FPlanWaypoint::pathcode_none);
	set_pathname("");
	set_coord(el.get_coord());
	set_note(el.get_usage_name());
}

void FPlanWaypoint::set_mapelement(const MapelementsDb::Mapelement& el)
{
	set_type(type_mapelement);
	set_name(el.get_name());
	set_coord(el.get_coord());
	set_note(el.get_typename());
	set_pathcode(FPlanWaypoint::pathcode_none);
	set_pathname("");
}

void FPlanWaypoint::set_wpt(const FPlanWaypoint& el)
{
	*this = el;
}

FPlanRoute::GFSResult::GFSResult(void)
	: m_minreftime(std::numeric_limits<gint64>::max()), m_maxreftime(std::numeric_limits<gint64>::min()),
	  m_minefftime(std::numeric_limits<gint64>::max()), m_maxefftime(std::numeric_limits<gint64>::min()),
	  m_modified(false), m_partial(false)
{
}

FPlanRoute::FPlanRoute(FPlan & fpp)
        : m_fp(&fpp), m_id(0), m_note(), m_time_offblock(0), m_savetime_offblock(0), m_time_onblock(0), m_savetime_onblock(0), m_curwpt(0), m_wpts(), m_dirty(false)
{
}

FPlanWaypoint& FPlanRoute::operator[](uint32_t nr)
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return m_wpts[nr];
}

const FPlanWaypoint& FPlanRoute::operator[](uint32_t nr) const
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        return m_wpts[nr];
}

void FPlanRoute::insert_wpt(uint32_t nr, const FPlanWaypoint & wpt)
{
	m_dirty = true;
        if (nr >= m_wpts.size()) {
                m_wpts.push_back(wpt);
                return;
        }
        m_wpts.insert(m_wpts.begin() + nr, wpt);
        if (m_id && m_fp) {
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR + 100000 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR - 99999 WHERE PLANID = ? AND NR >= ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT INTO waypoints (PLANID,NR) VALUES (?,?);");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
}

void FPlanRoute::delete_wpt(uint32_t nr)
{
        if (nr >= m_wpts.size())
                throw std::runtime_error("Waypoint index out of range");
        if (m_id && m_fp) {
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR + 99999 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET NR = NR - 100000 WHERE PLANID = ? AND NR > ?;");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        m_wpts.erase(m_wpts.begin() + nr);
        m_dirty = true;
}

void FPlanRoute::clear_wpt(void)
{
        if (!m_wpts.size())
		return;
        if (m_id && m_fp) {
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ?;");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        m_wpts.clear();
        m_dirty = true;
}

bool FPlanRoute::is_dirty(void) const
{
        if (m_dirty)
                return true;
        for (waypoints_t::const_iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; i++)
                if ((*i).m_dirty)
                        return true;
        return false;
}

void FPlanRoute::set_dirty(bool d)
{
        m_dirty = d;
        for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; i++)
                (*i).m_dirty = d;
}

void FPlanRoute::load_fp(id_t nid)
{
        m_id = nid;
	load_fp();
}

void FPlanRoute::load_fp(void)
{
        m_wpts.clear();
	if (!m_fp)
		throw std::runtime_error("load_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT TIMEOFFBLOCK,SAVETIMEOFFBLOCK,TIMEONBLOCK,SAVETIMEONBLOCK,"
                                              "CURWPT,NOTE FROM fplan WHERE ID = ?;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        throw std::runtime_error("load_fp: plan ID not found");
                m_time_offblock = cursor.getint64(0);
                m_savetime_offblock = cursor.getint64(1);
                m_time_onblock = cursor.getint64(2);
                m_savetime_onblock = cursor.getint64(3);
                m_curwpt = cursor.getint(4);
                m_note = cursor.getstring(5);

        }
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE PLANID = ? ORDER BY NR ASC;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        m_wpts.push_back(FPlanWaypoint(cursor));
        }
        m_dirty = false;
}

void FPlanRoute::save_fp(void)
{
        if (!is_dirty())
                return;
	if (!m_fp)
		throw std::runtime_error("save_fp: no database");
        if (!m_id) {
                {
                        //sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan ORDER BY ID DESC LIMIT 1;");
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT MAX(ID) FROM fplan;");
                        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                        if (cursor.step())
                                m_id = cursor.getint(0) + 1;
                        else
                                m_id = 1;
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT INTO fplan (ID) VALUES(?);");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
        }
        sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ? AND (NR < 0 OR NR >= ?);");
                cmd.bind(1, m_id);
                cmd.bind(2, get_nrwpt());
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE fplan SET TIMEOFFBLOCK = ?,"
                                              "SAVETIMEOFFBLOCK = ?, TIMEONBLOCK = ?, SAVETIMEONBLOCK = ?,"
                                              "CURWPT = ?, NOTE = ? WHERE ID = ?;");
                cmd.bind(1, (long long int)m_time_offblock);
                cmd.bind(2, (long long int)m_savetime_offblock);
                cmd.bind(3, (long long int)m_time_onblock);
                cmd.bind(4, (long long int)m_savetime_onblock);
                cmd.bind(5, (int)m_curwpt);
                cmd.bind(6, m_note);
                cmd.bind(7, m_id);
                cmd.executenonquery();
        }
        for (unsigned int nr = 0; nr < get_nrwpt(); nr++) {
                FPlanWaypoint& w(m_wpts[nr]);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "INSERT OR IGNORE INTO waypoints (PLANID,NR) VALUES(?,?);");
                        cmd.bind(1, m_id);
                        cmd.bind(2, (int)nr);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "UPDATE waypoints SET ICAO  = ?, NAME = ?, "
                                                      "LON = ?, LAT = ?, ALT = ?, FREQ = ?, TIME = ?, SAVETIME = ?, "
                                                      "FLAGS = ?, PATHNAME = ?, PATHCODE = ?, NOTE = ?, "
						      "WINDDIR = ?, WINDSPEED = ?, QFF = ?, ISAOFFS = ?, "
						      "TRUEALT = ?, TRUETRACK = ?, TRUEHEADING = ?, DECLINATION = ?, "
						      "DIST = ?, FUEL = ?, TAS = ?, RPM = ?, MP = ?, TYPE = ?, NAVAID = ?, "
						      "FLIGHTTIME = ? "
						      "WHERE PLANID = ? AND NR = ?;");
                        cmd.bind(1, w.get_icao());
                        cmd.bind(2, w.get_name());
                        cmd.bind(3, w.get_lon());
                        cmd.bind(4, w.get_lat());
                        cmd.bind(5, w.get_altitude());
                        cmd.bind(6, (int)w.get_frequency());
                        cmd.bind(7, (long long int)w.get_time());
                        cmd.bind(8, (long long int)w.get_save_time());
                        cmd.bind(9, w.get_flags());
                        cmd.bind(10, w.get_pathname());
                        cmd.bind(11, w.get_pathcode());
                        cmd.bind(12, w.get_note());
			cmd.bind(13, w.get_winddir());
			cmd.bind(14, w.get_windspeed());
			cmd.bind(15, w.get_qff());
			cmd.bind(16, w.get_isaoffset());
			cmd.bind(17, w.get_truealt());
			cmd.bind(18, w.get_truetrack());
			cmd.bind(19, w.get_trueheading());
			cmd.bind(20, w.get_declination());
			cmd.bind(21, (long long int)w.get_dist());
			cmd.bind(22, (long long int)w.get_fuel());
			cmd.bind(23, w.get_tas());
			cmd.bind(24, w.get_rpm());
			cmd.bind(25, w.get_mp());
			cmd.bind(26, w.get_type());
			cmd.bind(27, w.get_navaid());
			cmd.bind(28, (long long int)w.get_flighttime());
                        cmd.bind(29, m_id);
                        cmd.bind(30, (int)nr);
                        cmd.executenonquery();
                }
        }
        tran.commit();
        set_dirty(false);
}

void FPlanRoute::duplicate_fp(void)
{
        m_id = 0;
        m_dirty = true;
}

void FPlanRoute::new_fp(void)
{
        duplicate_fp();
        m_wpts.clear();
        m_note.clear();
        m_time_offblock = 0;
        m_savetime_offblock = 0;
        m_time_onblock = 0;
        m_savetime_onblock = 0;
        m_curwpt = 0;
        set_dirty(false);
}

void FPlanRoute::delete_fp(void)
{
        if (m_id) {
		if (!m_fp)
			throw std::runtime_error("delete_fp: no database");
                sqlite3x::sqlite3_transaction tran(m_fp->get_db(), true);
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM waypoints WHERE PLANID = ?;");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
                {
                        sqlite3x::sqlite3_command cmd(m_fp->get_db(), "DELETE FROM fplan WHERE ID = ?;");
                        cmd.bind(1, m_id);
                        cmd.executenonquery();
                }
                tran.commit();
        }
        new_fp();
}

void FPlanRoute::reverse_fp(void)
{
        if (m_wpts.empty())
                return;
        reverse(m_wpts.begin(), m_wpts.end());
        set_dirty(true);
}

bool FPlanRoute::load_first_fp(void)
{
	if (!m_fp)
		throw std::runtime_error("load_first_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan ORDER BY ID ASC LIMIT 1;");
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
		m_id = cursor.getint(0);
        }
        load_fp();
        return true;
}

bool FPlanRoute::load_next_fp(void)
{
	if (!m_fp)
		throw std::runtime_error("load_next_fp: no database");
        {
                sqlite3x::sqlite3_command cmd(m_fp->get_db(), "SELECT ID FROM fplan WHERE ID > ? ORDER BY ID ASC LIMIT 1;");
                cmd.bind(1, m_id);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                if (!cursor.step())
                        return false;
                m_id = cursor.getint(0);
        }
        load_fp();
        return true;
}

float FPlanRoute::total_distance_km(void) const
{
        float sum(0);
	for (unsigned int i(0), n(m_wpts.size()); i < n; ) {
		const Point& pt0(m_wpts[i].get_coord());
		if (pt0.is_invalid()) {
			++i;
			continue;
		}
		unsigned int j(i + 1);
		for (; j < n; ++j) {
			const Point& pt1(m_wpts[j].get_coord());
			if (pt1.is_invalid())
				continue;
			sum += pt0.spheric_distance_km(pt1);
			break;
		}
		i = j;
	}
        return sum;
}

double FPlanRoute::total_distance_km_dbl(void) const
{
        double sum(0);
	for (unsigned int i(0), n(m_wpts.size()); i < n; ) {
		const Point& pt0(m_wpts[i].get_coord());
		if (pt0.is_invalid()) {
			++i;
			continue;
		}
		unsigned int j(i + 1);
		for (; j < n; ++j) {
			const Point& pt1(m_wpts[j].get_coord());
			if (pt1.is_invalid())
				continue;
			sum += pt0.spheric_distance_km_dbl(pt1);
			break;
		}
		i = j;
	}
        return sum;
}

float FPlanRoute::gc_distance_km(void) const
{
	if (m_wpts.empty())
		return 0;
	return m_wpts.front().get_coord().spheric_distance_km(m_wpts.back().get_coord());
}

double FPlanRoute::gc_distance_km_dbl(void) const
{
	if (m_wpts.empty())
		return 0;
	return m_wpts.front().get_coord().spheric_distance_km_dbl(m_wpts.back().get_coord());
}

int32_t FPlanRoute::max_altitude(void) const
{
        int32_t a(std::numeric_limits<int32_t>::min());
        for (unsigned int i = m_wpts.size(); i > 0; )
                a = std::max(a, m_wpts[--i].get_altitude());
        return a;
}

int32_t FPlanRoute::max_altitude(uint16_t& flags) const
{
        int32_t a(std::numeric_limits<int32_t>::min());
	uint16_t f(0);
        for (unsigned int i = m_wpts.size(); i > 0; ) {
		const FPlanWaypoint& wpt(m_wpts[--i]);
		int16_t aw(wpt.get_altitude());
		if (aw < a)
			continue;
                a = aw;
		f = wpt.get_flags();
	}
	flags = f;
        return a;
}

Rect FPlanRoute::get_bbox(void) const
{
	Rect bbox;
	bool first(true);
        for (unsigned int i = 0; i < m_wpts.size(); ++i) {
                Point p(m_wpts[i].get_coord());
		if (p.is_invalid())
			continue;
		if (first) {
			bbox = Rect(p, p);
			first = false;
			continue;
		}
		bbox = bbox.add(p);
        }
	return bbox;
}

char FPlanRoute::get_flightrules(void) const
{
	unsigned int i(m_wpts.size());
	if (!i)
		return ' ';
	--i;
	if (!i)
		return m_wpts[0].is_ifr() ? 'I' : 'V';
	uint16_t vfr(0), ifr(0);
	while (i > 0) {
		uint16_t f(m_wpts[--i].get_flags());
		vfr |= ~f;
		ifr |= f;
		if (vfr & ifr & FPlanWaypoint::altflag_ifr)
			return m_wpts[0].is_ifr() ? 'Y' : 'Z';
	}
	return m_wpts[0].is_ifr() ? 'I' : 'V';
}

void FPlanRoute::save_time(uint32_t nr)
{
        if (!nr) {
                save_time_offblock();
                return;
        }
        if (nr > m_wpts.size()) {
                save_time_onblock();
                return;
        }
        m_wpts[nr-1].save_time();
}

void FPlanRoute::restore_time(uint32_t nr)
{
        if (!nr) {
                restore_time_offblock();
                return;
        }
        if (nr > m_wpts.size()) {
                restore_time_onblock();
                return;
        }
        m_wpts[nr-1].restore_time();
}

void FPlanRoute::set_time(uint32_t nr, uint32_t t)
{
        if (!nr) {
                set_time_offblock(t);
                return;
        }
        if (nr > m_wpts.size()) {
                set_time_onblock(t);
                return;
        }
        m_wpts[nr-1].set_time(t);
}

uint32_t FPlanRoute::get_time(uint32_t nr) const
{
        if (!nr)
                return get_time_offblock();
        if (nr > m_wpts.size())
                return get_time_onblock();
        return m_wpts[nr-1].get_time();
}

FPlanLeg FPlanRoute::get_leg(uint32_t nr)
{
        bool future = nr > get_curwpt();
        const FPlanWaypoint *wpt0(0), *wpt1(0);
        if (nr > 0 && nr <= get_nrwpt())
                wpt0 = &(*this)[nr - 1];
        if (nr < get_nrwpt())
                wpt1 = &(*this)[nr];
        if (!wpt0 && !wpt1) {
                if (nr > 0 && get_nrwpt()) {
                        wpt0 = &(*this)[get_nrwpt()-1];
                        return FPlanLeg("", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                        "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
				        wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_invalid, future);
                }
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                0, 0, 0, 0, FPlanLeg::legtype_invalid, future);
        }
        if (!wpt0)
                return FPlanLeg("", _("Off Block"), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_offblock(), wpt1->get_frequency(),
                                wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), wpt1->get_time(), wpt1->get_frequency(),
				wpt1->get_winddir_deg(), wpt1->get_windspeed_kts(), wpt1->get_qff_hpa(), wpt1->get_isaoffset_kelvin(), FPlanLeg::legtype_offblock_takeoff, future);
        if (!wpt1)
                return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                "", _("On Block"), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                                wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_landing_onblock, future);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
			wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_normal, future);
}

FPlanLeg FPlanRoute::get_directleg(void)
{
        uint32_t nr(get_nrwpt());
        bool future = false;
        if (!nr)
                return FPlanLeg("", _("Off Block"), Point(), 0, 0, 0, 0,
                                "", _("On Block"), Point(), 0, 0, 0, 0,
                                0, 0, 0, 0, FPlanLeg::legtype_invalid, future);
        const FPlanWaypoint *wpt0(&(*this)[0]), *wpt1(&(*this)[nr - 1]);
        return FPlanLeg(wpt0->get_icao(), wpt0->get_name(), wpt0->get_coord(), wpt0->get_altitude(), wpt0->get_flags(), wpt0->get_time(), wpt0->get_frequency(),
                        wpt1->get_icao(), wpt1->get_name(), wpt1->get_coord(), wpt1->get_altitude(), wpt1->get_flags(), get_time_onblock(), wpt1->get_frequency(),
                        wpt0->get_winddir_deg(), wpt0->get_windspeed_kts(), wpt0->get_qff_hpa(), wpt0->get_isaoffset_kelvin(), FPlanLeg::legtype_normal, future);
}

bool FPlanRoute::has_pathcodes(void) const
{
        for (unsigned int i = 0; i < m_wpts.size(); ++i)
		if (m_wpts[i].get_pathcode() != FPlanWaypoint::pathcode_none)
			return true;
	return false;
}

bool FPlanRoute::enforce_pathcode_vfrdeparr(void)
{
	unsigned int nr(m_wpts.size());
	if (nr <= 2U)
		return false;
	bool work(false);
	unsigned int i(0);
	for (; i < nr - 1; ++i) {
		const FPlanWaypoint& wpt(m_wpts[i]);
		if (wpt.get_icao() != m_wpts[0].get_icao() ||
		    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrarrival ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
			break;
	}
	if (i > 1) {
		work = true;
		while (i > 0) {
			--i;
			m_wpts[i].set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
		}
	}
	i = nr - 1;
	for (; i > 0; --i) {
		const FPlanWaypoint& wpt(m_wpts[i]);
		if (wpt.get_icao() != m_wpts[nr - 1].get_icao() ||
		    (wpt.get_flags() & FPlanWaypoint::altflag_ifr) ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrdeparture ||
		    wpt.get_pathcode() == FPlanWaypoint::pathcode_vfrtransition)
			break;
	}
	if (i < nr - 2) {
		work = true;
		while (i < nr - 1) {
			++i;
			m_wpts[i].set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
		}
	}
	return work;
}

bool FPlanRoute::enforce_pathcode_vfrifr(void)
{
	unsigned int nr(m_wpts.size());
	bool work(false);
	for (unsigned int i = 0; i < nr; ++i) {
		FPlanWaypoint& wpt(m_wpts[i]);
		work = wpt.enforce_pathcode_vfrifr() || work;
	}
	if (nr >= 2) {
		const FPlanWaypoint& wpt1(m_wpts[nr - 1]);
		FPlanWaypoint& wpt0(m_wpts[nr - 2]);
		work = work || ((wpt0.get_flags() ^ wpt1.get_flags()) & FPlanWaypoint::altflag_ifr);
		wpt0.set_flags(wpt0.get_flags() ^ ((wpt0.get_flags() ^ wpt1.get_flags()) & FPlanWaypoint::altflag_ifr));
	}
	if (nr >= 1) {
		FPlanWaypoint& wpt(m_wpts[nr - 1]);
		work = work || wpt.get_pathcode() != FPlanWaypoint::pathcode_none ||
			!wpt.get_pathname().empty();
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		wpt.set_pathname("");
	}
	return work;
}

void FPlanRoute::recompute_dist(void)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e;) {
		FPlanWaypoint& wpto(*i);
		for (;;) {
			++i;
			if (i == e)
				break;
			FPlanWaypoint& wpt(*i);
			if (!wpt.get_coord().is_invalid())
				break;
			wpt.set_dist(0);
			wpt.set_truetrack(0);
		}
		if (i == e) {
			wpto.set_dist(0);
			wpto.set_truetrack(0);
			break;
		}
		const FPlanWaypoint& wpt(*i);
		wpto.set_dist_nmi(wpto.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
		wpto.set_truetrack_deg(wpto.get_coord().spheric_true_course_dbl(wpt.get_coord()));
	}
}

void FPlanRoute::recompute_decl(void)
{
	WMM wmm;
	time_t curtime(time(0));
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i) {
		FPlanWaypoint& wpt(*i);
		if (wpt.get_coord().is_invalid()) {
			wpt.set_declination(0);
			continue;
		}
		wmm.compute(wpt.get_truealt() * (FPlanWaypoint::ft_to_m / 1000.0), wpt.get_coord(), curtime);
		wpt.set_declination(wmm.get_dec());
	}
}

void FPlanRoute::set_winddir(uint16_t dir)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_winddir(dir);
}

void FPlanRoute::set_windspeed(uint16_t speed)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_windspeed(speed);
}

void FPlanRoute::set_qff(uint16_t press)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_qff(press);
}

void FPlanRoute::set_isaoffset(int16_t temp)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_isaoffset(temp);
}

void FPlanRoute::set_declination(uint16_t decl)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_declination(decl);
}

void FPlanRoute::set_rpm(uint16_t rpm)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_rpm(rpm);
}

void FPlanRoute::set_mp(uint16_t mp)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_mp(mp);
}

void FPlanRoute::set_tas(uint16_t tas)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->set_tas(tas);
}

FPlanWaypoint FPlanRoute::insert_prepare(uint32_t nr)
{
	FPlanWaypoint wpt;
	{
		time_t t(time(0));
		wpt.set_time_unix(t);
                wpt.set_save_time_unix(t);
	}
	wpt.set_flags(0);
	wpt.set_altitude(0);
        unsigned int nrwpt(get_nrwpt());
	if (!nrwpt)
		return wpt;
	if (nr >= 1 && nr <= nrwpt) {
		const FPlanWaypoint& wpt1((*this)[nr - 1]);
		wpt.set_flags(wpt1.get_flags() & (FPlanWaypoint::altflag_ifr | FPlanWaypoint::altflag_standard));
		wpt.set_altitude(wpt1.get_altitude());
	}
	return wpt;
}

unsigned int FPlanRoute::insert(const AirportsDb::Airport& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(nr));
	wpt.set(el);
	insert_wpt(nr, wpt);
	return 1;
}

unsigned int FPlanRoute::insert(unsigned int vfrrtenr, const AirportsDb::Airport& el, uint32_t nr)
{
	if (vfrrtenr >= el.get_nrvfrrte())
		return insert(el, nr);
	const AirportsDb::Airport::VFRRoute& vfrrte(el.get_vfrrte(vfrrtenr));
	FPlanWaypoint wpt(insert_prepare(nr));
	if (!vfrrte.size()) {
		wpt.set(el);
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
		wpt.set_pathname(vfrrte.get_name());
		insert_wpt(nr, wpt);
		return 1;
	}
	nr = std::max(nr, static_cast<uint32_t>(get_nrwpt()));
	for (unsigned int nrr(0); nrr < vfrrte.size(); ++nrr) {
		const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(vfrrte[nrr]);
		if (vfrpt.is_at_airport()) {
			wpt.set(el);
		} else {
			wpt.set_icao(el.get_icao());
			wpt.set_name(vfrpt.get_name());
			wpt.set_coord(vfrpt.get_coord());
			wpt.set_altitude(vfrpt.get_altitude());
			wpt.set_frequency(0);
			Glib::ustring note(vfrpt.get_altcode_string());
			if (vfrpt.is_compulsory())
				note += " Compulsory";
			note += "\n" + vfrpt.get_pathcode_string();
			wpt.set_note(note);
		}
		switch (vfrpt.get_pathcode()) {
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
			wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
			break;
							
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
		case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
		default:
			wpt.set_pathcode(FPlanWaypoint::pathcode_none);
			break;
		}
		wpt.set_pathname(vfrrte.get_name());
		insert_wpt(nr, wpt);
		++nr;
	}
	return vfrrte.size();
}

unsigned int FPlanRoute::insert(unsigned int vfrrtenr, unsigned int vfrrteptnr, const AirportsDb::Airport& el, uint32_t nr)
{
	if (vfrrtenr >= el.get_nrvfrrte())
		return insert(el, nr);
	const AirportsDb::Airport::VFRRoute& vfrrte(el.get_vfrrte(vfrrtenr));
	if (vfrrteptnr >= vfrrte.size())
		return insert(vfrrtenr, el, nr);
	FPlanWaypoint wpt(insert_prepare(nr));
	const AirportsDb::Airport::VFRRoute::VFRRoutePoint& vfrpt(vfrrte[vfrrteptnr]);
	if (vfrpt.is_at_airport()) {
		wpt.set(el);
	} else {
		wpt.set_icao(el.get_icao());
		wpt.set_name(vfrpt.get_name());
		wpt.set_coord(vfrpt.get_coord());
		wpt.set_altitude(vfrpt.get_altitude());
		wpt.set_frequency(0);
		Glib::ustring note(vfrpt.get_altcode_string());
		if (vfrpt.is_compulsory())
			note += " Compulsory";
		note += "\n" + vfrpt.get_pathcode_string();
		wpt.set_note(note);
	}
	switch (vfrpt.get_pathcode()) {
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_arrival:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrarrival);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_departure:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrdeparture);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_vfrtransition:
		wpt.set_pathcode(FPlanWaypoint::pathcode_vfrtransition);
		break;

	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_holding:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_trafficpattern:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_other:
	case AirportsDb::Airport::VFRRoute::VFRRoutePoint::pathcode_invalid:
	default:
		wpt.set_pathcode(FPlanWaypoint::pathcode_none);
		break;
	}
	wpt.set_pathname(vfrrte.get_name());
	insert_wpt(nr, wpt);
	return 1;
}

unsigned int FPlanRoute::insert(const NavaidsDb::Navaid& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(nr));
	wpt.set(el);
	insert_wpt(nr, wpt);
	return 1;
}

unsigned int FPlanRoute::insert(const WaypointsDb::Waypoint& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(nr));
	wpt.set(el);
	insert_wpt(nr, wpt);
	return 1;
}

unsigned int FPlanRoute::insert(const MapelementsDb::Mapelement& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(nr));
	wpt.set(el);
	insert_wpt(nr, wpt);
	return 1;
}

unsigned int FPlanRoute::insert(const FPlanWaypoint& el, uint32_t nr)
{
	FPlanWaypoint wpt(insert_prepare(nr));
	wpt.set(el);
	insert_wpt(nr, wpt);
	return 1;
}

void FPlanRoute::recompute(const Aircraft& acft, float qnh, float tempoffs, double bhp, double rpm, double mp)
{
	if (!get_nrwpt())
		return;
	AirData ad;
	if (!std::isnan(qnh) && !std::isnan(tempoffs))
		ad.set_qnh_tempoffs(qnh, tempoffs);
        WMM wmm;
        time_t curtime(time(0));
	{
		FPlanWaypoint& wpt((*this)[0]);
		if (std::isnan(qnh) || std::isnan(tempoffs))
			ad.set_qnh_tempoffs(std::isnan(qnh) ? wpt.get_qff_hpa() : qnh, std::isnan(tempoffs) ? wpt.get_isaoffset_kelvin() : tempoffs);
		wpt.frob_flags(0, FPlanWaypoint::altflag_climb | FPlanWaypoint::altflag_descent);
		wpt.set_tas(0);
		wpt.set_flighttime(0);
		wpt.set_fuel(0);
		wpt.set_rpm(0);
		wpt.set_mp(0);
		if (wpt.is_standard())
			ad.set_pressure_altitude(wpt.get_altitude());
		else
			ad.set_true_altitude(wpt.get_altitude());
		wpt.set_truealt(ad.get_true_altitude());
		{
			double tm(acft.get_climb().altitude_to_time(ad.get_density_altitude()));
			wpt.set_tas(acft.get_climb().time_to_tas(tm));
		}
		wmm.compute(ad.get_true_altitude() * (FPlanWaypoint::ft_to_m / 1000.0), wpt.get_coord(), curtime);
		wpt.set_declination(wmm.get_dec());
		{
			double da(ad.get_density_altitude());
			double t(acft.get_climb().altitude_to_time(da));
			wpt.set_tas_kts(acft.get_climb().time_to_tas(t));
		}
	}
	{
		FPlanWaypoint& wpt((*this)[get_nrwpt() - 1]);
		wpt.set_tas(0);
		wpt.set_truetrack(0);
		wpt.set_trueheading(0);
		wpt.set_dist(0);
	}
	for (unsigned int i = 1; i < get_nrwpt(); ++i) {
		FPlanWaypoint& wpt((*this)[i]);
		FPlanWaypoint& wpto((*this)[i - 1]);
		if (std::isnan(qnh) || std::isnan(tempoffs))
			ad.set_qnh_tempoffs(std::isnan(qnh) ? wpt.get_qff_hpa() : qnh, std::isnan(tempoffs) ? wpt.get_isaoffset_kelvin() : tempoffs);
		wpt.frob_flags(0, FPlanWaypoint::altflag_climb | FPlanWaypoint::altflag_descent);
		wpt.set_tas(0);
		wpt.set_rpm(0);
		wpt.set_mp(0);
		double dist(wpto.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
		double tt(wpto.get_coord().spheric_true_course_dbl(wpt.get_coord()));
		wpto.set_dist_nmi(dist);
		wpto.set_truetrack_deg(tt);
		double prevda(ad.get_density_altitude());
		double prevta(ad.get_true_altitude());
		if (wpt.is_standard())
			ad.set_pressure_altitude(wpt.get_altitude());
		else
			ad.set_true_altitude(wpt.get_altitude());
		double nextda(ad.get_density_altitude());
		double ltime(wpto.get_flighttime()), fuel(wpto.get_fuel_usg());
		double th(tt);
		Wind<double> wind;
		wind.set_wind(wpto.get_winddir_deg(), wpto.get_windspeed_kts(), wpt.get_winddir_deg(), wpt.get_windspeed_kts());
		if (nextda - prevda > 100)
			wpto.frob_flags(FPlanWaypoint::altflag_climb, FPlanWaypoint::altflag_climb);
		if (prevda - nextda > 100)
			wpto.frob_flags(FPlanWaypoint::altflag_descent, FPlanWaypoint::altflag_descent);
		// check for climb segment
		if (wpto.is_climb()) {
			double prevt(acft.get_climb().altitude_to_time(prevda));
			double nextt(acft.get_climb().altitude_to_time(nextda));
			double prevd(acft.get_climb().time_to_distance(prevt));
			double nextd(acft.get_climb().time_to_distance(nextt));
			double climbd(nextd - prevd);
			// check for wind
			double speedfactor(1);
			if (wind.get_windspeed() > 0) {
				double tas(acft.get_climb().time_to_tas(prevt));
				wind.set_crs_tas(tt, tas);
				if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
					climbd = 0;
				} else {
					th = wind.get_hdg();
					speedfactor = wind.get_gs() / tas;
					climbd *= speedfactor;
					if (false)
						std::cerr << "FPlanRoute::recompute: climb: windcorr: TT " << tt << " TH " << th
							  << " TAS " << tas << " GS " << wind.get_gs()
							  << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
				}
			}
			if (climbd <= dist) {
				dist -= climbd;
			} else {
				// climb not finished at waypoint
				nextt = acft.get_climb().distance_to_time(prevd + dist / speedfactor);
				dist = 0;
				ad.set_density_altitude(acft.get_climb().time_to_altitude(nextt));
			}
			ltime += nextt - prevt;
			fuel += acft.get_climb().time_to_fuel(nextt) -
				acft.get_climb().time_to_fuel(prevt);
			wpto.set_tas_kts(acft.get_climb().time_to_tas(prevt));
		}
		// check for cruise segment
		unsigned int staynr(0), staytime(0);
		bool stay(wpto.is_stay(staynr, staytime));
		if (stay || dist > 0) {
			double tas(0), fuelflow(0), bhp1(bhp), rpm1(rpm), mp1(mp);
			if (std::isnan(bhp) && std::isnan(rpm) && std::isnan(mp)) {
				bhp1 = std::numeric_limits<double>::quiet_NaN();
				rpm1 = wpt.get_rpm();
				mp1 = wpt.get_mp_inhg();
			}
			double da(ad.get_density_altitude());
			acft.get_cruise().calculate(tas, fuelflow, bhp1, rpm1, mp1, da);
			if (!wpto.is_climb())
				wpto.set_tas_kts(tas);
			if (stay) {
				ltime += staytime;
				fuel +=  staytime * (1.0 / 3600.0) * fuelflow;
			} else {
				double gs(tas);
				if (wind.get_windspeed() > 0) {
					wind.set_crs_tas(tt, tas);
					if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
						gs = 1;
					} else if (!wpto.is_climb()) {
						th = wind.get_hdg();
						gs = wind.get_gs();
						if (false)
							std::cerr << "FPlanRoute::recompute: cruise: windcorr: TT " << tt << " TH " << th
								  << " dist " << dist << " TAS " << tas << " GS " << gs
								  << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
					}
				}
				double ct(dist / gs);
				ltime += ct * 3600.0;
				fuel += ct * fuelflow;
			}
			if (!std::isnan(rpm1) && rpm1 > 0)
				wpto.set_rpm(rpm1);
			if (!std::isnan(mp1) && mp1 > 0)
				wpto.set_mp_inhg(mp1);
		}
		wpt.set_flighttime(ltime);
		wpt.set_fuel_usg(fuel);
		{
			double nextta(ad.get_true_altitude());
			wpt.set_truealt(nextta);
			wmm.compute(nextta * (FPlanWaypoint::ft_to_m / 1000.0), wpt.get_coord(), curtime);
			wpt.set_declination_deg(wmm.get_dec());
		}
		wpto.set_trueheading_deg(th);
	}
}

namespace {
	void nwxweather_cb(const NWXWeather::windvector_t& winds, FPlanRoute& route, const sigc::slot<void,bool>& cb)
	{
		bool fplmodif(false);
		for (unsigned int i = 0; i < winds.size(); ++i) {
			const NWXWeather::WindsAloft& wind(winds[i]);
			if (true) {
				std::cerr << "Coord " << wind.get_coord().get_lat_str2() << ' ' << wind.get_coord().get_lon_str2() << " Alt ";
				if (wind.get_flags() & FPlanWaypoint::altflag_standard)
					std::cerr << "F" << wind.get_altitude();
				else
					std::cerr << wind.get_altitude();
				{
					struct tm utm;
					char buf[64];
					time_t t(wind.get_time());
					strftime(buf, sizeof(buf), "%Y-%m-%d %H:%M:%S", gmtime_r(&t, &utm));
					std::cerr << ' ' << buf;
				}
				std::cerr << ' ' << wind.get_winddir() << '/' << wind.get_windspeed()
					  << ' ' << wind.get_qff() << ' ' << wind.get_temp() << std::endl;
			}
			if (i >= route.get_nrwpt())
				continue;
			FPlanWaypoint& wpt(route[i]);
			{
				float d(wpt.get_coord().simple_distance_nmi(wind.get_coord()));
				if (d > 10) {
					std::cerr << "nwxweather: skipping " << wpt.get_icao_name() << " distance " << d << "nmi" << std::endl;
					continue;
				}
			}
			wpt.set_winddir_deg(wind.get_winddir());
			wpt.set_windspeed_kts(wind.get_windspeed());
			wpt.set_qff_hpa(wind.get_qff());
			wpt.set_oat_kelvin(wind.get_temp());
			fplmodif = true;
		}
		cb(fplmodif);
	}
};

void FPlanRoute::nwxweather(NWXWeather& nwx, const sigc::slot<void,bool>& cb)
{
        nwx.abort();
        nwx.winds_aloft(*this, sigc::bind(sigc::bind(sigc::ptr_fun(&nwxweather_cb), cb), *this));
}

FPlanRoute::GFSResult FPlanRoute::gfs(GRIB2& grib2)
{
	Rect bbox(get_bbox().oversize_nmi(100.f));
	GFSResult r;
	Glib::RefPtr<GRIB2::LayerInterpolateResult> windu, windv, prmsl, temperature;
	for (unsigned int wptnr = 0; wptnr < get_nrwpt(); ++wptnr) {
		FPlanWaypoint& wpt((*this)[wptnr]);
		float press(0);
		IcaoAtmosphere::std_altitude_to_pressure(&press, 0, wpt.get_altitude() * Point::ft_to_m);
		press *= 100;
		if (!windu || press < windu->get_minsurface1value() || press > windu->get_maxsurface1value() ||
		    wpt.get_time_unix() < windu->get_minefftime() || wpt.get_time_unix() > windu->get_maxefftime()) {
			{
				GRIB2::layerlist_t ll(grib2.find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_ugrd),
									wpt.get_time_unix(), GRIB2::surface_isobaric_surface, press));
				windu = GRIB2::interpolate_results(bbox, ll, wpt.get_time_unix(), press);
			}
			{
				GRIB2::layerlist_t ll(grib2.find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_momentum_vgrd),
									wpt.get_time_unix(), GRIB2::surface_isobaric_surface, press));
				windv = GRIB2::interpolate_results(bbox, ll, wpt.get_time_unix(), press);
			}
		}
		if (!temperature || press < temperature->get_minsurface1value() || press > temperature->get_maxsurface1value() ||
		    wpt.get_time_unix() < temperature->get_minefftime() || wpt.get_time_unix() > temperature->get_maxefftime()) {
			GRIB2::layerlist_t ll(grib2.find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_temperature_tmp),
								wpt.get_time_unix(), GRIB2::surface_isobaric_surface, press));
			temperature = GRIB2::interpolate_results(bbox, ll, wpt.get_time_unix(), press);
		}
		if (!prmsl ||
		    wpt.get_time_unix() < prmsl->get_minefftime() || wpt.get_time_unix() > prmsl->get_maxefftime()) {
			GRIB2::layerlist_t ll(grib2.find_layers(GRIB2::find_parameter(GRIB2::param_meteorology_mass_prmsl),
								wpt.get_time_unix()));
			prmsl = GRIB2::interpolate_results(bbox, ll, wpt.get_time_unix());
		}
		if (windu && windv) {
			if (wpt.get_time_unix() >= windu->get_minefftime() && wpt.get_time_unix() <= windu->get_maxefftime() &&
			    wpt.get_time_unix() >= windv->get_minefftime() && wpt.get_time_unix() <= windv->get_maxefftime()) {
				float wu(windu->operator()(wpt.get_coord(), wpt.get_time_unix(), press));
				float wv(windv->operator()(wpt.get_coord(), wpt.get_time_unix(), press));
				{
					std::pair<float,float> w(windu->get_layer()->get_grid()->transform_axes(wu, wv));
					wu = w.first;
					wv = w.second;
				}
				// convert from m/s to kts
				wu *= (-1e-3f * Point::km_to_nmi * 3600);
				wv *= (-1e-3f * Point::km_to_nmi * 3600);
				wpt.set_windspeed_kts(sqrtf(wu * wu + wv * wv));
				wpt.set_winddir_rad(M_PI - atan2f(wu, wv));
				r.set_modified();
				r.add_efftime(windu->get_minefftime());
				r.add_efftime(windu->get_maxefftime());
				r.add_reftime(windu->get_minreftime());
				r.add_reftime(windu->get_maxreftime());
				r.add_efftime(windv->get_minefftime());
				r.add_efftime(windv->get_maxefftime());
				r.add_reftime(windv->get_minreftime());
				r.add_reftime(windv->get_maxreftime());
			} else {
				r.set_partial();
			}
		}
		if (temperature) {
			if (wpt.get_time_unix() >= temperature->get_minefftime() && wpt.get_time_unix() <= temperature->get_maxefftime()) {
				wpt.set_oat_kelvin(temperature->operator()(wpt.get_coord(), wpt.get_time_unix(), press));
				r.set_modified();
				r.add_efftime(temperature->get_minefftime());
				r.add_efftime(temperature->get_maxefftime());
				r.add_reftime(temperature->get_minreftime());
				r.add_reftime(temperature->get_maxreftime());
			} else {
				r.set_partial();
			}
		}
		if (prmsl) {
			if (wpt.get_time_unix() >= prmsl->get_minefftime() && wpt.get_time_unix() <= prmsl->get_maxefftime()) {
				wpt.set_qff_hpa(prmsl->operator()(wpt.get_coord(), wpt.get_time_unix(), press) * 0.01f);
				r.set_modified();
				r.add_efftime(prmsl->get_minefftime());
				r.add_efftime(prmsl->get_maxefftime());
				r.add_reftime(prmsl->get_minreftime());
				r.add_reftime(prmsl->get_maxreftime());
			} else {
				r.set_partial();
			}
		}
	}
	return r;
}

void FPlanRoute::turnpoints(bool include_dct, float maxdev)
{
	for (waypoints_t::iterator i(m_wpts.begin()), e(m_wpts.end()); i != e; ++i)
		i->frob_flags(FPlanWaypoint::altflag_turnpoint, FPlanWaypoint::altflag_turnpoint);
	for (unsigned int nr = 1U; nr + 2U < get_nrwpt(); ) {
		unsigned int nr1(nr);
		for (; nr1 + 2U < get_nrwpt(); ++nr1) {
			const FPlanWaypoint& wpt2((*this)[nr1 + 2U]);
			const FPlanWaypoint& wpt1((*this)[nr1 + 1U]);
			const FPlanWaypoint& wpt0((*this)[nr]);
			if (!(wpt0.get_flags() & wpt1.get_flags() & FPlanWaypoint::altflag_ifr) ||
			    ((wpt0.get_flags() ^ wpt1.get_flags()) & FPlanWaypoint::altflag_standard) ||
			    (wpt0.get_altitude() != wpt1.get_altitude()))
				break;
			if (((wpt0.get_pathcode() != FPlanWaypoint::pathcode_airway &&
			      wpt0.get_pathcode() != FPlanWaypoint::pathcode_sid &&
			      wpt0.get_pathcode() != FPlanWaypoint::pathcode_star) ||
			     wpt1.get_pathcode() != wpt0.get_pathcode() ||
			     wpt0.get_pathname().empty() || wpt0.get_pathname() != wpt1.get_pathname()) &&
			    (!include_dct || wpt0.get_pathcode() != FPlanWaypoint::pathcode_directto ||
			     wpt1.get_pathcode() != FPlanWaypoint::pathcode_directto))
				break;
			if (maxdev < std::numeric_limits<float>::max()) {
				bool ok(true);
				for (unsigned int i = nr; i <= nr1; ) {
					++i;
					const FPlanWaypoint& wptx((*this)[i]);
					Point pt(wptx.get_coord().spheric_nearest_on_line(wpt0.get_coord(), wpt2.get_coord()));
					if (pt.spheric_distance_nmi(wptx.get_coord()) <= maxdev)
						continue;
					ok = false;
					break;
				}
				if (!ok)
					break;
			}
		}
		++nr;
		for (; nr <= nr1; ++nr)
			(*this)[nr].frob_flags(0, FPlanWaypoint::altflag_turnpoint);
	}
}

void FPlanRoute::fix_invalid_altitudes(unsigned int nr)
{
	if (nr >= m_wpts.size())
		return;
	FPlanWaypoint& wpt(m_wpts[nr]);
	if (wpt.is_altitude_valid())
		return;
	if (nr) {
		const FPlanWaypoint& wptp(m_wpts[nr-1]);
		if (wpt.is_ifr() || wptp.is_ifr())
			wpt.set_altitude(10000);
		else if (wptp.get_magnetictrack() < 0x8000)
			wpt.set_altitude(5500);
		else
			wpt.set_altitude(4500);
		return;
	}
	if (wpt.is_ifr())
		wpt.set_altitude(10000);
	else if (wpt.get_magnetictrack() < 0x8000)
		wpt.set_altitude(5500);
	else
		wpt.set_altitude(4500);
}

void FPlanRoute::fix_invalid_altitudes(unsigned int nr, TopoDb30& topodb, AirspacesDb& aspcdb)
{
	if (nr >= m_wpts.size())
		return;
	FPlanWaypoint& wpt(m_wpts[nr]);
	if (wpt.is_altitude_valid())
		return;
	if (wpt.get_coord().is_invalid()) {
		fix_invalid_altitudes(nr);
		return;
	}
	int32_t base(500);
	int32_t terrain(0);
	uint16_t magcrs(wpt.get_magnetictrack());
	if (wpt.is_ifr()) {
		base = 0;
		terrain = 10000;
	}
	if (nr) {
		const FPlanWaypoint& wptp(m_wpts[nr-1]);
		if (wptp.is_ifr()) {
			base = 0;
			terrain = 10000;
		}
		magcrs = wptp.get_magnetictrack();
	}
	bool italy_portugal(false);
	bool newzealand(false);
	bool england(false);
	{
		{
			AirspacesDb::elementvector_t ev(aspcdb.find_by_icao("LI", 0, AirspacesDb::comp_exact, 0, AirspacesDb::element_t::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
				if (evi->get_icao() != "LI" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_nas)
					continue;
				if (!evi->get_bbox().is_inside(wpt.get_coord()))
					continue;
				if (!evi->get_polygon().windingnumber(wpt.get_coord()))
					continue;
				italy_portugal = true;
				break;
			}
		}
		if (!italy_portugal) {
			AirspacesDb::elementvector_t ev(aspcdb.find_by_icao("LPPC", 0, AirspacesDb::comp_exact, 0, AirspacesDb::element_t::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
				if (evi->get_icao() != "LPPC" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
					continue;
				if (!evi->get_bbox().is_inside(wpt.get_coord()))
					continue;
				if (!evi->get_polygon().windingnumber(wpt.get_coord()))
					continue;
				italy_portugal = true;
				break;
			}
		}
		{
        		AirspacesDb::elementvector_t ev(aspcdb.find_by_icao("NZ", 0, AirspacesDb::comp_exact, 0, AirspacesDb::element_t::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
				if (evi->get_icao() != "NZ" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_nas)
					continue;
				if (!evi->get_bbox().is_inside(wpt.get_coord()))
					continue;
				if (!evi->get_polygon().windingnumber(wpt.get_coord()))
					continue;
			        newzealand = true;
				break;
			}
		}
		{
			// London FIR
			AirspacesDb::elementvector_t ev(aspcdb.find_by_icao("EGTT", 0, AirspacesDb::comp_exact, 0, AirspacesDb::element_t::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
				if (evi->get_icao() != "EGTT" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
					continue;
				if (!evi->get_bbox().is_inside(wpt.get_coord()))
					continue;
				if (!evi->get_polygon().windingnumber(wpt.get_coord()))
					continue;
			        england = true;
				break;
			}
		}
		if (!england) {
			// Scottish FIR
			AirspacesDb::elementvector_t ev(aspcdb.find_by_icao("EGPX", 0, AirspacesDb::comp_exact, 0, AirspacesDb::element_t::subtables_all));
			for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
				if (evi->get_icao() != "EGPX" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
				    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
					continue;
				if (!evi->get_bbox().is_inside(wpt.get_coord()))
					continue;
				if (!evi->get_polygon().windingnumber(wpt.get_coord()))
					continue;
			        england = true;
				break;
			}
		}
		if (nr) {
			TopoDb30::Profile profile(topodb.get_profile(m_wpts[nr-1].get_coord(), wpt.get_coord(), 5.0f));
			TopoDb30::elev_t maxelev(profile.get_maxelev());
			if (maxelev != TopoDb30::nodata)				
				terrain = std::max(terrain, Point::round<int32_t,double>(Point::m_to_ft_dbl * maxelev));
		}
		if (nr + 1 < m_wpts.size()) {
			TopoDb30::Profile profile(topodb.get_profile(wpt.get_coord(), m_wpts[nr+1].get_coord(), 5.0f));
			TopoDb30::elev_t maxelev(profile.get_maxelev());
			if (maxelev != TopoDb30::nodata)				
				terrain = std::max(terrain, Point::round<int32_t,double>(Point::m_to_ft_dbl * maxelev));
		}
	}
	if (england) {
		// quadrantal rules - not binding for VFR, but apply nevertheless
		base = 0;
		if (magcrs < 0x8000) {
			base += 1000;
			if (magcrs >= 0x4000)
				base += 500;
		} else {
			if (magcrs >= 0xc000)
				base += 500;
		}
	} else {
		if (italy_portugal) {
			if (magcrs >= 0x4000 && magcrs < 0xc000)
				base += 1000;
		} else if (newzealand) {
			if (magcrs < 0x4000 || magcrs >= 0xc000)
				base += 1000;
		} else {
			if (magcrs < 0x8000)
				base += 1000;
		}
	}
	{
		int32_t minalt(terrain);
		if (minalt >= 5000)
			minalt += 1000;
		minalt += 1000;
		int32_t alt = ((minalt + 1000 - base) / 2000) * 2000 + base;
		alt = std::max(std::min(alt, (int32_t)32000), (int32_t)0);
		wpt.set_altitude(alt);
	}
}

void FPlanRoute::fix_invalid_altitudes(unsigned int nr, Engine& eng)
{
	if (nr >= m_wpts.size())
		return;
	FPlanWaypoint& wpt(m_wpts[nr]);
	if (wpt.is_altitude_valid())
		return;
	if (wpt.get_coord().is_invalid()) {
		fix_invalid_altitudes(nr);
		return;
	}
	int32_t base(500);
	int32_t terrain(0);
	uint16_t magcrs(wpt.get_magnetictrack());
	if (wpt.is_ifr()) {
		base = 0;
		terrain = 10000;
	}
	if (nr) {
		const FPlanWaypoint& wptp(m_wpts[nr-1]);
		if (wptp.is_ifr()) {
			base = 0;
			terrain = 10000;
		}
		magcrs = wptp.get_magnetictrack();
	}
	bool italy_portugal(false);
	bool newzealand(false);
	bool england(false);
	{
		Glib::RefPtr<Engine::AirspaceResult> qli(eng.async_airspace_find_by_icao("LI", ~0, 0, AirspacesDb::comp_exact, AirspacesDb::element_t::subtables_all));
		Glib::RefPtr<Engine::AirspaceResult> qlppc(eng.async_airspace_find_by_icao("LPPC", ~0, 0, AirspacesDb::comp_exact, AirspacesDb::element_t::subtables_all));
		Glib::RefPtr<Engine::AirspaceResult> qnz(eng.async_airspace_find_by_icao("NZ", ~0, 0, AirspacesDb::comp_exact, AirspacesDb::element_t::subtables_all));
		Glib::RefPtr<Engine::AirspaceResult> qegtt(eng.async_airspace_find_by_icao("EGTT", ~0, 0, AirspacesDb::comp_exact, AirspacesDb::element_t::subtables_all));
		Glib::RefPtr<Engine::AirspaceResult> qegpx(eng.async_airspace_find_by_icao("EGPX", ~0, 0, AirspacesDb::comp_exact, AirspacesDb::element_t::subtables_all));
		Glib::RefPtr<Engine::ElevationProfileResult> qprof0, qprof1;
		if (nr)
			qprof0 = eng.async_elevation_profile(m_wpts[nr-1].get_coord(), wpt.get_coord());
		if (nr + 1 < m_wpts.size())
			qprof1 = eng.async_elevation_profile(wpt.get_coord(), m_wpts[nr+1].get_coord());
		Glib::Mutex mutex;
		Glib::Cond cond;
		if (qli)
			qli->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qlppc)
			qlppc->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qnz)
			qnz->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qegtt)
			qegtt->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qegpx)
			qegpx->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qprof0)
			qprof0->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		if (qprof1)
			qprof1->connect(sigc::mem_fun(cond, &Glib::Cond::broadcast));
		{
			Glib::Mutex::Lock lock(mutex);
			for (;;) {
				if (qli) {
					bool ok(qli->is_done());
					if (ok && !italy_portugal) {
						const AirspacesDb::elementvector_t& ev(qli->get_result());
						for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
							if (evi->get_icao() != "LI" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
							    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_nas)
								continue;
							if (!evi->get_bbox().is_inside(wpt.get_coord()))
								continue;
							if (!evi->get_polygon().windingnumber(wpt.get_coord()))
								continue;
							italy_portugal = true;
							break;
						}
					}
					if (ok || qli->is_error()) {
						Glib::RefPtr<Engine::AirspaceResult> q(qli);
						qli.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qlppc) {
					bool ok(qlppc->is_done());
					if (ok && !italy_portugal) {
						const AirspacesDb::elementvector_t& ev(qlppc->get_result());
						for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
							if (evi->get_icao() != "LPPC" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
							    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
								continue;
							if (!evi->get_bbox().is_inside(wpt.get_coord()))
								continue;
							if (!evi->get_polygon().windingnumber(wpt.get_coord()))
								continue;
							italy_portugal = true;
							break;
						}
					}
					if (ok || qlppc->is_error()) {
						Glib::RefPtr<Engine::AirspaceResult> q(qlppc);
						qlppc.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qnz) {
					bool ok(qnz->is_done());
					if (ok) {
						const AirspacesDb::elementvector_t& ev(qnz->get_result());
						for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
							if (evi->get_icao() != "NZ" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
							    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_nas)
								continue;
							if (!evi->get_bbox().is_inside(wpt.get_coord()))
								continue;
							if (!evi->get_polygon().windingnumber(wpt.get_coord()))
								continue;
							newzealand = true;
							break;
						}
					}
					if (ok || qnz->is_error()) {
						Glib::RefPtr<Engine::AirspaceResult> q(qnz);
						qnz.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qegtt) {
					bool ok(qegtt->is_done());
					if (ok && !england) {
						const AirspacesDb::elementvector_t& ev(qegtt->get_result());
						for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
							if (evi->get_icao() != "EGTT" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
							    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
								continue;
							if (!evi->get_bbox().is_inside(wpt.get_coord()))
								continue;
							if (!evi->get_polygon().windingnumber(wpt.get_coord()))
								continue;
							england = true;
							break;
						}
					}
					if (ok || qegtt->is_error()) {
						Glib::RefPtr<Engine::AirspaceResult> q(qegtt);
						qegtt.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qegpx) {
					bool ok(qegpx->is_done());
					if (ok && !england) {
						const AirspacesDb::elementvector_t& ev(qegpx->get_result());
						for (AirspacesDb::elementvector_t::const_iterator evi(ev.begin()), eve(ev.end()); evi != eve; ++evi) {
							if (evi->get_icao() != "EGPX" || evi->get_typecode() != AirspacesDb::Airspace::typecode_ead ||
							    evi->get_bdryclass() != AirspacesDb::Airspace::bdryclass_ead_fir)
								continue;
							if (!evi->get_bbox().is_inside(wpt.get_coord()))
								continue;
							if (!evi->get_polygon().windingnumber(wpt.get_coord()))
								continue;
							england = true;
							break;
						}
					}
					if (ok || qegpx->is_error()) {
						Glib::RefPtr<Engine::AirspaceResult> q(qegpx);
						qegpx.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qprof0) {
					bool ok(qprof0->is_done());
					if (ok && !england) {
						const TopoDb30::Profile& profile(qprof0->get_result());
						TopoDb30::elev_t maxelev(profile.get_maxelev());
						if (maxelev != TopoDb30::nodata)				
							terrain = std::max(terrain, Point::round<int32_t,double>(Point::m_to_ft_dbl * maxelev));
					}
					if (ok || qprof0->is_error()) {
						Glib::RefPtr<Engine::ElevationProfileResult> q(qprof0);
						qprof0.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (qprof1) {
					bool ok(qprof1->is_done());
					if (ok && !england) {
						const TopoDb30::Profile& profile(qprof1->get_result());
						TopoDb30::elev_t maxelev(profile.get_maxelev());
						if (maxelev != TopoDb30::nodata)				
							terrain = std::max(terrain, Point::round<int32_t,double>(Point::m_to_ft_dbl * maxelev));
					}
					if (ok || qprof1->is_error()) {
						Glib::RefPtr<Engine::ElevationProfileResult> q(qprof1);
						qprof1.reset();
						Engine::AirspaceResult::cancel(q);
					}
				}
				if (!qli && !qlppc && !qnz && !qegtt && !qegpx && !qprof0 && !qprof1)
					break;
				cond.wait(mutex);
			}
		}
	}
	if (england) {
		// quadrantal rules - not binding for VFR, but apply nevertheless
		base = 0;
		if (magcrs < 0x8000) {
			base += 1000;
			if (magcrs >= 0x4000)
				base += 500;
		} else {
			if (magcrs >= 0xc000)
				base += 500;
		}
	} else {
		if (italy_portugal) {
			if (magcrs >= 0x4000 && magcrs < 0xc000)
				base += 1000;
		} else if (newzealand) {
			if (magcrs < 0x4000 || magcrs >= 0xc000)
				base += 1000;
		} else {
			if (magcrs < 0x8000)
				base += 1000;
		}
	}
	{
		int32_t minalt(terrain);
		if (minalt >= 5000)
			minalt += 1000;
		minalt += 1000;
		int32_t alt = ((minalt + 1000 - base) / 2000) * 2000 + base;
		alt = std::max(std::min(alt, (int32_t)32000), (int32_t)0);
		wpt.set_altitude(alt);
	}
}

void FPlanRoute::fix_invalid_altitudes(void)
{
	for (waypoints_t::size_type i(0), n(m_wpts.size()); i < n; ++i)
		fix_invalid_altitudes(i);
}

void FPlanRoute::fix_invalid_altitudes(TopoDb30& topodb, AirspacesDb& aspc)
{
	for (waypoints_t::size_type i(0), n(m_wpts.size()); i < n; ++i)
		fix_invalid_altitudes(i, topodb, aspc);
}

void FPlanRoute::fix_invalid_altitudes(Engine& eng)
{
	for (waypoints_t::size_type i(0), n(m_wpts.size()); i < n; ++i)
		fix_invalid_altitudes(i, eng);
}

Glib::ustring FPlan::get_userdbpath(void)
{
#ifdef __MINGW32__
	Glib::ustring dbname(Glib::build_filename(Glib::get_user_config_dir(), "vfrnav"));
#else
        Glib::ustring dbname;
	if (!GLIB_CHECK_VERSION(2, 36, 0)) {
		bool found(false);
		std::string x(Glib::getenv("HOME", found));
		if (found)
			dbname = x;
	}
	if (dbname.empty())
		dbname = Glib::get_home_dir();
	dbname = Glib::build_filename(dbname, USERDBPATH);
#endif
#if defined(HAVE_GIOMM) && 0
        {
                Glib::RefPtr<Gio::File> dir(Gio::File::create_for_path(dbname));
                try {
#ifdef HAVE_GIOMM_MAKE_DIRECTORY_WITH_PARENTS
                        if (dir->make_directory_with_parents())
#else
                        if (dir->make_directory())
#endif
                                std::cerr << "vfrnav directory (" << dbname << ") created" << std::endl;
                        else
                                throw std::runtime_error("vfrnav directory creation failed");
                } catch (const Gio::Error& e) {
                        if (e.code() != Gio::Error::EXISTS)
                                throw;
                        std::cerr << "vfrnav directory (" << dbname << ") already exists" << std::endl;
                }
        }
#else
        if (g_mkdir_with_parents(dbname.c_str(), 0755)) {
                throw std::runtime_error("vfrnav directory (" + dbname +") creation failed");
        }
#endif
        return dbname;
}

FPlan::FPlan(void)
{
        Glib::ustring dbname(Glib::build_filename(get_userdbpath(), "fplan.db"));
        db.open(dbname);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::FPlan(sqlite3 *dbh)
{
        db.take(dbh);
        if (sqlite3_create_function(db.db(), "simpledist", 4, /*SQLITE_UTF8*/SQLITE_ANY, 0, &dbfunc_simpledist, 0, 0) != SQLITE_OK)
                throw std::runtime_error("FPlanDb: cannot add user function simpledist");
        init_database();
}

FPlan::~FPlan() throw()
{
}

void FPlan::dbfunc_simpledist(sqlite3_context *ctxt, int, sqlite3_value **values)
{
        sqlite3_result_double(ctxt, Point(sqlite3_value_int(values[0]), sqlite3_value_int(values[1]))
                        .spheric_distance_km(Point(sqlite3_value_int(values[2]), sqlite3_value_int(values[3]))));
}

void FPlan::init_database(void)
{
        {
                sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS fplan (ID INTEGER PRIMARY KEY NOT NULL, "
                                              "TIMEOFFBLOCK INTEGER,"
                                              "SAVETIMEOFFBLOCK INTEGER,"
                                              "TIMEONBLOCK INTEGER,"
                                              "SAVETIMEONBLOCK INTEGER,"
                                              "CURWPT INTEGER,"
                                              "NOTE TEXT);");
                cmd.executenonquery();
        }
	{
		static const struct field {
			const char *fldname;
			const char *fldtype;
		} fields[] = {
			{ "ICAO", "TEXT" },
			{ "NAME", "TEXT" },
			{ "LON", "INTEGER" },
			{ "LAT", "INTEGER" },
			{ "ALT", "INTEGER" },
			{ "FREQ", "INTEGER" },
			{ "TIME", "INTEGER" },
			{ "SAVETIME", "INTEGER" },
			{ "FLAGS", "INTEGER" },
			{ "PATHNAME", "TEXT" },
			{ "PATHCODE", "INTEGER" },
			{ "NOTE", "TEXT" },
			{ "WINDDIR", "INTEGER" },
			{ "WINDSPEED", "INTEGER" },
			{ "QFF", "INTEGER" },
			{ "ISAOFFS", "INTEGER" },
			{ "TRUEALT", "INTEGER" },
			{ "TRUETRACK", "INTEGER" },
			{ "TRUEHEADING", "INTEGER" },
			{ "DECLINATION", "INTEGER" },
			{ "DIST", "INTEGER" },
			{ "FUEL", "INTEGER" },
			{ "TAS", "INTEGER" },
			{ "RPM", "INTEGER" },
			{ "MP", "INTEGER" },
			{ "TYPE", "INTEGER" },
			{ "NAVAID", "INTEGER" },
			{ "FLIGHTTIME", "INTEGER" },
			{ 0, 0 }
		};
		{
			std::string flds;
			for (const struct field *f(fields); f->fldname && f->fldtype; ++f)
				flds += std::string(f->fldname) + " " + f->fldtype + ",";
			sqlite3x::sqlite3_command cmd(db, "CREATE TABLE IF NOT EXISTS waypoints (PLANID INTEGER NOT NULL "
						      "CONSTRAINT fk_planid REFERENCES fplan(ID) ON DELETE CASCADE,"
						      "NR INTEGER NOT NULL," + flds +
						      "PRIMARY KEY(PLANID,NR));");
			cmd.executenonquery();
		}
		// Auto Add new columns if they do not yet exist
		{
			sqlite3x::sqlite3_command cmd(db, "SELECT sql FROM sqlite_master WHERE type = \"table\" AND name = \"waypoints\";");
			sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
			if (!cursor.step())
				throw std::runtime_error("FPlan::init_database: unable to determine schema");
			std::string sql(cursor.getstring(0));
			unsigned int fldmask(0);
			for (std::string::const_iterator si(sql.begin()), se(sql.end()); si != se;) {
				if (!std::isalnum(*si)) {
					++si;
					continue;
				}
				std::string::const_iterator si2(si);
				++si2;
				while (si2 != se && std::isalnum(*si2))
					++si2;
				Glib::ustring tok(si, si2);
				si = si2;
				tok = tok.casefold();
				for (unsigned int i = 0; fields[i].fldname && fields[i].fldtype; ++i) {
					if (tok != Glib::ustring(fields[i].fldname).casefold())
						continue;
					fldmask |= 1U << i;
					break;
				}
			}
			for (unsigned int i = 0; fields[i].fldname && fields[i].fldtype; ++i) {
				if (fldmask & (1U << i))
					continue;
				sqlite3x::sqlite3_command cmd(db, std::string("ALTER TABLE waypoints ADD COLUMN ") + fields[i].fldname +
							      " " + fields[i].fldtype + ";");
				cmd.executenonquery();
			}
		}
	}
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanindex ON waypoints (PLANID);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptplanidnr ON waypoints (PLANID, NR);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlon ON waypoints (LON);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptlat ON waypoints (LAT);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wptname ON waypoints (NAME);"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "CREATE INDEX IF NOT EXISTS wpticao ON waypoints (ICAO);"); cmd.executenonquery(); }
}

void FPlan::delete_database(void)
{
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS fplan;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP TABLE IF EXISTS waypoints;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanindex;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptplanidnr;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlon;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptlat;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wptname;"); cmd.executenonquery(); }
        { sqlite3x::sqlite3_command cmd(db, "DROP INDEX IF EXISTS wpticao;"); cmd.executenonquery(); }
        init_database();
}

std::string FPlan::escape_string(const std::string& s)
{
        std::string r;
        r.reserve(s.size());
        for (std::string::const_iterator i1(s.begin()), ie(s.end()); i1 != ie; i1++) {
                std::string::value_type c(*i1);
                if (c == '_' || c == '%')
                        r.push_back('%');
                r.push_back(c);
        }
        return r;
}

std::vector<FPlanWaypoint> FPlan::find_by_name(const Glib::ustring & nm, unsigned int limit, unsigned int skip, comp_t comp)
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE NAME LIKE ? ORDER BY NAME ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(nm));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector<FPlanWaypoint> FPlan::find_by_icao(const Glib::ustring & ic, unsigned int limit, unsigned int skip, comp_t comp)
{
        std::vector<FPlanWaypoint> wpts;
        {
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE ICAO LIKE ? ORDER BY ICAO ASC LIMIT ? OFFSET ?;");
                Glib::ustring pattern(escape_string(ic));
                switch (comp) {
                        case comp_contains:
                                pattern = '%' + pattern;
                                /* fall through */

                        case comp_startswith:
                                pattern += '%';
                                break;

                        case comp_exact:
                        default:
                                break;
                }
                cmd.bind(1, pattern);
                cmd.bind(2, (int)limit);
                cmd.bind(3, (int)skip);
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}

std::vector<FPlanWaypoint> FPlan::find_nearest(const Point& pt, unsigned int limit, unsigned int skip, const Rect& rect)
{
        std::vector<FPlanWaypoint> wpts;
        {
                std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND LON >= ?5 AND LON <= ?7");
                if (rect.get_northeast().get_lon() < rect.get_southwest().get_lon())
                        std::string rectexpr("LAT >= ?6 AND LAT <= ?8 AND (LON <= ?5 OR LON >= ?7)");
                sqlite3x::sqlite3_command cmd(db, "SELECT " + FPlanWaypoint::reader_fields + " FROM waypoints WHERE " + rectexpr + " ORDER BY simpledist(LON,LAT,?1,?2) ASC LIMIT ?3 OFFSET ?4;");
                cmd.bind(1, pt.get_lon());
                cmd.bind(2, pt.get_lat());
                cmd.bind(3, (int)limit);
                cmd.bind(4, (int)skip);
                cmd.bind(5, rect.get_southwest().get_lon());
                cmd.bind(6, rect.get_southwest().get_lat());
                cmd.bind(7, rect.get_northeast().get_lon());
                cmd.bind(8, rect.get_northeast().get_lat());
                sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
                while (cursor.step())
                        wpts.push_back(FPlanWaypoint(cursor));
        }
        return wpts;
}
