/***************************************************************************
 *   Copyright (C) 2007, 2009, 2012, 2013, 2014 by Thomas Sailer           *
 *   t.sailer@alumni.ethz.ch                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "sysdeps.h"

#include <limits>
#include <sqlite3x.hpp>
#include <cstdlib>
#include <cmath>
#include <cstdio>
#include <iomanip>
#include <sstream>
#include "geom.h"

const Point::coord_t Point::lon_min;
const Point::coord_t Point::lon_max;
const Point::coord_t Point::lat_min;
const Point::coord_t Point::lat_max;
const Point Point::invalid = Point(0, 0x80000000);

const float Point::to_deg;
const float Point::to_rad;
const float Point::from_deg;
const float Point::from_rad;

const double Point::to_deg_dbl;
const double Point::to_rad_dbl;
const double Point::from_deg_dbl;
const double Point::from_rad_dbl;

const int32_t Point::pole_lat;

const float Point::earth_radius;
const float Point::km_to_nmi;
const float Point::m_to_ft;
const float Point::ft_to_m;
const float Point::mps_to_kts;

const double Point::earth_radius_dbl;
const double Point::km_to_nmi_dbl;
const double Point::m_to_ft_dbl;
const double Point::ft_to_m_dbl;
const double Point::mps_to_kts_dbl;

const unsigned int Point::setstr_lat;
const unsigned int Point::setstr_lon;
const unsigned int Point::setstr_excess;

void Point::set_invalid(void)
{
	lon = 0;
	lat = 0x80000000;
}

/*
 * cf. Joseph O'Rourke, Computational Geometry in C
 */

/* 
 * softSurfer's (www.softsurfer.com) variant uses less multiplications,
 * but computes the same value
 * http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm
 */

/* twice the signed area of the triangle p0, p1, p2 */

int64_t Point::area2(const Point& p0, const Point& p1, const Point& p2)
{
#if 1
        int64_t p2x = p2.get_lon();
        int64_t p2y = p2.get_lat();
        int64_t p0x = p0.get_lon() - p2x;
        int64_t p0y = p0.get_lat() - p2y;
        int64_t p1x = p1.get_lon() - p2x;
        int64_t p1y = p1.get_lat() - p2y;
        return p0x * (p1y - p0y) - p0y * (p1x - p0x);
#else
        return p0.get_lon() * (int64_t)p1.get_lat() - p0.get_lat() * (int64_t)p1.get_lon() +
                        p2.get_lon() * (int64_t)p0.get_lat() - p2.get_lat() * (int64_t)p0.get_lon() +
                        p1.get_lon() * (int64_t)p2.get_lat() - p1.get_lat() * (int64_t)p2.get_lon();
#endif
}

inline bool Point::is_between_helper(const Point& p0, const Point& p1) const
{
#if 1
	Rect bbox(p0, p0);
	bbox = bbox.add(p1);
	return bbox.is_inside(*this);
#else
        if (p0.get_lon() != p1.get_lon())
                return ((get_lon() >= p0.get_lon() && get_lon() <= p1.get_lon()) ||
                                (get_lon() >= p1.get_lon() && get_lon() <= p0.get_lon()));
        return ((get_lat() >= p0.get_lat() && get_lat() <= p1.get_lat()) ||
                        (get_lat() >= p1.get_lat() && get_lat() <= p0.get_lat()));
#endif
}

inline bool Point::is_strictly_between_helper(const Point& p0, const Point& p1) const
{
	Rect bbox(p0, p0);
	bbox = bbox.add(p1);
	return bbox.is_strictly_inside(*this);
}

inline bool Point::intersect_precheck(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
	int32_t xmin = l0a.get_lon();
	int32_t xmax = l0b.get_lon();
	int32_t ymin = l0a.get_lat();
	int32_t ymax = l0b.get_lat();
	if (xmin > xmax)
		std::swap(xmin, xmax);
	if (ymin > ymax)
		std::swap(ymin, ymax);
	if (l1a.get_lon() > xmax && l1b.get_lon() > xmax)
		return false;
	if (l1a.get_lon() < xmin && l1b.get_lon() < xmin)
		return false;
	if (l1a.get_lat() > ymax && l1b.get_lat() > ymax)
		return false;
	if (l1a.get_lat() < ymin && l1b.get_lat() < ymin)
		return false;
	return true;
}

/*
 * "proper" intersection, i.e. not one endpoint collinear with the other line
 */

bool Point::is_proper_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        if (!intersect_precheck(l0a, l0b, l1a, l1b))
                return false;
        int64_t abc = area2(l0a, l0b, l1a);
        int64_t abd = area2(l0a, l0b, l1b);
        int64_t cda = area2(l1a, l1b, l0a);
        int64_t cdb = area2(l1a, l1b, l0b);
        if (!abc || !abd || !cda || !cdb)
                return false;
        return ((abc >= 0) ^ (abd >= 0)) && ((cda >= 0) ^ (cdb >= 0));
}

/*
 * intersection, including one endpoint lying on the other line
 */

bool Point::is_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        if (!intersect_precheck(l0a, l0b, l1a, l1b))
                return false;
        int64_t abc = area2(l0a, l0b, l1a);
        int64_t abd = area2(l0a, l0b, l1b);
        int64_t cda = area2(l1a, l1b, l0a);
        int64_t cdb = area2(l1a, l1b, l0b);
        if (abc && abd && cda && cdb && ((abc >= 0) ^ (abd >= 0)) && ((cda >= 0) ^ (cdb >= 0)))
                return true;
        if (!abc && l1a.is_between_helper(l0a, l0b))
                return true;
        if (!abd && l1b.is_between_helper(l0a, l0b))
                return true;
        if (!cda && l0a.is_between_helper(l1a, l1b))
                return true;
        if (!cdb && l0b.is_between_helper(l1a, l1b))
                return true;
        return false;
}

/*
 * intersection, excluding one endpoint lying on the other line
 */

bool Point::is_strict_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        if (!intersect_precheck(l0a, l0b, l1a, l1b))
                return false;
        int64_t abc = area2(l0a, l0b, l1a);
        int64_t abd = area2(l0a, l0b, l1b);
        int64_t cda = area2(l1a, l1b, l0a);
        int64_t cdb = area2(l1a, l1b, l0b);
        if (abc && abd && cda && cdb && ((abc >= 0) ^ (abd >= 0)) && ((cda >= 0) ^ (cdb >= 0)))
                return true;
        if (!abc && l1a.is_strictly_between_helper(l0a, l0b))
                return true;
        if (!abd && l1b.is_strictly_between_helper(l0a, l0b))
                return true;
        if (!cda && l0a.is_strictly_between_helper(l1a, l1b))
                return true;
        if (!cdb && l0b.is_strictly_between_helper(l1a, l1b))
                return true;
        return false;
}

bool Point::is_between(const Point& p0, const Point& p1) const
{
        if (area2(p0, p1, *this))
                return false;
        return is_between_helper(p0, p1);
}

bool Point::is_strictly_between(const Point& p0, const Point& p1) const
{
        if (area2(p0, p1, *this))
                return false;
        return is_strictly_between_helper(p0, p1);
}

bool Point::is_convex(const Point& vprev, const Point& vnext) const
{
        return vnext.is_left_or_on(vprev, *this);
}

bool Point::is_in_cone(const Point& vprev, const Point& v, const Point& vnext) const
{
        if (v.is_convex(vprev, vnext))
                return vprev.is_strictly_left(v, *this) && 
                                vnext.is_strictly_left(*this, v);
        return !(vnext.is_left_or_on(v, *this) && 
                        vprev.is_left_or_on(*this, v));
}

/* 
 * based on an algorithm by softSurfer (www.softsurfer.com)
 * http://geometryalgorithms.com/Archive/algorithm_0103/algorithm_0103.htm
 */

/*
 * left_of: tests if a point is Left|On|Right of an infinite line.
 *    Input:  three points P0, P1, and P2
 *    Return: >0 for P2 left of the line through P0 and P1
 *            =0 for P2 on the line
 *            <0 for P2 right of the line
 *    See: the January 2001 Algorithm "Area of 2D and 3D Triangles and Polygons"
 */

inline bool Point::is_left_of(const Point& p0, const Point& p1) const
{
        int64_t lon = area2(p0, p1, *this);
        if (lon < 0)
                return -1;
        if (lon > 0)
                return 1;
        return 0;
}

bool Point::left_of(const Point& p0, const Point& p1) const
{
        int64_t p0x = p0.get_lon() - get_lon();
        int64_t p0y = p0.get_lat() - get_lat();
        int64_t p1x = p1.get_lon() - get_lon();
        int64_t p1y = p1.get_lat() - get_lat();
        int64_t lon = p0x * (p1y - p0y) - p0y * (p1x - p0x);
        if (lon < 0)
                return -1;
        if (lon > 0)
                return 1;
        return 0;
}

bool Point::in_box(const Point& p0, const Point& p1) const
{
	Point x(*this - p0), y(p1 - p0);
	if (y.get_lon() < 0) {
		if (x.get_lon() < y.get_lon() || x.get_lon() > 0)
			return false;
	} else {
		if (x.get_lon() > y.get_lon() || x.get_lon() < 0)
			return false;
	}
	if (y.get_lat() < 0) {
		if (x.get_lat() < y.get_lat() || x.get_lat() > 0)
			return false;
	} else {
		if (x.get_lat() > y.get_lat() || x.get_lat() < 0)
			return false;
	}
	return true;
}

/* http://mathworld.wolfram.com/Line-LineIntersection.html */

bool Point::intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b)
{
        int64_t x1 = l0a.get_lon();
        int64_t y1 = l0a.get_lat();
        int64_t x2 = l0b.get_lon();
        int64_t y2 = l0b.get_lat();
        int64_t x3 = l1a.get_lon();
        int64_t y3 = l1a.get_lat();
        int64_t x4 = l1b.get_lon();
        int64_t y4 = l1b.get_lat();
        int64_t detd = (x1 - x2) * (y3 - y4) - (x3 - x4) * (y1 - y2);
        if (!detd)
                return false;
        int64_t det1 = (x1 * y2) - (x2 * y1);
        int64_t det2 = (x3 * y4) - (x4 * y3);
	if (false)
		std::cerr << "detd " << detd << " det1 " << det1 << " det2 " << det2
			  << " x1-x2 " << (x1 - x2) << " y1-y2 " << (y1 - y2)
			  << " x3-x4 " << (x3 - x4) << " y3-y4 " << (y3 - y4) << std::endl;
#if 1
        set_lon(Point::round<Point::coord_t,double>((det1 * (double)(x3 - x4) - det2 * (double)(x1 - x2)) / detd));
        set_lat(Point::round<Point::coord_t,double>((det1 * (double)(y3 - y4) - det2 * (double)(y1 - y2)) / detd));
#else
        set_lon((det1 * (x3 - x4) - det2 * (x1 - x2) + detd / 2) / detd);
        set_lat((det1 * (y3 - y4) - det2 * (y1 - y2) + detd / 2) / detd);
#endif
        return true;
}

// return the point on an infinite line through la and lb nearest to *this

Point Point::nearest_on_line(const Point& la, const Point& lb)
{
	// a: vector from point to line
	// n: line normal vector
	// x: nearest point on line
	// x = a - (n^T * a) / (n^T * n) * n
	if (is_invalid()) {
		Point pt;
		pt.set_invalid();
		return pt;
	}
	Point pa(la - *this);
	Point pn(lb - la);
	double c(cos(this->get_lat_rad_dbl()));
	if (c < 1e-10) {
		Point pt;
		pt.set_invalid();
		return pt;
	}
	double cc(c * c);
	double ax(pa.get_lon());
	double ay(pa.get_lat());
	double nx(pn.get_lon());
	double ny(pn.get_lat());
	double t(cc * nx * nx + ny * ny);
	if (fabs(t) < 1e-10) {
		Point pt;
		pt.set_invalid();
		return pt;
	}
	t = (cc * nx * ax + ny * ay) / t;
	Point p(ax - t * nx, ay - t * ny);
	return p + *this;
}

Point Point::spheric_nearest_on_line(const Point& la, const Point& lb) const
{
	// http://stackoverflow.com/questions/1299567/how-to-calculate-distance-from-a-point-to-a-line-segment-on-a-sphere
	CartesianVector3Dd A((CartesianVector3Dd)PolarVector3Dd(la));
	CartesianVector3Dd B((CartesianVector3Dd)PolarVector3Dd(lb));
	CartesianVector3Dd C((CartesianVector3Dd)PolarVector3Dd(*this));
	// G = A x B
	CartesianVector3Dd G(A.crossprod(B));
	// F = C x G
	CartesianVector3Dd F(C.crossprod(G));
	// T = G x F
	PolarVector3Dd T(G.crossprod(F));
	Point pt1(T.getcoord());
	Point pt2(-pt1.get_lon(), -pt1.get_lat());
	return (simple_distance_rel(pt1) < simple_distance_rel(pt2)) ? pt1 : pt2;
}

Point Point::simple_unwrap(const Point& p2) const
{
        float r1 = cosf(((*this) + p2).get_lat_rad() * 0.5);
        Point p(p2 - *this);
        p.set_lon((coord_t)(r1 * p.get_lon()));
        return p;
}

float Point::simple_true_course(const Point& p2) const
{
        Point p(simple_unwrap(p2));
        float v = atan2f(p.get_lon(), p.get_lat()) * (180.0 / M_PI);
        if (v < 0)
                v += 360;
        return v;
}

uint64_t Point::simple_distance_rel( const Point & p2 ) const
{
        Point p(simple_unwrap(p2));
        return p.length2();
}

float Point::simple_distance_km(const Point& p2) const
{
        Point p(simple_unwrap(p2));
        return p.length() * (earth_radius * to_rad);
}

Point Point::simple_course_distance_nmi(float course, float dist) const
{
        float r1 = cosf(get_lat_rad());
        float r2 = course * (M_PI / 180.0);
        float x = sinf(r2);
        float y = cosf(r2);
        r2 = r1 * r1 * x * x + y * y;
        if (r2 < 0.001)
                return *this;
        r2 = dist * (1.0 / km_to_nmi / earth_radius / to_rad) / sqrt(r2);
        x *= r2;
        y *= r2;
        return (*this) + Point((coord_t)x, (coord_t)y);
}

Rect Point::simple_box_km(float centerdist) const
{
        centerdist *= (from_rad / earth_radius);
        float distlon = centerdist;
        if (abs(get_lat()) < pole_lat)
                distlon /= cosf(get_lat_rad());
        Point pdist((int32_t)distlon, (int32_t)centerdist);
        Point sw(*this - pdist), ne(*this + pdist);
        if (sw.get_lat() < -pole_lat)
                sw.set_lat(-pole_lat);
        if (ne.get_lat() > pole_lat)
                ne.set_lat(pole_lat);
        return Rect(sw, ne);
}

Rect Point::simple_box_nmi(float centerdist) const
{
        return simple_box_km(centerdist * (float)(1.0 / km_to_nmi));
}

/*
 * Spheric Geometry
 * http://de.wikipedia.org/wiki/Orthodrome
 * http://williams.best.vwh.net/avform.htm
 */

#define USE_HAVERSINE

float Point::spheric_true_course(const Point& p2) const
{
        Point pdiff(p2 - *this);
        float lat1 = get_lat_rad(), lat2 = p2.get_lat_rad();
        float sinlat1 = sinf(lat1), coslat1 = cosf(lat1);
        float sinlat2 = sinf(lat2), coslat2 = cosf(lat2);
        float londiff = pdiff.get_lon_rad();
        float crs = atan2f(sinf(londiff) * coslat2, coslat1 * sinlat2 - sinlat1 * coslat2 * cosf(londiff));
        crs *= (180.0 / M_PI);
        if (crs < 0)
                crs += 360;
        return crs;
}

float Point::spheric_distance_km(const Point& p2) const
{
        Point pdiff(p2 - *this);
#if defined(USE_HAVERSINE)
        float slatd = sinf(pdiff.get_lat_rad() * 0.5f);
        float slond = sinf(pdiff.get_lon_rad() * 0.5f);
        return 2.0f * earth_radius * asinf(std::min(1.f, sqrtf(slatd * slatd + cosf(get_lat_rad()) * cosf(p2.get_lat_rad()) * slond * slond)));
#else
        float lat = pdiff.get_lat_rad();
        float lon = pdiff.get_lon_rad() * cosf(get_lat_rad());
        float dist = acosf(cosf(lat) * cosf(lon));
        //double dist = acos(cos(lat) * cos(lon));
        return dist * earth_radius;
#endif
}

Point Point::spheric_course_distance_nmi(float course, float dist) const
{
        //std::cerr << "spheric_course_distance_nmi: crs " << course << " d " << dist << " pt " << (*this) << std::endl;
        dist *= (1.0f / km_to_nmi / earth_radius);
        course *= (M_PI / 180.f);
        float sintc = sinf(course), costc = cosf(course);
        float lat1 = get_lat_rad();
        float sinlat1 = sinf(lat1), coslat1 = cosf(lat1);
        float cosd = cosf(dist), sind = sinf(dist);
        float sinlat = sinlat1 * cosd + coslat1 * sind * costc;
        float lat = asinf(sinlat);
#if 1
        float dlon = atan2f(sintc * sind * coslat1, cosd - sinlat1 * sinlat);
#else
        float dlon = 0;
        float coslat = cosf(lat);
        if (fabsf(coslat) > 1e-10)
                dlon = asinf(sintc * sind / coslat);
#endif
        Point p;
        p.set_lat_rad(lat);
        p.set_lon_rad(dlon);
        p.set_lon(get_lon() + p.get_lon());
        return p;
}

double Point::spheric_true_course_dbl(const Point& p2) const
{
        Point pdiff(p2 - *this);
        double lat1 = get_lat_rad_dbl(), lat2 = p2.get_lat_rad_dbl();
        double sinlat1 = sin(lat1), coslat1 = cos(lat1);
        double sinlat2 = sin(lat2), coslat2 = cos(lat2);
        double londiff = pdiff.get_lon_rad();
        double crs = atan2(sin(londiff) * coslat2, coslat1 * sinlat2 - sinlat1 * coslat2 * cos(londiff));
        crs *= (180.0 / M_PI);
        if (crs < 0)
                crs += 360;
        return crs;
}

double Point::spheric_distance_km_dbl(const Point& p2) const
{
        Point pdiff(p2 - *this);
#if defined(USE_HAVERSINE)
        double slatd = sin(pdiff.get_lat_rad_dbl() * 0.5);
        double slond = sin(pdiff.get_lon_rad_dbl() * 0.5);
        return 2.0 * earth_radius_dbl * asin(std::min(1., sqrt(slatd * slatd + cos(get_lat_rad_dbl()) * cos(p2.get_lat_rad_dbl()) * slond * slond)));
#else
        double lat = pdiff.get_lat_rad_dbl();
        double lon = pdiff.get_lon_rad_dbl() * cos(get_lat_rad_dbl());
        double dist = acos(cos(lat) * cos(lon));
        return dist * earth_radius_dbl;
#endif
}

Point Point::halfway(const Point& pt2) const
{
        Point pd(pt2 - *this);
        pd = Point(pd.get_lon() / 2, pd.get_lat() / 2);
        return *this + pd;
}

Glib::ustring Point::to_ustring(const std::wstring& str)
{
	GError *error(0);

#if SIZEOF_WCHAR_T == 4
	glong n_bytes(0);
	const Glib::ScopedPtr<char> buf(g_ucs4_to_utf8(reinterpret_cast<const gunichar*>(str.data()), str.size(), 0, &n_bytes, &error));
#elif defined(G_OS_WIN32) && SIZEOF_WCHAR_T == 2
	glong n_bytes(0);
	const Glib::ScopedPtr<char> buf(g_utf16_to_utf8(reinterpret_cast<const gunichar2*>(str.data()), str.size(), 0, &n_bytes, &error));
#else
	gsize n_bytes(0);
	const Glib::ScopedPtr<char> buf(g_convert(reinterpret_cast<const char*>(str.data()), str.size() * sizeof(std::wstring::value_type), "UTF-8", "WCHAR_T", 0, &n_bytes, &error));
#endif
	if (error)
		Glib::Error::throw_exception(error);
	return Glib::ustring(buf.get(), buf.get() + n_bytes);
}

Glib::ustring Point::get_lon_str(void) const
{
	int32_t min(get_lon());
	char sgn((min < 0) ? 'W' : 'E');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 6000000));
	int32_t deg(min / 6000000);
	min -= deg * 6000000;
	std::wostringstream oss;
	oss << sgn << std::setw(3) << std::setfill(L'0') << deg << (wchar_t)0xb0
	    << std::fixed << std::setw(8) << std::setfill(L'0') << std::setprecision(5) << (min * 0.00001) << '\'';
	return to_ustring(oss.str());
}

Glib::ustring Point::get_lat_str(void) const
{
	int32_t min(get_lat());
	char sgn((min < 0) ? 'S' : 'N');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 6000000));
	int32_t deg(min / 6000000);
	min -= deg * 6000000;
	std::wostringstream oss;
	oss << sgn << std::setw(2) << std::setfill(L'0') << deg << (wchar_t)0xb0
	    << std::fixed << std::setw(8) << std::setfill(L'0') << std::setprecision(5) << (min * 0.00001) << '\'';
	return to_ustring(oss.str());
}

Glib::ustring Point::get_lon_str2(void) const
{
	int32_t min(get_lon());
	char sgn((min < 0) ? 'W' : 'E');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 6000000));
	int32_t deg(min / 6000000);
	min -= deg * 6000000;
	std::wostringstream oss;
	oss << sgn << std::setw(3) << std::setfill(L'0') << deg << L' '
	    << std::fixed << std::setw(8) << std::setfill(L'0') << std::setprecision(5) << (min * 0.00001);
	return to_ustring(oss.str());
}

Glib::ustring Point::get_lat_str2(void) const
{
	int32_t min(get_lat());
	char sgn((min < 0) ? 'S' : 'N');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 6000000));
	int32_t deg(min / 6000000);
	min -= deg * 6000000;
	std::wostringstream oss;
	oss << sgn << std::setw(2) << std::setfill(L'0') << deg << L' '
	    << std::fixed << std::setw(8) << std::setfill(L'0') << std::setprecision(5) << (min * 0.00001);
	return to_ustring(oss.str());
}

Glib::ustring Point::get_lon_fpl_str(void) const
{
	int32_t min(get_lon());
	char sgn((min < 0) ? 'W' : 'E');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 60));
	int32_t deg(min / 60);
	min -= deg * 60;
	std::wostringstream oss;
	oss << std::setw(3) << std::setfill(L'0') << deg << std::setw(2) << std::setfill(L'0') << min << sgn;
	return to_ustring(oss.str());
}

Glib::ustring Point::get_lat_fpl_str(void) const
{
	int32_t min(get_lat());
	char sgn((min < 0) ? 'S' : 'N');
	min = round<int32_t,double>(abs(min) * (to_deg_dbl * 60));
	int32_t deg(min / 60);
	min -= deg * 60;
	std::wostringstream oss;
	oss << std::setw(2) << std::setfill(L'0') << deg << std::setw(2) << std::setfill(L'0') << min << sgn;
	return to_ustring(oss.str());
}

Glib::ustring Point::get_fpl_str(void) const
{
	return get_lat_fpl_str() + get_lon_fpl_str();
}

std::ostream& Point::print(std::ostream& os) const
{
        os << '(' << (std::string)get_lon_str() << ',' << (std::string)get_lat_str() << ')';
        return os;
}

bool Point::set_lon_str(const Glib::ustring& str)
{
        char hemi;
        double deg, min, sec;
        Point::coord_t c(0);
        bool sok = sscanf(str.c_str(), "%c%lf\302\260%lf'%lf\"", &hemi, &deg, &min, &sec) == 4;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf %lf", &hemi, &deg, &min, &sec) == 4;
        if (sok) {
                if (hemi != 'E' && hemi != 'e' && hemi != 'W' && hemi != 'w')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min + (1.0 / 60.0 / 60.0) * sec) * Point::from_deg);
                if (hemi == 'W' || hemi == 'w')
                        c = -c;
                goto ok;
        }
        sok = sscanf(str.c_str(), "%c%lf\302\260%lf'", &hemi, &deg, &min) == 3;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf", &hemi, &deg, &min) == 3;
        if (sok) {
                if (hemi != 'E' && hemi != 'e' && hemi != 'W' && hemi != 'w')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min) * Point::from_deg);
                if (hemi == 'W' || hemi == 'w')
                        c = -c;
                goto ok;
        }
err:
        return false;

ok:
        lon = c;
        return true;
}

bool Point::set_lat_str(const Glib::ustring& str)
{
        char hemi;
        double deg, min, sec;
        Point::coord_t c(0);
        bool sok = sscanf(str.c_str(), "%c%lf\302\260%lf'%lf\"", &hemi, &deg, &min, &sec) == 4;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf %lf", &hemi, &deg, &min, &sec) == 4;
        if (sok) {
                if (hemi != 'N' && hemi != 'n' && hemi != 'S' && hemi != 's')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min + (1.0 / 60.0 / 60.0) * sec) * Point::from_deg);
                if (hemi == 'S' || hemi == 's')
                        c = -c;
                goto ok;
        }
        sok = sscanf(str.c_str(), "%c%lf\302\260%lf'", &hemi, &deg, &min) == 3;
        if (!sok)
                sok = sscanf(str.c_str(), "%c%lf %lf", &hemi, &deg, &min) == 3;
        if (sok) {
                if (hemi != 'N' && hemi != 'n' && hemi != 'S' && hemi != 's')
                        goto err;
                c = (Point::coord_t)((deg + (1.0 / 60.0) * min) * Point::from_deg);
                if (hemi == 'S' || hemi == 's')
                        c = -c;
                goto ok;
        }
err:
        return false;

ok:
        lat = c;
        return true;
}

void Point::set_ch1903(float x, float y)
{
        if (y < x)
		std::swap(x, y);
        y = (y - 600000) * 1e-6;
        x = (x - 200000) * 1e-6;
        float lambda = 2.6779094
                        + 4.728982 * y
                        + 0.791484 * y * x
                        + 0.1306 * y * x * x
                        - 0.0436 * y * y * y;
        float phi = 16.9023892
                        + 3.238272 * x
                        - 0.270978 * y * y
                        - 0.002528 * x * x
                        - 0.0447 * y * y * x
                        - 0.0140 * x * x * x;
	set_lon(round<coord_t,float>(lambda * (100.0 / 36.0 * Point::from_deg)));
	set_lat(round<coord_t,float>(phi * (100.0 / 36.0 * Point::from_deg)));
}

bool Point::get_ch1903(float& x, float& y) const
{
        float phi = (get_lat() - ((int32_t)(169028.66 * Point::from_deg / 60.0 / 60.0))) * (Point::to_deg * 60.0 * 60.0 / 10000.0);
        float lambda = (get_lon() - ((int32_t)(26782.5 * Point::from_deg / 60.0 / 60.0))) * (Point::to_deg * 60.0 * 60.0 / 10000.0);
        float yy = 600072.37
                 + 211455.93 * lambda
                 -  10938.51 * lambda * phi
                 -      0.36 * lambda * phi * phi
                 -     44.54 * lambda * lambda * lambda;
        float xx = 200147.07
                 + 308807.95 * phi
                 +   3745.25 * lambda * lambda
                 +     76.63 * phi * phi
                 -    194.56 * lambda * lambda * phi
                 +    119.79 * lambda * lambda * lambda;
	x = xx;
	y = yy;
	return (xx >= 0) && (xx < 400000) && (yy >= 400000) && (yy < 1000000);
}

const unsigned int Point::maidenhead_maxdigits;

const unsigned int Point::maidenhead_max[maidenhead_maxdigits] = {
	18, 10, 24, 10, 24
};

const double Point::maidenhead_factors[maidenhead_maxdigits] = {
	20 * from_deg_dbl,
	2 * from_deg_dbl,
	2.0/24.0 * from_deg_dbl,
	2.0/240.0 * from_deg_dbl,
	2.0/240.0/24.0 * from_deg_dbl
};

Glib::ustring Point::get_locator(unsigned int digits) const
{
	digits = std::min(digits >> 1, maidenhead_maxdigits);
	if (!digits)
		return Glib::ustring();
	double lon(get_lon()), lat(get_lat());
	lat *= 2;
	{
		double rnd(0.5 * maidenhead_factors[digits-1] + 180 * from_deg_dbl);
		lon += rnd;
		lat += rnd;
	}
	Glib::ustring r;
	for (unsigned int i = 0; i < digits; ++i) {
		int londig, latdig;
		{
			double f(1 / maidenhead_factors[i]);
			londig = floor(lon * f);
			latdig = floor(lat * f);
		}
		londig = std::max(std::min(londig, (int)maidenhead_max[i]-1), 0);
		latdig = std::max(std::min(latdig, (int)maidenhead_max[i]-1), 0);
		lon -= londig * maidenhead_factors[i];
		lat -= latdig * maidenhead_factors[i];
		if (i & 1) {
			r += (char)('0' + londig);
			r += (char)('0' + latdig);
			continue;
		}
		r += (char)('A' + londig);
		r += (char)('A' + latdig);
	}
	return r;	
}

unsigned int Point::set_str_ch1903(Glib::ustring::const_iterator& si, Glib::ustring::const_iterator se)
{
	Glib::ustring::const_iterator si2(si);
	if (si2 == se || (*si2 != 'C' && *si2 != 'c'))
		return 0;
	++si2;
	if (si2 == se || (*si2 != 'H' && *si2 != 'h'))
		return 0;
	++si2;
	float x[2];
	unsigned int nr(0);
	while (si2 != se) {
		if (std::isspace(*si2)) {
			++si2;
			continue;
		}
		if (*si2 < '0' || *si2 > '9')
			break;
		Glib::ustring::const_iterator si3(si2);
		bool dot(true);
		do {
			if (*si3 == '.')
				dot = false;
			++si3;
		} while (si3 != se && ((*si3 >= '0' && *si3 <= '9') || (*si3 == '.' && dot)));
		x[nr++] = Glib::Ascii::strtod(Glib::ustring(si2, si3));
		si2 = si3;
		if (nr == 2)
			break;
	}
	if (nr < 2 || x[0] >= 1000000 || x[1] >= 1000000)
		return 0;
	set_ch1903(x[0], x[1]);
	return setstr_lat | setstr_lon;
}

unsigned int Point::set_str_locator(Glib::ustring::const_iterator& si, Glib::ustring::const_iterator se)
{
	unsigned int buf[2 * maidenhead_maxdigits];
	{
		unsigned int cnt(0);
		Glib::ustring::const_iterator si2(si);
		for (; si2 != se && cnt < 2 * maidenhead_maxdigits; ++cnt, ++si2) {
			if (cnt & 2) {
				if (*si2 < '0' || *si2 > '9')
					break;
				buf[cnt] = *si2 - '0';
			} else {
				if (*si2 >= 'a' && *si2 <= 'z')
					buf[cnt] = *si2 - 'a';
				else if (*si2 >= 'A' && *si2 <= 'Z')
					buf[cnt] = *si2 - 'A';
				else
					break;
			}
			if (buf[cnt] >= maidenhead_max[cnt >> 1])
				break;
		}
		if (false)
			std::cerr << "set_str_locator: " << cnt << " digits" << std::endl;
		if (cnt < 6)
			return 0;
		cnt &= ~1;
		for (unsigned int i = 0; i < cnt; ++i)
			++si;
		for (; cnt < 2 * maidenhead_maxdigits; ++cnt)
			buf[cnt] = maidenhead_max[cnt >> 1] >> 1;
	}
	double lon(0.5 * maidenhead_factors[maidenhead_maxdigits-1] - 180 * from_deg_dbl);
	double lat(lon);
	for (unsigned int i = 0; i < maidenhead_maxdigits; ++i) {
		lon += maidenhead_factors[i] * buf[2 * i + 0];
		lat += maidenhead_factors[i] * buf[2 * i + 1];
	}
	lat *= 0.5;
	set_lat(round<coord_t,double>(lat));
	set_lon(round<coord_t,double>(lon));
	return setstr_lat | setstr_lon;
}

unsigned int Point::set_str(const Glib::ustring& str)
{
	unsigned int r(0);
	{
		static const double factors[3] = {
			Point::from_deg_dbl, Point::from_deg_dbl / 60.0, Point::from_deg_dbl / 60.0 / 60.0
		};
		Glib::ustring::const_iterator si(str.begin()), se(str.end());
		unsigned int pre(0), precnt(0), cnt(0);
		double prefrac(0);
		int hemimul(0);
		bool hemilon(false), dot(true);
		while (si != se) {
			if (std::isspace(*si)) {
				++si;
				continue;
			}
			if (!hemimul && !precnt && ((*si >= 'A' && *si <= 'R') || (*si >= 'a' && *si <= 'r'))) {
				unsigned int r2(set_str_locator(si, se));
				if (r2) {
					r |= r2;
					continue;
				}
			}
			if (!hemimul && !precnt && (*si == 'c' || *si == 'C')) {
				unsigned int r2(set_str_ch1903(si, se));
				if (r2) {
					r |= r2;
					continue;
				}
			}
			if (*si >= '0' && *si <= '9' && !hemimul) {
				if (precnt)
					break;
				bool predot(true);
				double fracmul(1);
				do {
					if (*si == '.') {
						++si;
						predot = false;
						continue;
					}
					if (predot) {
						pre *= 10;
						pre += (*si++) - '0';
						++precnt;
						continue;
					}
					fracmul *= 0.1;
					prefrac += fracmul * ((*si++) - '0');
				} while (si != se && ((*si >= '0' && *si <= '9') || (predot && *si == '.')));
				continue;
			}
			if (hemimul && cnt < sizeof(factors)/sizeof(factors[0]) &&
			    ((*si >= '0' && *si <= '9') || (dot && *si == '.'))) {
				{
					Glib::ustring::const_iterator si2(si);
					do {
						if (*si2 == '.')
							dot = false;
						++si2;
					} while (si2 != se && ((*si2 >= '0' && *si2 <= '9') || (dot && *si2 == '.')));
					Point::coord_t c(round<coord_t,double>(Glib::Ascii::strtod(Glib::ustring(si, si2)) * factors[cnt]));
					c *= hemimul;
					si = si2;
					if (hemilon) {
						if (cnt)
							c += get_lon();
						set_lon(c);
						r |= setstr_lon;
					} else {
						if (cnt)
							c += get_lat();
						set_lat(c);
						r |= setstr_lat;
					}
				}
				while (si != se && (std::isspace(*si) ||
						    (*si == (gunichar)176 && cnt == 0) ||
						    (*si == '\'' && cnt == 1) ||
						    (*si == '"' && cnt == 2)))
					++si;
				if (false && si != se)
					std::cerr << "Coordinate parsing level " << cnt << " stopped at " << (uint32_t)*si << std::endl;
				++cnt;
				continue;
			}
			if (*si == 'N' || *si == 'n') {
				hemimul = 1;
				hemilon = false;
			} else if (*si == 'S' || *si == 's') {
				hemimul = -1;
				hemilon = false;
			} else if (*si == 'E' || *si == 'e') {
				hemimul = 1;
				hemilon = true;
			} else if (*si == 'W' || *si == 'w') {
				hemimul = -1;
				hemilon = true;
			} else {
				if (r)
					r |= setstr_excess;
				break;
			}
			++si;
			cnt = 0;
			if (precnt == 6U + hemilon || precnt == 4U + hemilon || precnt == 2U + hemilon) {
				Point::coord_t c;
				if (precnt >= 6) {
					unsigned int deg(pre / 10000);
					pre -= deg * 10000;
					unsigned int min(pre / 100);
					pre -= min * 100;
					c = round<coord_t,double>(deg * factors[0] + min * factors[1] + (pre + prefrac) * factors[2]);
				} else if (precnt >= 4) {
					unsigned int deg(pre / 100);
					pre -= deg * 100;
					c = round<coord_t,double>(deg * factors[0] + (pre + prefrac) * factors[1]);
				} else {
					c = round<coord_t,double>((pre + prefrac) * factors[0]);
				}
				c *= hemimul;
				if (hemilon) {
					set_lon(c);
					r |= setstr_lon;
				} else {
					set_lat(c);
					r |= setstr_lat;
				}
				pre = precnt = 0;
				prefrac = 0;
				hemimul = 0;
				hemilon = false;
				dot = true;
				continue;
			}
			if (precnt) {
				if (r)
					r |= setstr_excess;
				break;
			}
			dot = true;
		}
		if (hemimul && !cnt && r)
			r |= setstr_excess;
	}
	return r;
}

int64_t Rect::get_east_unwrapped(void) const
{
        int64_t r(get_east());
        if (get_east() < get_west())
                r += 0x100000000LL;
        return r;
}

bool Rect::is_inside(const Point& pt) const
{
	Point p0(pt - get_southwest());
	Point p1(get_northeast() - get_southwest());
	return ((uint32_t)p0.get_lon() <= (uint32_t)p1.get_lon()) &&
		((uint32_t)p0.get_lat() <= (uint32_t)p1.get_lat());
}

bool Rect::is_strictly_inside(const Point& pt) const
{
	Point p0(pt - get_southwest());
	Point p1(get_northeast() - get_southwest());
	return ((uint32_t)p0.get_lon() > 0) &&
		((uint32_t)p0.get_lat() > 0) &&
		((uint32_t)p0.get_lon() < (uint32_t)p1.get_lon()) &&
		((uint32_t)p0.get_lat() < (uint32_t)p1.get_lat());
}

bool Rect::is_lon_inside(const Point& pt) const
{
        Point p0(pt - get_southwest());
        Point p1(get_northeast() - get_southwest());
        return ((uint32_t)p0.get_lon() <= (uint32_t)p1.get_lon());
}

bool Rect::is_lat_inside(const Point& pt) const
{
        Point p0(pt - get_southwest());
        Point p1(get_northeast() - get_southwest());
        return ((uint32_t)p0.get_lat() <= (uint32_t)p1.get_lat());
}

uint64_t Rect::simple_distance_rel(const Point& pt) const
{
        if (is_inside(pt))
                return 0;
        uint64_t r0 = pt.simple_distance_rel(get_southwest());
        uint64_t r1 = pt.simple_distance_rel(get_northeast());
        uint64_t r2 = pt.simple_distance_rel(get_southeast());
        uint64_t r3 = pt.simple_distance_rel(get_northwest());
        uint64_t r = std::min(std::min(r0, r1), std::min(r2, r3));
        if (is_lon_inside(pt)) {
                uint64_t r0 = pt.simple_distance_rel(Point(pt.get_lon(), get_north()));
                uint64_t r1 = pt.simple_distance_rel(Point(pt.get_lon(), get_south()));
                r = std::min(r, std::min(r0, r1));
        }
        if (is_lat_inside(pt)) {
                uint64_t r0 = pt.simple_distance_rel(Point(get_east(), pt.get_lat()));
                uint64_t r1 = pt.simple_distance_rel(Point(get_west(), pt.get_lat()));
                r = std::min(r, std::min(r0, r1));
        }
        return r;
}

float Rect::simple_distance_km(const Point& pt) const
{
        return sqrt(simple_distance_rel(pt)) * (Point::earth_radius * Point::to_rad);
}

float Rect::spheric_distance_km(const Point& pt) const
{
        if (is_inside(pt))
                return 0;
        float r0 = pt.spheric_distance_km(get_southwest());
        float r1 = pt.spheric_distance_km(get_northeast());
        float r2 = pt.spheric_distance_km(get_southeast());
        float r3 = pt.spheric_distance_km(get_northwest());
        float r = std::min(std::min(r0, r1), std::min(r2, r3));
        if (is_lon_inside(pt)) {
                float r0 = pt.spheric_distance_km(Point(pt.get_lon(), get_north()));
                float r1 = pt.spheric_distance_km(Point(pt.get_lon(), get_south()));
                r = std::min(r, std::min(r0, r1));
        }
        if (is_lat_inside(pt)) {
                float r0 = pt.spheric_distance_km(Point(get_east(), pt.get_lat()));
                float r1 = pt.spheric_distance_km(Point(get_west(), pt.get_lat()));
                r = std::min(r, std::min(r0, r1));
        }
        return r;
}

std::pair<bool,bool> Rect::is_overlap_helper(const Rect& r) const
{
	Point p0(r.get_southwest() - get_southwest());
	Point p1(r.get_northeast() - get_southwest());
	Point pdim(get_northeast() - get_southwest());
	return std::pair<bool,bool>((uint32_t)p0.get_lon() <= (uint32_t)pdim.get_lon() ||
				    (uint32_t)p1.get_lon() <= (uint32_t)pdim.get_lon(),
				    (uint32_t)p0.get_lat() <= (uint32_t)pdim.get_lat() ||
				    (uint32_t)p1.get_lat() <= (uint32_t)pdim.get_lat());
}

bool Rect::is_intersect(const Rect& r) const
{
	std::pair<bool,bool> ol0(is_overlap_helper(r));
	std::pair<bool,bool> ol1(r.is_overlap_helper(*this));
	return (ol0.first || ol1.first) && (ol0.second  || ol1.second);
}

bool Rect::is_intersect(const Point& la, const Point& lb) const
{
	return Point::is_intersection(la, lb, get_southwest(), get_southeast()) ||
		Point::is_intersection(la, lb, get_southeast(), get_northeast()) ||
		Point::is_intersection(la, lb, get_northeast(), get_northwest()) ||
		Point::is_intersection(la, lb, get_northwest(), get_southwest());
}

uint64_t Rect::simple_distance_rel(const Rect& r) const
{
        if (is_intersect(r))
                return 0;
        uint64_t r0 = simple_distance_rel(get_southwest());
        uint64_t r1 = simple_distance_rel(get_northeast());
        uint64_t r2 = simple_distance_rel(get_southeast());
        uint64_t r3 = simple_distance_rel(get_northwest());
        return std::min(std::min(r0, r1), std::min(r2, r3));
}

float Rect::simple_distance_km(const Rect& r) const
{
        return sqrt(simple_distance_rel(r)) * (Point::earth_radius * Point::to_rad);
}

float Rect::spheric_distance_km(const Rect& r) const
{
        if (is_intersect(r))
                return 0;
        float r0 = spheric_distance_km(get_southwest());
        float r1 = spheric_distance_km(get_northeast());
        float r2 = spheric_distance_km(get_southeast());
        float r3 = spheric_distance_km(get_northwest());
        return std::min(std::min(r0, r1), std::min(r2, r3));
}

std::ostream & Rect::print(std::ostream& os) const
{
        return os << get_southwest() << '-' << get_northeast();
}

Rect Rect::oversize_nmi(float nmi) const
{
	return oversize_km(nmi * (float)(1.0 / Point::km_to_nmi));
}

Rect Rect::oversize_km(float km) const
{
	Rect r(*this);
        km *= (Point::from_rad / Point::earth_radius);
	uint64_t londiff((uint32_t)(r.get_east() - r.get_west()));
	{
		float distlon = km;
		if (abs(r.get_south()) < Point::pole_lat)
			distlon /= cosf(r.get_southwest().get_lat_rad());
		Point::coord_t dl(distlon);
		r.set_west(r.get_west() - dl);
		londiff += dl;
	}
	{
		float distlon = km;
		if (abs(r.get_north()) < Point::pole_lat)
			distlon /= cosf(r.get_northeast().get_lat_rad());
		Point::coord_t dl(distlon);
		londiff += dl;
	}
	londiff = std::min(londiff, static_cast<uint64_t>(std::numeric_limits<uint32_t>::max()));
	r.set_east(r.get_west() + londiff);
	r.set_north(std::min((Point::coord_t)(r.get_north() + km), Point::pole_lat));
	r.set_south(std::max((Point::coord_t)(r.get_south() - km), -Point::pole_lat));
        return r;
}

Rect Rect::add(const Point& pt) const
{
	Point sw(get_southwest());
	Point p0(get_northeast() - sw);
	Point p1(pt - sw);
	Rect ret(Point(std::min((Point::coord_t)0, p1.get_lon()), std::min((Point::coord_t)0, p1.get_lat())) + sw,
		 Point(std::max(p0.get_lon(), p1.get_lon()), std::max(p0.get_lat(), p1.get_lat())) + sw);
	return ret;
}

bool Rect::is_empty(void) const
{
	return get_south() > get_north() || get_west() > get_east();
}

void Rect::set_empty(void)
{
	southwest = Point(std::numeric_limits<Point::coord_t>::max(),
			  std::numeric_limits<Point::coord_t>::max());
	northeast = Point(std::numeric_limits<Point::coord_t>::min(),
			  std::numeric_limits<Point::coord_t>::min());
}

int64_t Rect::area2(void) const
{
        /* http://mathworld.wolfram.com/PolygonArea.html */
        /* loop through all edges of the rect, in the order sw se ne nw - sw */
	/* area += (x1 * y2 - x2 * y1) */
	int64_t area = get_west() * (int64_t)get_south() - get_east() * (int64_t)get_south() + // sw se
		get_east() * (int64_t)get_north() - get_east() * (int64_t)get_south() + // se ne
		get_east() * (int64_t)get_north() - get_west() * (int64_t)get_north() + // ne nw
		get_west() * (int64_t)get_south() - get_west() * (int64_t)get_north(); // nw sw
        return area;
}

uint64_t Rect::simple_circumference_rel(void) const
{
	uint64_t d = get_southwest().simple_distance_rel(get_southeast()) +
		get_southeast().simple_distance_rel(get_northeast()) +
		get_northeast().simple_distance_rel(get_northwest()) +
		get_northwest().simple_distance_rel(get_southwest());
	return d;
}

float Rect::simple_circumference_km(void) const
{
	float d = get_southwest().simple_distance_km(get_southeast()) +
		get_southeast().simple_distance_km(get_northeast()) +
		get_northeast().simple_distance_km(get_northwest()) +
		get_northwest().simple_distance_km(get_southwest());
	return d;
}

float Rect::spheric_circumference_km(void) const
{
	float d = get_southwest().spheric_distance_km(get_southeast()) +
		get_southeast().spheric_distance_km(get_northeast()) +
		get_northeast().spheric_distance_km(get_northwest()) +
		get_northwest().spheric_distance_km(get_southwest());
	return d;
}

double Rect::spheric_circumference_km_dbl(void) const
{
	double d = get_southwest().spheric_distance_km_dbl(get_southeast()) +
		get_southeast().spheric_distance_km_dbl(get_northeast()) +
		get_northeast().spheric_distance_km_dbl(get_northwest()) +
		get_northwest().spheric_distance_km_dbl(get_southwest());
	return d;
}

double Rect::get_simple_area_km2_dbl(void) const
{
	Point center(boxcenter());
	if (center.is_invalid())
		return std::numeric_limits<double>::quiet_NaN();
	// area2 delivers the double area
	double a(area2());
	a *= (0.5 * Point::earth_radius_dbl * Point::to_rad_dbl *
	      Point::earth_radius_dbl * Point::to_rad_dbl);
	a *= cos(center.get_lat_rad());
	return a;
}

void PolygonSimple::ScanLine::insert(Point::coord_t lon, int wn)
{
	std::pair<iterator,bool> i(this->insert(value_type(lon, 0)));
	i.first->second += wn;
}

PolygonSimple::ScanLine PolygonSimple::ScanLine::integrate(Point::coord_t lon) const
{
	const_iterator istart(this->lower_bound(lon)), b(this->begin()), e(this->end());
	if (istart == e) {
		istart = b;
		if (istart == e)
			return ScanLine();
	}
	const_iterator icur(istart);
	int wn(0);
	ScanLine sl;
	const Point::coord_t min(std::numeric_limits<Point::coord_t>::min());
	do {
		// compute next iterator (wrap around)
		const_iterator inext(icur);
		++inext;
		if (inext == e)
			inext = b;
		// update winding number
		wn += icur->second;
		sl.insert(value_type(icur->first, wn));
		// check wrap around
		if (wn && inext->first < icur->first && inext->first != min)
			sl.insert(value_type(min, wn));
		// advance to next
		icur = inext;
	} while (icur != istart);
	if (wn) {
		std::ostringstream oss;
		oss << "PolygonSimple::ScanLine::integrate: nonzero winding number integral: " << wn
		    << " lon " << lon;
		for (const_iterator i(b); i != e; ++i)
			oss << " (" << i->first << ',' << i->second << ')';
		throw std::runtime_error(oss.str());
	}
	return sl;
}

PolygonSimple::ScanLine PolygonSimple::ScanLine::combine(const ScanLine& sl2, int factor1, int factor2) const
{
	int wn1(0), wn2(0);
	const_iterator i1(this->begin()), e1(this->end());
	const_iterator i2(sl2.begin()), e2(sl2.end());
	ScanLine sl;
	int wn(0);
	for (;;) {
		Point::coord_t lon;
		{
			bool set1(false), set2(false);
			if (i1 != e1) {
				set1 = true;
				lon = i1->first;
			}
			if (i2 != e2) {
				if (!set1 || i2->first < lon) {
					set2 = true;
					set1 = false;
					lon = i2->first;
				} else if (i2->first == lon) {
					set2 = true;
				}
			}
			if (!(set1 && set2))
				break;
			if (set1) {
				wn1 = i1->second;
				++i1;
			}
			if (set2) {
				wn2 = i2->second;
				++i2;
			}
		}
		int wnnew(wn1 * factor1 + wn2 * factor2);
		if (wnnew != wn) {
			wn = wnnew;
			sl.insert(value_type(lon, wn));
		}
	}
	return sl;
}

Rect::operator PolygonSimple(void) const
{
	PolygonSimple p;
	p.push_back(get_southwest());
	p.push_back(get_southeast());
	p.push_back(get_northeast());
	p.push_back(get_northwest());
	return p;
}

std::ostream& operator<<(std::ostream& os, const Point& p)
{
        return p.print(os);
}

std::ostream & operator<<(std::ostream & os, const Rect & r)
{
        return r.print(os);
}

PolygonSimple::reference PolygonSimple::operator[](size_type x)
{
        size_type y(size());
        if (x >= y && y > 0)
                x %= y;
        return base_t::operator[](x);
}

PolygonSimple::const_reference PolygonSimple::operator[](size_type x) const
{
        size_type y(size());
        if (x >= y && y > 0)
                x %= y;
        return base_t::operator[](x);
}

bool PolygonSimple::is_on_boundary( const Point & pt ) const
{
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, ++i)
                if (pt.is_between(operator[](i), operator[](j)))
                        return true;
        return false;
}

bool PolygonSimple::is_intersection(const Point & la, const Point & lb) const
{
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, ++i)
                if (Point::is_intersection(la, lb, operator[](i), operator[](j)))
                        return true;
        return false;
}

bool PolygonSimple::is_strict_intersection(const Point & la, const Point & lb) const
{
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, ++i)
                if (Point::is_strict_intersection(la, lb, operator[](i), operator[](j)))
                        return true;
        return false;
}

bool PolygonSimple::is_intersection(const Rect & r) const
{
        return is_intersection(r.get_southwest(), r.get_northwest())
            || is_intersection(r.get_northwest(), r.get_northeast())
            || is_intersection(r.get_northeast(), r.get_southeast())
            || is_intersection(r.get_southeast(), r.get_southwest());
}

int PolygonSimple::windingnumber(const Point & pt) const
{
        int wn = 0; /* the winding number counter */

        /* loop through all edges of the polygon */
	const unsigned int sz(size());
        unsigned int j = sz - 1;
        for (unsigned int i = 0; i < sz; j = i, ++i) {
		const Point& pti(operator[](i));
		const Point& ptj(operator[](j));
                int64_t lon = Point::area2(ptj, pti, pt);
                /* check if pt lies on edge from V[j] to V[i] */
                if (!lon && pt.is_between_helper(ptj, pti))
                        return std::numeric_limits<int>::max();
                /* edge from V[j] to V[i] */
                if (ptj.get_lat() <= pt.get_lat()) {
                        if (pti.get_lat() > pt.get_lat())
                                /* an upward crossing */
                                if (lon > 0)
                                        /* P left of edge: have a valid up intersect */
                                        wn++;
                } else {
                        if (pti.get_lat() <= pt.get_lat())
                                /* a downward crossing */
                                if (lon < 0)
                                        /* P right of edge: have a valid down intersect */
                                        wn--;
                }
        }
        return wn;
}

PolygonSimple::ScanLine PolygonSimple::scanline(Point::coord_t lat) const
{
	const unsigned int sz(size());
	if (!sz)
		return ScanLine();
	ScanLine sl;
	Rect bbox(operator[](0), operator[](0));
        unsigned int j = sz - 1;
        for (unsigned int i = 0; i < sz; j = i, ++i) {
		const Point& pti(operator[](i));
		const Point& ptj(operator[](j));
		bbox = bbox.add(pti);
		if (false)
			std::cerr << "scanline: lat " << lat << " edge " << ptj.get_lon() << ',' << ptj.get_lat()
				  << ' ' << pti.get_lon() << ',' << pti.get_lat() << std::endl;
		if (pti.get_lat() == lat && ptj.get_lat() == lat) {
			// left->right: +1
			// right->left: -1
			int wn(1 | -(ptj.get_lon() >= pti.get_lon()));
			sl.insert(ptj.get_lon(), wn);
			sl.insert(pti.get_lon(), -wn);
			if (false)
				std::cerr << "scanline: insert " << ptj.get_lon() << ',' << wn << ' '
					  << pti.get_lon() << ',' << -wn << std::endl;
			continue;
		}
		int wn(0);
		if (ptj.get_lat() <= lat) {
			if (pti.get_lat() > lat)
				// upward edge: +1
				wn = 1;
		} else {
			if (pti.get_lat() <= lat)
				// downward edge: -1
				wn = -1;
		}
		if (!wn)
			continue;
		Point ptdiff(pti - ptj);
		Point::coord_t lon(ptj.get_lon() + (ptdiff.get_lon() * (lat - ptj.get_lat()) + ptdiff.get_lat() / 2) / ptdiff.get_lat());
		sl.insert(lon, wn);
		if (false)
			std::cerr << "scanline: insert " << lon << ',' << wn << std::endl;
	}
	try {
		return sl.integrate(bbox.get_west());
	} catch (const std::runtime_error& e) {
		std::ostringstream oss;
		oss << e.what() << " lat " << lat << " poly";
		for (unsigned int i = 0; i < sz; ++i) {
			const Point& pti(operator[](i));
			oss << ' ' << pti.get_lon() << ',' << pti.get_lat();
		}
		throw std::runtime_error(oss.str());
	}
}

/*
 * negative, if polygon is clockwise oriented
 */
int64_t PolygonSimple::area2(void) const
{
        int64_t area = 0;

        /* http://mathworld.wolfram.com/PolygonArea.html */
        /* loop through all edges of the polygon */
        unsigned int j = size() - 1;
        for (unsigned int i = 0; i < size(); j = i, i++) {
                /* edge from V[j] to V[i]  (j = (i + 1) mod size) */
                int64_t x1 = operator[](j).get_lon();
                int64_t y1 = operator[](j).get_lat();
                int64_t x2 = operator[](i).get_lon();
                int64_t y2 = operator[](i).get_lat();
                area += (x1 * y2 - x2 * y1);
        }
        return area;
}

void PolygonSimple::make_counterclockwise(void)
{
	if (area2() >= 0)
		return;
	reverse();
}

void PolygonSimple::make_clockwise(void)
{
	if (area2() <= 0)
		return;
	reverse();
}

unsigned int PolygonSimple::next(unsigned int it) const
{
	if (size() < 2U)
		return 0U;
	++it;
	if (it >= size())
		it = 0;
	return it;
}

unsigned int PolygonSimple::prev(unsigned int it) const
{
	if (size() < 2U)
		return 0U;
	if (!it)
		it = size();
	--it;
	return it;
}

void PolygonSimple::reverse(void)
{
	std::reverse(begin(), end());
}

void PolygonSimple::remove_redundant_polypoints(void)
{
        unsigned int sz = size();
        if (sz >= 2) {
                if (operator[](0) == operator[](sz-1)) {
                        sz--;
                        resize(sz);
                }
        }
	if (sz < 3) {
		clear();
		return;
	}
        bool first = true;
        for (unsigned int i = 0; i < sz;) {
                unsigned int j = (i + 1) % sz;
                unsigned int k = (j + 1) % sz;
                if (!operator[](j).is_between(operator[](i), operator[](k))) {
                        i++;
                        continue;
                }
#if 0
                if (first) {
                        std::cerr << "remove_redundant_polypoints: poly:";
                        for (unsigned int m = 0; m < sz; m++)
                                operator[](m).print(std::cerr << ' ');
                        std::cerr << std::endl;
                }
                first = false;
                operator[](j).print(std::cerr << "remove_redundant_polypoints: removing: ");
                operator[](i).print(std::cerr << " between: ");
                operator[](k).print(std::cerr << ' ') << std::endl;
#endif
                erase(begin() + j);
                sz--;
		if (sz < 3)
			break;
        }
        if (sz < 3)
                clear();
}

Rect PolygonSimple::get_bbox(void) const
{
        if (!size())
                return Rect();
        Point pt(operator[](0));
        Point sw(0, 0), ne(0, 0);
        for (unsigned int i = 1; i < size(); i++) {
                Point p(operator[](i) - pt);
                sw = Point(std::min(sw.get_lon(), p.get_lon()), std::min(sw.get_lat(), p.get_lat()));
                ne = Point(std::max(ne.get_lon(), p.get_lon()), std::max(ne.get_lat(), p.get_lat()));
        }
        return Rect(sw + pt, ne + pt);
}

uint64_t PolygonSimple::simple_circumference_rel(void) const
{
	uint64_t d(0);
	for (unsigned int i = 0, j = size() - 1, k = size(); i < k; j = i, ++i)
		d += operator[](i).simple_distance_rel(operator[](j));
	return d;
}

float PolygonSimple::simple_circumference_km(void) const
{
	float d(0);
	for (unsigned int i = 0, j = size() - 1, k = size(); i < k; j = i, ++i)
		d += operator[](i).simple_distance_km(operator[](j));
	return d;
}

float PolygonSimple::spheric_circumference_km(void) const
{
	double d(0);
	for (unsigned int i = 0, j = size() - 1, k = size(); i < k; j = i, ++i)
		d += operator[](i).spheric_distance_km(operator[](j));
	return d;
}

double PolygonSimple::spheric_circumference_km_dbl(void) const
{
	double d(0);
	for (unsigned int i = 0, j = size() - 1, k = size(); i < k; j = i, ++i)
		d += operator[](i).spheric_distance_km_dbl(operator[](j));
	return d;
}

double PolygonSimple::get_simple_area_km2_dbl(void) const
{
	Point center(get_bbox().boxcenter());
	if (center.is_invalid())
		return std::numeric_limits<double>::quiet_NaN();
	// area2 delivers the double area
	double a(area2());
	a *= (0.5 * Point::earth_radius_dbl * Point::to_rad_dbl *
	      Point::earth_radius_dbl * Point::to_rad_dbl);
	a *= cos(center.get_lat_rad());
	return a;
}

PolygonSimple PolygonSimple::line_width_rectangular(const Point& l0, const Point& l1, double width_nmi)
{
	double crs(0);
	if (l0 != l1)
		crs = l0.spheric_true_course_dbl(l1);
	width_nmi *= 0.5;
	PolygonSimple r;
	r.push_back(l0.spheric_course_distance_nmi(crs + 135, width_nmi));
	r.push_back(l0.spheric_course_distance_nmi(crs + 225, width_nmi));
	r.push_back(l1.spheric_course_distance_nmi(crs + 315, width_nmi));
	r.push_back(l1.spheric_course_distance_nmi(crs + 45, width_nmi));
	if (true && (!r.windingnumber(l0) || !r.windingnumber(l1)))
		throw std::runtime_error("line_width_rectangular points not inside box");
	return r;
}

void PolygonSimple::snap_bits(unsigned int nrbits)
{
	if (!nrbits)
		return;
	Point::coord_t mask(~((1U << nrbits) - 1U));
	Point::coord_t rnd((1U << nrbits) >> 1);
	for (iterator i(begin()), e(end()); i != e; ++i) {
		Point& pt(*i);
		pt.set_lon((pt.get_lon() + rnd) & mask);
		pt.set_lat((pt.get_lat() + rnd) & mask);
	}
	remove_redundant_polypoints();
}

void PolygonSimple::randomize_bits(unsigned int nrbits)
{
	if (!nrbits)
		return;
	Point::coord_t rmask((1U << nrbits) - 1U);
	Point::coord_t mask(~rmask);
	for (iterator i(begin()), e(end()); i != e; ++i) {
		Point& pt(*i);
		pt.set_lon((pt.get_lon() & mask) | (rand() & rmask));
		pt.set_lat((pt.get_lat() & mask) | (rand() & rmask));
	}
	remove_redundant_polypoints();
}

unsigned int PolygonSimple::zap_leastimportant(Point& pt)
{
	pt.set_invalid();
	unsigned int sz(size());
	if (sz < 4) {
		clear();
		return 0;
	}
	unsigned int idx(0);
	int64_t area(std::numeric_limits<int64_t>::max());
        for (unsigned int i = 0; i < sz; ++i) {
                unsigned int j = (i + 1) % sz;
                unsigned int k = (j + 1) % sz;
		int64_t a(Point::area2(operator[](i), operator[](j), operator[](k)));
		if (a > area)
			continue;
		area = a;
		idx = j;
        }
        pt = operator[](idx);
	erase(begin() + idx);
	return idx;	
}

std::ostream& PolygonSimple::print_nrvertices(std::ostream& os, bool prarea) const
{
	os << size();
	if (prarea)
		os << " [" << area2() << ']';
	return os;
}

std::ostream& PolygonSimple::print(std::ostream& os) const
{
	os << '(';
	bool subseq(false);
	for (const_iterator i(begin()), e(end()); i != e; ++i) {
		if (subseq)
			os << ", ";
		subseq = true;
		os << i->get_lon_str2() << ' ' << i->get_lat_str2();
	}
	return os << ')';
}

void PolygonSimple::bindblob(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(size() * 8);
        if (!sz) {
                cmd.bind(index);
                return;
        }
        uint8_t data[sz];
        for (size_type i(0); i < size(); i++) {
                Point pt(operator[](i));
                data[8*i+0] = pt.get_lon() >> 0;
                data[8*i+1] = pt.get_lon() >> 8;
                data[8*i+2] = pt.get_lon() >> 16;
                data[8*i+3] = pt.get_lon() >> 24;
                data[8*i+4] = pt.get_lat() >> 0;
                data[8*i+5] = pt.get_lat() >> 8;
                data[8*i+6] = pt.get_lat() >> 16;
                data[8*i+7] = pt.get_lat() >> 24;
        }
        cmd.bind(index, data, sz);
}

void PolygonSimple::getblob(sqlite3x::sqlite3_cursor & cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        if (sz <= 7)
                return;
        sz >>= 3;
        reserve(sz);
        for (size_type i(0); i < (size_type)sz; i++) {
                uint32_t lon, lat;
                lon = data[8*i+0] | (data[8*i+1] << 8) | (data[8*i+2] << 16) | (data[8*i+3] << 24);
                lat = data[8*i+4] | (data[8*i+5] << 8) | (data[8*i+6] << 16) | (data[8*i+7] << 24);
                push_back(Point(lon, lat));
        }
}

bool PolygonHole::is_on_boundary(const Point & pt) const
{
        if (m_exterior.is_on_boundary(pt))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                if (ii->is_on_boundary(pt))
                        return true;
        return false;
}

bool PolygonHole::is_intersection(const Point & la, const Point & lb) const
{
        if (m_exterior.is_intersection(la, lb))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                if (ii->is_intersection(la, lb))
                        return true;
        return false;
}

bool PolygonHole::is_strict_intersection(const Point & la, const Point & lb) const
{
        if (m_exterior.is_strict_intersection(la, lb))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                if (ii->is_strict_intersection(la, lb))
                        return true;
        return false;
}

bool PolygonHole::is_intersection(const Rect & r) const
{
        if (m_exterior.is_intersection(r))
                return true;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                if (ii->is_intersection(r))
                        return true;
        return false;
}

int PolygonHole::windingnumber(const Point & pt) const
{
#if 1
        int wn(m_exterior.windingnumber(pt));
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                wn += ii->windingnumber(pt);
        return wn;
#else
        int wn(m_exterior.windingnumber(pt));
        if (!wn)
                return wn;
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                if (ii->windingnumber(pt))
                        return 0;
        return wn;
#endif
}

PolygonHole::ScanLine PolygonHole::scanline(Point::coord_t lat) const
{
	ScanLine sl(m_exterior.scanline(lat));
	for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                sl = sl.combine(ii->scanline(lat));
	return sl;
}

int64_t PolygonHole::area2(void) const
{
        int64_t r(m_exterior.area2());
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                r += ii->area2();
        return r;
}

double PolygonHole::get_simple_area_km2_dbl(void) const
{
	double r(m_exterior.get_simple_area_km2_dbl());
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                r += ii->get_simple_area_km2_dbl();
        return r;
}

void PolygonHole::remove_redundant_polypoints(void)
{
        m_exterior.remove_redundant_polypoints();
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                ii->remove_redundant_polypoints();
}

void PolygonHole::normalize(void)
{
        bool es(m_exterior.area2() < 0);
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii) {
                bool is(ii->area2() < 0);
                if (is == es)
			ii->reverse();
        }
}

void PolygonHole::normalize_boostgeom(void)
{
	m_exterior.make_counterclockwise();
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
		ii->make_clockwise();
}

void PolygonHole::snap_bits(unsigned int nrbits)
{
	m_exterior.snap_bits(nrbits);
	if (m_exterior.empty())
		m_interior.clear();
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
		ii->snap_bits(nrbits);
}

void PolygonHole::randomize_bits(unsigned int nrbits)
{
	m_exterior.randomize_bits(nrbits);
	if (m_exterior.empty())
		m_interior.clear();
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
		ii->randomize_bits(nrbits);
}

unsigned int PolygonHole::zap_holes(uint64_t abslim, double rellim)
{
	if (m_interior.empty())
		return 0;
	if (rellim > 0) {
		uint64_t a(Point::round<int64_t,double>(std::abs(m_exterior.area2()) * rellim));
		abslim = std::max(abslim, a);
	}
	unsigned int nrerase(0);
        for (interior_t::iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie;) {
		if (std::abs(ii->area2()) >= abslim) {
			++ii;
			continue;
		}
		ii = m_interior.erase(ii);
		ie = m_interior.end();
		++nrerase;
	}
	if (true)
		std::cerr << "PolygonHole::zap_holes: removed " << nrerase << " holes, area limit " << abslim << std::endl;
	return nrerase;
}

std::ostream& PolygonHole::print_nrvertices(std::ostream& os, bool prarea) const
{
	m_exterior.print_nrvertices(os, prarea);
	if (m_interior.empty())
		return os;
	os << " (";
	bool subseq(false);
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii) {
		if (subseq)
			os << ' ';
		subseq = true;
		ii->print_nrvertices(os, prarea);
	}
	return os << ')';
}

std::ostream& PolygonHole::print(std::ostream& os) const
{
	m_exterior.print(os);
	if (m_interior.empty())
		return os;
	os << " [";
	bool subseq(false);
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii) {
		if (subseq)
			os << ", ";
		subseq = true;
		ii->print(os);
	}
	return os << ']';
}

unsigned int PolygonHole::bloblength(const PolygonSimple& p)
{
        return p.size() * 8 + 4;
}

uint8_t * PolygonHole::blobencode(const PolygonSimple& p, uint8_t * data)
{
        unsigned int sz(p.size());
        data[0] = sz;
        data[1] = sz >> 8;
        data[2] = sz >> 16;
        data[3] = sz >> 24;
        data += 4;
        for (unsigned int i = 0; i < sz; ++i) {
                Point pt(p[i]);
                data[0] = pt.get_lon() >> 0;
                data[1] = pt.get_lon() >> 8;
                data[2] = pt.get_lon() >> 16;
                data[3] = pt.get_lon() >> 24;
                data[4] = pt.get_lat() >> 0;
                data[5] = pt.get_lat() >> 8;
                data[6] = pt.get_lat() >> 16;
                data[7] = pt.get_lat() >> 24;
                data += 8;
        }
        return data;
}

uint8_t const *PolygonHole::blobdecode(PolygonSimple& p, uint8_t const *data, uint8_t const *end)
{
        p.clear();
        if (data + 4 > end)
                return end;
        uint32_t sz(data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24));
        data += 4;
        if (sz >= 0x01000000 || data + 8 * sz > end)
                return end;
        p.reserve(sz);
        for (unsigned int i = 0; i < sz; ++i) {
                uint32_t lon, lat;
                lon = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
                lat = data[4] | (data[5] << 8) | (data[6] << 16) | (data[7] << 24);
                data += 8;
                p.push_back(Point(lon, lat));
        }
        return data;
}

void PolygonHole::getblob(sqlite3x::sqlite3_cursor & cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        uint8_t const *end(data + sz);
        data = blobdecode(m_exterior, data, end);
        while (data < end) {
                m_interior.push_back(PolygonSimple());
                data = blobdecode(m_interior.back(), data, end);
        }
}

void PolygonHole::bindblob(sqlite3x::sqlite3_command & cmd, int index) const
{
        unsigned int sz(bloblength(m_exterior));
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                sz += bloblength(*ii);
        if (sz <= 4) {
                cmd.bind(index);
                return;
        }
        uint8_t data[sz];
        uint8_t *d = blobencode(m_exterior, data);
        for (interior_t::const_iterator ii(m_interior.begin()), ie(m_interior.end()); ii != ie; ++ii)
                d = blobencode(*ii, d);
        cmd.bind(index, data, sz);
}

MultiPolygonHole::MultiPolygonHole()
{
}

MultiPolygonHole::MultiPolygonHole(const PolygonHole& p)
{
	push_back(p);
}

bool MultiPolygonHole::is_on_boundary(const Point& pt) const
{
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		if (i->is_on_boundary(pt))
			return true;
	return false;
}

bool MultiPolygonHole::is_intersection(const Point& la, const Point& lb) const
{
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		if (i->is_intersection(la, lb))
			return true;
	return false;
}

bool MultiPolygonHole::is_strict_intersection(const Point& la, const Point& lb) const
{
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		if (i->is_strict_intersection(la, lb))
			return true;
	return false;
}

bool MultiPolygonHole::is_intersection(const Rect& r) const
{
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		if (i->is_intersection(r))
			return true;
	return false;
}

int MultiPolygonHole::windingnumber(const Point& pt) const
{
	int wn(0);
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		wn += i->windingnumber(pt);
	return wn;
}

MultiPolygonHole::ScanLine MultiPolygonHole::scanline(Point::coord_t lat) const
{
	ScanLine sl;
	for (const_iterator i(begin()), e(end()); i != e; ++i)
               sl = sl.combine(i->scanline(lat));
	return sl;
}

int64_t MultiPolygonHole::area2(void) const
{
	int64_t a(0);
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		a += i->area2();
	return a;
}

void MultiPolygonHole::remove_redundant_polypoints(void)
{
	for (iterator i(begin()); i != end();) {
		i->remove_redundant_polypoints();
		if (i->get_exterior().size() < 3) {
			i = erase(i);
			continue;
		}
		++i;
	}
}

void MultiPolygonHole::snap_bits(unsigned int nrbits)
{
	for (iterator i(begin()); i != end();) {
		i->snap_bits(nrbits);
		if (i->get_exterior().size() < 3) {
			i = erase(i);
			continue;
		}
		++i;
	}
}

void MultiPolygonHole::randomize_bits(unsigned int nrbits)
{
	for (iterator i(begin()); i != end();) {
		i->randomize_bits(nrbits);
		if (i->get_exterior().size() < 3) {
			i = erase(i);
			continue;
		}
		++i;
	}
}

unsigned int MultiPolygonHole::zap_holes(uint64_t abslim, double rellim)
{
	unsigned int r(0);
	for (iterator i(begin()), e(end()); i != e; ++i)
		r += i->zap_holes(abslim, rellim);
	return r;
}

unsigned int MultiPolygonHole::zap_polys(uint64_t abslim, double rellim)
{
	if (empty())
		return 0;
	if (rellim > 0) {
		uint64_t a(Point::round<int64_t,double>(std::abs(area2()) * rellim));
		abslim = std::max(abslim, a);
	}
	unsigned int nrerase(0);
        for (iterator i(begin()), e(end()); i != e;) {
		if (std::abs(i->area2()) >= abslim) {
			++i;
			continue;
		}
		i = erase(i);
		e = end();
		++nrerase;
	}
	if (true)
		std::cerr << "MultiPolygonHole::zap_polys: removed " << nrerase << " polygons, area limit " << abslim << std::endl;
	return nrerase;
}

std::ostream& MultiPolygonHole::print_nrvertices(std::ostream& os, bool prarea) const
{
	os << '{';
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		i->print_nrvertices(os << ' ', prarea);
	return os << " }";
}

std::ostream& MultiPolygonHole::print(std::ostream& os) const
{
	os << '{';
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		i->print(os << ' ');
	return os << " }";
}

Rect MultiPolygonHole::get_bbox(void) const
{
	Rect bbox;
	bbox.set_empty();
	const_iterator i(begin()), e(end());
	if (i == e)
		return bbox;
	bbox = i->get_bbox();
	for (++i; i != e; ++i)
		bbox = bbox.add(i->get_bbox());
	return bbox;
}

void MultiPolygonHole::getblob(sqlite3x::sqlite3_cursor& cursor, int index)
{
        clear();
        if (cursor.isnull(index))
                return;
        int sz;
        uint8_t const *data((uint8_t const *)cursor.getblob(index, sz));
        uint8_t const *end(data + sz);
        while (data < end) {
                push_back(PolygonHole());
                data = blobdecode(back(), data, end);
        }
}

void MultiPolygonHole::bindblob(sqlite3x::sqlite3_command& cmd, int index) const
{
        unsigned int sz(0);
        for (const_iterator i(begin()), e(end()); i != e; ++i)
                sz += bloblength(*i);
        if (sz <= 4) {
                cmd.bind(index);
                return;
        }
        uint8_t data[sz];
        uint8_t *d = data;
        for (const_iterator i(begin()), e(end()); i != e; ++i)
                d = blobencode(*i, d);
        cmd.bind(index, data, sz);
}

void MultiPolygonHole::normalize(void)
{
	for (iterator i(begin()), e(end()); i != e; ++i)
		i->normalize();
}

void MultiPolygonHole::normalize_boostgeom(void)
{
	for (iterator i(begin()), e(end()); i != e; ++i)
		i->normalize_boostgeom();
}

double MultiPolygonHole::get_simple_area_km2_dbl(void) const
{
	double a(0);
	for (const_iterator i(begin()), e(end()); i != e; ++i)
		a += i->get_simple_area_km2_dbl();
	return a;
}

unsigned int MultiPolygonHole::bloblength(const PolygonHole& p)
{
	unsigned int len(2 + PolygonHole::bloblength(p.get_exterior()));
	for (unsigned int n(p.get_nrinterior()), i(0); i < n; ++i)
		len += PolygonHole::bloblength(p[i]);
	return len;
}

uint8_t *MultiPolygonHole::blobencode(const PolygonHole& p, uint8_t *data)
{
	unsigned int n(p.get_nrinterior());
	data[0] = n;
	data[1] = n >> 8;
	data += 2;
	data = PolygonHole::blobencode(p.get_exterior(), data);
	for (unsigned int i = 0; i < n; ++i)
		data = PolygonHole::blobencode(p[i], data);
	return data;	
}

uint8_t const *MultiPolygonHole::blobdecode(PolygonHole& p, uint8_t const *data, uint8_t const *end)
{
	p.clear();
	if (data + 2 > end)
		return end;
	unsigned int n(data[0] | (data[1] << 8));
     	data += 2;
	{
		PolygonSimple ps;
		data = PolygonHole::blobdecode(ps, data, end);
		p.set_exterior(ps);
	}
	for (; n > 0; --n) {
		PolygonSimple ps;
		data = PolygonHole::blobdecode(ps, data, end);
		p.add_interior(ps);
	}
	return data;
}

template <typename T> const typename CartesianVector3D<T>::float_t CartesianVector3D<T>::from_deg;

template <typename T>
CartesianVector3D<T>& CartesianVector3D<T>::operator+=(const CartesianVector3D<T>& v)
{
        m_v[0] += v.m_v[0];
        m_v[1] += v.m_v[1];
        m_v[2] += v.m_v[2];
        return *this;
}

template <typename T>
CartesianVector3D<T>& CartesianVector3D<T>::operator-=(const CartesianVector3D<T>& v)
{
        m_v[0] -= v.m_v[0];
        m_v[1] -= v.m_v[1];
        m_v[2] -= v.m_v[2];
        return *this;
}

template <typename T>
CartesianVector3D<T>& CartesianVector3D<T>::operator*=(float_t v)
{
        m_v[0] *= v;
        m_v[1] *= v;
        m_v[2] *= v;
        return *this;
}

template <typename T>
typename CartesianVector3D<T>::float_t CartesianVector3D<T>::scalarprod(const CartesianVector3D<T>& v) const
{
        return m_v[0] * v.m_v[0] + m_v[1] * v.m_v[1] + m_v[2] * v.m_v[2];
}

template <typename T>
CartesianVector3D<T> CartesianVector3D<T>::crossprod(const CartesianVector3D<T>& v) const
{
	return CartesianVector3D<T>(m_v[1] * v.m_v[2] - m_v[2] * v.m_v[1],
				    m_v[2] * v.m_v[0] - m_v[0] * v.m_v[2],
				    m_v[0] * v.m_v[1] - m_v[1] * v.m_v[0]);
}

template <typename T>
typename CartesianVector3D<T>::float_t CartesianVector3D<T>::length2(void) const
{
        return m_v[0] * m_v[0] + m_v[1] * m_v[1] + m_v[2] * m_v[2];
}

template <typename T>
typename CartesianVector3D<T>::float_t CartesianVector3D<T>::nearest_fraction(const CartesianVector3D<T>& v0, const CartesianVector3D<T>& v1)
{
        CartesianVector3D<T> n(v1 - v0);
        float_t t = n.length2();
        if (t < (float_t)1e-6)
                return 0.5;
        t = n.scalarprod((*this) - v0) / t;
        return std::min(std::max(t, (float_t)0.0), (float_t)1.0);
}

template <typename T>
void CartesianVector3D<T>::rotate_xz(float_t ang)
{
        float_t c = cos(ang), s = sin(ang);
        float_t t = m_v[0] * c - m_v[2] * s;
        m_v[2] = m_v[0] * s + m_v[2] * c;
        m_v[0] = t;
}

template <typename T>
void CartesianVector3D<T>::rotate_xy(float_t ang)
{
        float_t c = cos(ang), s = sin(ang);
        float_t t = m_v[0] * c - m_v[1] * s;
        m_v[1] = m_v[0] * s + m_v[1] * c;
        m_v[0] = t;
}

template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::earth_radius;
template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::km_to_nmi;
template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::m_to_ft;
template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::ft_to_m;
template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::from_deg;
template <typename T> const typename PolarVector3D<T>::float_t PolarVector3D<T>::to_deg;

template <typename T>
PolarVector3D<T>::PolarVector3D(const Point& pt, int16_t altft)
	: m_lat(pt.get_lat_rad()), m_lon(pt.get_lon_rad()), m_r(altft * ft_to_m + earth_radius)
{
}

template <typename T>
Point PolarVector3D<T>::getcoord(void) const
{
	return Point((int32_t)(m_lon * Point::from_rad), (int32_t)(m_lat * Point::from_rad));
}

template <typename T>
CartesianVector3D<T>::operator PolarVector3D<T>(void) const
{
        float_t r = length();
        float_t lon = atan2(m_v[1], m_v[0]);
        float_t lat = 0;
        if (r >= 0.1)
                lat = asin(m_v[2] / r);
        return PolarVector3D<T>(lat, lon, r);
}

template <typename T>
PolarVector3D<T>::operator CartesianVector3D<T>(void) const
{
        float_t r = m_r;
        float_t z = r * sin(m_lat);
        r *= cos(m_lat);
        float_t x = r * cos(m_lon);
        float_t y = r * sin(m_lon);
        return CartesianVector3D<T>(x, y, z);
}

template class CartesianVector3D<float>;
template class PolarVector3D<float>;
template class CartesianVector3D<double>;
template class PolarVector3D<double>;
