//
// C++ Interface: geom
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009, 2012, 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef GEOM_H
#define GEOM_H

//#include <cstdint>
#include <stdint.h>
#include <cmath>
#include <iostream>
#include <vector>
#include <glibmm.h>

#include "sysdeps.h"

namespace sqlite3x {
        class sqlite3_cursor;
        class sqlite3_command;
};

class Rect;
class PolygonSimple;

class Point {
  public:
	typedef int32_t coord_t;
	static const coord_t lon_min = 0x80000000;
	static const coord_t lon_max = 0x7FFFFFFF;
	static const coord_t lat_min = 0xC0000000;
	static const coord_t lat_max = 0x40000000;
	static const Point invalid;

	Point(coord_t lo = 0, coord_t la = 0) : lon(lo), lat(la) {}
	coord_t get_lon(void) const { return lon; }
	coord_t get_lat(void) const { return lat; }
	void set_lon(coord_t lo) { lon = lo; }
	void set_lat(coord_t la) { lat = la; }
	bool operator==(const Point &p) const { return lon == p.lon && lat == p.lat; }
	bool operator!=(const Point &p) const { return !operator==(p); }
	Point operator+(const Point &p) const { return Point(lon + (uint32_t)p.lon, lat + (uint32_t)p.lat); }
	Point operator-(const Point &p) const { return Point(lon - (uint32_t)p.lon, lat - (uint32_t)p.lat); }
	Point operator-() const { return Point(-lon, -lat); }
	Point& operator+=(const Point &p) { lon += (uint32_t)p.lon; lat += (uint32_t)p.lat; return *this; }
	Point& operator-=(const Point &p) { lon -= (uint32_t)p.lon; lat -= (uint32_t)p.lat; return *this; }
	bool operator<(const Point &p) const { return lon < p.lon && lat < p.lat; }
	bool operator<=(const Point &p) const { return lon <= p.lon && lat <= p.lat; }
	bool operator>(const Point &p) const { return lon > p.lon && lat > p.lat; }
	bool operator>=(const Point &p) const { return lon >= p.lon && lat >= p.lat; }
	uint64_t length2(void) const { return (uint64_t)(get_lon() * (int64_t)get_lon()) + (uint64_t)(get_lat() * (int64_t)get_lat()); }
	double length(void) const { return sqrt(length2()); }
	bool left_of(const Point& p0, const Point& p1) const;
	bool intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);
	Point nearest_on_line(const Point& la, const Point& lb);
	Point spheric_nearest_on_line(const Point& la, const Point& lb) const;

	bool is_strictly_left(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) > 0; }
	bool is_left_or_on(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) >= 0; }
	bool is_collinear(const Point& p0, const Point& p1) const { return area2(p0, p1, *this) == 0; }
	bool is_between(const Point& p0, const Point& p1) const;
	bool is_strictly_between(const Point& p0, const Point& p1) const;
	bool is_convex(const Point& vprev, const Point& vnelont) const;
	bool is_in_cone(const Point& vprev, const Point& v, const Point& vnelont) const;
	bool is_left_of(const Point& p0, const Point& p1) const;
	bool in_box(const Point& p0, const Point& p1) const;
	bool is_invalid(void) const { return lat < lat_min || lat > lat_max; }
	void set_invalid(void);

	std::ostream& print(std::ostream& os) const;

	static int64_t area2(const Point& p0, const Point& p1, const Point& p2);
	static bool is_proper_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);
	static bool is_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);
	static bool is_strict_intersection(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);

	Point simple_unwrap(const Point& p2) const;

  public:
	static const float to_deg = 90.0 / (1UL << 30);
	static const float to_rad = M_PI / (1UL << 31);
	static const float from_deg = (1UL << 30) / 90.0;
	static const float from_rad = (1UL << 31) / M_PI;

	static const double to_deg_dbl = 90.0 / (1UL << 30);
	static const double to_rad_dbl = M_PI / (1UL << 31);
	static const double from_deg_dbl = (1UL << 30) / 90.0;
	static const double from_rad_dbl = (1UL << 31) / M_PI;

	static const int32_t pole_lat = (1UL << 30);

	static const float earth_radius = 6378.155;
	static const float km_to_nmi = 1.0 / 1.852;
	static const float m_to_ft = 1.0 / 0.3048;
	static const float ft_to_m = 0.3048;
	static const float mps_to_kts = 3.6 / 1.852;

	static const double earth_radius_dbl = 6378.155;
	static const double km_to_nmi_dbl = 1.0 / 1.852;
	static const double m_to_ft_dbl = 1.0 / 0.3048;
	static const double ft_to_m_dbl = 0.3048;
	static const double mps_to_kts_dbl = 3.6 / 1.852;

	template<typename I, typename F> static I round(F x) { return static_cast<I>((x < 0) ? (x - 0.5) : (x + 0.5)); }

	float get_lon_rad(void) const { return lon * to_rad; }
	float get_lat_rad(void) const { return lat * to_rad; }
	void set_lon_rad(float lo) { lon = round<coord_t,float>(lo * from_rad); }
	void set_lat_rad(float la) { lat = round<coord_t,float>(la * from_rad); }

	double get_lon_rad_dbl(void) const { return lon * to_rad_dbl; }
	double get_lat_rad_dbl(void) const { return lat * to_rad_dbl; }
	void set_lon_rad_dbl(double lo) { lon = round<coord_t,double>(lo * from_rad_dbl); }
	void set_lat_rad_dbl(double la) { lat = round<coord_t,double>(la * from_rad_dbl); }

	float get_lon_deg(void) const { return lon * to_deg; }
	float get_lat_deg(void) const { return lat * to_deg; }
	void set_lon_deg(float lo) { lon = round<coord_t,float>(lo * from_deg); }
	void set_lat_deg(float la) { lat = round<coord_t,float>(la * from_deg); }

	double get_lon_deg_dbl(void) const { return lon * to_deg_dbl; }
	double get_lat_deg_dbl(void) const { return lat * to_deg_dbl; }
	void set_lon_deg_dbl(double lo) { lon = round<coord_t,double>(lo * from_deg_dbl); }
	void set_lat_deg_dbl(double la) { lat = round<coord_t,double>(la * from_deg_dbl); }

	float simple_true_course(const Point& p2) const;
	uint64_t simple_distance_rel(const Point& p2) const;
	float simple_distance_km(const Point& p2) const;
	float simple_distance_nmi(const Point& p2) const { return simple_distance_km(p2) * km_to_nmi; }
	Point simple_course_distance_nmi(float course, float dist) const;
	Rect simple_box_km(float centerdist) const;
	Rect simple_box_nmi(float centerdist) const;

	float spheric_true_course(const Point& p2) const;
	float spheric_distance_km(const Point& p2) const;
	float spheric_distance_nmi(const Point& p2) const { return spheric_distance_km(p2) * km_to_nmi; }
	Point spheric_course_distance_nmi(float course, float dist) const;

	double spheric_true_course_dbl(const Point& p2) const;
	double spheric_distance_km_dbl(const Point& p2) const;
	double spheric_distance_nmi_dbl(const Point& p2) const { return spheric_distance_km_dbl(p2) * km_to_nmi_dbl; }

	Point halfway(const Point& pt2) const;

	static Glib::ustring to_ustring(const std::wstring& str);

	Glib::ustring get_lon_str(void) const;
	Glib::ustring get_lat_str(void) const;
	Glib::ustring get_lon_str2(void) const;
	Glib::ustring get_lat_str2(void) const;
	Glib::ustring get_lon_fpl_str(void) const;
	Glib::ustring get_lat_fpl_str(void) const;
	Glib::ustring get_fpl_str(void) const;

	bool set_lon_str(const Glib::ustring& str);
	bool set_lat_str(const Glib::ustring& str);
	static const unsigned int setstr_lat    = 1 << 0;
	static const unsigned int setstr_lon    = 1 << 1;
	static const unsigned int setstr_excess = 1 << 2;
	unsigned int set_str(const Glib::ustring& str);

	void set_ch1903(float x, float y);
	bool get_ch1903(float& x, float& y) const;
	Glib::ustring get_locator(unsigned int digits = 10) const;

	template<class Archive> void hibernate_binary(Archive& ar) {
		ar.io(lat);
		ar.io(lon);
	}

  private:
	static const unsigned int maidenhead_maxdigits = 5;
	static const unsigned int maidenhead_max[maidenhead_maxdigits];
	static const double maidenhead_factors[maidenhead_maxdigits];

	coord_t lon, lat;

	bool is_between_helper(const Point& p0, const Point& p1) const;
	bool is_strictly_between_helper(const Point& p0, const Point& p1) const;

	static bool intersect_precheck(const Point& l0a, const Point& l0b, const Point& l1a, const Point& l1b);
	unsigned int set_str_ch1903(Glib::ustring::const_iterator& si, Glib::ustring::const_iterator se);
	unsigned int set_str_locator(Glib::ustring::const_iterator& si, Glib::ustring::const_iterator se);

	friend class PolygonSimple;
};

class Rect {
  public:
	Rect(void) : southwest(Point(0, -Point::pole_lat)), northeast(Point(-1, Point::pole_lat)) {}
	Rect(const Point& sw, const Point& ne) : southwest(sw), northeast(ne) {}
	Point get_southwest(void) const { return southwest; }
	Point get_northeast(void) const { return northeast; }
	Point get_southeast(void) const { return Point(northeast.get_lon(), southwest.get_lat()); }
	Point get_northwest(void) const { return Point(southwest.get_lon(), northeast.get_lat()); }
	Point::coord_t get_north(void) const { return northeast.get_lat(); }
	Point::coord_t get_south(void) const { return southwest.get_lat(); }
	Point::coord_t get_east(void) const { return northeast.get_lon(); }
	int64_t get_east_unwrapped(void) const;
	Point::coord_t get_west(void) const { return southwest.get_lon(); }
	void set_north(Point::coord_t c) { northeast.set_lat(c); }
	void set_south(Point::coord_t c) { southwest.set_lat(c); }
	void set_east(Point::coord_t c) { northeast.set_lon(c); }
	void set_west(Point::coord_t c) { southwest.set_lon(c); }
	bool operator==(const Rect &r) const { return get_southwest() == r.get_southwest() && get_northeast() == r.get_northeast(); }
	bool operator!=(const Rect &r) const { return !operator==(r); }
	Rect& operator+=(const Point &p) { southwest += p; northeast += p; return *this; }
	Rect& operator-=(const Point &p) { southwest -= p; northeast -= p; return *this; }
	Rect operator+(const Point &p) { Rect r(*this); r += p; return r; }
	Rect operator-(const Point &p) { Rect r(*this); r -= p; return r; }
	bool is_inside(const Point& pt) const;
	bool is_inside(const Rect& r) const { return is_inside(r.get_southwest()) && is_inside(r.get_northeast()); }
	bool is_strictly_inside(const Point& pt) const;
	bool is_strictly_inside(const Rect& r) const { return is_strictly_inside(r.get_southwest()) && is_strictly_inside(r.get_northeast()); }
	bool is_lon_inside(const Point& pt) const;
	bool is_lat_inside(const Point& pt) const;
	uint64_t simple_distance_rel(const Point& pt) const;
	float simple_distance_km(const Point& pt) const;
	float simple_distance_nmi(const Point& pt) const { return simple_distance_km(pt) * Point::km_to_nmi; }
	float spheric_distance_km(const Point& pt) const;
	float spheric_distance_nmi(const Point& pt) const { return spheric_distance_km(pt) * Point::km_to_nmi; }
	bool is_intersect(const Rect& r) const;
	bool is_intersect(const Point& la, const Point& lb) const;
	uint64_t simple_distance_rel(const Rect& r) const;
	float simple_distance_km(const Rect& r) const;
	float simple_distance_nmi(const Rect& r) const { return simple_distance_km(r) * Point::km_to_nmi; }
	float spheric_distance_km(const Rect& r) const;
	float spheric_distance_nmi(const Rect& r) const { return spheric_distance_km(r) * Point::km_to_nmi; }
	Point boxcenter(void) const { return southwest.halfway(northeast); }
	Rect oversize_nmi(float nmi) const;
	Rect oversize_km(float km) const;
	Rect add(const Point& pt) const;
	Rect add(const Rect& r) const { return add(r.get_northeast()).add(r.get_southwest()); }
	int64_t area2(void) const;
	uint64_t simple_circumference_rel(void) const;
	float simple_circumference_km(void) const;
	float simple_circumference_nmi(void) const { return simple_circumference_km() * Point::km_to_nmi; }
	float spheric_circumference_km(void) const;
	float spheric_circumference_nmi(void) const { return spheric_circumference_km() * Point::km_to_nmi; }
	double spheric_circumference_km_dbl(void) const;
	double spheric_circumference_nmi_dbl(void) const { return spheric_circumference_km_dbl() * Point::km_to_nmi_dbl; }
	double get_simple_area_km2_dbl(void) const;
	double get_simple_area_nmi2_dbl(void) const { return get_simple_area_km2_dbl() * (Point::km_to_nmi_dbl * Point::km_to_nmi_dbl); }
	operator PolygonSimple(void) const;

	bool is_empty(void) const;
	void set_empty(void);

	std::ostream& print(std::ostream& os) const;

	template<class Archive> void hibernate_binary(Archive& ar) {
		southwest.hibernate_binary(ar);
		northeast.hibernate_binary(ar);
	}

  private:
	Point southwest;
	Point northeast;

	std::pair<bool,bool> is_overlap_helper(const Rect& r) const;
};

class MultiPolygonHole;

class PolygonSimple : public std::vector<Point> {
  private:
	typedef std::vector<Point> base_t;

  public:
	class ScanLine : public std::map<Point::coord_t,int> {
	  private:
		typedef std::map<Point::coord_t,int> base_t;

	  public:
		using base_t::insert;
		void insert(Point::coord_t lon, int wn);
		ScanLine integrate(Point::coord_t lon) const;
		ScanLine combine(const ScanLine& sl2, int factor1 = 1, int factor2 = 1) const;
	};

	reference operator[](size_type x);
	const_reference operator[](size_type x) const;
	bool is_on_boundary(const Point& pt) const;
	bool is_intersection(const Point& la, const Point& lb) const;
	bool is_intersection(const Rect& r) const;
	bool is_strict_intersection(const Point& la, const Point& lb) const;
	int windingnumber(const Point& pt) const;
	ScanLine scanline(Point::coord_t lat) const;
	int64_t area2(void) const;
	bool is_counterclockwise(void) const { return area2() >= 0; }
	void make_counterclockwise(void);
	void make_clockwise(void);
	unsigned int next(unsigned int it) const;
	unsigned int prev(unsigned int it) const;
	void reverse(void);
	void remove_redundant_polypoints(void);
	Rect get_bbox(void) const;
	uint64_t simple_circumference_rel(void) const;
	float simple_circumference_km(void) const;
	float simple_circumference_nmi(void) const { return simple_circumference_km() * Point::km_to_nmi; }
	float spheric_circumference_km(void) const;
	float spheric_circumference_nmi(void) const { return spheric_circumference_km() * Point::km_to_nmi; }
	double spheric_circumference_km_dbl(void) const;
	double spheric_circumference_nmi_dbl(void) const { return spheric_circumference_km_dbl() * Point::km_to_nmi_dbl; }
	// get_simple_area is signed depending on the turn direction
	double get_simple_area_km2_dbl(void) const;
	double get_simple_area_nmi2_dbl(void) const { return get_simple_area_km2_dbl() * (Point::km_to_nmi_dbl * Point::km_to_nmi_dbl); }
	static PolygonSimple line_width_rectangular(const Point& l0, const Point& l1, double width_nmi);
	void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
	void getblob(sqlite3x::sqlite3_cursor& cursor, int index);
	void snap_bits(unsigned int nrbits);
	void randomize_bits(unsigned int nrbits);
	unsigned int zap_leastimportant(Point& pt);
	unsigned int zap_leastimportant(void) { Point pt; return zap_leastimportant(pt); }
	std::ostream& print_nrvertices(std::ostream& os, bool prarea = false) const;
	std::ostream& print(std::ostream& os) const;
	bool geos_is_valid(void) const;
	void geos_make_valid(void);
	MultiPolygonHole geos_simplify(void);

	template<class Archive> void hibernate_binary(Archive& ar) {
		uint32_t sz(size());
		ar.io(sz);
		if (ar.is_load())
			resize(sz);
		for (iterator i(begin()), e(end()); i != e; ++i)
			i->hibernate_binary(ar);
	}

  private:
	bool geos_make_valid_helper(void);
};

class PolygonHole {
  public:
	typedef PolygonSimple::ScanLine ScanLine;

	PolygonHole(const PolygonSimple& extr = PolygonSimple(), const std::vector<PolygonSimple>& intr = std::vector<PolygonSimple>()) : m_exterior(extr), m_interior(intr) {}
	void clear(void) { m_exterior.clear(); m_interior.clear(); }
	PolygonSimple& get_exterior(void) { return m_exterior; }
	const PolygonSimple& get_exterior(void) const { return m_exterior; }
	void set_exterior(const PolygonSimple& extr) { m_exterior = extr; }
	unsigned int get_nrinterior(void) const { return m_interior.size(); }
	PolygonSimple& operator[](unsigned int x) { return m_interior[x]; }
	const PolygonSimple& operator[](unsigned int x) const { return m_interior[x]; }
	void add_interior(const PolygonSimple& intr) { m_interior.push_back(intr); }
	void insert_interior(const PolygonSimple& intr, int index) { m_interior.insert(m_interior.begin() + index, intr); }
	void erase_interior(int index) { m_interior.erase(m_interior.begin() + index); }
	void clear_interior(void) { m_interior.clear(); }
	bool is_on_boundary(const Point& pt) const;
	bool is_intersection(const Point& la, const Point& lb) const;
	bool is_intersection(const Rect& r) const;
	bool is_strict_intersection(const Point& la, const Point& lb) const;
	int windingnumber(const Point& pt) const;
	ScanLine scanline(Point::coord_t lat) const;
	int64_t area2(void) const;
	void remove_redundant_polypoints(void);
	Rect get_bbox(void) const { return m_exterior.get_bbox(); }
	void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
	void getblob(sqlite3x::sqlite3_cursor& cursor, int index);
	void normalize(void);
	void normalize_boostgeom(void);
	// get_simple_area is signed depending on the turn direction
	double get_simple_area_km2_dbl(void) const;
	double get_simple_area_nmi2_dbl(void) const { return get_simple_area_km2_dbl() * (Point::km_to_nmi_dbl * Point::km_to_nmi_dbl); }
	void snap_bits(unsigned int nrbits);
	void randomize_bits(unsigned int nrbits);
	unsigned int zap_holes(uint64_t abslim = 0, double rellim = 0);
	std::ostream& print_nrvertices(std::ostream& os, bool prarea = false) const;
	std::ostream& print(std::ostream& os) const;
	bool geos_is_valid(void) const;

	// needed for boost adaptation
	typedef std::vector<PolygonSimple> interior_t;
	interior_t& get_interior(void) { return m_interior; }
	const interior_t& get_interior(void) const { return m_interior; }

	template<class Archive> void hibernate_binary(Archive& ar) {
		m_exterior.hibernate_binary(ar);
		uint32_t sz(m_interior.size());
		ar.io(sz);
		if (ar.is_load())
			m_interior.resize(sz);
		for (interior_t::iterator i(m_interior.begin()), e(m_interior.end()); i != e; ++i)
			i->hibernate_binary(ar);
	}

  private:
	PolygonSimple m_exterior;
	interior_t m_interior;

	friend class MultiPolygonHole;
	static unsigned int bloblength(const PolygonSimple& p);
	static uint8_t *blobencode(const PolygonSimple& p, uint8_t *data);
	static uint8_t const *blobdecode(PolygonSimple& p, uint8_t const *data, uint8_t const *end);
};

class MultiPolygonHole : public std::vector<PolygonHole> {
  public:
	typedef PolygonSimple::ScanLine ScanLine;

	MultiPolygonHole();
	MultiPolygonHole(const PolygonHole& p);
	bool is_on_boundary(const Point& pt) const;
	bool is_intersection(const Point& la, const Point& lb) const;
	bool is_intersection(const Rect& r) const;
	bool is_strict_intersection(const Point& la, const Point& lb) const;
	int windingnumber(const Point& pt) const;
	ScanLine scanline(Point::coord_t lat) const;
	int64_t area2(void) const;
	void remove_redundant_polypoints(void);
	Rect get_bbox(void) const;
	void bindblob(sqlite3x::sqlite3_command& cmd, int index) const;
	void getblob(sqlite3x::sqlite3_cursor& cursor, int index);
	void normalize(void);
	void normalize_boostgeom(void);
	// get_simple_area is signed depending on the turn direction
	double get_simple_area_km2_dbl(void) const;
	double get_simple_area_nmi2_dbl(void) const { return get_simple_area_km2_dbl() * (Point::km_to_nmi_dbl * Point::km_to_nmi_dbl); }
	void snap_bits(unsigned int nrbits);
	void randomize_bits(unsigned int nrbits);
	unsigned int zap_holes(uint64_t abslim = 0, double rellim = 0);
	unsigned int zap_polys(uint64_t abslim = 0, double rellim = 0);
	std::ostream& print_nrvertices(std::ostream& os, bool prarea = false) const;
	std::ostream& print(std::ostream& os) const;
	bool geos_is_valid(void) const;
	void geos_make_valid(void);
	void geos_union(const MultiPolygonHole& m);
	void geos_subtract(const MultiPolygonHole& m);
	void geos_intersect(const MultiPolygonHole& m);

	template<class Archive> void hibernate_binary(Archive& ar) {
		uint32_t sz(size());
		ar.io(sz);
		if (ar.is_load())
			resize(sz);
		for (iterator i(begin()), e(end()); i != e; ++i)
			i->hibernate_binary(ar);
	}

  private:
	static unsigned int bloblength(const PolygonHole& p);
	static uint8_t *blobencode(const PolygonHole& p, uint8_t *data);
	static uint8_t const *blobdecode(PolygonHole& p, uint8_t const *data, uint8_t const *end);
};

template <typename T>
class PolarVector3D;

template <typename T>
class CartesianVector3D {
  public:
	typedef T float_t;
	static const float_t from_deg     = M_PI / 180.0;

	CartesianVector3D(float_t x = 0.0, float_t y = 0.0, float_t z = 0.0) { m_v[0] = x; m_v[1] = y; m_v[2] = z; }
	float_t getx(void) const { return m_v[0]; }
	float_t gety(void) const { return m_v[1]; }
	float_t getz(void) const { return m_v[2]; }
	void setx(float_t v) { m_v[0] = v; }
	void sety(float_t v) { m_v[1] = v; }
	void setz(float_t v) { m_v[2] = v; }
	CartesianVector3D<T>& operator+=(const CartesianVector3D<T>& v);
	CartesianVector3D<T>& operator-=(const CartesianVector3D<T>& v);
	CartesianVector3D<T>& operator*=(float_t v);
	CartesianVector3D<T> operator+(const CartesianVector3D<T>& v) const { CartesianVector3D<T> pv(*this); pv += v; return pv; }
	CartesianVector3D<T> operator-(const CartesianVector3D<T>& v) const { CartesianVector3D<T> pv(*this); pv -= v; return pv; }
	CartesianVector3D<T> operator-(void) const { CartesianVector3D<T> pv; pv -= *this; return pv; }
	CartesianVector3D<T> operator*(float_t v) const { CartesianVector3D<T> pv(*this); pv *= v; return pv; }
	float_t scalarprod(const CartesianVector3D<T>& v) const;
	CartesianVector3D<T> crossprod(const CartesianVector3D<T>& v) const;
	float_t length2(void) const;
	float_t length(void) const { return sqrt(length2()); }
	float_t nearest_fraction(const CartesianVector3D<T>& v0, const CartesianVector3D<T>& v1);
	void rotate_xz(float_t ang);
	void rotate_xy(float_t ang);
	operator PolarVector3D<T>(void) const;

	template<class Archive> void hibernate_binary(Archive& ar) {
		for (int i = 0; i < 3; ++i)
			ar.io(m_v[i]);
	}

  private:
	float_t m_v[3];
};

template <typename T>
class PolarVector3D {
  public:
	typedef T float_t;
	static const float_t earth_radius = 6378.155 * 1000.0; /* earth_radius in m */
	static const float_t km_to_nmi    = 0.5399568;
	static const float_t m_to_ft      = 3.2808399;
	static const float_t ft_to_m      = 0.3048;
	static const float_t from_deg     = M_PI / 180.0;
	static const float_t to_deg       = 180.0 / M_PI;

  PolarVector3D(float_t lat = 0, float_t lon = 0, float_t r = 0) : m_lat(lat), m_lon(lon), m_r(r) {}
	PolarVector3D(const Point& pt, int16_t altft = 0);
	float_t getlat(void) const { return m_lat; }
	float_t getlatdeg(void) const { return m_lat * to_deg; }
	float_t getlon(void) const { return m_lon; }
	float_t getlondeg(void) const { return m_lon * to_deg; }
	Point getcoord(void) const;
	float_t getalt(void) const { return m_r - earth_radius; }
	float_t getaltft(void) const { return (m_r - earth_radius) * m_to_ft; }
	float_t getr(void) const { return m_r; }
	void setlat(float_t v) { m_lat = v; }
	void setlatdeg(float_t v) { m_lat = v * from_deg; }
	void setlon(float_t v) { m_lon = v; }
	void setlondeg(float_t v) { m_lon = v * from_deg; }
	void setcoord(const Point& pt) { m_lon = pt.get_lon_rad(); m_lat = pt.get_lat_rad(); }
	void setr(float_t v) { m_r = v; }
	void setalt(float_t v) { m_r = v + earth_radius; }
	void setaltft(float_t v) { m_r = v * ft_to_m + earth_radius; }
	PolarVector3D<T>& operator*=(float_t s) { m_r *= s; return *this; }
	PolarVector3D<T> operator*(float_t s) const { PolarVector3D<T> pv(*this); pv *= s; return pv; }
	operator CartesianVector3D<T>(void) const;

	template<class Archive> void hibernate_binary(Archive& ar) {
		ar.io(m_lat);
		ar.io(m_lon);
		ar.io(m_r);
	}

  private:
	float_t m_lat, m_lon, m_r;
};

typedef CartesianVector3D<float> CartesianVector3Df;
typedef PolarVector3D<float> PolarVector3Df;
typedef CartesianVector3D<double> CartesianVector3Dd;
typedef PolarVector3D<double> PolarVector3Dd;

std::ostream& operator<<(std::ostream& os, const Point& p);
std::ostream& operator<<(std::ostream& os, const Rect& r);

#endif /* GEOM_H */
