/*****************************************************************************/

/*
 *      navlog.cc  --  Navigation Log Creation.
 *
 *      Copyright (C) 2012, 2013, 2014
 *          Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include "aircraft.h"
#include "fplan.h"
#include "airdata.h"
#include "engine.h"
#include "baro.h"
#include "wmm.h"
#include "wind.h"
#include "icaofpl.h"

#include <cmath>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <ext/stdio_filebuf.h>

#ifdef __WIN32__
#include <windows.h>
#endif

class Aircraft::NavFPlan {
public:
	NavFPlan(const Aircraft& acft, Engine& engine, double cruise_rpm, double cruise_mp, double cruise_bhp);
	~NavFPlan();
	void begin(const WeightBalance::elementvalues_t& wbev = WeightBalance::elementvalues_t());
	void fpl(const FPlanRoute& fpl);
	void profile(const FPlanRoute& fpl);
	void end(const Glib::ustring& templatefile);

protected:
	class Profile {
	  public:
		Profile(Engine& engine);
		std::ostream& profile(std::ostream& os, const FPlanRoute& fpl, double corridor_width = 5);

	  protected:
		TopoDb30::Profile m_profile;
		Glib::RefPtr<Engine::ElevationProfileResult> m_topoquery;
		Glib::Cond m_cond;
		Glib::Mutex m_mutex;
		Engine& m_engine;

		void async_cancel(void);
		void async_clear(void);
		void async_connectdone(void);
		void async_done(void);
		void async_waitdone(void);
	};


	const Aircraft& m_aircraft;
	Engine& m_engine;
	__gnu_cxx::stdio_filebuf<char> *m_tempfile;
	std::string m_tempfilename;
	double m_cruise_rpm;
	double m_cruise_mp;
	double m_cruise_bhp;
};

Aircraft::NavFPlan::Profile::Profile(Engine& engine)
	: m_engine(engine)
{
}

std::ostream& Aircraft::NavFPlan::Profile::profile(std::ostream& os, const FPlanRoute& fpl, double corridor_width)
{
	double dist(0);
	unsigned int nrwpt(fpl.get_nrwpt());
	for (unsigned int i = 1; i < nrwpt;) {
		const FPlanWaypoint& wpt0(fpl[i - 1]);
		const FPlanWaypoint& wpt1(fpl[i]);
		++i;
		bool lastleg(i >= nrwpt);
		async_cancel();
		async_clear();
		m_topoquery = m_engine.async_elevation_profile(wpt0.get_coord(), wpt1.get_coord(), corridor_width);
		async_connectdone();
		async_waitdone();
		unsigned int profsize(m_profile.size());
		for (unsigned int j = 0; j < profsize;) {
			const TopoDb30::ProfilePoint& pp(m_profile[j]);
			TopoDb30::elev_t el(pp.get_elev()), maxel(pp.get_maxelev());
			if (el == TopoDb30::nodata) {
				++j;
				continue;
			}
			if (el == TopoDb30::ocean)
				el = 0;
			if (maxel == TopoDb30::nodata)
				maxel = el;
			else if (maxel == TopoDb30::ocean)
				maxel = 0;
			os << "VPROF|";
			++j;
			if (j == 1)
				os << wpt0.get_icao_name() << '|' << wpt0.get_true_altitude();
			else if (lastleg && j >= profsize)
				os << wpt1.get_icao_name() << '|' << wpt1.get_true_altitude();
			else
				os << '|';
			os << '|' << (dist + pp.get_dist()) << '|' << el << '|' << maxel << std::endl;
		}
		dist += wpt0.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord());
	}
	return os;
}

void Aircraft::NavFPlan::Profile::async_cancel(void)
{
        Engine::ElevationProfileResult::cancel(m_topoquery);
}

void Aircraft::NavFPlan::Profile::async_clear(void)
{
	m_topoquery = Glib::RefPtr<Engine::ElevationProfileResult>();
	m_profile.clear();
}

void Aircraft::NavFPlan::Profile::async_connectdone(void)
{
	if (m_topoquery)
		m_topoquery->connect(sigc::mem_fun(*this, &Profile::async_done));
}

void Aircraft::NavFPlan::Profile::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

void Aircraft::NavFPlan::Profile::async_waitdone(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        for (;;) {
		if (m_topoquery && m_topoquery->is_done()) {
			m_profile.clear();
			if (!m_topoquery->is_error())
				m_profile = m_topoquery->get_result();
			m_topoquery = Glib::RefPtr<Engine::ElevationProfileResult>();
		}
		if (!m_topoquery)
			break;
		m_cond.wait(m_mutex);
	}
}

Aircraft::NavFPlan::NavFPlan(const Aircraft& acft, Engine& engine, double cruise_rpm, double cruise_mp, double cruise_bhp)
	: m_aircraft(acft), m_engine(engine), m_tempfile(0), m_cruise_rpm(cruise_rpm), m_cruise_mp(cruise_mp), m_cruise_bhp(cruise_bhp)
{
}

Aircraft::NavFPlan::~NavFPlan()
{
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
	}
}

void Aircraft::NavFPlan::begin(const WeightBalance::elementvalues_t& wbev)
{	
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
		m_tempfile = 0;
	}
	int fd(Glib::file_open_tmp(m_tempfilename, "vfrnav_navlog_"));
	m_tempfile = new __gnu_cxx::stdio_filebuf<char>(fd, std::ios::out);
	std::ostream os(m_tempfile);
	os << "ACFT|" << m_aircraft.get_callsign()
	   << '|' << m_aircraft.get_manufacturer()
	   << '|' << m_aircraft.get_model()
	   << '|' << m_aircraft.get_year()
	   << '|' << m_aircraft.get_description()
	   << '|' << m_aircraft.get_icaotype()
	   << '|' << m_aircraft.get_equipment()
	   << '|' << m_aircraft.get_transponder()
	   << '|' << m_aircraft.get_pbn()
	   << '|' << m_aircraft.get_colormarking()
	   << '|' << m_aircraft.get_wakecategory() << std::endl
	   << "ACFTD|" << m_aircraft.get_mtom()
	   << '|' << m_aircraft.get_vr0()
	   << '|' << m_aircraft.get_vr1()
	   << '|' << m_aircraft.get_vx0()
	   << '|' << m_aircraft.get_vx1()
	   << '|' << m_aircraft.get_vy()
	   << '|' << m_aircraft.get_vs0()
	   << '|' << m_aircraft.get_vs1()
	   << '|' << m_aircraft.get_vno()
	   << '|' << m_aircraft.get_vne()
	   << '|' << m_aircraft.get_vref0()
	   << '|' << m_aircraft.get_vref1()
	   << '|' << m_aircraft.get_va()
	   << '|' << m_aircraft.get_vfe()
	   << '|' << m_aircraft.get_vgearext()
	   << '|' << m_aircraft.get_vgearret()
	   << '|' << m_aircraft.get_vbestglide()
	   << '|' << m_aircraft.get_glideslope()
	   << '|' << m_aircraft.get_fuelmass()
	   << '|' << m_aircraft.get_maxbhp() << std::endl;
	{
		const Aircraft::WeightBalance& wb(m_aircraft.get_wb());
		for (unsigned int i = 0; i < wb.get_nrelements(); ++i) {
			const Aircraft::WeightBalance::Element& el(wb.get_element(i));
			os << "WBEL|" << el.get_name();
			if (i >= wbev.size())
				os << "|0|0";
			else
				os << '|' << wbev[i].get_value() << '|' << wbev[i].get_unit();
			os << '|' << el.get_massmin()
			   << '|' << el.get_massmax()
			   << '|' << el.get_arm() << '|';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_fixed)
				os << 'F';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_binary)
				os << 'B';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_avgas)
				os << 'A';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_oil)
				os << 'O';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_gear)
				os << 'G';
			for (unsigned int j = 0; j < el.get_nrunits(); ++j) {
				const Aircraft::WeightBalance::Element::Unit& u(el.get_unit(j));
				os << '|' << u.get_name() 
				   << '|' << u.get_factor()
				   << '|' << u.get_offset();
			}
			os << std::endl;
		}
		for (unsigned int i = 0; i < wb.get_nrenvelopes(); ++i) {
			const Aircraft::WeightBalance::Envelope& env(wb.get_envelope(i));
			os << "WBENV|" << env.get_name();
			for (unsigned int j = 0; j < env.size(); ++j) {
				const Aircraft::WeightBalance::Envelope::Point& pt(env[j]);
				os << '|' << pt.get_mass() << '|' << pt.get_arm();
			}
			os << std::endl;
		}
	}
	{
		const Aircraft::Distances& dists(m_aircraft.get_dist());
		for (unsigned int i = 0; i < dists.get_nrtakeoffdist(); ++i) {
			const Aircraft::Distances::Distance& dist(dists.get_takeoffdist(i));
			os << "TKOFFDIST|" << dist.get_name()
			   << '|' << dist.get_vrotate()
			   << '|' << dist.get_mass() << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << std::endl;
		}
		for (unsigned int i = 0; i < dists.get_nrldgdist(); ++i) {
			const Aircraft::Distances::Distance& dist(dists.get_ldgdist(i));
			os << "LDGDIST|" << dist.get_name()
			   << '|' << dist.get_vrotate()
			   << '|' << dist.get_mass() << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << std::endl;
		}
	}
}

void Aircraft::NavFPlan::fpl(const FPlanRoute& fpl)
{
	if (!m_tempfile || fpl.get_nrwpt() < 1)
		return;
	std::ostream os(m_tempfile);
	{
		IcaoFlightPlan icaofpl(m_engine);
		icaofpl.populate(fpl, IcaoFlightPlan::awymode_keep);
		icaofpl.set_aircraft(m_aircraft, m_cruise_rpm, m_cruise_mp, m_cruise_bhp);
		os << "FPLH|" << fpl.get_time_offblock_unix()
		   << '|' << fpl.get_time_onblock_unix()
		   << '|' << icaofpl.get_fpl() << std::endl;
	}
	for (unsigned int i = 0; i < 2; ++i) {
		const FPlanWaypoint& wpt(fpl[(fpl.get_nrwpt() - 1) & -i]);
		{
			double tas(0), fuelflow(0), bhp(m_cruise_bhp), rpm(m_cruise_rpm), mp(m_cruise_mp), da(wpt.get_density_altitude() + 1500.0);
			m_aircraft.get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
			os << (i ? "FPLDESTHLD|" : "FPLDEPHLD|") << wpt.get_density_altitude()
			   << '|' << fuelflow << '|' << tas << '|' << bhp << std::endl;
		}
		IcaoFlightPlan::FindCoord fc(m_engine);
		if (!fc.find(wpt.get_icao(), wpt.get_name(), IcaoFlightPlan::FindCoord::flag_subtables | IcaoFlightPlan::FindCoord::flag_airport, wpt.get_coord()))
			continue;
		AirportsDb::Airport arpt;
		if (!fc.get_airport(arpt))
			continue;
		if (!arpt.is_valid())
			continue;
		for (unsigned int j = 0; j < arpt.get_nrrwy(); ++j) {
			const AirportsDb::Airport::Runway& rwy(arpt.get_rwy(j));
			os << (i ? "FPLDESTRWY" : "FPLDEPRWY")
			   << '|' << rwy.get_ident_he()
			   << '|' << rwy.get_ident_le()
			   << '|' << rwy.get_length()
			   << '|' << rwy.get_width()
			   << '|' << rwy.get_surface()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_he_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_he_coord().get_lon_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_le_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_le_coord().get_lon_deg_dbl()
			   << '|' << rwy.get_he_tda()
			   << '|' << rwy.get_he_lda()
			   << '|' << rwy.get_he_disp()
			   << '|' << rwy.get_he_hdg()
			   << '|' << rwy.get_he_elev()
			   << '|' << rwy.get_le_tda()
			   << '|' << rwy.get_le_lda()
			   << '|' << rwy.get_le_disp()
			   << '|' << rwy.get_le_hdg()
			   << '|' << rwy.get_le_elev()
			   << '|' << (int)rwy.is_he_usable()
			   << '|' << (int)rwy.is_le_usable();
			for (unsigned int k = 0; k < 8; ++k)
				os << '|' << AirportsDb::Airport::Runway::get_light_string(rwy.get_he_light(k));
			for (unsigned int k = 0; k < 8; ++k)
				os << '|' << AirportsDb::Airport::Runway::get_light_string(rwy.get_le_light(k));
			os << std::endl;
		}
		for (unsigned int j = 0; j < arpt.get_nrhelipads(); ++j) {
			const AirportsDb::Airport::Helipad& hp(arpt.get_helipad(j));
			os << (i ? "FPLDESTHP" : "FPLDEPHP")
			   << '|' << hp.get_ident()
			   << '|' << hp.get_length()
			   << '|' << hp.get_width()
			   << '|' << hp.get_surface()
			   << '|' << std::fixed << std::setprecision(9) << hp.get_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << hp.get_coord().get_lon_deg_dbl()
			   << '|' << hp.get_hdg()
			   << '|' << hp.get_elev()
			   << '|' << (int)hp.is_usable() << std::endl;
		}
		for (unsigned int j = 0; j < arpt.get_nrcomm(); ++j) {
			const AirportsDb::Airport::Comm& comm(arpt.get_comm(j));
			std::string sector(comm.get_sector());
			std::string oprhours(comm.get_oprhours());
			for (std::string::iterator si(sector.begin()), se(sector.end()); si != se; ++si)
				if (*si == '\n' || *si == '\r')
					*si = ' ';
			for (std::string::iterator si(oprhours.begin()), se(oprhours.end()); si != se; ++si)
				if (*si == '\n' || *si == '\r')
					*si = ' ';
			os << (i ? "FPLDESTCOMM" : "FPLDEPCOMM")
			   << '|' << comm.get_name()
			   << '|' << sector
			   << '|' << oprhours;
			for (unsigned int k = 0; k < 5; ++k)
				os << '|' << comm[k];
			os << std::endl;
		}
	}
        for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
		const FPlanWaypoint& wpt(fpl[i]);
		os << "FPLL|" << wpt.get_icao() << '|' << wpt.get_name() << '|' << wpt.get_pathname()
		   << '|' << wpt.get_pathcode_name() << '|';
		if (!i)
			os << (wpt.get_time_unix() - fpl.get_time_offblock_unix());
		else
			os << (wpt.get_flighttime() - fpl[i - 1].get_flighttime());
		os << '|' << wpt.get_frequency()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_coord().get_lat_deg_dbl()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_coord().get_lon_deg_dbl()
		   << '|' << wpt.get_altitude()
		   << '|' << ((wpt.get_flags() & FPlanWaypoint::altflag_standard) ? "STD" : "QNH")
		   << '|' << ((wpt.get_flags() & FPlanWaypoint::altflag_ifr) ? "IFR" : "VFR")
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_truealt()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_truetrack_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_trueheading_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_declination_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_magnetictrack_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_magneticheading_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_dist_nmi()
		   << '|';
		if (i)
			os << std::fixed << std::setprecision(9) << (wpt.get_fuel_usg() - fpl[i - 1].get_fuel_usg());
		else
			os << '0';
		double gs(wpt.get_tas_kts());
		if (wpt.get_windspeed() > 0) {
			Wind<double> wind;
			wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
			wind.set_crs_tas(wpt.get_truetrack_deg(), wpt.get_tas_kts());
			if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0)
				gs = 1;
			else
				gs = wind.get_gs();
		}
		os << '|' << std::fixed << std::setprecision(9) << wpt.get_tas_kts()
		   << '|' << std::fixed << std::setprecision(9) << gs
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_rpm()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_mp_inhg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_winddir_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_windspeed_kts()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_qff_hpa()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_isaoffset_kelvin()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_oat_kelvin()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_pressure_altitude()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_density_altitude()
		   << std::endl;
	}
}

void Aircraft::NavFPlan::profile(const FPlanRoute& fpl)
{
	if (!m_tempfile || fpl.get_nrwpt() < 2)
		return;
	std::ostream os(m_tempfile);
	Profile p(m_engine);
	p.profile(os, fpl);
}

void Aircraft::NavFPlan::end(const Glib::ustring& templatefile)
{
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
		m_tempfile = 0;
	}
	std::vector<std::string> argv;
	std::vector<std::string> env(Glib::listenv());
	{
		bool needfile(true);
		for (std::vector<std::string>::iterator i(env.begin()), e(env.end()); i != e; ++i) {
			if (*i == "VFRNAV_FPLFILE") {
				*i += "=" + m_tempfilename;
				needfile = false;
				continue;
			}
			*i += "=" + Glib::getenv(*i);
		}
		if (needfile)
			env.push_back("VFRNAV_FPLFILE=" + m_tempfilename);
	}
	if (false) {
		std::cerr << "Environment:" << std::endl;
		for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
			std::cerr << *i << std::endl;
	}
#ifdef __WIN32__
	{
		std::string version;
		std::string path;
		static const char * const regkeys[] = {
			"Software\\LibreOffice\\LibreOffice",
			"Software\\OpenOffice.org\\OpenOffice.org",
			0
		};
		for (const char * const *regkey = regkeys; *regkey; ++regkey) {
			HKEY hKey;
			if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, *regkey, 0, KEY_READ, &hKey) == ERROR_SUCCESS) {
				if (false)
					std::cerr << "Enumerating keys in HKLM\\" << *regkey << std::endl;
				for (DWORD dwIndex = 0;; ++dwIndex) {
					TCHAR lpName[1024];
					DWORD lpcbName(sizeof(lpName));
					FILETIME fTime;
					if (RegEnumKeyEx(hKey, dwIndex, lpName, &lpcbName, NULL, NULL, NULL, &fTime) != ERROR_SUCCESS)
						break;
					std::string ver1(lpName);
					if (false)
						std::cerr << "Subkey " << dwIndex << ": " << ver1 << std::endl;
					if (!path.empty() && ver1 < version)
						continue;
					//TCHAR lpData[1024];
					BYTE lpData[1024];
					DWORD cbData(sizeof(lpData));
					//LONG err(RegGetValue(hKey, lpName, "Path", RRF_RT_REG_SZ, 0, &lpData, &cbData));
					HKEY hKey2;
					LONG err(RegOpenKeyEx(hKey, lpName, 0, KEY_READ, &hKey2));
					if (err != ERROR_SUCCESS) {
						if (true)
							std::cerr << "Cannot open key HKLM\\" << *regkey << "\\" << lpName
								  << " Error " << g_win32_error_message(err) << " (" << err << ')' << std::endl;
						continue;
					}
					DWORD lpType;
					err = RegQueryValueEx(hKey2, "Path", 0, &lpType, lpData, &cbData);
					if (cbData < sizeof(lpData))
						cbData[lpData] = 0;
					RegCloseKey(hKey2);
					if (err != ERROR_SUCCESS) {
						if (true)
							std::cerr << "Cannot get key HKLM\\" << *regkey << "\\" << lpName
								  << " value: Error " << g_win32_error_message(err) << " (" << err << ')' << std::endl;
						continue;
					}
					version = ver1;
					path = (char *)lpData;
					
				}
				RegCloseKey(hKey);
			}
		}
		if (!path.empty()) {
			std::cerr << "Found libreoffice V" << version << ": " << path << std::endl;
			argv.push_back(path);
		} else {
			std::cerr << "Libreoffice not found; trying soffice.exe" << std::endl;
			argv.push_back("soffice.exe");
		}
	}
#else
	argv.push_back("soffice");
#endif
	argv.push_back("--calc");
	argv.push_back("--norestore");
	argv.push_back("--nologo");
	argv.push_back("-n");
	argv.push_back(templatefile);
	argv.push_back("macro://./Standard.FPLImport.Main()");
	Glib::spawn_async_with_pipes(Glib::get_home_dir(), argv, env, Glib::SPAWN_SEARCH_PATH);
}

void Aircraft::navfplan(const Glib::ustring& templatefile, Engine& engine, const FPlanRoute& fplan,
			double cruise_rpm, double cruise_mp, double cruise_bhp, const WeightBalance::elementvalues_t& wbev) const
{
	NavFPlan n(*this, engine, cruise_rpm, cruise_mp, cruise_bhp);
	n.begin(wbev);
	n.fpl(fplan);
	n.profile(fplan);
	n.end(templatefile);
}

class Aircraft::GnmCellSpec {
public:
	GnmCellSpec(const std::string& name = "", const std::string& value = "A1");
	const std::string& get_name(void) const { return m_name; }
	unsigned int get_row(void) const { return m_row; }
	unsigned int get_col(void) const { return m_col; }
	std::string get_rowstr(unsigned int wptnr = 0, unsigned int wptrows = 1, unsigned int pgbrkheader = 0,
			       unsigned int pgbrkinterval = 0, unsigned int pgbrkoffset = 0) const;
	std::string get_colstr(void) const;
	std::string get_colrow(bool absolute = false) const;
	static bool is_fplanrow(const std::string& name);
	bool is_fplanrow(void) const { return is_fplanrow(get_name()); }
	int compare(const GnmCellSpec& x) const { return get_name().compare(x.get_name()); }
	bool operator==(const GnmCellSpec& x) const { return compare(x) == 0; }
	bool operator!=(const GnmCellSpec& x) const { return compare(x) != 0; }
	bool operator<(const GnmCellSpec& x) const { return compare(x) < 0; }
	bool operator<=(const GnmCellSpec& x) const { return compare(x) <= 0; }
	bool operator>(const GnmCellSpec& x) const { return compare(x) > 0; }
	bool operator>=(const GnmCellSpec& x) const { return compare(x) >= 0; }

protected:
	std::string m_name;
	unsigned int m_row;
	unsigned int m_col;
};

Aircraft::GnmCellSpec::GnmCellSpec(const std::string& name, const std::string& value)
	: m_name(name), m_row(0), m_col(0)
{
	std::string::size_type n(value.rfind('!'));
	if (n == std::string::npos)
		n = 0;
	else
		++n;
	unsigned int rowdigits(0), coldigits(0);
	for (; n < value.size(); ++n) {
		char ch(value[n]);
		if (ch == '$')
			continue;
		if (ch >= '0' && ch <= '9') {
			m_row *= 10;
			m_row += ch - '0';
			++rowdigits;
			continue;
		}
		if (ch >= 'A' && ch <= 'Z') {
			m_col *= 26;
			m_col += ch - 'A';
			++coldigits;
			continue;
		}
		std::ostringstream oss;
		oss << "invalid row/col character: " << ch;
		throw std::runtime_error(oss.str());
	}
	if (!rowdigits)
		throw std::runtime_error(value + " has no row spec");
	if (!coldigits)
		throw std::runtime_error(value + " has no col spec");
	if (!m_row)
		throw std::runtime_error("row should start at one");
	--m_row;
	{
		unsigned int x(1), a(0);
		for (unsigned int i(0); i < coldigits; ++i) {
			a += x;
			x *= 26;
		}
		m_col += a - 1;
	}
}

std::string Aircraft::GnmCellSpec::get_rowstr(unsigned int wptnr, unsigned int wptrows, unsigned int pgbrkheader,
					      unsigned int pgbrkinterval, unsigned int pgbrkoffset) const
{
	std::ostringstream oss;
	unsigned int hdr(0);
	if (pgbrkinterval)
		hdr = (wptnr + pgbrkoffset) / pgbrkinterval;
	oss << (get_row() + wptnr * wptrows + hdr * pgbrkheader);
	return oss.str();
}

std::string Aircraft::GnmCellSpec::get_colstr(void) const
{
	std::ostringstream oss;
	oss << get_col();
	return oss.str();
}

std::string Aircraft::GnmCellSpec::get_colrow(bool absolute) const
{
	std::ostringstream oss;
	if (absolute)
		oss << '$';
	{
		unsigned int n(get_col() + 1);
		std::string c;
		for (;;) {
			if (n <= 26) {
				c = (char)('A' + n - 1) + c;
				break;
			}
			unsigned int nn(n);
			n /= 26;
			nn -= n * 26;
			c = (char)('A' + nn) + c;
		}
		oss << c;
	}
	if (absolute)
		oss << '$';
	oss << (get_row() + 1);
	return oss.str();
}

bool Aircraft::GnmCellSpec::is_fplanrow(const std::string& name)
{
	return name == "WPTICAO" || name == "WPTNAME" || name == "WPTICAONAME" || name == "WPTIDENT" || 
		name == "WPTPATHNAME" || name == "WPTPATHCODE" || name == "WPTNOTE" || name == "WPTTIME" || name == "WPTFLTTIME" || 
		name == "WPTFREQ" || name == "WPTLAT" || name == "WPTLON" || name == "WPTALT" || name == "WPTALTSTD" || 
		name == "WPTFRULES" || name == "WPTCLIMB" || name == "WPTDESCENT" || name == "WPTTURNPOINT" || 
		name == "WPTWINDDIR" || name == "WPTWINDSPEED" || name == "WPTQFF" || name == "WPTISAOFFS" || 
		name == "WPTSLT" || name == "WPTOAT" || name == "WPTPA" || name == "WPTTA" || name == "WPTDA" || name == "WPTTRUEALT" || 
		name == "WPTTT" || name == "WPTTH" || name == "WPTMT" || name == "WPTMH" || name == "WPTDECL" || 
		name == "WPTDIST" || name == "WPTFUEL" || name == "WPTTAS" || name == "WPTCRUISETAS" || name == "WPTAVGTAS" ||
		name == "WPTGS" || name == "WPTCRUISEGS" || name == "WPTAVGGS" ||
		name == "WPTRPM" || name == "WPTMP" || name == "WPTTYPE";
}

class Aircraft::GnmCellSpecs : public std::set<GnmCellSpec> {
public:
	typedef std::set<GnmCellSpec> base_t;
	using base_t::find;
	GnmCellSpecs(void);
	GnmCellSpecs(const xmlpp::Node *sheet);
	const_iterator find(const std::string& name) const { return base_t::find(GnmCellSpec(name)); }
	const_iterator find(const std::string& pfx, const std::string& name) const { return find(pfx + name); }
	const_iterator find(const char *name) const { if (!name) return end(); return find(std::string(name)); }
	const_iterator find(const std::string& pfx, const char *name) const { if (!name) return end(); return find(pfx, std::string(name)); }
	void parse(const xmlpp::Node *sheet);
	bool computefplrows(void);
	unsigned int get_fplfirstrow(void) const { return m_fplfirstrow; }
	unsigned int get_fplrows(void) const { return m_fplrows; }
	unsigned int get_pgbrkheader(void) const { return m_pgbrkheader; }
	unsigned int get_pgbrkinterval(void) const { return m_pgbrkinterval; }
	unsigned int get_pgbrkoffset(void) const { return m_pgbrkoffset; }
	unsigned int get_pagelen(void) const { return m_pagelen; }

	std::ostream& print(std::ostream& os) const;
	unsigned int get_rowadjlinelen(void) const { return m_rowadjlinelen; }
	int get_rowadjlineoffs(void) const { return m_rowadjlineoffs; }
	double get_rowadjlineheight(void) const { return m_rowadjlineheight; }
	double get_rowadjlineheightoffs(void) const { return m_rowadjlineheightoffs; }
	bool is_rowadj(void) const {
		return !std::isnan(get_rowadjlineheight()) && !std::isnan(get_rowadjlineheightoffs()) &&
			get_rowadjlinelen() > 0;
	}

protected:
	unsigned int m_fplfirstrow;
	unsigned int m_fplrows;
	unsigned int m_pgbrkheader;
	unsigned int m_pgbrkinterval;
	unsigned int m_pgbrkoffset;
	unsigned int m_pagelen;
	unsigned int m_rowadjlinelen;
	int m_rowadjlineoffs;
	double m_rowadjlineheight;
	double m_rowadjlineheightoffs;
};

Aircraft::GnmCellSpecs::GnmCellSpecs(void)
	: m_fplfirstrow(0), m_fplrows(0), m_pgbrkheader(0), m_pgbrkinterval(0), m_pgbrkoffset(0),
	  m_pagelen(100), m_rowadjlinelen(0), m_rowadjlineoffs(0),
	  m_rowadjlineheight(std::numeric_limits<double>::quiet_NaN()),
	  m_rowadjlineheightoffs(std::numeric_limits<double>::quiet_NaN())
{
}

Aircraft::GnmCellSpecs::GnmCellSpecs(const xmlpp::Node *sheet)
	: m_fplfirstrow(0), m_fplrows(0), m_pgbrkheader(0), m_pgbrkinterval(0), m_pgbrkoffset(0),
	  m_pagelen(100), m_rowadjlinelen(0), m_rowadjlineoffs(0),
	  m_rowadjlineheight(std::numeric_limits<double>::quiet_NaN()),
	  m_rowadjlineheightoffs(std::numeric_limits<double>::quiet_NaN())
{
	parse(sheet);
	computefplrows();
}

void Aircraft::GnmCellSpecs::parse(const xmlpp::Node *sheet)
{
	if (!sheet)
		return;
	xmlpp::Node::NodeList nlnames(sheet->get_children("Names"));
	for (xmlpp::Node::NodeList::const_iterator ni(nlnames.begin()), ne(nlnames.end()); ni != ne; ++ni) {
		xmlpp::Node::NodeList nlname((*ni)->get_children("Name"));
		for (xmlpp::Node::NodeList::const_iterator ni(nlname.begin()), ne(nlname.end()); ni != ne; ++ni) {
			std::string name, value;
			xmlpp::Node::NodeList nll((*ni)->get_children("name"));
			for (xmlpp::Node::NodeList::const_iterator ni(nll.begin()), ne(nll.end()); ni != ne; ++ni) {
				const xmlpp::TextNode* txt(static_cast<xmlpp::Element *>(*ni)->get_child_text());
				if (!txt)
					continue;
				name = txt->get_content();
			}
			nll = (*ni)->get_children("value");
			for (xmlpp::Node::NodeList::const_iterator ni(nll.begin()), ne(nll.end()); ni != ne; ++ni) {
				const xmlpp::TextNode* txt(static_cast<const xmlpp::Element *>(*ni)->get_child_text());
				if (!txt)
					continue;
				value = txt->get_content();
			}
			if (name == "RowAdjust_LINEHEIGHT") {
				m_rowadjlineheight = strtod(value.c_str(), 0);
				continue;
			}
			if (name == "RowAdjust_LINEHEIGHTOFFSET") {
				m_rowadjlineheightoffs = strtod(value.c_str(), 0);
				continue;
			}
			if (name == "RowAdjust_LINELEN") {
				m_rowadjlinelen = strtoul(value.c_str(), 0, 0);
				continue;
			}
			if (name == "RowAdjust_LINEOFFS") {
				m_rowadjlineoffs = strtol(value.c_str(), 0, 0);
				continue;
			}
			if (name == "PageBreak_HEADER") {
				m_pgbrkheader = strtoul(value.c_str(), 0, 0);
				continue;
			}
			if (name == "PageBreak_INTERVAL") {
				m_pgbrkinterval = strtoul(value.c_str(), 0, 0);
				continue;
			}
			if (name == "PageBreak_OFFSET") {
				m_pgbrkoffset = strtoul(value.c_str(), 0, 0);
				continue;
			}
			if (name == "Page_LENGTH") {
				m_pagelen = strtoul(value.c_str(), 0, 0);
				continue;
			}
			if (name.compare(0, 5, "Cell_") || value.empty())
				continue;
			try {
				std::pair<iterator,bool> ins(insert(GnmCellSpec(name.substr(5), value)));
				if (!ins.second)
					std::cerr << "Gnumeric Name Error: " << name << '=' << value << ": duplicate name" << std::endl;
			} catch (const std::runtime_error& e) {
				std::cerr << "Gnumeric Name Error: " << name << '=' << value << ": " << e.what() << std::endl;
			}
		}
	}
}

bool Aircraft::GnmCellSpecs::computefplrows(void)
{
	m_fplfirstrow = m_fplrows = 0;
	unsigned int firstrow(std::numeric_limits<unsigned int>::max());
	unsigned int lastrow(std::numeric_limits<unsigned int>::min());
	for (const_iterator i(begin()), e(end()); i != e; ++i) {
		if (!i->is_fplanrow())
			continue;
		firstrow = std::min(firstrow, i->get_row());
		lastrow = std::max(lastrow, i->get_row());
	}
	if (firstrow > lastrow)
		return false;
	m_fplfirstrow = firstrow;
	m_fplrows = lastrow - firstrow + 1;
	return true;
}

std::ostream& Aircraft::GnmCellSpecs::print(std::ostream& os) const
{
	for (const_iterator i(begin()), e(end()); i != e; ++i) {
		os << "Cell: " << i->get_name() << " R" << i->get_row() << " C" << i->get_col()
		   << ' ' << i->get_colrow();
		if (i->is_fplanrow())
			os << " WPT";
		os << std::endl;
	}
	return os;
}

void Aircraft::navfplan_gnumeric(const std::string& outfile, Engine& engine, const FPlanRoute& fplan,
				 double cruise_rpm, double cruise_mp, double cruise_bhp,
				 const WeightBalance::elementvalues_t& wbev, const std::string& templatefile,
				 gint64 gfsminreftime, gint64 gfsmaxreftime,
				 gint64 gfsminefftime, gint64 gfsmaxefftime, bool hidewbpage) const
{
	xmlpp::DomParser p(templatefile);
	if (!p) {
		std::ostringstream oss;
		oss << "cannot open template file " << templatefile;
		throw std::runtime_error(oss.str());
	}
	xmlpp::Document *doc(p.get_document());
	xmlpp::Element *root(doc->get_root_node());
	if (root->get_name() != "Workbook") {
		std::ostringstream oss;
		oss << "invalid template file " << templatefile << " (root node " << root->get_name() << ')';
		throw std::runtime_error(oss.str());
	}
	xmlpp::Node::NodeList nlsheets(root->get_children("Sheets"));
	for (xmlpp::Node::NodeList::const_iterator ni(nlsheets.begin()), ne(nlsheets.end()); ni != ne; ++ni) {
		xmlpp::Node::NodeList nlsheet((*ni)->get_children("Sheet"));
		unsigned int sheetnr(0);
		for (xmlpp::Node::NodeList::const_iterator ni(nlsheet.begin()), ne(nlsheet.end()); ni != ne; ++ni, ++sheetnr) {
			double armmin(std::numeric_limits<double>::max());
			double armmax(std::numeric_limits<double>::min());
			double massmin(std::numeric_limits<double>::max());
			double massmax(std::numeric_limits<double>::min());
			unsigned int atcfpllen(0);
			unsigned int atcfplrow(~0U);
			GnmCellSpecs cellspecs(*ni);
			if (!false)
				cellspecs.print(std::cerr);
			// hide page 1 if the model does not have w&b info and takeoff/landing distances
			if (sheetnr == 1 && (hidewbpage || (!get_wb().get_nrelements() && !get_wb().get_nrenvelopes() &&
							    !get_dist().get_nrtakeoffdist() && !get_dist().get_nrldgdist()))) {
				xmlpp::Attribute *attr(static_cast<xmlpp::Element *>(*ni)->get_attribute("Visibility"));
				if (attr)
					attr->set_value("GNM_SHEET_VISIBILITY_HIDDEN");
			}
			xmlpp::Node::NodeList nlcells((*ni)->get_children("Cells"));
			for (xmlpp::Node::NodeList::const_iterator ni(nlcells.begin()), ne(nlcells.end()); ni != ne; ++ni) {
				switch (sheetnr) {
				case 0:
				{
					GnmCellSpecs::const_iterator csi;
					{
						IcaoFlightPlan::awymode_t awym(IcaoFlightPlan::awymode_keep);
						bool ok(false);
						if ((ok = (csi = cellspecs.find("ATCFPL")) != cellspecs.end()))
							awym = IcaoFlightPlan::awymode_keep;
						else if ((ok = (csi = cellspecs.find("ATCFPLC")) != cellspecs.end()))
							awym = IcaoFlightPlan::awymode_collapse;
						else if ((ok = (csi = cellspecs.find("ATCFPLCDCT")) != cellspecs.end()))
							awym = IcaoFlightPlan::awymode_collapse_dct;
						else if ((ok = (csi = cellspecs.find("ATCFPLCALL")) != cellspecs.end()))
							awym = IcaoFlightPlan::awymode_collapse_all;
						if (ok) {
							IcaoFlightPlan icaofpl(engine);
							icaofpl.populate(fplan, awym);
							icaofpl.set_aircraft(*this, cruise_rpm, cruise_mp, cruise_bhp);
							std::string fpl(icaofpl.get_fpl());
							atcfpllen = fpl.size();
							atcfplrow = csi->get_row();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr());
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(fpl);
						}
					}
					if ((csi = cellspecs.find("FPLID")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.get_id();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("FPLNOTE")) != cellspecs.end()) {
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "60");
						el->set_child_text(fplan.get_note());
					}
					if ((csi = cellspecs.find("NRWPT")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.get_nrwpt();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("OFFBLOCK")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.get_time_offblock_unix();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("ONBLOCK")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.get_time_onblock_unix();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("TOTDIST")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.total_distance_nmi_dbl();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("GCDIST")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.gc_distance_nmi_dbl();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("MAXALT")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.max_altitude();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("FRULES")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << fplan.get_flightrules();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "60");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("GFSMINREFTIME")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << gfsminreftime;
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("GFSMAXREFTIME")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << gfsmaxreftime;
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("GFSMINEFFTIME")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << gfsminefftime;
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					if ((csi = cellspecs.find("GFSMAXEFFTIME")) != cellspecs.end()) {
						std::ostringstream oss;
						oss << gfsmaxefftime;
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", csi->get_rowstr());
						el->set_attribute("Col", csi->get_colstr());
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					double cumdist(0);
					for (unsigned int i(0), n(fplan.get_nrwpt()); i < n + 2; ++i) {
						unsigned int wptidx(i);
						if (wptidx >= n)
							wptidx = 0;
						if (i == n + 1)
							wptidx = n - 1;
						const FPlanWaypoint& wpt(fplan[wptidx]);
						if (i >= n)
							wptidx = 0;
						std::string pfx("WPT");
						if (i == n)
							pfx = "DEP";
						else if (i == n + 1)
							pfx = "DEST";
						if ((csi = cellspecs.find(pfx, "ICAO")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_icao());
						}
						if ((csi = cellspecs.find(pfx, "NAME")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_name());
						}
						if ((csi = cellspecs.find(pfx, "ICAONAME")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_icao_name());
						}
						if ((csi = cellspecs.find(pfx, "IDENT")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_ident());
						}
						if ((csi = cellspecs.find(pfx, "PATHNAME")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_pathname());
						}
						if ((csi = cellspecs.find(pfx, "PATHCODE")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_pathcode_name());
						}
						if ((csi = cellspecs.find(pfx, "NOTE")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_note());
						}
						if ((csi = cellspecs.find(pfx, "TIME")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_time_unix();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "FLTTIME")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_flighttime();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "FREQ")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_frequency();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "LAT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_coord().get_lat_deg_dbl();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "LON")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_coord().get_lon_deg_dbl();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "ALT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "ALTSTD")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.is_standard() ? "STD" : "QNH");
						}
						if ((csi = cellspecs.find(pfx, "FRULES")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.is_ifr() ? "IFR" : "VFR");
						}
						if ((csi = cellspecs.find(pfx, "CLIMB")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(wpt.is_climb() ? "1" : "0");
						}
						if ((csi = cellspecs.find(pfx, "DESCENT")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(wpt.is_descent() ? "1" : "0");
						}
						if ((csi = cellspecs.find(pfx, "TURNPOINT")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(wpt.is_turnpoint() ? "1" : "0");
						}
						if ((csi = cellspecs.find(pfx, "WINDDIR")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_winddir_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "WINDSPEED")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_windspeed_kts();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "QFF")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_qff_hpa();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "ISAOFFS")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_isaoffset_kelvin();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "SLT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_sltemp_kelvin();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "OAT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_oat_kelvin();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "PA")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_pressure_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TA")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_true_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "DA")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_density_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TRUEALT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_truealt();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_truetrack_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TH")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_trueheading_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "MT")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_magnetictrack_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "MH")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_magneticheading_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "DECL")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_declination_deg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "DIST")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_dist_nmi();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if (i < n && (csi = cellspecs.find(pfx, "CUMDIST")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << cumdist;
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "FUEL")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_fuel_usg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TAS")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_tas_kts();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "CRUISETAS")) != cellspecs.end()) {
							std::ostringstream oss;
							bool val(!wpt.is_climb() && !wpt.is_descent());
							if (val)
								oss << wpt.get_tas_kts();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", val ? "40" : "60");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "AVGTAS")) != cellspecs.end()) {
							std::ostringstream oss;
							bool val(!wpt.is_climb() && !wpt.is_descent());
							if (val) {
								oss << wpt.get_tas_kts();
							} else if (i + 1 < n && wpt.get_dist() > 0) {
								const FPlanWaypoint& wptn(fplan[i + 1]);
								if (wptn.get_flighttime() > wpt.get_flighttime()) {
									Wind<double> wind;
									wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts(), wptn.get_winddir_deg(), wptn.get_windspeed_kts());
									wind.set_crs_gs(wpt.get_truetrack_deg(),
											wpt.get_dist_nmi() * 3600 / (wptn.get_flighttime() - wpt.get_flighttime()));
									if (false)
										std::cerr << "AVGTAS: Wind " << wind.get_winddir() << '/' << wind.get_windspeed()
											  << " TAS " << wind.get_tas() << " GS " << wind.get_gs()
											  << " HDG " << wind.get_hdg() << " CRS " << wind.get_crs() << std::endl;
									if (!std::isnan(wind.get_tas())) {
										oss << wind.get_tas();
										val = true;
									}
								}
							}
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", val ? "40" : "60");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "GS")) != cellspecs.end()) {
							std::ostringstream oss;
							double gs(wpt.get_tas_kts());
							if (wpt.get_windspeed() > 0) {
								Wind<double> wind;
								wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
								wind.set_crs_tas(wpt.get_truetrack_deg(), wpt.get_tas_kts());
								if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0)
									gs = 1;
								else
									gs = wind.get_gs();
							}
							oss << gs;
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "CRUISEGS")) != cellspecs.end()) {
							std::ostringstream oss;
							bool val(!wpt.is_climb() && !wpt.is_descent());
							if (val) {
								double gs(wpt.get_tas_kts());
								if (wpt.get_windspeed() > 0) {
									Wind<double> wind;
									wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
									wind.set_crs_tas(wpt.get_truetrack_deg(), wpt.get_tas_kts());
									if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0)
										gs = 1;
									else
										gs = wind.get_gs();
								}
								oss << gs;
							}
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", val ? "40" : "60");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "AVGGS")) != cellspecs.end()) {
							std::ostringstream oss;
							bool val(!wpt.is_climb() && !wpt.is_descent());
							if (val) {
								double gs(wpt.get_tas_kts());
								if (wpt.get_windspeed() > 0) {
									Wind<double> wind;
									wind.set_wind(wpt.get_winddir_deg(), wpt.get_windspeed_kts());
									wind.set_crs_tas(wpt.get_truetrack_deg(), wpt.get_tas_kts());
									if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0)
										val = false;
									else
										oss << wind.get_gs();
								} else {
									oss << gs;
								}
							} else if (i + 1 < n && wpt.get_dist() > 0) {
								const FPlanWaypoint& wptn(fplan[i + 1]);
								if (wptn.get_flighttime() > wpt.get_flighttime()) {
									double gs(wpt.get_dist_nmi() * 3600 / (wptn.get_flighttime() - wpt.get_flighttime()));
									oss << gs;
									val = true;
								}
							}
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", val ? "40" : "60");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "RPM")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_rpm();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "MP")) != cellspecs.end()) {
							std::ostringstream oss;
							oss << wpt.get_mp_inhg();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if ((csi = cellspecs.find(pfx, "TYPE")) != cellspecs.end()) {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", csi->get_rowstr(wptidx, cellspecs.get_fplrows(),
												 cellspecs.get_pgbrkheader(),
												 cellspecs.get_pgbrkinterval(),
												 cellspecs.get_pgbrkoffset()));
							el->set_attribute("Col", csi->get_colstr());
							el->set_attribute("ValueType", "60");
							el->set_child_text(wpt.get_type_string());
						}
						cumdist += wpt.get_dist_nmi();
					}
					break;
				}

				case 1:
				{
					{
						std::ostringstream oss;
						oss << get_fuelmass();
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", "3");
						el->set_attribute("Col", "24");
						el->set_attribute("ValueType", "40");
						el->set_child_text(oss.str());
					}
					std::string loadedfuelformula;
					bool fuelarmset(false);
					double kggain(std::numeric_limits<double>::quiet_NaN()), kgoffs(std::numeric_limits<double>::quiet_NaN());
					const Aircraft::WeightBalance& wb(get_wb());
					for (unsigned int i = 0; i < wb.get_nrelements(); ++i) {
						const Aircraft::WeightBalance::Element& wel(wb.get_element(i));
						std::ostringstream row;
						row << (15 + i);
						{
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "0");
							el->set_attribute("ValueType", "60");
							el->set_child_text(wel.get_name());
						}
						{
							std::ostringstream oss;
							if (wel.get_flags() & Aircraft::WeightBalance::Element::flag_fixed)
								oss << wel.get_massmin();
							else if (wel.get_flags() & Aircraft::WeightBalance::Element::flag_binary)
								oss << '0';
							else if (i >= wbev.size())
								oss << '0';
							else
								oss << wbev[i].get_value();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "1");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						{
							std::string unit;
							if (i < wbev.size() && wbev[i].get_unit() < wel.get_nrunits())
								unit = wel.get_unit(wbev[i].get_unit()).get_name();
							if (unit.empty() && wel.get_flags() & Aircraft::WeightBalance::Element::flag_gear) {
								for (unsigned int j(0); j < wel.get_nrunits(); ++j) {
									if (wel.get_unit(j).get_name() == "Retracted")
										unit = wel.get_unit(j).get_name();
								}
							}
							if (unit.empty() && wel.get_nrunits())
								unit = wel.get_unit(0).get_name();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "2");
							el->set_attribute("ValueType", "60");
							el->set_child_text(unit);
						}
						{
							std::ostringstream oss;
							oss << wel.get_arm();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "3");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						{
							std::ostringstream oss;
							oss << wel.get_massmin();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "5");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						{
							std::ostringstream oss;
							oss << wel.get_massmax();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "6");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						if (wel.get_flags() & Aircraft::WeightBalance::Element::flag_avgas) {
							{
								std::ostringstream oss;
								oss << "=F" << (16 + i);
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "23");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << "+E" << (16 + i) << "/Y4";
								loadedfuelformula += oss.str();
							}
							if (!fuelarmset) {
								std::ostringstream oss;
								oss << "=D" << (16 + i);
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", "3");
								el->set_attribute("Col", "23");
								el->set_child_text(oss.str());
							}
						} else {
							{
								std::ostringstream oss;
								oss << "=E" << (16 + i);
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "23");
								el->set_child_text(oss.str());
							}
						}
						std::string gainexpr("1");
						std::string offsexpr("0");
						for (unsigned int j = 0; j < wel.get_nrunits(); ++j) {
							const Aircraft::WeightBalance::Element::Unit& u(wel.get_unit(j));
							{
								std::ostringstream oss;
								oss << (52 + j);
						        	xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", oss.str());
								el->set_attribute("ValueType", "60");
								el->set_child_text(u.get_name());
							}
							{
								std::ostringstream oss;
								oss << "IF(C" << (16 + i) << "=\"" << u.get_name() << "\","
								    << u.get_factor() << ',' << gainexpr << ')';
								gainexpr = oss.str();
							}
							{
								std::ostringstream oss;
								oss << "IF(C" << (16 + i) << "=\"" << u.get_name() << "\","
								    << u.get_offset() << ',' << offsexpr << ')';
								offsexpr = oss.str();
							}
							if (u.get_name() == "kg" && std::isnan(kggain)) {
								kggain = u.get_factor();
								kgoffs = u.get_offset();
							}
						}
						if (wel.get_nrunits() == 1) {
							std::ostringstream oss0, oss1;
							oss0 << wel.get_unit(0).get_factor();
							gainexpr = oss0.str();
							oss0 << wel.get_unit(0).get_offset();
							offsexpr = oss1.str();
						}
						if (!gainexpr.empty() && gainexpr[0] == 'I') {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "25");
							el->set_child_text("=" + gainexpr);
						} else {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "25");
							el->set_attribute("ValueType", "40");
							el->set_child_text(gainexpr);
						}
						if (!offsexpr.empty() && offsexpr[0] == 'I') {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "26");
							el->set_child_text("=" + offsexpr);
						} else {
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row.str());
							el->set_attribute("Col", "26");
							el->set_attribute("ValueType", "40");
							el->set_child_text(offsexpr);
						}
					}
					if (!loadedfuelformula.empty()) {
						xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
						el->set_attribute("Row", "10");
						el->set_attribute("Col", "3");
						el->set_child_text("=" + loadedfuelformula.substr(1));
					}
					for (unsigned int i = 0; i < wb.get_nrenvelopes(); ++i) {
						const Aircraft::WeightBalance::Envelope& env(wb.get_envelope(i));
						std::ostringstream row0, row1;
						row0 << (50 + 2 * i);
						row1 << (51 + 2 * i);
						{
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row0.str());
							el->set_attribute("Col", "0");
							el->set_attribute("ValueType", "60");
							el->set_child_text(env.get_name());
						}
						{
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", row1.str());
							el->set_attribute("Col", "0");
							el->set_attribute("ValueType", "60");
							el->set_child_text(env.get_name());
						}
						for (unsigned int j = 0; j <= env.size(); ++j) {
							const Aircraft::WeightBalance::Envelope::Point& pt(env[(j >= env.size()) ? 0 : j]);
							armmin = std::min(armmin, pt.get_arm());
							armmax = std::max(armmax, pt.get_arm());
							massmin = std::min(massmin, pt.get_mass());
							massmax = std::max(massmax, pt.get_mass());
							std::ostringstream col;
							col << (1 + j);
							{
								std::ostringstream oss;
								oss << pt.get_mass();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row0.str());
								el->set_attribute("Col", col.str());
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << pt.get_arm();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row1.str());
								el->set_attribute("Col", col.str());
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
						}
					}
					{
						unsigned int rowcnt(100);
						const Aircraft::Distances& dists(get_dist());
						for (unsigned int i = 0; i < dists.get_nrtakeoffdist(); ++i, ++rowcnt) {
							const Aircraft::Distances::Distance& dist(dists.get_takeoffdist(i));
							std::ostringstream row;
							row << rowcnt;
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "0");
								el->set_attribute("ValueType", "60");
								el->set_child_text("Takeoff");
							}
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "1");
								el->set_attribute("ValueType", "60");
								el->set_child_text(dist.get_name());
							}
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "2");
								el->set_child_text("=X101");
							}
							{
								std::ostringstream oss;
								if (std::isnan(kggain))
									oss << "=AA" << (rowcnt + 1);
								else
									oss << "=(E43-" << kgoffs << ")*" << (1.0 / kggain);
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "4");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << dist.get_vrotate();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "25");
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << dist.get_mass();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "26");
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
							{
								const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
								unsigned int colcnt(30);
								for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi, ++colcnt) {
									std::ostringstream col, oss;
									col << colcnt;
									oss << *pi;
									xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
									el->set_attribute("Row", row.str());
									el->set_attribute("Col", col.str());
									el->set_attribute("ValueType", "40");
									el->set_child_text(oss.str());
								}
							}
							{
								const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
								unsigned int colcnt(40);
								for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi, ++colcnt) {
									std::ostringstream col, oss;
									col << colcnt;
									oss << *pi;
									xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
									el->set_attribute("Row", row.str());
									el->set_attribute("Col", col.str());
									el->set_attribute("ValueType", "40");
									el->set_child_text(oss.str());
								}
							}
						}
						for (unsigned int i = 0; i < dists.get_nrldgdist(); ++i, ++rowcnt) {
							const Aircraft::Distances::Distance& dist(dists.get_ldgdist(i));
							std::ostringstream row;
							row << rowcnt;
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "0");
								el->set_attribute("ValueType", "60");
								el->set_child_text("Landing");
							}
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "1");
								el->set_attribute("ValueType", "60");
								el->set_child_text(dist.get_name());
							}
							{
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "2");
								el->set_child_text("=Y101");
							}
							{
								std::ostringstream oss;
								if (std::isnan(kggain))
									oss << "=AA" << (rowcnt + 1);
								else
									oss << "=(E42-" << kgoffs << ")*" << (1.0 / kggain);
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "4");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << dist.get_vrotate();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "25");
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
							{
								std::ostringstream oss;
								oss << dist.get_mass();
								xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
								el->set_attribute("Row", row.str());
								el->set_attribute("Col", "26");
								el->set_attribute("ValueType", "40");
								el->set_child_text(oss.str());
							}
							{
								const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
								unsigned int colcnt(30);
								for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi, ++colcnt) {
									std::ostringstream col, oss;
									col << colcnt;
									oss << *pi;
									xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
									el->set_attribute("Row", row.str());
									el->set_attribute("Col", col.str());
									el->set_attribute("ValueType", "40");
									el->set_child_text(oss.str());
								}
							}
							{
								const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
								unsigned int colcnt(40);
								for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi, ++colcnt) {
									std::ostringstream col, oss;
									col << colcnt;
									oss << *pi;
									xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
									el->set_attribute("Row", row.str());
									el->set_attribute("Col", col.str());
									el->set_attribute("ValueType", "40");
									el->set_child_text(oss.str());
								}
							}
						}
					}
					{
						const FPlanWaypoint& wpt(fplan[0]);
						{
							std::ostringstream oss;
							oss << wpt.get_density_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", "100");
							el->set_attribute("Col", "23");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
					}
					{
						const FPlanWaypoint& wpt(fplan[fplan.get_nrwpt() - 1]);
						{
							std::ostringstream oss;
							oss << wpt.get_density_altitude();
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", "100");
							el->set_attribute("Col", "24");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
						double tas(0), fuelflow(0), bhp(cruise_bhp), rpm(cruise_rpm), mp(cruise_mp), da(wpt.get_density_altitude() + 1500.0);
						get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
						{
							std::ostringstream oss;
							oss << fuelflow;
							xmlpp::Element *el((*ni)->add_child("Cell", root->get_namespace_prefix()));
							el->set_attribute("Row", "7");
							el->set_attribute("Col", "1");
							el->set_attribute("ValueType", "40");
							el->set_child_text(oss.str());
						}
					}
					break;
				}

				default:
					break;
				}
			}
			xmlpp::Node::NodeList nlrows((*ni)->get_children("Rows"));
			for (xmlpp::Node::NodeList::const_iterator ni(nlrows.begin()), ne(nlrows.end()); ni != ne; ++ni) {
				switch (sheetnr) {
				case 0:
				{
					{
						unsigned int firstrow(cellspecs.get_fplfirstrow() + cellspecs.get_fplrows() * fplan.get_nrwpt());
						if (fplan.get_nrwpt() && cellspecs.get_pgbrkinterval())
							firstrow += ((fplan.get_nrwpt() - 1 + cellspecs.get_pgbrkoffset())
								     / cellspecs.get_pgbrkinterval()) * cellspecs.get_pgbrkheader();
						unsigned int lastrow(cellspecs.get_fplfirstrow() + cellspecs.get_fplrows() * cellspecs.get_pagelen());
						if (cellspecs.get_pgbrkinterval())
							lastrow += ((cellspecs.get_pagelen() - 1 + cellspecs.get_pgbrkoffset())
								    / cellspecs.get_pgbrkinterval()) * cellspecs.get_pgbrkheader();
						std::ostringstream nr, count;
						nr << firstrow;
						count << (lastrow-firstrow);
						xmlpp::Element *el((*ni)->add_child("RowInfo", root->get_namespace_prefix()));
						el->set_attribute("No", nr.str());
						el->set_attribute("Count", count.str());
						el->set_attribute("Unit", "12.1");
						el->set_attribute("Hidden", "1");
					}
					if (atcfpllen && atcfplrow != ~0U && cellspecs.is_rowadj() &&
					    (cellspecs.get_rowadjlineoffs() >= 0 || atcfpllen > (unsigned int)-cellspecs.get_rowadjlineoffs())) {
						unsigned int nrlines(atcfpllen + cellspecs.get_rowadjlineoffs() + cellspecs.get_rowadjlinelen() - 1U);
						nrlines /= cellspecs.get_rowadjlinelen();
						std::ostringstream nr, unit;
						nr << atcfplrow;
						unit << (cellspecs.get_rowadjlineheight() * nrlines + cellspecs.get_rowadjlineheightoffs());
						xmlpp::Element *el((*ni)->add_child("RowInfo", root->get_namespace_prefix()));
						el->set_attribute("No", nr.str());
						el->set_attribute("Unit", unit.str());
					}
					break;
				}

				case 1:
				{
					{
						const Aircraft::WeightBalance& wb(get_wb());
						{
							std::ostringstream nr, count;
							nr << (15 + wb.get_nrelements());
							count << (25 - wb.get_nrelements());
							xmlpp::Element *el((*ni)->add_child("RowInfo", root->get_namespace_prefix()));
							el->set_attribute("No", nr.str());
							el->set_attribute("Count", count.str());
							el->set_attribute("Unit", "12.1");
							el->set_attribute("Hidden", "1");
						}
						{
							std::ostringstream nr, count;
							nr << (50 + 2 * wb.get_nrenvelopes());
							count << (46 - 2 * wb.get_nrenvelopes());
							xmlpp::Element *el((*ni)->add_child("RowInfo", root->get_namespace_prefix()));
							el->set_attribute("No", nr.str());
							el->set_attribute("Count", count.str());
							el->set_attribute("Unit", "12.1");
							el->set_attribute("Hidden", "1");
						}
					}
					{
						const Aircraft::Distances& dists(get_dist());
						std::ostringstream nr, count;
						nr << (100 + dists.get_nrtakeoffdist() + dists.get_nrldgdist());
						count << (10 - dists.get_nrtakeoffdist() - dists.get_nrldgdist());
						xmlpp::Element *el((*ni)->add_child("RowInfo", root->get_namespace_prefix()));
						el->set_attribute("No", nr.str());
						el->set_attribute("Count", count.str());
						el->set_attribute("Unit", "12.1");
						el->set_attribute("Hidden", "1");
					}
					break;
				}

				default:
					break;
				}
			}
			xmlpp::Node::NodeList nlprintinfo((*ni)->get_children("PrintInformation"));
			for (xmlpp::Node::NodeList::const_iterator ni(nlprintinfo.begin()), ne(nlprintinfo.end()); ni != ne; ++ni) {
				xmlpp::Node::NodeList nlhpgbrk((*ni)->get_children("hPageBreaks"));
				for (xmlpp::Node::NodeList::const_iterator ni(nlhpgbrk.begin()), ne(nlhpgbrk.end()); ni != ne; ++ni) {
					switch (sheetnr) {
					case 0:
					{
						if (!cellspecs.get_pgbrkinterval())
							break;
						{
							xmlpp::Node::NodeList children((*ni)->get_children());
							for (xmlpp::Node::NodeList::const_iterator nni(children.begin()), nne(children.end()); nni != nne; ++nni)
								(*ni)->remove_child(*nni);
						}
						unsigned int count(0);
						for (unsigned int i(1), n(fplan.get_nrwpt()); i < n; ++i) {
							if (i > cellspecs.get_pagelen())
								break;
							unsigned int row;
							{
								unsigned int x(i - 1 + cellspecs.get_pgbrkoffset());
								row = x / cellspecs.get_pgbrkinterval();
								x -= row * cellspecs.get_pgbrkinterval();
								if (x)
									continue;
							}
							row *= cellspecs.get_pgbrkheader();
							row += cellspecs.get_fplfirstrow() + cellspecs.get_fplrows() * (i - 1);
							row -= cellspecs.get_pgbrkheader();
							std::ostringstream oss;
							oss << row;							
							xmlpp::Element *el((*ni)->add_child("break", root->get_namespace_prefix()));
							el->set_attribute("pos", oss.str());
							el->set_attribute("type", "manual");
							++count;
						}
						std::ostringstream oss;
						oss << count;
						static_cast<xmlpp::Element *>(*ni)->set_attribute("count", oss.str());
						break;
					}

					default:
						break;
					}
				}
			}
			xmlpp::Node::NodeList nlobjs((*ni)->get_children("Objects"));
			for (xmlpp::Node::NodeList::const_iterator ni(nlobjs.begin()), ne(nlobjs.end()); ni != ne; ++ni) {
				xmlpp::Node::NodeList nlobjgraphs((*ni)->get_children("SheetObjectGraph"));
				for (xmlpp::Node::NodeList::const_iterator ni(nlobjgraphs.begin()), ne(nlobjgraphs.end()); ni != ne; ++ni) {
					xmlpp::Node::NodeList nlgogobjgraph((*ni)->get_children("GogObject"));
					for (xmlpp::Node::NodeList::const_iterator ni(nlgogobjgraph.begin()), ne(nlgogobjgraph.end()); ni != ne; ++ni) {
						        xmlpp::Attribute *attr;
							if (!(attr = static_cast<xmlpp::Element *>(*ni)->get_attribute("type")))
								continue;
							if (attr->get_value() != "GogGraph")
								continue;
							xmlpp::Node::NodeList nlgogobjchart((*ni)->get_children("GogObject"));
							for (xmlpp::Node::NodeList::const_iterator ni(nlgogobjchart.begin()), ne(nlgogobjchart.end()); ni != ne; ++ni) {
								xmlpp::Attribute *attr;
								if (!(attr = static_cast<xmlpp::Element *>(*ni)->get_attribute("role")))
									continue;
								if (attr->get_value() != "Chart")
									continue;
								xmlpp::Node::NodeList nlgogobjax((*ni)->get_children("GogObject"));
								for (xmlpp::Node::NodeList::const_iterator ni(nlgogobjax.begin()), ne(nlgogobjax.end()); ni != ne; ++ni) {
									xmlpp::Attribute *attr;
									if (!(attr = static_cast<xmlpp::Element *>(*ni)->get_attribute("role")))
										continue;
									bool xaxis(attr->get_value() == "X-Axis");
									bool yaxis(attr->get_value() == "Y-Axis");
									if (!xaxis && !yaxis)
										continue;
									xmlpp::Node::NodeList nld((*ni)->get_children("data"));
									for (xmlpp::Node::NodeList::const_iterator ni(nld.begin()), ne(nld.end()); ni != ne; ++ni) {
										switch (sheetnr) {
										case 1:
											if (xaxis && !std::isnan(armmin) && !std::isnan(armmax) && armmin < armmax) {
												double scale(1);
												{
													double d(armmax - armmin);
													for (unsigned int i(0); i < 12; ++i) {
														double s(((i % 3) == 1) ? 2.5 : 2);
														d /= s;
														if (d < 20)
															break;
														scale *= s;
													}
												}
												{
													std::ostringstream oss;
													oss << (scale * std::floor((armmin - 0.1 * (armmax - armmin)) / scale));
													xmlpp::Element *el((*ni)->add_child("dimension"));
													el->set_attribute("id", "0");
													el->set_attribute("type", "GnmGODataScalar");
													el->set_child_text(oss.str());
												}
												{
													std::ostringstream oss;
													oss << (scale * std::ceil((armmax + 0.1 * (armmax - armmin)) / scale));
													xmlpp::Element *el((*ni)->add_child("dimension"));
													el->set_attribute("id", "1");
													el->set_attribute("type", "GnmGODataScalar");
													el->set_child_text(oss.str());
												}
											}
											if (yaxis && !std::isnan(massmin) && !std::isnan(massmax) && massmin < massmax) {
												double scale(1);
												{
													double d(massmax - massmin);
													for (unsigned int i(0); i < 12; ++i) {
														double s(((i % 3) == 1) ? 2.5 : 2);
														d /= s;
														if (d < 20)
															break;
														scale *= s;
													}
												}
												{
													std::ostringstream oss;
													oss << (scale * std::floor((massmin - 0.1 * (massmax - massmin)) / scale));
													xmlpp::Element *el((*ni)->add_child("dimension"));
													el->set_attribute("id", "0");
													el->set_attribute("type", "GnmGODataScalar");
													el->set_child_text(oss.str());
												}
												{
													std::ostringstream oss;
													oss << (scale * std::ceil((massmax + 0.1 * (massmax - massmin)) / scale));
													xmlpp::Element *el((*ni)->add_child("dimension"));
													el->set_attribute("id", "1");
													el->set_attribute("type", "GnmGODataScalar");
													el->set_child_text(oss.str());
												}
											}
											break;

										default:
											break;
										}
									}
								}
							}
					}
				}
			}
		}
	}
	doc->write_to_file_formatted(outfile);
}
